/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Module for remote display. Should work with any display driver.
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

/* must be after system includes */
#include "thinc.h"

#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"
#include "xf86Resources.h"
#include "xf86Version.h"
#include "xf86cmap.h"
#include "compiler.h"
#include "mi.h"
#include "mipointer.h"
#include "scrnintstr.h"
#include "pixmapstr.h"

/* for multiple servers */
#include <signal.h>
#include "xf86Priv.h"
#include <sys/vt.h>

#ifdef T_USEMITSHM
#define _XSHM_SERVER_
#include "X11/extensions/XShm.h"
#include "X11/extensions/shmstr.h"
#endif

/* xaa */
#ifdef T_USEXAA
#include "xaalocal.h"
#endif

/* leo suggested this so we can identifiy the version of a compiled server */
const char *thincProtocolVersionName = "THINC Protocol Version 0.1.2";

/*** Global variables ***/
static unsigned long ThincGeneration = 0;
int thincPrivateIndex = -1;
int thincGCPrivateIndex = -1;
int thincPixmapPrivateIndex = -1;
#ifdef T_APPSHARE
int thincWindowPrivateIndex = -1;
#endif

#ifdef T_DO_DEBUG
unsigned short thincDebugLevel = T_DBG_DEFAULT;
#else
unsigned short thincDebugLevel = 0;
#endif

/* local functions */

#ifdef T_USEMITSHM
static ShmFuncs thincShmFuncs = {thincShmCreatePixmap, thincShmPutImage};
static void thincSHMInit(ScreenPtr pScreen);
#endif

#ifdef T_USEXAA
static Bool thincXAAInit(ScreenPtr pScreen);
static void thincXAAUninit(thincScreenPtr pThinc);
#endif

#ifdef T_NODRAW
static Bool thincSaveScreen(ScreenPtr pScreen, int mode);
#endif

#if 0
static int thincInitNetwork(thincScreenPtr pThinc, const char *listen_addr,
	int port);
#else
static int thincInitNetwork(thincScreenPtr pThinc, const char *listen_addr,
	int port, int *find_port);
#endif
/* for multiple servers */
static void thincEnableDisableFBAccess (int scrnIndex, Bool enable);
static void thincSignal(int sig);

/*** Configuration Options ***/
typedef enum {
	OPTION_THINC_PORT
	,OPTION_THINC_LISTEN
	,OPTION_THINC_NOOFFSCREEN
	,OPTION_THINC_NOHWCURSOR
	,OPTION_THINC_NOWBUFFER
	,OPTION_THINC_NOCBUFFER
	,OPTION_THINC_NOCOPY
	,OPTION_THINC_NOSFILL
	,OPTION_THINC_NOPFILL
	,OPTION_THINC_NOGLYPH
	,OPTION_THINC_NOBILEVEL
	,OPTION_THINC_NOCOMPRESS
	,OPTION_THINC_COMPRESS_LEVEL
	,OPTION_THINC_VIDCOMPRESS
	,OPTION_THINC_NOVIDEO
	,OPTION_THINC_DEBUGLEVEL
	,OPTION_THINC_NOSSL
	,OPTION_THINC_NORESIZE
	,OPTION_THINC_NOSCHED
	,OPTION_THINC_NOCACHE
	,OPTION_THINC_CACHESZ
	,OPTION_THINC_NOBITCACHE
	,OPTION_THINC_BITCACHESZ
	,OPTION_THINC_NOPIXCACHE
	,OPTION_THINC_PIXCACHESZ
	,OPTION_THINC_UPDATEW
	,OPTION_THINC_UPDATEH
	,OPTION_THINC_FINDPORT
	,OPTION_THINC_NOAUTH
	,OPTION_THINC_PING
	/* 3 options: whether to record, where to put the files and whether
	 * to record only when clients are connected */
	,OPTION_THINC_RECORD
	,OPTION_THINC_RECORD_DIR
	,OPTION_THINC_RECORD_CONNONLY
	,OPTION_THINC_RECORD_RATE
#ifdef T_ENABLE_DEBUG_OPTIONS
	/* useful for testing. */
	,OPTION_THINC_NORAW
	,OPTION_THINC_EMPTYQUEUEOK
	,OPTION_THINC_OVERLAPOK
#endif
} THINCOpts;

static const OptionInfoRec ThincOptions[] = {
	/*token,		name,		type,		val, found */
	/* listen address */
	{OPTION_THINC_LISTEN,	"thincListen",	OPTV_STRING,	{0}, FALSE},
	{OPTION_THINC_PORT,	"thincPort",	OPTV_INTEGER,	{0}, FALSE},
	/* protocol options. By default they're all active, use these
	 * options to deactivate them. */
	{OPTION_THINC_NOCOPY,	"thincNoScreenCopy", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOSFILL,	"thincNoSolidFill", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOPFILL,	"thincNoPixmapFill", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOGLYPH,	"thincNoGlyphFill", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOBILEVEL, "thincNoBilevelFill", OPTV_BOOLEAN,{0},FALSE},
	{OPTION_THINC_NOHWCURSOR, "thincNoClientCursor", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOWBUFFER, "thincNoWriteBuffer", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOCBUFFER, "thincNoCommandBuffer", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOOFFSCREEN, "thincNoOffscreen", OPTV_BOOLEAN,{0},FALSE},
	{OPTION_THINC_NOCOMPRESS, "thincNoCompression", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_COMPRESS_LEVEL, "thincCompressionLevel", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_VIDCOMPRESS, "thincVideoCompression", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOVIDEO, "thincNoVideo", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOSSL, "thincNoSSL", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NORESIZE, "thincNoResize", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOSCHED, "thincNoScheduler", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOCACHE, "thincNoImageCache", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOBITCACHE, "thincNoBitmapCache", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_NOPIXCACHE, "thincNoPixmapCache", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_CACHESZ, "thincImageCacheBits", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_BITCACHESZ, "thincBitmapCacheBits", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_PIXCACHESZ, "thincPixmapCacheBits", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_UPDATEW, "thincUpdateWidth", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_UPDATEH, "thincUpdateHeight", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_DEBUGLEVEL, "thincDebugLevel", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_FINDPORT, "thincFindPort", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_NOAUTH, "thincNoAuth", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_PING, "thincPingFreq", OPTV_INTEGER, {0}, FALSE},
	{OPTION_THINC_RECORD, "thincRecord", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_RECORD_DIR, "thincRecordDir", OPTV_STRING, {0}, FALSE},
	{OPTION_THINC_RECORD_CONNONLY, "thincRecordConnOnly", OPTV_BOOLEAN, {0}, TRUE},
	{OPTION_THINC_RECORD_RATE, "thincRecordRate", OPTV_INTEGER, {0}, FALSE},
#ifdef T_ENABLE_DEBUG_OPTIONS
	/* testing options */
	{OPTION_THINC_NORAW, "thincNoRaw", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_EMPTYQUEUEOK, "thincEmptyQueuesOK", OPTV_BOOLEAN, {0}, FALSE},
	{OPTION_THINC_OVERLAPOK, "thincAllowOverlap", OPTV_BOOLEAN, {0}, FALSE},
#endif
	{-1,			NULL,		OPTV_NONE,	{0}, FALSE}
};

static const OptionInfoRec *THINCAvailableOptions(void *unused)
{
	return ThincOptions;
}

static MODULESETUPPROTO(ThincSetup);
static XF86ModuleVersionInfo ThincVersRec = {
	"thinc",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	THINC_VERSION_MAJOR, THINC_VERSION_MINOR, THINC_PATCHLEVEL,
	ABI_CLASS_VIDEODRV,	/* requires the video driver ABI */
	ABI_VIDEODRV_VERSION,
	MOD_CLASS_NONE,
	{0, 0, 0, 0}
};

XF86ModuleData thincModuleData = { &ThincVersRec, ThincSetup, NULL };
ModuleInfoRec THINC = {
	THINC_VERSION_MAJOR,
	"THINC",
	NULL,
	0,
	THINCAvailableOptions,
};

/* needed symbols */
/* FIXME: Doesn't work, we need to link with the library (static) */

/* Symbols needed for hardware cursor */
#if 1
static const char *cursorSymbols[] = {
	"xf86CreateCursorInfoRec",
	"xf86DestroyCursorInfoRec",
	"xf86InitCursor",
	NULL
};
#endif

/* for xaa */
#ifdef T_USEXAA
static const char *xaaSymbols[] = {
	"XAACreateInfoRec",
	"XAADestroyInfoRec",
	"XAAInit",
	NULL
};
#endif

static pointer
ThincSetup(pointer Module, pointer Options, int *ErrorMajor, int *ErrorMinor)
{
	static int initialised = T_FALSE;
	t_dbg_name(func)

	thincDbg(T_DBG_BASE, ("%s: Entering...\n", func));

	if (!initialised) {
		initialised = T_TRUE;
#ifndef REMOVE_LOADER_CHECK_MODULE_INFO
		if (xf86LoaderCheckSymbol("xf86AddModuleInfo"))
#endif
			xf86AddModuleInfo(&THINC, Module);

		thincDbg(T_DBG_BASE, ("%s: Loading additional dependencies\n",
			func));
		/* load additional symbols we need */
		/* FIXME: This doesn't work, we actually have to link
		 * with the respective libraries */
#if 0
		LoaderRefSymLists(cursorSymbols, xaaSymbols, NULL);
		xf86LoaderRefSymLists(cursorSymbols, xaaSymbols, NULL);
#endif
		LoaderRefSymLists(cursorSymbols, NULL);
	}
	thincDbg(T_DBG_BASE, ("%s: finished\n", func));

	return (pointer) TRUE;
}

/*** primary initialization fn for remote display.  Called by drivers. ***/
Bool ThincInit(ScreenPtr pScreen, char *FBStart)
{
	t_dbg_name(func)
	thincScreenPtr pThinc;
	ScrnInfoPtr xf86Scrn = xf86Screens[pScreen->myNum];
	int thincPort, index;
	OptionInfoPtr options;
	char *listen_addr;
	Bool opt;
	int err, find_port = T_NO, zlevel, record_flags=0;
	char *record_dir=NULL;

	thincDbg(T_DBG_BASE, ("%s: my generation %ld, server generation %ld\n", 
		func, ThincGeneration, serverGeneration));

	if (!xf86LoadSubModule(xf86Scrn, "ramdac"))
		return FALSE;
	xf86LoaderReqSymLists(cursorSymbols, NULL);

	if (ThincGeneration != serverGeneration) {
		thincDbg(T_DBG_BASE, ("%s: myGeneration != serverGeneration, "
			"allocating private indexes\n", func));
		/* allocate Screen private index */
		if ((thincPrivateIndex = AllocateScreenPrivateIndex()) < 0) {
			thincError("ERROR: Couldn't allocate screen private\n");
			return FALSE;
		}
		/* allocate Pixmap private */
		thincPixmapPrivateIndex = AllocatePixmapPrivateIndex();
		if( !AllocatePixmapPrivate(pScreen, thincPixmapPrivateIndex,
			sizeof(thincPixmapRec)) ) {
			thincError("ERROR: Couldn't allocate Pixmap private\n");
			return FALSE;
		}
		/* allocate GC private - doesn't fail */
		thincGCPrivateIndex = AllocateGCPrivateIndex();
		if(!AllocateGCPrivate(pScreen, thincGCPrivateIndex, 
			sizeof(thincGCRec))) {
			thincError("ERROR: Couldn't allocate GC private\n");
			return FALSE;
		}
#ifdef T_APPSHARE
		/* allocate window private */
		thincWindowPrivateIndex = AllocateWindowPrivateIndex();
		if(!AllocateWindowPrivate(pScreen, thincWindowPrivateIndex,
			sizeof(thincWindowRec)) ) {
			thincError("ERROR: Couldn't allocate Window private\n");
			return FALSE;
		}
#endif
		ThincGeneration = serverGeneration;
	}

	/* allocate the actual screen private */
	thincDbg(T_DBG_BASE, ("%s: allocating Screen Private struct\n", func));
	if (!(pThinc = malloc(sizeof(thincScreenRec)))) {
		thincError("Couldn't allocate ScreenRec\n");
		return FALSE;
	}

	pScreen->devPrivates[thincPrivateIndex].ptr = (pointer) pThinc;

	/* Process Options */
	options = malloc(sizeof(ThincOptions));
	if(!options) {
		thincError("Couldn't allocate options structure\n");
		free(pThinc);
		return FALSE;
	}

	/* network parameters */
	memcpy(options, ThincOptions, sizeof(ThincOptions));
	xf86ProcessOptions(xf86Scrn->scrnIndex, xf86Scrn->options, options);
	if(!xf86GetOptValInteger(options, OPTION_THINC_PORT, &thincPort)) {
		thincLog("THINC: using default port %d\n", _THINC_PORT);
		thincPort = _THINC_PORT;
	}

	if(xf86GetOptValInteger(options, OPTION_THINC_FINDPORT, &thincPort)) {
		thincLog("THINC: looking for free port starting at %d\n", 
			thincPort);
		find_port = T_YES;
	}
 
	/* no need to chk for NULL-ness since our listening option uses
	 * that as the default value */
	listen_addr = xf86GetOptValString(options, OPTION_THINC_LISTEN);

	if(xf86GetOptValInteger(options, OPTION_THINC_PING, 
		(int*)&pThinc->ping_freq)) {
		if(!pThinc->ping_freq) 
			thincLog("THINC: Disabling pings to clients\n");
		else 
			thincLog("THINC: sending pings every %dms\n", 
				pThinc->ping_freq);
	} else {
		pThinc->ping_freq = T_TIMER;
		thincLog("THINC: Using default ping frequency %dms\n",
			pThinc->ping_freq);
	}

	if(xf86GetOptValBool(options, OPTION_THINC_NOHWCURSOR, &opt)) {
		thincLog("THINC: Not using Client-side cursor\n");
		pThinc->useHWCursor = T_FALSE;
	} else
		pThinc->useHWCursor = T_TRUE;

	/* Server Features */
	if(xf86GetOptValBool(options, OPTION_THINC_NOOFFSCREEN, &opt)) {
		thincLog("THINC: Not optimizing offscreen operations\n");
		pThinc->useOffscreen = T_FALSE;
	} else 
		pThinc->useOffscreen = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_NOCOMPRESS, &opt)) {
		thincLog("THINC: Not using compression for updates\n");
		pThinc->compress = T_FALSE;
	} else
		pThinc->compress = T_TRUE;

	if(xf86GetOptValInteger(options, OPTION_THINC_COMPRESS_LEVEL,
		&zlevel)) {
		thincLog("THINC: Using compression level %d\n", zlevel);
	} else {
		thincLog("THINC: Using default compression level %d\n",
			T_ZCOMPRESS_LEVEL);
		zlevel = T_ZCOMPRESS_LEVEL;
	}

	if(xf86GetOptValBool(options, OPTION_THINC_VIDCOMPRESS, &opt)) {
		thincLog("THINC: Using compression for video\n");
		pThinc->compressVideo = T_TRUE;
	} else
		pThinc->compressVideo = T_FALSE;

	if(xf86GetOptValBool(options, OPTION_THINC_NOVIDEO, &opt)) {
		thincLog("THINC: Not enabling Video Support\n");
		pThinc->useVideo = T_FALSE;
	} else
		pThinc->useVideo = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_NOSSL, &opt)) {
		thincLog("THINC: Not enabling SSL Support\n");
		pThinc->beSecure = T_FALSE;
	} else
		pThinc->beSecure = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_NOAUTH, &opt)) {
		thincLog("THINC: Not enabling Authentication Support\n");
		pThinc->useAuth = T_FALSE;
	} else
		pThinc->useAuth = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_NORESIZE, &opt)) {
		thincLog("THINC: Not enabling Client Resize Support\n");
		pThinc->allowResize = T_FALSE;
	} else
		pThinc->allowResize = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_NOSCHED, &opt)) {
		thincLog("THINC: Not enabling command scheduling\n");
		pThinc->doSched = T_FALSE;
	} else
		pThinc->doSched = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_NOCACHE, &opt)) {
		thincLog("THINC: Not using image cache\n");
		pThinc->useRawCache = T_FALSE;
	} else
		pThinc->useRawCache = T_TRUE;
	if(!xf86GetOptValInteger(options, OPTION_THINC_CACHESZ, 
		(int*)&pThinc->raw_cache_bits)) {
		thincLog("THINC: using default image cache size: %d\n", 
			T_RAWCACHE_BITS);
		pThinc->raw_cache_bits = T_RAWCACHE_BITS;
	}

	if(xf86GetOptValBool(options, OPTION_THINC_NOBITCACHE, &opt)) {
		thincLog("THINC: Not using bitmap cache\n");
		pThinc->useBitCache = T_FALSE;
	} else
		pThinc->useBitCache = T_TRUE;
	if(!xf86GetOptValInteger(options, OPTION_THINC_BITCACHESZ, 
		(int*)&pThinc->bit_cache_bits)) {
		thincLog("THINC: using default bitmap cache size: %d\n", 
			T_BITCACHE_BITS);
		pThinc->bit_cache_bits = T_BITCACHE_BITS;
	}
	if(xf86GetOptValBool(options, OPTION_THINC_NOPIXCACHE, &opt)) {
		thincLog("THINC: Not using pixmap cache\n");
		pThinc->usePixCache = T_FALSE;
	} else
		pThinc->usePixCache = T_TRUE;
	if(!xf86GetOptValInteger(options, OPTION_THINC_PIXCACHESZ, 
		(int*)&pThinc->pix_cache_bits)) {
		thincLog("THINC: using default pixmap cache size: %d\n", 
			T_PIXCACHE_BITS);
		pThinc->pix_cache_bits = T_PIXCACHE_BITS;
	}

	if(xf86GetOptValBool(options, OPTION_THINC_NOWBUFFER, &opt)) {
		thincLog("THINC: Not using write buffering\n");
		pThinc->doWriteBuffer = T_FALSE;
	} else
		pThinc->doWriteBuffer = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_NOCBUFFER, &opt)) {
		thincLog("THINC: Not using command buffering\n");
		pThinc->doCmdBuffer = T_FALSE;
	} else
		pThinc->doCmdBuffer = T_TRUE;

#ifdef T_DO_DEBUG
	if(xf86GetOptValInteger(options, OPTION_THINC_DEBUGLEVEL, 
		(int*)&thincDebugLevel)) {
		thincLog("THINC: Using debug level: %d\n", thincDebugLevel);
	} else {
		thincLog("THINC: Using default debug level: %d\n", 
			thincDebugLevel);
		thincDebugLevel = T_DBG_DEFAULT;
	}
#else
	if(xf86GetOptValInteger(options, OPTION_THINC_DEBUGLEVEL, 
		(int*)&thincDebugLevel)) {
		thincLog("THINC: Ignoring DebugLevel directive, debugging "
			"was disabled at compile time\n");
	}
#endif

#ifdef T_ENABLE_DEBUG_OPTIONS
	if(xf86GetOptValBool(options, OPTION_THINC_NORAW, &opt)) {
		thincLog("THINC: Not doing raw updates\n");
		pThinc->doRaw = T_FALSE;
	} else
		pThinc->doRaw = T_TRUE;

	if(xf86GetOptValBool(options, OPTION_THINC_EMPTYQUEUEOK, &opt)) {
		thincLog("THINC: Replacing empty queues with RAW Updates\n");
		pThinc->emptyQueuesOK = T_TRUE;
	} else
		pThinc->emptyQueuesOK = T_FALSE;

	if(xf86GetOptValBool(options, OPTION_THINC_OVERLAPOK, &opt)) {
		thincLog("THINC: Allowing updates to overlap\n");
		pThinc->allowOverlap = T_TRUE;
	} else
		pThinc->allowOverlap = T_FALSE;
#endif

	if(!xf86GetOptValInteger(options, OPTION_THINC_UPDATEW, 
		(int*)&pThinc->updateWidth)) {
		thincLog("THINC: using default update width\n");
		pThinc->updateWidth = 0;
	}
	if(!xf86GetOptValInteger(options, OPTION_THINC_UPDATEH, 
		(int*)&pThinc->updateHeight)) {
		thincLog("THINC: using default update height\n");
		pThinc->updateHeight = 0;
	}

	/* protocol options */
	/* NOTE: Some of these options may not have an actual effect, or
	 * may make things not work correctly (for example, using offscreen 
	 * pixmaps but not enabling glyphs results in bad updates being sent).
	 * Not easily fixable.
	 */
	if(xf86GetOptValBool(options, OPTION_THINC_NOCOPY, &opt)) {
		thincLog("THINC: Not using ScreenToScreen Copies\n");
		pThinc->useCopy = T_FALSE;
	} else {
		pThinc->useCopy = T_TRUE;
	}

	if(xf86GetOptValBool(options, OPTION_THINC_NOSFILL, &opt)) {
		thincLog("THINC: Not using Solid Fills\n");
		pThinc->useSFill = T_FALSE;
	} else {
		pThinc->useSFill = T_TRUE;
	}

	if(xf86GetOptValBool(options, OPTION_THINC_NOPFILL, &opt)) {
		thincLog("THINC: Not using Pixmap Fills\n");
		pThinc->usePFill = T_FALSE;
	} else {
		pThinc->usePFill = T_TRUE;
	}

	if(xf86GetOptValBool(options, OPTION_THINC_NOGLYPH, &opt)) {
		thincLog("THINC: Not using Glyph Fills\n");
		pThinc->useGlyph = T_FALSE;
	} else {
		pThinc->useGlyph = T_TRUE;
	}

	if(xf86GetOptValBool(options, OPTION_THINC_NOBILEVEL, &opt)) {
		thincLog("THINC: Not using Bilevel Fills\n");
		pThinc->useBilevel = T_FALSE;
	} else {
		pThinc->useBilevel = T_TRUE;
	}

#ifdef T_RECORD
	/* always do this, so we don't have to do extra check for doRecord */
	if(!xf86GetOptValInteger(options, OPTION_THINC_RECORD_RATE, 
		(int*)&pThinc->rate_ms)) {
		thincLog("THINC: Not rate limiting recording\n");
		pThinc->rate_ms = T_REC_DEFAULT_RATE;
	} else {
		thincLog("THINC: Limiting recording to once every %ums\n",
			pThinc->rate_ms);
	}
	if(xf86GetOptValBool(options, OPTION_THINC_RECORD, &opt)) {
		thincLog("THINC: Recording your session\n");
		pThinc->doRecord = T_TRUE;
	} else {
		thincLog("THINC: Not recording your session\n");
		pThinc->doRecord = T_FALSE;
		goto no_record;
	}
	
	if(xf86GetOptValBool(options, OPTION_THINC_RECORD_CONNONLY, &opt)) {
		thincLog("THINC: Only recording when user is connected\n");
		record_flags = 0;
	} else {
		thincLog("THINC: Always recording commands\n");
		record_flags = T_REC_ALWAYS;
	}

	record_dir = xf86GetOptValString(options, OPTION_THINC_RECORD_DIR);
	if(record_dir) {
		thincLog("THINC: Using %s as base directory for recording "
			"commands\n", record_dir);
	} else {
		thincLog("THINC: Using default basedirectory %s for "
			"recording commands\n", T_REC_DEFAULT_DIR);
	}
no_record:
#endif	/* T_RECORD */
	free(options);

	/* Fill command handlers */
	index = thincCmd_idx(T_FB_RUPDATE);
	pThinc->cmdHandlerList[index] = thincAllocCmdInfo(T_FB_RUPDATE,
		"thincRawUpdate", 14, thincRawCreate, thincRawDestroy,
		thincRawCopy, thincRawModify, thincRawExecute, 
		thincRawGetInfo, thincRawFlush);
	if(!pThinc->cmdHandlerList[index]) {
		thincError("ERROR: Coulnd't allocate memory for "
			"Offscreen Pixmap Handler\n");
		return FALSE;
	}
	index = thincCmd_idx(T_FB_COPY);
	pThinc->cmdHandlerList[index] = thincAllocCmdInfo(T_FB_COPY,
		"thincCopy", 14, thincCopyCreate, thincCopyDestroy,
		thincCopyCopy, thincCopyModify, thincCopyExecute, 
		thincCopyGetInfo, thincCopyFlush);
	if(!pThinc->cmdHandlerList[index]) {
		thincError("ERROR: Coulnd't allocate memory for "
			"Offscreen Pixmap Handler\n");
		return FALSE;
	}
	index = thincCmd_idx(T_FB_SFILL);
	pThinc->cmdHandlerList[index] = thincAllocCmdInfo(T_FB_SFILL, 
		"thincSolidFill", 14, thincSFillCreate, 
		thincSFillDestroy, thincSFillCopy, thincSFillModify, 
		thincSFillExecute, thincSFillGetInfo,
		thincSFillFlush);
	if(!pThinc->cmdHandlerList[index]) {
		thincError("ERROR: Coulnd't allocate memory for "
			"Offscreen Pixmap Handler\n");
		return FALSE;
	}
	index = thincCmd_idx(T_FB_PFILL);
	pThinc->cmdHandlerList[index] = thincAllocCmdInfo(T_FB_PFILL, 
		"thincPixmapFill", 15, thincPFillCreate, 
		thincPFillDestroy, thincPFillCopy, thincPFillModify, 
		thincPFillExecute, thincPFillGetInfo,
		thincPFillFlush);
	if(!pThinc->cmdHandlerList[index]) {
		thincError("ERROR: Coulnd't allocate memory for "
			"Offscreen Pixmap Handler\n");
		return FALSE;
	}
	index = thincCmd_idx(T_FB_GLYPH);
	pThinc->cmdHandlerList[index] = thincAllocCmdInfo(T_FB_GLYPH, 
		"thincGlyph", 10, thincGlyphCreate, thincGlyphDestroy,
		thincGlyphCopy, thincGlyphModify, thincGlyphExecute, 
		thincGlyphGetInfo, thincGlyphFlush);
	if(!pThinc->cmdHandlerList[index]) {
		thincError("ERROR: Coulnd't allocate memory for "
			"Offscreen Pixmap Handler\n");
		return FALSE;
	}
	index = thincCmd_idx(T_FB_BILEVEL);
	pThinc->cmdHandlerList[index] = thincAllocCmdInfo(T_FB_BILEVEL, 
		"thincBilevel", 12, thincBilevelCreate, 
		thincBilevelDestroy, thincBilevelCopy, 
		thincBilevelModify, thincBilevelExecute, 
		thincBilevelGetInfo, thincBilevelFlush);
	if(!pThinc->cmdHandlerList[index]) {
		thincError("ERROR: Coulnd't allocate memory for "
			"Offscreen Pixmap Handler\n");
		return FALSE;
	}
	index = thincCmd_idx(T_FB_CURSOR);
	pThinc->cmdHandlerList[index] = thincAllocCmdInfo(T_FB_CURSOR, 
		"thincCursor", 11, thincCursorCreate, 
		thincCursorDestroy, thincCursorCopy, 
		thincCursorModify, thincCursorExecute, 
		thincCursorGetInfo, thincCursorFlush);
	if(!pThinc->cmdHandlerList[index]) {
		thincError("ERROR: Coulnd't allocate memory for "
			"Offscreen Pixmap Handler\n");
		return FALSE;
	}

	/* init network */
#if 0
	if(!thincInitNetwork(pThinc, listen_addr, thincPort)) {
#else
	if(!thincInitNetwork(pThinc, listen_addr, thincPort, &find_port)) {
#endif
		thincError("THINC ERROR: Could not initialize network. "
			"Maybe you forgot to copy the SSL certificates to "
			"/etc/X11/thinc?\n");
		return FALSE;
	} else if(find_port) {
		thincMsg("THINC: Listening on port %d\n", find_port);
	}

	t_list_init(&pThinc->clients);

	/* add our mouse and keyboard */
	thincInitMouse();
	thincInitKeyboard();

	/* initialize private struct */
	pThinc->theFB = FBStart;
	pThinc->width = xf86Scrn->virtualX;
	pThinc->height = xf86Scrn->virtualY;
	pThinc->depth = xf86Scrn->depth;
	pThinc->bpp = xf86Scrn->bitsPerPixel;
	pThinc->Bpp = pThinc->bpp >> 3;
	switch(pThinc->Bpp) {
		case 1: pThinc->Bpp_shift = 0; break;
		case 2: pThinc->Bpp_shift = 1; break;
		case 4: pThinc->Bpp_shift = 2; break;
		default: 
			T_BUG_ARG("Have no support for %d bpp. Tell "
				"developers to stop being lazy...",
				pThinc->bpp);
			return FALSE;
	}
	pThinc->widthBytes = t_pix_to_bytes(pThinc->width, pThinc);
	pThinc->pScreen = pScreen;

	if(!pThinc->updateWidth || pThinc->updateWidth > pThinc->width)
		pThinc->updateWidth = pThinc->width;
	if(!pThinc->updateHeight || pThinc->updateHeight > pThinc->height)
		pThinc->updateHeight = pThinc->height;
	pThinc->updateSize = pThinc->updateWidth * pThinc->updateHeight;

	if(pThinc->useRawCache && !pThinc->raw_cache_bits)
		pThinc->raw_cache_bits = T_RAWCACHE_BITS;
	if(pThinc->useBitCache && !pThinc->bit_cache_bits)
		pThinc->bit_cache_bits = T_BITCACHE_BITS;
	if(pThinc->usePixCache && !pThinc->pix_cache_bits)
		pThinc->bit_cache_bits = T_PIXCACHE_BITS;

	/**** wrap functions ****/
	thinc_wrap(pThinc, pScreen, WakeupHandler, thincWakeupHandler);
	thinc_wrap(pThinc, pScreen, BlockHandler, thincBlockHandler);
	thinc_wrap(pThinc, pScreen, CloseScreen, thincCloseScreen);

#ifdef T_NODRAW
	thinc_wrap(pThinc, pScreen, SaveScreen, thincSaveScreen);
#endif

	/* for multiple servers */
	thinc_wrap(pThinc, xf86Scrn, EnableDisableFBAccess, 
		thincEnableDisableFBAccess);

	/* window funcs */
#ifdef T_APPSHARE	/* no need to wrap this normally */
	thinc_wrap(pThinc, pScreen, CreateWindow, thincCreateWindow);
	thinc_wrap(pThinc, pScreen, DestroyWindow, thincDestroyWindow);
	thinc_wrap(pThinc, pScreen, ChangeWindowAttributes, 
		thincChangeWindowAttributes);
	thinc_wrap(pThinc, pScreen, RealizeWindow, thincMapWindow);
	thinc_wrap(pThinc, pScreen, UnrealizeWindow, thincUnmapWindow);
	thinc_wrap(pThinc, pScreen, PositionWindow, thincPositionWindow);
	thinc_wrap(pThinc, pScreen, WindowExposures, thincWindowExposures);
#endif	/* T_APPSHARE */
	thinc_wrap(pThinc, pScreen, PaintWindowBackground, 
		thincPaintWindowBackground);
	thinc_wrap(pThinc, pScreen, PaintWindowBorder, thincPaintWindowBorder);
	thinc_wrap(pThinc, pScreen, CopyWindow, thincCopyWindow);
	/* Pixmap Funcs */
	thinc_wrap(pThinc, pScreen, CreatePixmap, thincCreatePixmap);
	thinc_wrap(pThinc, pScreen, DestroyPixmap, thincDestroyPixmap);
	thinc_wrap(pThinc, pScreen, ModifyPixmapHeader, 
		thincModifyPixmapHeader);
	/* GC funcs */
	thinc_wrap(pThinc, pScreen, CreateGC, thincCreateGC);

	/* in case the cursor initialization function is not called */
	pThinc->cursor = NULL;

	pThinc->resizeCount = 0;
	pThinc->clientCount = 0;

	/* init stats */
	thincStatsInit(pThinc);

	/* init instrumentation */
	if(thincInstrInit(pThinc) == T_ERROR) {
		thincError("ERROR: Couldn't initialize instrumentation "
			"code\n");
		return FALSE;
	}

#ifdef T_USEMITSHM
	/* Since there's no way to wrap, we just override the driver's
	 * shm funcs. Hopefully we're doing all that they need to do :)
	 */
	if(pThinc->useOffscreen)
		thincSHMInit(pScreen);
#endif

	/* Initialize XVideo */
	if(pThinc->useVideo && !thincInitVideo(pThinc)) {
		thincWarn("Couldn't initialize XVideo support.\n");
		pThinc->useVideo = T_FALSE;
		pThinc->thincVid = NULL;
	} else
		pThinc->droppedFrames = 0;

	if(pThinc->compress || pThinc->compressVideo) {
		thincDbg(T_DBG_BASE, ("%s: initialiazing compression "
			"engine\n", func));
		err = thinc_compressInit(pThinc, zlevel);
		if(err == T_ERROR) {
			thincError("ERROR: Compression initialization "
				"failed\n");
			return FALSE;
		}
	}

	/* Initialize bitmaps */
	if(thincInitBitmaps(pThinc) == T_ERROR) {
		thincError("ERROR: Couldn't initialize bitmaps\n");
		return FALSE;
	}

	/* initialize timer */
	if(thincInitTimer(pThinc) == T_ERROR) {
		thincError("ERROR: Couldn't initialize timer\n");
		return FALSE;
	}

	pThinc->last_mouse_client = NULL;

#ifdef T_USERENDER
	if(thincPictureInit(pScreen) == T_ERROR) {
		thincError("ERROR: Couldn't initialize RENDER code\n");
		return FALSE;
	}
#endif

#ifdef T_USEXAA
	/* Initialize XAA */
	/* TODO: What happens if driver also wants to initialize it?
	 * How can we wrap around the driver's functions? */
	if(!thincXAAInit(pScreen)) {
		thincError("ERROR: Couldn't initialize XAA code\n");
		return FALSE;
	}
#endif

	thincDbg(T_DBG_BASE, ("%s: getting session owner information "
		"uid is %d\n", func, getuid()));

	/* always get owner, even if we don't do auth */
	if(thinc_get_owner(getuid(),&pThinc->owner,&pThinc->owner_len)
		== T_AUTH_ERROR) 
	{
		thincError("ERROR: Couldn't get session owner\n");
		return FALSE;
	}
	thincDbg(T_DBG_BASE, ("%s: session owner is %s (%d chars)\n", 
		func, pThinc->owner, pThinc->owner_len));

#ifdef T_NODRAW
	/* initialize client 0 */
	if(thincClient0Init(pThinc) == T_ERROR) {
		thincError("ERROR: Couldn't initialize Client 0\n");
		return FALSE;
	}
#endif

	if(thincAppshareInit(pThinc) == T_ERROR) {
		thincError("ERROR: Couldn't initialize application sharing\n");
		return FALSE;
	}

	/* initialiaze thincback stats. need to do it before RecordInit */
#if 0
	pThinc->mouse_down = 0;
	pThinc->mouse_up = 0;
	pThinc->keyb_down = 0;
	pThinc->keyb_up = 0;
	pThinc->display_count = 0;
	REGION_INIT(pThinc->pScreen, &pThinc->display_change, NullBox, 1);
#else
	pThinc->screensaver = 0;	/* global */
	pThinc->total_area = (double)pThinc->width*pThinc->height;

	t_list_init(&pThinc->stats);
	pThinc->stats_next_id = 0;
#endif

#ifdef T_RECORD
	if(pThinc->doRecord && 
		(thincRecordInit(pThinc, record_dir, record_flags) == T_ERROR )
	) {
		thincError("ERROR: Couldn't initialize command recorder\n");
		return FALSE;
	}
#endif

	/* Load the extension */
	thincLoadExtension(pThinc);

	/* for multiple servers */
	signal(SIGUSR1, thincSignal);

	/* the end */
	thincDbg(T_DBG_BASE, ("%s: Finished. Framebuffer is %dx%d@(%d/%d)\n",
		func, pThinc->width, pThinc->height, pThinc->depth, 
		pThinc->bpp));
	return TRUE;
}

/* software cursor should have been initialized by the time this function is
 * called */
/* TODO: Seems like there can be only 1 hw cursor. therefore, if the video
 * driver has one, we should wrap around it instead of initializing our
 * own and probably messing everything up.. but, will wrap work??
 * There may be a problem with the Flags and specially with the way
 * the driver may ask for the cursor info. We would have to support all
 * of the possible packing formats (not too bad, but definitely for
 * the ToDo) */
int ThincCursorInit(ScreenPtr pScreen)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = THINCPtr(pScreen);

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	pThinc->moveCount = 0;

	/* initialize hw cursor */
	if(!pThinc->useHWCursor) {
		thincDbg(T_DBG_INPUT, ("%s: Hardware Cursor is deactivated\n", 
			func));
		return T_OK;
	}

	if(THINCHWCursorInit(pScreen) == T_ERROR) {
		thincError("ERROR: Hardware cursor initialization failed\n");
		pThinc->useHWCursor = T_FALSE;
		return T_ERROR;
	}
	thincDbg(T_DBG_INPUT, ("%s: Hardware Cursor "
		"initialized (%d x %d). size: %d\n", func,
		pThinc->cursor->width, pThinc->cursor->height, 
		pThinc->cursor->size));

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
	return T_OK;
}

/* initialize network. If being secure, initialize security related stuff
 * too */
#if 0
static int thincInitNetwork(thincScreenPtr pThinc, const char *listen_addr,
	int port)
#else
static int thincInitNetwork(thincScreenPtr pThinc, const char *listen_addr,
	int port, int *find_port)
#endif
{
	/* FIXME FIXME */
	const char *cert_path = "/etc/X11/thinc/certs/cacert.pem";
	const char *key_path = "/etc/X11/thinc/certs/privkey.pem";

#if 0
	pThinc->listenSocket = tcpThincListen(listen_addr, port);
#else
	pThinc->listenSocket = tcpThincListen(listen_addr, port, find_port);
#endif
	if(pThinc->listenSocket == T_ERROR) {
		return T_FALSE;
	}

	if(pThinc->beSecure) {
		pThinc->ctx = thinc_ssl_init_server(cert_path, key_path);
		if(!pThinc->ctx) {
			tcpThincClose(pThinc->listenSocket);
			return T_FALSE;
		}
	} else {
		pThinc->ctx = NULL;
	}

	FD_ZERO(&pThinc->fds);
	pThinc->fdmax = -1;
	thincAddSocket(pThinc, pThinc->listenSocket);

	return T_TRUE;
}

/**************************************/
/*** check for input in our sockets ***/
/**************************************/
void thincWakeupHandler(int i, pointer blockData, unsigned long err, 
	pointer pReadmask)
{
	ScreenPtr pScreen = screenInfo.screens[i];
	ScrnInfoPtr xf86Scrn = xf86Screens[pScreen->myNum];
	thincScreenPtr pThinc = THINCPtr(pScreen);
	int sigstate;
	int res;
	fd_set socks;
	t_dbg_name(func)

	instr_add("WAKEUP", TI_START);

	thincDbg(T_DBG_ANNOY, ("%s: Checking for data on our sockets, "
		"fdmax %d\n", func, pThinc->fdmax));

	/* hack so that multiple servers can run concurrently. this way
	 * we don't have to modify the X server */
	if(!xf86Scrn->vtSema)
		xf86Scrn->vtSema = 1;

	/* flush clients' buffer */
	thincFlushBufferAll(pThinc);

	sigstate = xf86BlockSIGIO(); 
	/* chk listening socket */
	res = thincDoSelect(&pThinc->fds, &socks, pThinc->fdmax);

	/* TODO: Break this loop into function calls, getting ugly */
	if (res == 0) {
		thincDbg(T_DBG_ANNOY, ("%s: no input for us\n", func));
		goto end;
	} else if (res == T_ERROR) {
		thincError("ERROR: Error checking for input on sockets\n");
		goto end;
	} else {
		if(thincHaveNewClients(pThinc, &socks) == T_YES)
			/* one less socket to check */
			--res;

		/* are we done? */
		if (res == 0)
			goto end;

		/* better this than a segfault... */
		t_bug_on_2(
			t_list_is_empty(&pThinc->clients),
			T_BUG_ARG("select() says we have more active clients "
				"than we can find %d", res),
			goto end
		);
		/* screw it, i won't check for errors. not today */
		thincCheckClients(pThinc, &socks, res);
	}
  end:
	xf86UnblockSIGIO(sigstate);

	thincDbg(T_DBG_ANNOY, ("%s: calling driver's function\n", func));
	thinc_unwrap(pThinc, pScreen, WakeupHandler);
	(*pScreen->WakeupHandler) (i, blockData, err, pReadmask);
	thinc_wrap(pThinc, pScreen, WakeupHandler, thincWakeupHandler);
	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));
	instr_add("WAKEUP", TI_END);
}

/* We're about to block, flush the clients' buffer */
void thincBlockHandler(int i, pointer blockData, pointer pTimeout,
	pointer pReadmask)
{
	t_dbg_name(func)
	ScreenPtr pScreen = screenInfo.screens[i];
	ScrnInfoPtr xf86Scrn = xf86Screens[pScreen->myNum];
	thincScreenPtr pThinc = THINCPtr(pScreen);
	int res;

	instr_add("BLOCK", TI_START);
	thincDbg(T_DBG_ANNOY, ("%s\n", func));

	/* hack so that multiple servers can run concurrently. this way
	 * we don't have to modify the X server */
	if(!xf86Scrn->vtSema)
		xf86Scrn->vtSema = 1;

	/* flush command buffer, then write buffer */
	res = thincFlushCmdAll(pThinc);
	if(res == T_ERROR) {
		thincError("ERROR: Couldn't flush command buffer\n");
	} else if(res == T_PART) {
		/* set timer in case we go quiet for too long */
#ifdef T_RECORD
		if(pThinc->rate_ms)
			thincSetTimer(pThinc, TimerForceOld, pThinc->rate_ms, 
				thincTimerCallback, (pointer)pThinc);
		else
			thincSetTimer(pThinc, TimerForceOld, T_TIMER, 
				thincTimerCallback, (pointer)pThinc);
#else
			thincSetTimer(pThinc, TimerForceOld, T_TIMER, 
				thincTimerCallback, (pointer)pThinc);
#endif	/* T_RECORD */
	} else if(pThinc->ping_freq) {
		thincSetTimer(pThinc, TimerForceOld, pThinc->ping_freq, 
			thincTimerCallback, (pointer)pThinc);
	}

	res = thincFlushBufferAll(pThinc);
	if(res == T_ERROR) {
		thincError("ERROR: Couldn't flush the clients' buffers\n");
	}
	
	/* call driver's function */
	thincDbg(T_DBG_ANNOY, ("%s: calling driver's funciton\n", func));
	thinc_unwrap(pThinc, pScreen, BlockHandler);
	(*pScreen->BlockHandler)(i, blockData, pTimeout, pReadmask);
	thinc_wrap(pThinc, pScreen, BlockHandler, thincBlockHandler);

	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));
	instr_add("BLOCK", TI_END);
}

CARD32 thincTimerCallback(OsTimerPtr timer, CARD32 now, pointer arg)
{
	thincScreenPtr pThinc = (thincScreenPtr)arg;
	int err;

	err = thincFlushCmdAll(pThinc);
	if(err == T_ERROR) {
		/* error, but caller is not really expecting one. FlushAll
		 * should take care of killing whichever client caused
		 * this so nothing else for us to do */
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush command "
			"buffer\n", __func__));
		goto notimer;
	} else if(err == T_PART) {
		/* couldn't finish, set timer again */
		goto timer;
	} else {
		/* flush the actual data */
		err = thincFlushBufferAll(pThinc);
		if(err == T_ERROR) {
			/* same as above */
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush data "
				"buffer\n", __func__));
			goto notimer;
		}
		/* all cool, no need for timer */
		goto notimer;
	}

timer:
	/* FIXME: if we do both ping and rate limit, here we should return
	 * the min of the two... */
	if(pThinc->ping_freq) {
		thincSendPingAll(pThinc);
		return pThinc->ping_freq;
	} else {
		/* we're not doing pings but we need to be woken up again */
#ifdef T_RECORD
		return pThinc->rate_ms;
#else
		return T_TIMER;
#endif
	}

notimer:
	if(pThinc->ping_freq) {
		thincSendPingAll(pThinc);
		return pThinc->ping_freq;
	} else
		return 0;
}

/* Shutdown the module */
Bool thincCloseScreen (int index, ScreenPtr pScreen)
{
	thincScreenPtr pThinc = THINCPtr(pScreen);
	thincClientPtr cl, next;
	ScrnInfoPtr xf86Scrn = xf86Screens[pScreen->myNum];
	struct thincScreenStats *s, *ns;

	/**** print stats ****/
	thincStatsPrint(pThinc);

	/**** print instrumentation ****/
	thincInstrPrint(pThinc);

#ifdef T_RECORD
	/* this should be done before closing clients */
	if(pThinc->doRecord) 
		thincRecordDestroy(pThinc);
#endif

	/* uninit thincback stuff */
#if 0
	REGION_UNINIT(pThinc->pScreen, &pThinc->display_change);
#else
	t_list_for_each_entry(s, ns, struct thincScreenStats, &pThinc->stats, 
		list) {
		thincSStatsFreeIt(pThinc, s);
	}
	pThinc->stats_next_id = 0;
#endif


	/* close all clients */
	t_list_for_each_entry(cl, next, thincClientRec, &pThinc->clients, link)
	{
		cl->close_cl(cl);
	}
	t_list_init(&pThinc->clients);

	/* close listening sockets */
	if(tcpThincClose(pThinc->listenSocket) != 0)
		thincError("Error: Couldn't close listening socket\n");

	thincAppshareUninit(pThinc);

	/* Free HW Cursor structure */
	if(pThinc->useHWCursor) {
		THINCHWCursorUninit(pThinc);
	}

#ifdef T_USERENDER
	thincPictureUninit(pScreen);
#endif

#ifdef T_USEXAA
	thincXAAUninit(pThinc);
#endif

	/* Destroy offscreen pixmap structures */
	if(pThinc->useOffscreen) {
		thincDestroyCmdInfo(
			pThinc->cmdHandlerList[thincCmd_idx(T_FB_BILEVEL)]);
		thincDestroyCmdInfo(
			pThinc->cmdHandlerList[thincCmd_idx(T_FB_GLYPH)]);
		thincDestroyCmdInfo(
			pThinc->cmdHandlerList[thincCmd_idx(T_FB_PFILL)]);
		thincDestroyCmdInfo(
			pThinc->cmdHandlerList[thincCmd_idx(T_FB_SFILL)]);
		thincDestroyCmdInfo(
			pThinc->cmdHandlerList[thincCmd_idx(T_FB_RUPDATE)]);
	}

	if(pThinc->compress || pThinc->compressVideo) {
		thinc_compressEnd(pThinc);
	}
	/* unwrap functions */
	thinc_unwrap(pThinc, pScreen, WakeupHandler);
	thinc_unwrap(pThinc, pScreen, BlockHandler);
	thinc_unwrap(pThinc, pScreen, CloseScreen);
#ifdef T_NODRAW
	thinc_unwrap(pThinc, pScreen, SaveScreen);
#endif

	/* for multiple servers */
	thinc_unwrap(pThinc, xf86Scrn, EnableDisableFBAccess);

#ifdef T_APPSHARE
	thinc_unwrap(pThinc, pScreen, CreateWindow);
	thinc_unwrap(pThinc, pScreen, DestroyWindow);
	thinc_unwrap(pThinc, pScreen, ChangeWindowAttributes);
	thinc_unwrap(pThinc, pScreen, RealizeWindow);
	thinc_unwrap(pThinc, pScreen, UnrealizeWindow);
	thinc_unwrap(pThinc, pScreen, PositionWindow);
	thinc_unwrap(pThinc, pScreen, WindowExposures);
#endif
	thinc_unwrap(pThinc, pScreen, PaintWindowBackground);
	thinc_unwrap(pThinc, pScreen, PaintWindowBorder);
	thinc_unwrap(pThinc, pScreen, CopyWindow);
	thinc_unwrap(pThinc, pScreen, CreatePixmap);
	thinc_unwrap(pThinc, pScreen, DestroyPixmap);
	thinc_unwrap(pThinc, pScreen, ModifyPixmapHeader);
	thinc_unwrap(pThinc, pScreen, CreateGC);

	if(pThinc->useAuth) {
		free(pThinc->owner);
		pThinc->owner_len = 0;
	}

	/* free memory */
	free(pThinc);

	return (*pScreen->CloseScreen)(index, pScreen);
}

/* Shared Memory code */
#ifdef T_USEMITSHM
extern void ShmRegisterFuncs(ScreenPtr pScreen, ShmFuncsPtr funcs);

static void thincSHMInit(ScreenPtr pScreen)
{
	ShmRegisterFuncs(pScreen, &thincShmFuncs);
}
#endif /* MIT-SHM */

#ifdef T_USEXAA
/* Initialize XAA. See XAA.HOWTO (../xaa/XAA.HOWTO) for details */
static Bool thincXAAInit(ScreenPtr pScreen)
{
	const char *func = __func__;
	thincScreenPtr pThinc = THINCPtr(pScreen);
	XAAInfoRecPtr xaaInfo;
	int flags;
	/* XXX: For now, we only support GXCOPY. Maybe we could support
	 * ROP_NEEDS_SOURCE? But then, do that on the server and have
	 * the client do gxcopy only, or have the client actually
	 * support those modes? */
	const int base_flags = GXCOPY_ONLY;

	thincDbg(T_DBG_DRAW, ("%s: Creating XAAInfo structure\n", func));

	/*** Create XAA Info structure ***/
	xaaInfo = pThinc->xaaInfo = XAACreateInfoRec();
	if(!xaaInfo) {
		thincDbg(T_DBG_ERROR, ("%s: Couldn't create XAA info struct\n",
			func));
		return FALSE;
	}

	/*** Fill the XAA Info struct with our capabilities ***/
	thincDbg(T_DBG_DRAW, ("%s: Filling in XAAInfo struct\n", func));

	xaaInfo->Sync = thincSync;

	/* Things we definitely don't support:
	 * => arbitrary solid lines
	 * => arbitrary dashed lines
	 */

	/* TODO: We don't need to provide all of this, since some of them
	 * will just fall back to our default GC/Window funcs (which is
	 * what we want). Chk which are really needed and delete the
	 * rest */

	/* Screen to Screen Copies */
	/* XXX: NOT NEEDED */
#if 1
	flags = base_flags;
	xaaInfo->SetupForScreenToScreenCopy = thincSetupForScreenToScreenCopy;
	xaaInfo->ScreenToScreenCopyFlags = flags;
	xaaInfo->SubsequentScreenToScreenCopy = 
		thincSubsequentScreenToScreenCopy;
#endif

	/* Solid Fills */
	/* XXX: NOT NEEDED */
#if 1
	xaaInfo->SetupForSolidFill = thincSetupForSolidFill;
	xaaInfo->SubsequentSolidFillRect = thincSubsequentSolidFillRect;
	xaaInfo->SolidFillFlags = flags;
#endif
	
	/* Color Expansion Fills */
	/* XXX: NOT NEEDED. Used for bilevel text */
#if 0
	/* For now we use bitmaps with least significant bit in rightmost
	 * position */
	flags = base_flags;
	flags |= BIT_ORDER_IN_BYTE_MSBFIRST;

	xaaInfo->SetupForScreenToScreenColorExpandFill = 
		thincSetupForScreenToScreenColorExpandFill;
	xaaInfo->SubsequentScreenToScreenColorExpandFill = 
		thincSubsequentScreenToScreenColorExpandFill;
	xaaInfo->ScreenToScreenColorExpandFillFlags = flags;

	/* TODO: support LEFT_EDGE_CLIPPING_NEGATIVE_X ? */

	/* CPUToScreen Expansion Fills:
	 * Direct -> need to create memory space so that XAA can write
	 * the bitmap there (pThinc->xaaAperture). Max size should be
	 * max size of bitmap we have defined
	 * Indirect: Not supported
	 */
	pThinc->xaaApertureSize = T_MAX_BITMAP_SIZE;
	pThinc->xaaAperture = malloc(pThinc->xaaApertureSize);
	if(!pThinc->xaaAperture) {
		thincDbg(T_DBG_ERROR, ("%s: ERROR: Couldn't allocate memory "
			"for XAA aperture\n", func));
		return FALSE;
	}
	memset(pThinc->xaaAperture, 1, 8);
	xaaInfo->ColorExpandBase = pThinc->xaaAperture;
	xaaInfo->ColorExpandRange = pThinc->xaaApertureSize;
	flags |= CPU_TRANSFER_PAD_DWORD; /* padding of the aperture */ 
	flags |= SCANLINE_PAD_DWORD; /* padding of each scanline */
	flags |= CPU_TRANSFER_BASE_FIXED; /* since we're cheating and this
					     is not a real apperture :) */
	flags |= SYNC_AFTER_COLOR_EXPAND;

	xaaInfo->SetupForCPUToScreenColorExpandFill =
		thincSetupForCPUToScreenColorExpandFill;
	xaaInfo->SubsequentCPUToScreenColorExpandFill =
		thincSubsequentCPUToScreenColorExpandFill;
	xaaInfo->CPUToScreenColorExpandFillFlags = flags;
#endif

	/* 8x8 Mono Pattern Fills */
	/* XXX: NOT NEEDED. What are they used for? */
#if 1
	flags = base_flags;

	flags |= BIT_ORDER_IN_BYTE_MSBFIRST;
	/* for now asking XAA to pass us the DWORDS while we figure out
	 * how to enable the pixmap cache */
	flags |= HARDWARE_PATTERN_PROGRAMMED_BITS;
	flags |= HARDWARE_PATTERN_SCREEN_ORIGIN;

	xaaInfo->SetupForMono8x8PatternFill = thincSetupForMono8x8PatternFill;
	xaaInfo->SubsequentMono8x8PatternFillRect =
		thincSubsequentMono8x8PatternFillRect;
#endif
	
	/* 8x8 Color Pattern Fills */
#if 1
	flags = base_flags;
	flags |= HARDWARE_PATTERN_SCREEN_ORIGIN;
	/*
	flags |= NO_TRANSPARENCY;
	 */

	xaaInfo->SetupForColor8x8PatternFill = 
		thincSetupForColor8x8PatternFill;
	xaaInfo->SubsequentColor8x8PatternFillRect =
		thincSubsequentColor8x8PatternFillRect;
	xaaInfo->Color8x8PatternFillFlags = flags;
#endif

	/* Image Writes. Use direct mode */
#if 1
	flags = base_flags;
	/*
	flags |= NO_TRANSPARENCY;
	 */

	pThinc->xaaApertureSize = 4*1024*1024;
	pThinc->xaaAperture = malloc(pThinc->xaaApertureSize);
	if(!pThinc->xaaAperture) {
		thincDbg(T_DBG_ERROR, ("%s: ERROR: Couldn't allocate memory "
			"for XAA aperture\n", func));
		return FALSE;
	}
	xaaInfo->ImageWriteBase = pThinc->xaaAperture;
	xaaInfo->ImageWriteRange = pThinc->xaaApertureSize;

	flags |=  CPU_TRANSFER_PAD_DWORD;
	flags |= SCANLINE_PAD_DWORD;
	flags |= CPU_TRANSFER_BASE_FIXED;
	flags |= SYNC_AFTER_IMAGE_WRITE;

	xaaInfo->ImageWriteFlags = flags;
	xaaInfo->SetupForImageWrite = thincSetupForImageWrite;
	xaaInfo->SubsequentImageWriteRect = thincSubsequentImageWriteRect;
#endif

	/*** Initialize XAA ***/
	thincDbg(T_DBG_DRAW, ("%s: Initializing XAA\n", func));
	if(!XAAInit(pScreen, xaaInfo)) {
		thincDbg(T_DBG_ERROR, ("%s: ERROR: XAA Initialization call "
			"failed\n", func));
		return FALSE;
	}

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	return TRUE;
}

/* Call from CloseScreen() */
static void thincXAAUninit(thincScreenPtr pThinc)
{
	XAADestroyInfoRec(pThinc->xaaInfo);
#if 0
	free(pThinc->xaaAperture);
	pThinc->xaaApertureSize = -1;
#endif
}
#endif /* T_USEXAA */

/* for multiple servers */
static void thincEnableDisableFBAccess (int scrnIndex, Bool enable)
{
	t_dbg_name(func)
        ScrnInfoPtr xf86Scrn = xf86Screens[scrnIndex];
	thincScreenPtr pThinc = THINCPtr(xf86Scrn->pScreen);

	thincDbg(T_DBG_BASE, ("%s: entering\n", func));

	/* this used to work in xfree86, but with x.org the hack is no
	 * longer useful. we either let all calls pass or none, as the
	 * server will segfault otherwise. for now, let all calls
	 * pass as this doesn't seem to affect multiple servers running */
#ifdef XORG_VERSION_CURRENT
	thinc_unwrap(pThinc, xf86Scrn, EnableDisableFBAccess);
	(*xf86Scrn->EnableDisableFBAccess)(scrnIndex, enable);
	thinc_wrap(pThinc, xf86Scrn, EnableDisableFBAccess,
		thincEnableDisableFBAccess);
#else	
	/* XFree86 (last tested with 4.3.0) */
	/* only pass through calls to enable framebuffer access */
	if (enable == TRUE) {
		thinc_unwrap(pThinc, xf86Scrn, EnableDisableFBAccess);
		(*xf86Scrn->EnableDisableFBAccess)(scrnIndex, enable);
		thinc_wrap(pThinc, xf86Scrn, EnableDisableFBAccess,
			thincEnableDisableFBAccess);
	}
#endif

	thincDbg(T_DBG_BASE, ("%s: finished\n", func));
}

/* this is a hack but it works. we basically get rid of all the x crap
 * that would make us very unstable when running multiple servers (if one
 * exits, it'd bring down another with it, etc). we just replace the
 * normal X behavior with this simple function. 
 * The way things work:
 * when a new server starts up it signals other running servers (SIGUSR1).
 * it then asks the kernel to give it a vt and waits until the vt is given
 * to it. so we, as any ugly nasty person would do, replace the sigusr 
 * handler so that X doesn't try to do all its funky stuff, and just tell
 * the kernel "sure, take away our vt. al cabo q ni la queria" (that's the
 * ioctl call). seems to work just fine (TM) */
static void thincSignal(int sig)
{
	ioctl(xf86Info.consoleFd, VT_RELDISP, (void*)1);
	signal(SIGUSR1, thincSignal);
}

int thincSetDebugLevel(thincScreenPtr pThinc, int debugLevel)
{
#ifdef T_DO_DEBUG
	if(debugLevel < T_DBG_MIN || debugLevel > T_DBG_MAX) {
		thincError("%d not good\n", debugLevel);
		return T_ERROR;
	}

	thincDebugLevel = (unsigned short)debugLevel;

	thincMsg("Changed debug level to %#x\n", thincDebugLevel);
#endif
	/* silently ignore if not doing debug */
	return T_OK;
}

int thincGetDebugLevel(thincScreenPtr pThinc)
{
#ifdef T_DO_DEBUG
	return (int)thincDebugLevel;
#else
	return 0;
#endif
}

#ifdef T_NODRAW
static Bool thincSaveScreen(ScreenPtr pScreen, int mode)
{
	t_dbg_name(func)
	Bool res;
	thincScreenPtr pThinc = THINCPtr(pScreen);
		
	thincDbg(T_DBG_ANNOY, ("%s: mode %d\n", func, mode));

	if(thincClient0Purge(pThinc) == T_ERROR) {
		T_BUG("Couldn't purge client0's buffer. we may be screwed");
		return FALSE;
	}

	thinc_unwrap(pThinc, pScreen, SaveScreen);
	res = pScreen->SaveScreen(pScreen, mode);
	thinc_wrap(pThinc, pScreen, SaveScreen, thincSaveScreen);

	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));

	return res;
}
#endif

#define t_def_inc_get_and_reset(name, value) \
inline void thincInc##name(thincScreenPtr pThinc)	\
{							\
	struct thincScreenStats *s;			\
	t_list_for_each_entry_fast(s, struct thincScreenStats, &pThinc->stats, \
		list) { 				\
		s->value++;				\
	}						\
}							\
\
inline int thincGet##name(thincScreenPtr pThinc, int id)	\
{								\
	struct thincScreenStats *s = thincSStatsFind(pThinc, id);	\
	return s ? s->value : T_ERROR;				\
}								\
\
inline int thincReset##name(thincScreenPtr pThinc, int id)	\
{								\
	struct thincScreenStats *s = thincSStatsFind(pThinc, id);	\
	if(s) {							\
		s->value = 0;					\
		return T_OK;					\
	} else {						\
		return T_ERROR;					\
	}							\
}

t_def_inc_get_and_reset(MouseDown, mouse_down)
t_def_inc_get_and_reset(MouseUp, mouse_up)
t_def_inc_get_and_reset(KeybDown, keyb_down)
t_def_inc_get_and_reset(KeybUp, keyb_up)
t_def_inc_get_and_reset(DisplayCount, display_count)

int thincGetScreensaver(thincScreenPtr pThinc)
{
	/* TODO: Need to figure out how to get this info */
	return pThinc->screensaver;
}

#if 0
int thincGetDisplayPercent(thincScreenPtr pThinc, RegionPtr reg)
{
	int perc, area = 0, numrects, i;
	BoxPtr c;
	float p;

	c = REGION_RECTS(reg);
	numrects = REGION_NUM_RECTS(reg);

	for(i=0; i<numrects; ++i, ++c) {
		area += (box_width(c) * box_height(c));
	}

	p = area/pThinc->total_area;
	perc = p*100;

	return perc;
}

int thincResetDisplayPercent(thincScreenPtr pThinc, RegionPtr reg)
{
	/* now empty the region */
	REGION_UNINIT(pThinc->pScreen, reg);
	REGION_INIT(pThinc->pScreen, reg, NullBox, 1);

	return T_OK;
}
#else
void thincAddDisplayChange(thincScreenPtr pThinc, RegionPtr reg)
{
	struct thincScreenStats *s;			
	t_list_for_each_entry_fast(s, struct thincScreenStats, &pThinc->stats, 
		list) { 				
		REGION_UNION(pThinc->pScreen, &s->display_change,
			&s->display_change, reg);
	}						
}

int thincGetDisplayPercent(thincScreenPtr pThinc, int id)
{
	int perc, area = 0, numrects, i;
	BoxPtr c;
	float p;
	RegionPtr reg;
	struct thincScreenStats *s = thincSStatsFind(pThinc, id);

	if(!s)
		return T_ERROR;

	reg = &s->display_change;

	c = REGION_RECTS(reg);
	numrects = REGION_NUM_RECTS(reg);

	for(i=0; i<numrects; ++i, ++c) {
		area += (box_width(c) * box_height(c));
	}

	p = area/pThinc->total_area;
	perc = p*100;

	return perc;
}

int thincResetDisplayPercent(thincScreenPtr pThinc, int id)
{
	struct thincScreenStats *s = thincSStatsFind(pThinc, id);
	if(!s)
		return T_ERROR;

	/* now empty the region */
	REGION_UNINIT(pThinc->pScreen, &s->display_change);
	REGION_INIT(pThinc->pScreen, &s->display_change, NullBox, 1);

	return T_OK;
}
#endif

int thincSStatsNew(thincScreenPtr pThinc)
{
	struct thincScreenStats *s;

	s = malloc(sizeof(*s));
	if(!s)
		return T_ERROR;

	t_list_init(&s->list);

	s->id = pThinc->stats_next_id++;
	/* TODO :) */
	t_bug_on_2(pThinc->stats_next_id < 0,
		T_BUG("stats identifier wrapped around, fix me!"),
		return T_ERROR
	);

	s->mouse_down = 0;
	s->mouse_up = 0;
	s->keyb_down = 0;
	s->keyb_up = 0;
	s->display_count = 0;
	REGION_INIT(pThinc->pScreen, &s->display_change, NullBox, 1);

	t_list_add_head(&s->list, &pThinc->stats);


	return s->id;
}

inline struct thincScreenStats* thincSStatsFind(thincScreenPtr pThinc, 
	int stats_id)
{
	struct thincScreenStats *s = NULL;

	t_list_for_each_entry_fast(s, struct thincScreenStats, &pThinc->stats, 
		list) 
	{
		if(s->id == stats_id)
			break;
	}

	return s;
}

inline void thincSStatsFree(thincScreenPtr pThinc, int stats_id)
{
	struct thincScreenStats *s = thincSStatsFind(pThinc, stats_id);

	if(!s)
		return;
	thincSStatsFreeIt(pThinc, s);
}

inline void thincSStatsFreeIt(thincScreenPtr pThinc, struct thincScreenStats *s)
{
	t_list_del_el(&s->list);
	REGION_UNINIT(pThinc->pScreen, &s->display_change);

	free(s);
}
