/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for application sharing
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/*
 * we don't share windows. instead we share whole applications (x clients).
 * so instead of marking each particular window as shared we mark the
 * application to which it belongs as shared. 
 * then, when drawing happens on a window, we get its corresponding app and
 * check if it's shared or not. if it's not shared, we just ignore the 
 * draw request.
 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#ifdef T_APPSHARE

static thincAppPtr thincAppGet(thincScreenPtr pThinc, WindowPtr pWin);

int thincAppshareInit(thincScreenPtr pThinc) 
{
	t_dbg_name(func)
	register thincAppPtr curr;
	register unsigned int i;

	thincDbg(T_DBG_WIN, ("%s: have room for %d apps\n", func, T_APP_MAX));

	pThinc->apps = malloc(T_APP_MAX * sizeof(*pThinc->apps));
	if(!pThinc->apps) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Can't allocate memory "
			"for %d apps\n", func, T_APP_MAX));
		return T_ERROR;
	}

	curr = pThinc->apps;
	/* app 0 is desktop */
	curr->id = T_APP_ID_DESKTOP;
	curr->flags = 0;
	curr->count = 0;
	t_list_init(&curr->windows);
	curr->name = strdup("DESKTOP");
	curr->name_len = strlen(curr->name);
	++curr;
	
	for(i=1; i<T_APP_MAX; ++i, ++curr) {
		curr->id = i;
		curr->flags = 0;
		curr->count = 0;
		t_list_init(&curr->windows);
		curr->name = NULL;
		curr->name_len = 0;
	}
	pThinc->app_count = 0;

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));
	return T_OK;
}

void thincAppshareUninit(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincAppPtr a;
	unsigned int i;
	thincWindowPtr prev, w;

	thincDbg(T_DBG_WIN, ("%s\n", func));

	a = pThinc->apps;
	for(i=0; i<T_APP_MAX; ++i, ++a) {
		t_list_for_each_entry(w, prev, thincWindowRec,&a->windows,list) 
		{
			t_list_del_el(&w->list);
			/* in case the window gets destroyed after us,
			 * make sure it doesn't think it's still part of
			 * some app */
			w->app = NULL;
		}
		t_list_init(&a->windows);
		a->count = 0;
		a->flags = 0;
	}

	free(pThinc->apps);

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));
}

#define t_app_get(app, pThinc, pWin) do {	\
	app = thincAppGet(pThinc, pWin);	\
	t_bug_on_2(				\
		!app,				\
		T_BUG_ARG("Window %#lx has no valid application\n",\
			t_winXID(pWin)),	\
		return T_ERROR			\
	);					\
} while(0)

static thincAppPtr thincAppGet(thincScreenPtr pThinc, WindowPtr pWin)
{
	t_bug_on_2(
		t_window_app(pWin) >= T_APP_MAX,
		T_BUG_ARG_2("Window has 'impossible' X client id %d."
			"I only know up to %d\n", t_window_app(pWin), 
			T_APP_MAX),
		return NULL
	);
	
	return &pThinc->apps[t_window_app(pWin)];
}

/* start sharing the application to which pWin belongs */
int thincWindowShare(thincScreenPtr pThinc, WindowPtr pWin)
{
	t_dbg_name(func)
	thincAppPtr app;

	thincDbg(T_DBG_WIN, ("%s: sharing window %#lx\n", func, 
		t_winXID(pWin)));

	t_app_get(app, pThinc, pWin);

	app->flags |= T_APP_SHARED;

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));
	return T_OK;
}

/* checks if a window is shared or not */
int thincWindowIsShared(thincScreenPtr pThinc, WindowPtr pWin)
{
	thincAppPtr app;

	t_app_get(app, pThinc, pWin);

	if(app->flags & T_APP_SHARED)
		return T_YES;
	else
		return T_NO;
}

/* make sure window's linked list and app member have been initialized
 * before calling us */
int thincAppWindowAdd(WindowPtr pWindow)
{
	t_dbg_name(func)
	thincAppPtr a;
	thincWindowPtr pPriv;
	thincScreenPtr pThinc = THINCPtr(pWindow->drawable.pScreen);

	thincDbg(T_DBG_WIN, ("%s: adding window %#lx to application %d\n",
		func, t_winXID(pWindow), t_window_app(pWindow)));

	pPriv = THINCWINPtr(pWindow);
	t_bug_on_2(
		(!t_list_is_empty(&pPriv->list)) || pPriv->app,
		T_BUG_ARG_2("Window %#lx (%#lx) already belongs to another "
			"app\n", t_winXID(pWindow), t_winID(pWindow)),
		return T_ERROR
	);

	t_app_get(a, pThinc, pWindow);
	t_list_add_tail(&pPriv->list, &a->windows);
	pPriv->app = a;
	a->count++;
	if(a->count == 1) {
		a->flags |= T_APP_ACTIVE;
		pThinc->app_count++;
		thincDbg(T_DBG_WIN, ("%s: activated app %#x: %s, now have "
			"%d apps\n", func, a->id, a->name, 
			pThinc->app_count));
	}

	thincDbg(T_DBG_WIN, ("%s: finished, app has %d windows\n", func,
		a->count));
	return T_OK;
}

void thincAppWindowRemove(WindowPtr pWindow)
{
	t_dbg_name(func)
	thincAppPtr app;
	thincWindowPtr pPriv = THINCWINPtr(pWindow);
	thincScreenPtr pThinc = THINCPtr(pWindow->drawable.pScreen);

	thincDbg(T_DBG_WIN, ("%s: removing window %#lx from application %d\n",
		func, t_winXID(pWindow), t_window_app(pWindow)));

	t_list_del_el(&pPriv->list);

	if(pPriv->app) {
		app = pPriv->app;
		pPriv->app = NULL;
		app->count--;
		/* if there are no more windows, unmark this app as shared */
		if(!app->count) {
			thincDbg(T_DBG_WIN, ("%s: no windows left, assuming "
				"app %#x (%s) is gone, unsharing and "
				"unactivating\n", func, app->id, app->name));
			app->flags &= ~T_APP_SHARED;
			app->flags &= ~T_APP_ACTIVE;
			pThinc->app_count--;
		}
		thincDbg(T_DBG_WIN, ("%s: finished, app has %d windows\n", 
			func, app->count));
	} else {
		thincDbg(T_DBG_WIN, ("%s: finished\n", func));
	}
}

#if 0
int thincWindowShareSimple(thincScreenPtr pThinc, WindowPtr pWin)
{
	thincWindowPtr pPriv = THINCWINPtr(pWin);
	pPriv->flags |= T_SHARE;
	return T_OK;
}

/* finds all top level window's coming from same client as pWin and marks them 
 * as shared */
int thincWindowShareAll(thincScreenPtr pThinc, WindowPtr pWin)
{
	t_dbg_name(func)
	unsigned long client_id = t_app_id(pWin->drawable.id);
	register WindowPtr curr, root = thincGetRootWindow(pThinc);

	thincDbg(T_DBG_WIN, ("%s: window's id: %#lx, client_id %#lx\n",
		func, pWin->drawable.id, client_id));

	curr = root->firstChild;
	while(curr != root->lastChild) {
		thincDbg(T_DBG_WIN, ("%s: checking window %#lx\n", func,
			curr->drawable.id));
		if(t_app_id(curr->drawable.id) == client_id) {
			thincDbg(T_DBG_WIN, ("%s: found shareable window "
				"%#lx\n", func, curr->drawable.id));
			thincWindowShareSimple(pThinc, curr);
		}
		curr = curr->nextSib;
	}
	thincWindowShareSimple(pThinc, pWin);

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));
	return T_OK;
}

int thincWindowIsShared(WindowPtr pWin)
{
	thincWindowPtr pPriv;
	
	while(pWin) {
		pPriv = THINCWINPtr(pWin);
		if(pPriv->flags & T_SHARE) 
			return T_TRUE;
		pWin = pWin->parent;
	}

	return T_FALSE;
}

/* finds window that's closest to the root, it's a parent of pWin and it 
 * belongs to the same x client (ie. not a window manager window) */
static WindowPtr thincWindowFindTop(WindowPtr pWin)
{
	WindowPtr root = thincGetRootWindow(pThinc);
	WindowPtr curr = pWin;
	int id = t_window_app(pWin);

	while(1) {
		if(curr->parent == root)
			return curr;
		else if(t_window_app(curr->parent) != id)
			return curr;
		curr = curr->parent;
	}

	return NULL;
}
#endif

#endif /* T_APPSHARE */
