/* Copyright 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for Glyph and Bilevel messages
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/*
 * Flush a GLYPH/BILEVEL command:
 * Since we define a maximum size for bitmaps, then the command needs to
 * be broken up before we can execute the stack. The disadvantage here is 
 * that some processing is done that's not accounted for in the statistics 
 * (I don't want to break up statistics collecting to multiple places).
 * We use the same message stack for GLYPHs and BILEVELs and internally
 * figure out which one to use. 
 * Code here is (structurally speaking) very similar to raw message
 * processing.
 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include <netinet/in.h>

/* needed for WriteFile module */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* the methods */
#ifdef T_GETSTATS
static int thinc_bitmapPreStats(thincClientPtr cl, void *msg);
static int thinc_bitmapPostStats(thincClientPtr cl, void *msg);
#endif
static int thinc_bitmapSend(thincClientPtr cl, void *msg);
static int thinc_bitmapWriteFile(thincClientPtr cl, void *msg);
static int thinc_bitmapChkCache(thincClientPtr cl, void *msg);
static int thinc_bitmapAddCache(thincClientPtr cl, void *msg);
#ifdef T_RECORD
static int thinc_bitmapRecord(thincClientPtr cl, void *msg);
static int thinc_glyphRecordCb(thincClientPtr cl, char *buf, unsigned int *len,
	void *ptr);
static int thinc_bilevelRecordCb(thincClientPtr cl, char *buf, 
	unsigned int *len, void *ptr);
#endif

/* the priorities. current order:
 * PRESTATS
 * CHKCACHE
 * SEND
 * ADDCACHE
 * POSTSTATS
 */
#ifdef T_GETSTATS
#define T_BITPRIO_PRESTATS	T_PRIO_MIN
#define T_BITPRIO_POSTSTATS	T_PRIO_MAX
#endif
#define T_BITPRIO_SEND		T_PRIO_SEND
#define T_BITPRIO_COMPRESS	T_BITPRIO_SEND-10
#define T_BITPRIO_CHKCACHE	T_BITPRIO_SEND-50
#define T_BITPRIO_ADDCACHE	T_BITPRIO_SEND+10
/* change to modify where in the pipeline you want to save the data */
#define T_BITPRIO_WRITEFILE	T_BITPRIO_SEND-60
#ifdef T_RECORD
#define T_BITPRIO_RECORD	T_BITPRIO_SEND+1
#endif

/* the static modules. From here, we make copies to each client's stack */
#ifdef T_GETSTATS
thincDefModule(bitmapPreStats, "BIT_PRESTATS", 13, T_BITPRIO_PRESTATS);
thincDefModule(bitmapPostStats, "BIT_POSTSTATS", 14, T_BITPRIO_POSTSTATS);
#endif
thincDefModule(bitmapSend, "BIT_SEND", 9, T_BITPRIO_SEND);
thincDefModule(bitmapWriteFile, "BIT_WRITEFILE", 14, T_BITPRIO_WRITEFILE);
thincDefModule(bitmapChkCache, "BIT_CHKCACHE", 13, T_BITPRIO_CHKCACHE);
thincDefModule(bitmapAddCache, "BIT_ADDCACHE", 13, T_BITPRIO_ADDCACHE);
#ifdef T_RECORD
thincDefModule(bitmapRecord, "BITMAP_RECORD", 13, T_BITPRIO_RECORD);
#endif

/* local variables */
#ifdef T_GETSTATS
static unsigned long t_start;	/* needed for statistics */
#endif

#define T_BIT_EXTRALEN	32

/* NOTES ON BITMAP OBJECT
 * Modules cannot set the data pointer to memory to which there are no other 
 * references, since modules will blindly replace that pointer as needed. DON'T
 * ASSUME THAT YOU CAN CREATE A POST_* MODULE AND FREE THE POINTER THERE.
 * You may end up freeing another pointer and cause a nuclear meltdown!
 */
typedef struct thincBitObj_S {
	unsigned int x:16;
	unsigned int y:16;
	unsigned int width:16;
	unsigned int height:16;
	unsigned int fg:32;
	unsigned int bg:32;

	/* information for write and stats. Basically how much total data
	 * was sent as a result of this bitmap object */
	unsigned int total_len:32;

	/* the bitmap data. Always stored contiguously */
	char *data;
	char *orig_data;
	unsigned int data_len:32;
	unsigned int data_width:32;	/* width of one line of the bitmap 
					   (bitmap_width(width)) */

	/* the rectangles to fill. may be NULL */
	BoxPtr rects;
	unsigned int numrects:16;

	/* extra headers for the message */
	unsigned int flags:8;		/* message flags */
	unsigned int extra_len:8;	/* current length <= T_BIT_EXTRALEN */
	char *extra;		/* current end */
	char extra_data[T_BIT_EXTRALEN];

	unsigned int hash:32;
	unsigned int which:8;
} thincBitObj;

/***** local functions ******/
static int thincGBLFlushPerRect(thincScreenPtr pThinc, thincClientPtr cl,
	int which, char *bits, BoxPtr bits_area, BoxPtr rects,
	unsigned int numrects, unsigned long fg, unsigned long bg);
static int thincGBLFlushFill(thincScreenPtr pThinc, thincClientPtr cl,
	int which, char *bits, BoxPtr bits_area, BoxPtr rects,
	unsigned int numrects, unsigned long fg, unsigned long bg);
static int thincSendGBLRect(thincClientPtr cl, int which, char *bits, 
	BoxPtr bits_area, BoxPtr rect, unsigned long fg, unsigned long bg);
static int thincSendGBLRectSmall(thincClientPtr cl, int which, 
	char *bits, BoxPtr bits_area, BoxPtr rect, unsigned long fg, 
	unsigned long bg);
static int thincSendGBLRectBig(thincClientPtr cl, int which, 
	char *bits, BoxPtr bits_area, BoxPtr rect, int rect_size, 
	unsigned long fg, unsigned long bg);
static int thincGetBitmap(char *bits, BoxPtr bits_area, BoxPtr rect,
	char *bitmap);
static int thincSendGBLRectSimple(thincClientPtr cl, int which,
	char *bits, BoxPtr bits_area, BoxPtr rect, 
	unsigned long fg, unsigned long bg);
static int thincSendGBLRectUnaligned(thincClientPtr cl, int which, 
	char *bits, BoxPtr bits_area, BoxPtr rect, int align, 
	unsigned long fg, unsigned long bg);
static int thincGBLExecStack(thincClientPtr cl, int which, 
	char *bits, BoxPtr rect, unsigned long fg, unsigned long bg);
static int thincBitmapSendRects(thincClientPtr cl, char *buffer,
	BoxPtr rects, int numrects);

/*****  module helper functions ****/
static int thinc_bilevelSend(thincClientPtr cl, thincBitObj *obj);
static int thinc_glyphSend(thincClientPtr cl, thincBitObj *obj);

/**** useful macros ****/
/* checks that we haven't overflowed the extra data buffer. Shouldn't happen
 * as long as extra messages and the buffer size are kept in sync. Should
 * be called every time something is added to the extra buffer */
#define checkExtraBuffer(obj) \
	if((obj)->extra_len > T_BIT_EXTRALEN) { \
		thincError("BUG: %s: %d: %s(): BIT update extra headers " \
			"overflowed static data buffer in BITMAP message " \
			"object! Current lenght is %u but buffer's size is " \
			"%u\n", __FILE__, __LINE__, modname, obj->extra_len, \
			T_BIT_EXTRALEN); \
		return T_ERROR; \
	}

#define bitmap_addmodule(cl, mod) { \
	err = thinc_msgAddUsingModule((cl), (cl)->bitmapStack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add BITMAP module"\
			" %s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}

/* initializes and adds appropriate modules to the BITMAP message stack */
int thinc_bitmapInitStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;
	
	thincDbg(T_DBG_CLIENT, ("%s: creating BITMAP message stack\n", func));
	
	cl->bitmapStack = thinc_msgStackCreate(cl);
	if(!cl->bitmapStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create BITMAP "
			"message stack\n", func));
		return T_ERROR;
	}

	/* add the modules */
#ifdef T_GETSTATS
	bitmap_addmodule(cl, &bitmapPreStats);
	bitmap_addmodule(cl, &bitmapPostStats);
#endif
	bitmap_addmodule(cl, &bitmapSend);

	if(cl->parent->useBitCache) {
		/* cache needs the data contiguously */
		bitmap_addmodule(cl, &bitmapChkCache);
		bitmap_addmodule(cl, &bitmapAddCache);
	}

	/* only for debugging */
#if 0
	bitmap_addmodule(cl, &bitmapWriteFile);
#endif	

#if 0
#ifdef T_RECORD
	/* only for first client */
	if(cl->parent->doRecord) {
		thincScreenPtr pThinc = cl->parent;
		thincRecordPtr record = t_recordPtr(pThinc);
		/* this mess because when we get called, clientCount hasn't
		 * been set */
		if( 
		    ((record->cl_enabled) && (pThinc->clientCount == 1) ) ||
		    (pThinc->clientCount == 0)
		) {
			bitmap_addmodule(cl, &bitmapRecord);
		}
	}
#endif
#endif

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

void thinc_bitmapDestroyStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->bitmapStack);
	cl->bitmapStack = NULL;
}

/************ Flush BITMAP command ********/
int thincGlyphFlush(thincScreenPtr pThinc, thincClientPtr cl, 
	thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct glyphCmd *glyph = cmd_data(cmd, struct glyphCmd*);
	char *bits;
	BoxPtr rects, area;
	RegionPtr reg;
	int err;
	unsigned int numrects;

	thincDbg(T_DBG_PROTO, ("%s: flushing command %#lx\n", func, cmd->id));

	reg = &glyph->reg;
	numrects = REGION_NUM_RECTS(reg);
	rects = REGION_RECTS(reg);
	bits = (char*) (glyph+1);
	if(cl->parent->allowResize && (cl->resized == T_CL_RESZ_LESS) ) {
		/* ugh, need to punt to a RAW */
		thincCmdPtr cmd;
		RegionRec r;
		area = &glyph->orig_area;
		REGION_INIT(pThinc->pScreen, &r, area, 1);
		cmd = thincRawCreateV(pThinc, &r, T_RAW_VA_NONE);
		if(!cmd) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create RAW "
				"command for resized GLYPH update\n", func));
			return T_ERROR;
		}
		REGION_UNINIT(pThinc->pScreen, &r);
		err = thincRawFlush(pThinc, cl, cmd);
		thincRawDestroy(pThinc, cmd);
	} else {
		if(cmd->flags & T_BIT_FILL) {
			BoxRec b;
			coords_to_box(b, glyph->x, glyph->y, glyph->w,
				glyph->h);
			area = &b;
			err = thincGBLFlushFill(pThinc, cl, T_FB_GLYPH, bits,
				area, rects, numrects, glyph->color, 0);
		} else {
			area = &glyph->orig_area;
			err = thincGBLFlushPerRect(pThinc, cl, T_FB_GLYPH, bits, 
				area, rects, numrects, glyph->color, 0);
		}
	}
	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return err;
}

int thincBilevelFlush(thincScreenPtr pThinc, thincClientPtr cl, 
	thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct bilevelCmd *bilevel = cmd_data(cmd, struct bilevelCmd*);
	char *bits;
	BoxPtr rects, area;
	RegionPtr reg;
	unsigned int numrects;
	int err;

	thincDbg(T_DBG_PROTO, ("%s: flushing command %#lx\n", func, cmd->id));
	reg = &bilevel->reg;
	numrects = REGION_NUM_RECTS(reg);
	rects = REGION_RECTS(reg);
	bits = (char*) (bilevel+1);

	if(cmd->flags & T_BIT_FILL) {
		BoxRec b;
		coords_to_box(b, bilevel->x, bilevel->y, bilevel->w,
			bilevel->h);
		area = &b;
		thincDbg(T_DBG_PROTO, ("%s: flushing simple bitmap "
			"command\n", func));
		err = thincGBLFlushFill(pThinc, cl, T_FB_BILEVEL, bits,
			area, rects, numrects, bilevel->fg, bilevel->bg);
	} else {
		thincDbg(T_DBG_PROTO, ("%s: flushing bitmap fill command "
			"have %u rects\n", func, numrects));
		area = &bilevel->orig_area;
		err = thincGBLFlushPerRect(pThinc, cl, T_FB_BILEVEL, bits, 
			area, rects, numrects, bilevel->fg, bilevel->bg);
	}
	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return err;
}

static int thincGBLFlushPerRect(thincScreenPtr pThinc, thincClientPtr cl,
	int which, char *bits, BoxPtr bits_area, BoxPtr rects,
	unsigned int numrects, unsigned long fg, unsigned long bg)
{
	t_dbg_name(func)
	register unsigned int i;
	BoxPtr rect = rects;
	int err;

	thincDbg(T_DBG_PROTO, ("%s: have %d rects to fill\n", func,
		numrects));

	for(i=0; i<numrects; ++i, ++rect) {
		err = thincSendGBLRect(cl, which, bits, bits_area, rect, 
			fg, bg);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush "
				"BITMAP command\n", func));
			return T_ERROR;
		}
	}
	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;
}

/* bits covers the whole area of bits_area. */
static int thincSendGBLRect(thincClientPtr cl, int which, char *bits, 
	BoxPtr bits_area, BoxPtr rect, unsigned long fg, unsigned long bg)
{
	t_dbg_name(func)
	int bsize = box_bitmap_size(rect), err;

	thincDbg(T_DBG_PROTO, ("%s: width %d, height %d, size of rectangle's "
		"bitmap: %d bytes\n", func, box_width(rect), 
		box_height(rect), bsize));
	print_box(T_DBG_PROTO, bits_area, "area of bitmap:");
	print_box(T_DBG_PROTO, rect, "area of rect:");

	if(bsize <= T_MAX_BITMAP_SIZE)
		err = thincSendGBLRectSmall(cl, which, bits, bits_area, rect, 
			fg, bg);
	else
		/* need to divide rectangle in appropiately sized pieces */
		err = thincSendGBLRectBig(cl, which, bits, bits_area, rect, 
			bsize, fg, bg);
	if(err != T_OK) {
		thincDbg(T_DBG_PROTO, ("ERROR: %s: Couldn't send "
			"rectangle GLYPH/BILEVEL update: %d\n", func, 
			err));
		return T_ERROR;
	}

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;
}

static int thincSendGBLRectSmall(thincClientPtr cl, int which, 
	char *bits, BoxPtr bits_area, BoxPtr rect, 
	unsigned long fg, unsigned long bg)
{
	t_dbg_name(func)
	int x_pos, align, err, x_off;

	thincDbg(T_DBG_PROTO, ("%s: sending small rectangle GLYPH/BILEVEL "
		"update\n", func));
	
	x_off = bits_area->x1;
	x_pos = rect->x1 - x_off;

	/* check if rectangle is on byte boundary */
	align = x_pos % T_PIXELS_PER_BYTE;
	if(align == 0) {
		/* yes! */
		thincDbg(T_DBG_PROTO, ("%s: rectangle is aligned "
			"properly and has right size, done!\n", func));
		err = thincSendGBLRectSimple(cl, which, bits, bits_area, 
			rect, fg, bg);
		if(err != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"send Bilevel message: %d\n", 
				func, err));
			return T_ERROR;
		}
	} else {
		/* Ok, need to divide the rectangle in 2 updates:
		 * small one which is not aligned, and the rest
		 * which starts at the next byte boundary */
		thincDbg(T_DBG_PROTO, ("%s: rectangle not properly "
			"aligned (%d). Sending 2 updates\n", 
			func, align));
		err= thincSendGBLRectUnaligned(cl, which, bits,
			bits_area, rect, align, fg, bg);
		if(err != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"send Bilevel message for unaligned "
				"rectangle: %d\n", func, err));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;
}

static int thincSendGBLRectBig(thincClientPtr cl, int which,
	char *bits, BoxPtr bits_area, BoxPtr rect, int rect_size, 
	unsigned long fg, unsigned long bg)
{
	t_dbg_name(func)
	xRectangle curr, area;
	BoxRec box;
	int rect_width, curr_size, limit, err;
	int done;

	thincDbg(T_DBG_PROTO, ("%s: dividing rectangle of size %d bytes. "
		"max size is %d bytes\n", func, rect_size, T_MAX_BITMAP_SIZE));

	rect_width = box_bitmap_width(rect);
	box_to_xrect(rect, area);

	if(rect_width <= T_MAX_BITMAP_SIZE) {
		/* at least one line fits */
		curr.x = area.x;
		curr.y = area.y;
		curr.width = area.width;
		done = T_FALSE;
		do {
			curr.height = 1;
			curr_size = rect_width;
			limit = area.height + area.y - curr.y;

			while(	(curr_size <= T_MAX_BITMAP_SIZE) && 
				(curr.height <= limit) ) {
				++curr.height;
				curr_size += rect_width;
			}

			/* tsk, tsk, always eating too much!!! */
			--curr.height;
			
			xrect_to_box(curr, &box);

			err = thincSendGBLRectSmall(cl, which, bits, bits_area,
				&box, fg, bg);
			if(err != T_OK)
				goto error;
			/* the end? */
			if(curr.height == limit) {
				done = T_TRUE;
				continue;
			}
			
			/* next! */
			curr.y += curr.height;
		} while(!done);
	} else {
		/* not even a line fits, ugly :( */
		int k,j,height = area.height;
		/* how many pieces per line? (actually 1 less, last one
		 * is handled separatedly) */
		int chunks = rect_width/T_MAX_BITMAP_SIZE;
		/* width of last chunk */
		int lastw= area.width % (T_MAX_BITMAP_SIZE*T_PIXELS_PER_BYTE);
		/* width, in pixels, of every chunk */
		int chunkw = T_MAX_BITMAP_SIZE*T_PIXELS_PER_BYTE;
		area.height = 1;

		for(k=0; k<height; k++) {
			area.x = rect->x1;
			area.width = chunkw;
			for(j=0; j<chunks; j++) {
				xrect_to_box(area, &box);
				err = thincSendGBLRectSmall(cl, which, bits,
					bits_area, &box, fg, bg);
				if(err != T_OK)
					goto error;
				area.x += chunkw;
			}
			/* send last chunk */
			if(lastw) {
				area.width = lastw;
				xrect_to_box(area, &box);
				err = thincSendGBLRectSmall(cl, which, bits,
					bits_area, &box, fg, bg);
				if(err != T_OK)
					goto error;
			}
			/* next line */				
			area.y++;
		}
	}

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;

error:
	thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send a piece of the "
		"GLYPH/BILEVEL rectangle\n", func));
	return T_ERROR;
}

/* copies the bitmap corresponding to rectangle rect. Assumes rectangle
 * starts at byte boundary */ 
static int thincGetBitmap(char *bits, BoxPtr bits_area, BoxPtr rect,
	char *bitmap)
{
	int x_off, y_off;
	int x_pos, y_pos;
	int bwidth, tbwidth, i, height;
	char *from, *to;

	x_off = bits_area->x1;
	y_off = bits_area->y1;
	x_pos = rect->x1 - x_off;
	y_pos = rect->y1 - y_off;
	x_pos /= T_PIXELS_PER_BYTE;

	bwidth = box_bitmap_width(rect);
	tbwidth = box_bitmap_width(bits_area);
	from = bits + x_pos + (y_pos * tbwidth);
	to = bitmap;
	height = box_height(rect);

	for(i=0; i<height; ++i) {
		memcpy(to, from, bwidth);
		to += bwidth;
		from += tbwidth;
	}

	return T_OK;
}

/* so how it works:
 * align has how many bytes we need to shift left
 * the first byte of each line to align the
 * rectangle properly. So we divide the rectangle
 * in two updates:
 * (rect->x1, rect->y1, firstw, box_height(rect)
 * and
 * (rect->x1+firstw, rect->y1, box_width() - firstw,
 * box_height(rect))
 * where firstw is T_PIXELS_PER_BYTE - align
 */
static int thincSendGBLRectUnaligned(thincClientPtr cl, int which,
	char *bits, BoxPtr bits_area, BoxPtr rect, int align, 
	unsigned long fg, unsigned long bg)
{
	t_dbg_name(func)
	BoxRec first, second;
	int firstw;
	int x_pos, y_pos, x_off, y_off, err = T_ERROR;
	int tbwidth, i, height, width;
	char bitmap[T_MAXH], *from, *to;

	thincDbg(T_DBG_PROTO, ("%s: sending bilevel update for unaligned "
		"rectangle: %d\n", func, align));
	print_box(T_DBG_PROTO, rect, "unaligned rect:");

	width = box_width(rect);
	firstw = MIN(T_PIXELS_PER_BYTE - align, width);

	thincDbg(T_DBG_PROTO, ("%s: firstw %d, width %d\n", func, firstw, 
		width));
	first.x1 = first.x2 = rect->x1;
	first.x2 += firstw;
	first.y1 = rect->y1;
	first.y2 = rect->y2;

	print_box(T_DBG_PROTO, &first, "first rectangle:");

	/* get bitmap for first rectangle. Need to copy 1 byte for each
	 * row and then shift it appropiately */
	x_off = bits_area->x1;
	y_off = bits_area->y1;
	x_pos = first.x1 - x_off;
	x_pos /= T_PIXELS_PER_BYTE;
	y_pos = first.y1 - y_off;

	tbwidth = box_bitmap_width(bits_area);
	from = bits + x_pos + (y_pos * tbwidth);
	to = bitmap;
	height = box_height(&first);

	for(i=0; i<height; ++i) {
		*to = from[0] >> align;
		from += tbwidth;
		++to;
	}

	/* send it */
	err = thincGBLExecStack(cl, which, bitmap, &first, fg, bg);
	if(err != T_OK) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send first "
			"rectangle: %d\n", func, err));
		return T_ERROR;
	}

	thincDbg(T_DBG_PROTO, ("%s: sent first rectangle. sending second "
		"one\n", func));

	/* now do the second update */
	if(first.x2 == rect->x2) {
		/* no need */
		thincDbg(T_DBG_PROTO, ("%s: unaligned rectangle is at right "
			"edge of udpate. No need for second update\n", func));
		return T_OK;
	}
	
	second.x1 = rect->x1+firstw;
	second.x2 = rect->x2;
	second.y1 = rect->y1;
	second.y2 = rect->y2;

	/* this one we can send normal way */
	err = thincSendGBLRectSimple(cl, which, bits, bits_area, &second, 
		fg, bg);
	if(err != T_OK) {
		thincDbg(T_DBG_PROTO, ("ERROR: %s: Couldn't send second "
			"rectangle: %d\n", func, err));
		return T_ERROR;
	}
	
	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;
}

/* bitmap for rectangle is small enough, and it's properly aligned */
static int thincSendGBLRectSimple(thincClientPtr cl, int which,
	char *bits, BoxPtr bits_area, BoxPtr rect, 
	unsigned long fg, unsigned long bg)
{
	t_dbg_name(func)
	char bitmap[T_MAX_BITMAP_SIZE], *dabitmap;
	int err = T_ERROR;

	thincDbg(T_DBG_PROTO, ("%s: sending bilevel for rectangle "
		"x %d, y %d, w %d, h %d\n", func, rect->x1, rect->y1,
		box_width(rect), box_height(rect)));

	if(!box_equal(bits_area, rect)) {
		err = thincGetBitmap(bits, bits_area, rect, bitmap);
		if(err != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"get bitmap for rectangle: %d\n", 
				func, err));
			return T_ERROR;
		}
		dabitmap = bitmap;
	} else {
		thincDbg(T_DBG_PROTO, ("%s; bitmap area same as rectangle\n",
			func));
		dabitmap = bits;
	}
	/* now send it */
	err = thincGBLExecStack(cl, which, dabitmap, rect, fg, bg);
	if(err != T_OK) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send GLYPH/BILEVEL "
			"message for rectangle: %d\n", func, err));
		return T_ERROR;
	} else {
		thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
		return T_OK;
	}
}

static int thincGBLExecStack(thincClientPtr cl, int which, 
	char *bits, BoxPtr rect, unsigned long fg, unsigned long bg)
{
	thincBitObj obj;

	obj.x = rect->x1;
	obj.y = rect->y1;
	obj.width = box_width(rect);
	obj.height = box_height(rect);
	obj.fg = fg;
	obj.bg = bg;

	obj.data = bits;
	obj.orig_data = bits;

	obj.data_len = bitmap_size(obj.width, obj.height);
	obj.data_width = bitmap_width(obj.width);

	obj.rects = NULL;
	obj.numrects = 0;

	obj.flags = 0;
	obj.extra_len = 0;
	obj.extra = obj.extra_data;
	obj.total_len = 0;

	obj.which = which;

	return thinc_msgStackExec(cl, cl->bitmapStack, &obj);
}

static int thincGBLFlushFill(thincScreenPtr pThinc, thincClientPtr cl,
	int which, char *bits, BoxPtr bits_area, BoxPtr rects,
	unsigned int numrects, unsigned long fg, unsigned long bg)
{
	t_dbg_name(func)
	thincBitObj obj;

	thincDbg(T_DBG_PROTO, ("%s: sending bitmap fill. numrects %d\n",
		func, numrects));

	obj.x = bits_area->x1;
	obj.y = bits_area->y1;
	obj.width = box_width(bits_area);
	obj.height = box_height(bits_area);
	print_box(T_DBG_PROTO, bits_area, "bitmap region");

	obj.fg = fg;
	obj.bg = bg;

	obj.data = bits;
	obj.orig_data = bits;
	obj.data_len = bitmap_size(obj.width, obj.height);
	obj.data_width = bitmap_width(obj.width);

	obj.rects = rects;
	obj.numrects = numrects;

	obj.flags = 0;
	obj.extra_len = 0;
	obj.extra = obj.extra_data;
	obj.total_len = 0;

	obj.which = which;

	thincDbg(T_DBG_PROTO, ("%s: exec'ing message stack. finished\n",func));

	return thinc_msgStackExec(cl, cl->bitmapStack, &obj);
}

/************ BITMAP MODULES *************/
#ifdef T_GETSTATS
static int thinc_bitmapPreStats(thincClientPtr cl, void *msg)
{
	stats_start_nostr(t_start);
	stats_init_simple(writeTime);
	stats_init_simple(blockTime);
	return T_OK;
}

static int thinc_bitmapPostStats(thincClientPtr cl, void *msg)
{
	unsigned long t_end;
	thincBitObj *obj = msg;
	struct thincStats *stats;

	stats = ( (obj->which == T_FB_GLYPH) ? &cl->parent->glyphStats :
		&cl->parent->bilevelStats );

	/* TODO: Maybe stats should be per client?? */
	stats_end(stats, t_start, t_end, flush);
	stats_inc(stats, count);
	stats_add(stats, write, writeTime);
	stats_add(stats, block, blockTime);
	stats_add(stats, size, obj->total_len);

	return T_OK;
}
#endif

static int thinc_bitmapSend(thincClientPtr cl, void *msg)
{
	thincBitObj *obj = msg;

	/* these two functions are very similar but maybe too messy to
	 * handle the differences? */
	if(obj->which == T_FB_GLYPH)
		return thinc_glyphSend(cl, obj);
	else
		return thinc_bilevelSend(cl, obj);
}

static int thinc_glyphSend(thincClientPtr cl, thincBitObj *obj)
{
	t_dbg_modname(modname, bitmapSend.name)
	thinc_fbGlyph *g;
	int dobuffer, err;
	char *msg;
	unsigned int write_len;
	
	thincDbg(T_DBG_PROTO, ("%s: sending GLYPH message to client %s\n", 
		modname, cl->id));

	write_len = obj->extra_len + obj->data_len;
	if(obj->numrects)
		write_len += obj->numrects*T_RECTREGION_SIZE;
	obj->total_len = THINC_HEADERSIZE + T_FB_GLYPH_SIZE + write_len;

	/* any chance of buffering it? */
	g = (thinc_fbGlyph*) thincBufferMsg(cl, T_FB_GLYPH, obj->flags, 
		T_FB_GLYPH_SIZE, write_len);
	if(!g) {
		thincDbg(T_DBG_PROTO, ("%s: Can't buffer GLYPH message\n",
			modname));
		thincFlushBuffer(cl);
		g = (thinc_fbGlyph*)t_msgBuffer;
		dobuffer = T_NO;
	} else
		dobuffer = T_YES;

	g->x = htons(obj->x);
	g->y = htons(obj->y);
	g->width = htons(obj->width);
	g->height = htons(obj->height);
	g->color = htonl(obj->fg);
	g->numrects = htons(obj->numrects);

	if(dobuffer) {
		msg = (((char*)g)+T_FB_GLYPH_SIZE);
		memcpy(msg, obj->extra_data, obj->extra_len);
		msg += obj->extra_len;
		if(!obj->data_len)
			goto rects1;
		memcpy(msg, obj->data, obj->data_len);
rects1:
		msg += obj->data_len;
		if(!obj->numrects)
			goto end;
		/* ignore return value since it's just doing memcpy */
		thincBitmapSendRects(cl, msg, obj->rects, obj->numrects);
	} else {
		err = cl->send_msg(cl, T_FB_GLYPH, T_FB_GLYPH_SIZE,
			obj->flags);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"GLYPH message\n", modname));
			return T_ERROR;
		}
		err = thincWriteData(cl, obj->extra_data, obj->extra_len);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"GLYPH message\n", modname));
			return T_ERROR;
		}
		/* play it safe */
		if(!obj->data_len)
			goto rects2;
		err = thincWriteData(cl, obj->data, obj->data_len);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"GLYPH message\n", modname));
			return T_ERROR;
		}
rects2:
		if(!obj->numrects)
			goto end;
		err = thincBitmapSendRects(cl, NULL, obj->rects, 
			obj->numrects);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"GLYPH message\n", modname));
			return T_ERROR;
		}
	}
end:
	thincDbg(T_DBG_PROTO, ("%s: finished, sent %u bytes\n", modname,
		obj->total_len));
	return T_OK;

	return T_OK;
}

static int thinc_bilevelSend(thincClientPtr cl, thincBitObj *obj)
{
	t_dbg_modname(modname, bitmapSend.name)
	thinc_fbBilevel *b;
	int dobuffer, err;
	char *msg;
	unsigned int write_len;
	
	thincDbg(T_DBG_PROTO, ("%s: sending BILEVEL message to client %s\n", 
		modname, cl->id));

	write_len = obj->data_len + obj->extra_len;
	if(obj->numrects)
		write_len += obj->numrects*T_RECTREGION_SIZE;
	obj->total_len = THINC_HEADERSIZE + T_FB_BILEVEL_SIZE + write_len;

	thincDbg(T_DBG_PROTO, ("%s: bitmap size is %u, extra size is %u "
		"and total size is %u\n", modname, obj->data_len,
		obj->extra_len, obj->total_len));

	/* any chance of buffering it? */
	b = (thinc_fbBilevel*) thincBufferMsg(cl, T_FB_BILEVEL,	obj->flags, 
		T_FB_BILEVEL_SIZE, write_len);
	if(!b) {
		thincDbg(T_DBG_PROTO, ("%s: Can't buffer BILEVEL message\n",
			modname));
		thincFlushBuffer(cl);
		b = (thinc_fbBilevel*)t_msgBuffer;
		dobuffer = T_NO;
	} else
		dobuffer = T_YES;

	b->x = htons(obj->x);
	b->y = htons(obj->y);
	b->width = htons(obj->width);
	b->height = htons(obj->height);
	b->fg = htonl(obj->fg);
	b->bg = htonl(obj->bg);
	b->numrects = htons(obj->numrects);

	if(dobuffer) {
		msg = (((char*)b)+T_FB_BILEVEL_SIZE);
		memcpy(msg, obj->extra_data, obj->extra_len);
		msg += obj->extra_len;
		if(!obj->data_len)
			goto rects1;
		memcpy(msg, obj->data, obj->data_len);
		msg+= obj->data_len;
rects1:
		if(!obj->numrects)
			goto end;
		/* ignore return value since it's just doing memcpy */
		thincBitmapSendRects(cl, msg, obj->rects, obj->numrects);
	} else {
		err = cl->send_msg(cl, T_FB_BILEVEL, T_FB_BILEVEL_SIZE,
			obj->flags);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"BILEVEL message\n", modname));
			return T_ERROR;
		}
		err = thincWriteData(cl, obj->extra_data, obj->extra_len);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"BILEVEL message\n", modname));
			return T_ERROR;
		}
		if(!obj->data_len)
			goto rects2;
		err = thincWriteData(cl, obj->data, obj->data_len);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"BILEVEL message\n", modname));
			return T_ERROR;
		}
rects2:
		if(!obj->numrects)
			goto end;
		err = thincBitmapSendRects(cl, NULL, obj->rects, 
			obj->numrects);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"BILEVEL message\n", modname));
			return T_ERROR;
		}
	}
end:
	thincDbg(T_DBG_PROTO, ("%s: finished, sent %u bytes\n", modname,
		obj->total_len));
	return T_OK;
}

static int thincBitmapSendRects(thincClientPtr cl, char *buffer,
	BoxPtr rects, int numrects)
{
	t_dbg_name(func)
	register int i;

	if(buffer) {
		register thinc_rectRegion *rp = (void*) buffer;
		for(i=0; i<numrects; ++i, ++rects, ++rp) {
			box_to_rectp_hton(rects, rp);
		}
	} else {
		int err;
		thinc_rectRegion r;
		for(i=0; i<numrects; ++i, ++rects) {
			box_to_rect_hton(rects, r);
			err = thincWriteData(cl, &r, T_RECTREGION_SIZE);
			if(err == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"send bitmap rectangle %d\n",
					func, i));
				return T_ERROR;
			}
		}
	}

	return T_OK;
}

/* Writes the bitmap to a local file. useful for debugging. */
#define do_write(fd, ptr, len) \
	err = write(fd, ptr, len); \
	if(err == -1) { \
		thincDbg(T_DBG_ERROR, ("%s: Couldn't write BITMAP update to " \
			"file: %s\n", modname, strerror(errno))); \
		return T_ERROR; \
	}

static int thinc_bitmapWriteFile(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, bitmapWriteFile.name)
	static int count = 0;
	unsigned int a;
	char fn[255];
	int err, dump; /* fd */
	const char *bl_base = "/tmp/bitmap/bl";
	const char *glyph_base = "/tmp/bitmap/glyph";
	const char *base;
	thincBitObj *obj = msg;
	const int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

	/* Write a header?? */
	const unsigned char write_header = T_TRUE;

	if(obj->which == T_FB_GLYPH) {
		base = glyph_base;
		strncpy(fn, base, 17);
		sprintf(fn+17, "%d", count++);
	} else {
		base = bl_base;
		strncpy(fn, base, 14);
		sprintf(fn+14, "%d", count++);
	}

	thincMsg("writing bitmap to file %s\n", fn);
	dump = creat(fn, mode);
	if(dump == -1) {
		thincDbg(T_DBG_ERROR, ("%s: Couldn't open bitmap file %s. "
			"count %d: %s\n", modname,fn,count,strerror(errno)));
		return T_ERROR;
	}

	if(write_header) {
		a = obj->width;
		a = htonl(a);
		do_write(dump, &a, 4);
		a = obj->height;
		a = htonl(a);
		do_write(dump, &a, 4);
#if 0		/* view-glyph doesn't know about fg and bg */
		a = obj->fg;
		a = htonl(a);
		do_write(dump, &a, 4);
		if(obj->which == T_FB_GLYPH) {
			a = 0;	/* pad */
			do_write(dump, &a, 4);
		} else {
			a = obj->bg;
			a = htonl(a);
			do_write(dump, &a, 4);
		}
		a = htonl(obj->data_len);
		do_write(dump, &a, 4);
#else
		a = htonl(obj->data_len);
		do_write(dump, &a, 4);
		/* pad */
		a = 0;
		do_write(dump, &a, 4);
#endif
	}
	do_write(dump, obj->data, obj->data_len);

	close(dump);
	return T_OK;
}

static int thinc_bitmapChkCache(thincClientPtr cl, void *msg)
{
	const char *modname = bitmapChkCache.name;
	unsigned long hash;
	thincBitObj *obj = msg;
	int err;

	thincDbg(T_DBG_PROTO, ("%s: checking if BITMAP update is in cache\n",
		modname));

	err = thinc_cache_query_data(cl->bit_cache, obj->data, obj->data_len,
		&hash);
	obj->hash = hash;
	if(err == T_NO) {
		thinc_cacheAdd *a;
		thincDbg(T_DBG_PROTO, ("%s: update is not in cache :*( "
			"Telling client to add it, hash is %u\n", 
			modname, obj->hash));
		obj->extra_len += T_FB_CACHE_ADD_SIZE;
		checkExtraBuffer(obj);
		a = (thinc_cacheAdd*) obj->extra;
		a->cache_id = htonl(obj->hash);
		obj->extra += T_FB_CACHE_ADD_SIZE;
		obj->flags |= T_FB_BITMAP_ADDCACHE;
	} else {
		thinc_cacheData *c;
		thincDbg(T_DBG_PROTO, ("%s: sweeet, found cached data. hash "
			"is %u\n", modname, obj->hash));
		obj->extra_len += T_FB_CACHE_DATA_SIZE;
		checkExtraBuffer(obj);
		c = (thinc_cacheData*)obj->extra;
		c->cache_id = htonl(obj->hash);
		obj->extra += T_FB_CACHE_DATA_SIZE;
		obj->flags |= T_FB_BITMAP_CACHED;

		/* fix up object so nobody sends data */
		obj->data = NULL;
		obj->data_len=0;
	}
	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_bitmapAddCache(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, bitmapAddCache.name)
	thincBitObj *obj = msg;

	thincDbg(T_DBG_PROTO, ("%s: possibly adding BITMAP update to cache\n",
		modname));

	if(obj->flags & T_FB_BITMAP_CACHED) {
		thincDbg(T_DBG_PROTO, ("%s: update already cached, finished\n",
			modname));
		return T_OK;
	} else {
		int err;
		err = thinc_cache_add_hash(cl->bit_cache, obj->hash, NULL);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add "
				"BITMAP update to cache\n", modname));
			return T_ERROR;
		} else {
			thincDbg(T_DBG_PROTO, ("%s: Added update to cache with "
				"hash %u, finished\n", modname, obj->hash));
			return T_OK;
		}
	}
}

#ifdef T_RECORD
int thinc_bitmapInitRecordStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_RECORD, ("%s: creating SFILL message stack for "
		"record command\n", func));

	cl->bitmapStack = thinc_msgStackCreate(cl);
	if(!cl->bitmapStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create SFILL "
			"message stack\n", func));
		return T_ERROR;
	}
	
	bitmap_addmodule(cl, &bitmapRecord);

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

void thinc_bitmapDestroyRecordStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->bitmapStack);
	cl->bitmapStack = NULL;
}

/* wrapper obj in case we need to call callback more than once */
struct thincBitObjRec {
	thincBitObj*	obj;
	char*		data;	/* next data to write */
	unsigned int	data_len;	/* how much we have yet to write.
					if 0, then only rects are missing
					(which we assume to fit all in
					buffer) */
};

static int thinc_glyphRecord(thincClientPtr cl, thincBitObj *obj);
static int thinc_bilevelRecord(thincClientPtr cl, thincBitObj *obj);

static int thinc_bitmapRecord(thincClientPtr cl, void *m)
{
	thincBitObj *obj = m;

	if(obj->which == T_FB_GLYPH)
		return thinc_glyphRecord(cl, obj);
	else
		return thinc_bilevelRecord(cl, obj);
}

static int thinc_glyphRecord(thincClientPtr cl, thincBitObj *obj)
{
	t_dbg_modname(modname, bitmapRecord.name)
	unsigned int o_total = 0;
	int err;
	char cached = T_FALSE;
	struct thincBitObjRec o_rec;

	thincDbg(T_DBG_RECORD, ("%s: \n", modname));

	/* if we're the recorder client, then total_len hasn't been computed.
	 * merde */
	o_total = obj->total_len;
	if(!obj->total_len) {
		obj->total_len = THINC_HEADERSIZE + T_FB_GLYPH_SIZE + 
			(obj->numrects*T_RECTREGION_SIZE) + obj->data_len +
			obj->extra_len;
	} else if(!obj->data_len) {
		/* it's cached. fix things up */ 
		thincDbg(T_DBG_RECORD, ("%s: it's cached. need to get it "
			"and fix some things up\n", modname));
		cached = T_TRUE;
		obj->data = obj->orig_data;
		obj->data_len = obj->data_width * obj->height;
		thincDbg(T_DBG_RECORD, ("%s: recovered cached update. data "
			"length is %u\n", modname, obj->data_len));
		obj->total_len += obj->data_len;
	}
	obj->total_len -= THINC_HEADERSIZE;
	
	thincDbg(T_DBG_RECORD, ("%s: update's total length is %d bytes\n", 
		modname, obj->total_len));

	o_rec.obj = obj;
	o_rec.data = NULL;

	err = thincRecord(cl, obj->total_len, T_FB_GLYPH, obj->flags,
		thinc_glyphRecordCb, &o_rec);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't record GLYPH "
			"command\n", modname));
		return T_ERROR;
	}

	if(cached) {
		/* revert our changes */
		obj->data = NULL;
		obj->data_len = 0;
	}
	obj->total_len = o_total;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_bilevelRecord(thincClientPtr cl, thincBitObj *obj)
{
	t_dbg_modname(modname, bitmapRecord.name)
	unsigned int o_total = 0;
	int err;
	char cached = T_FALSE;
	struct thincBitObjRec o_rec;

	thincDbg(T_DBG_RECORD, ("%s: \n", modname));

	/* if we're the recorder client, then total_len hasn't been computed.
	 * merde */
	o_total = obj->total_len;
	if(!obj->total_len) {
		obj->total_len = THINC_HEADERSIZE + T_FB_BILEVEL_SIZE + 
			(obj->numrects*T_RECTREGION_SIZE) + obj->data_len +
			obj->extra_len;
	} else if(!obj->data_len) {
		/* it's cached. fix things up */ 
		thincDbg(T_DBG_RECORD, ("%s: it's cached. need to get it "
			"and fix some things up\n", modname));
		cached = T_TRUE;
		obj->data = obj->orig_data;
		obj->data_len = obj->data_width * obj->height;
		thincDbg(T_DBG_RECORD, ("%s: recovered cached update. data "
			"length is %u\n", modname, obj->data_len));
		obj->total_len += obj->data_len;
	}
	obj->total_len -= THINC_HEADERSIZE;
	
	thincDbg(T_DBG_RECORD, ("%s: update's total length is %d bytes\n", 
		modname, obj->total_len));

	o_rec.obj = obj;
	o_rec.data = NULL;

	err = thincRecord(cl, obj->total_len, T_FB_BILEVEL, obj->flags,
		thinc_bilevelRecordCb, &o_rec);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't record BILEVEL "
			"command\n", modname));
		return T_ERROR;
	}

	if(cached) {
		/* revert our changes */
		obj->data = NULL;
		obj->data_len = 0;
	}
	obj->total_len = o_total;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

/* such a mess... */
static int thinc_bitmapRecordCbData(thincClientPtr cl, char *buf, 
	unsigned int *len, struct thincBitObjRec *rec);
static int thinc_bitmapRecordCbJustRects(thincClientPtr cl, char *buf, 
	unsigned int *len, thincBitObj *obj);

static int thinc_glyphRecordCb(thincClientPtr cl, char *buf, 
	unsigned int *len, void *ptr)
{
	t_dbg_modname(modname, bitmapRecord.name)
	struct thincBitObjRec *rec = ptr;
	thincBitObj *obj = rec->obj;
	unsigned int d_len, h_len, a_len, i;
	thinc_fbGlyph *g;
	char *msg;
	register BoxPtr rects;
	register thinc_rectRegion *rp;

	thincDbg(T_DBG_RECORD, ("%s\n", modname));

	/* see if first time around, and send msg. we save all the flags, too
	 * much pain to set only those we would care about on
	 * playback. note that we assume that at the most basic, len
	 * has enough space for the message and the extra info */
	if(rec->data) {
		return thinc_bitmapRecordCbData(cl, buf, len, rec);
	}

	thincDbg(T_DBG_RECORD, ("%s: first time, recording header\n",
		modname));

	/* TODO: Add T_BUG_ON check between h_len and len */
	h_len = T_FB_GLYPH_SIZE + obj->extra_len;

	g = (thinc_fbGlyph*) buf;
	g->x = htons(obj->x);
	g->y = htons(obj->y);
	g->width = htons(obj->width);
	g->height = htons(obj->height);
	g->color = htonl(obj->fg);
	g->numrects = htons(obj->numrects);

	msg = ((char*)g) + T_FB_GLYPH_SIZE;
	memcpy(msg, obj->extra_data, obj->extra_len);
	msg += obj->extra_len;

	/* we still have to write all this ... */
	d_len = obj->data_len + (obj->numrects*T_RECTREGION_SIZE);
	/* ... and this is how much space we have available */
	a_len = *len - h_len;

	thincDbg(T_DBG_RECORD, ("%s: trying to record %u bytes, have space "
		"for %u\n", modname, d_len, a_len));

	if(a_len >= d_len) {
		/* sweetness */
		thincDbg(T_DBG_RECORD, ("%s: we fit just fine\n", modname));
		memcpy(msg, obj->data, obj->data_len);
		msg += obj->data_len;

		thincDbg(T_DBG_RECORD, ("%s: recording %d rectangles\n", 
			modname, obj->numrects));
		rp = (thinc_rectRegion*)msg;
		rects = obj->rects;

		for(i=0; i<obj->numrects; ++i, ++rects, ++rp) {
			box_to_rectp_hton(rects, rp);
		}

		*len = obj->total_len; /* wrote it all */
		thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u bytes\n", 
			modname, *len));

		return T_REC_DONE;
	} else {
		/* damn. ok, see if data fits... */
		if(a_len >= obj->data_len) {
			thincDbg(T_DBG_RECORD, ("%s: data fits. problem is "
				"damn rectangles\n", modname));
			memcpy(msg, obj->data, obj->data_len);
			msg += obj->data_len;

			/* next time, just send rects */
			rec->data = obj->data;
			rec->data_len = 0;

			*len = h_len + obj->data_len;
			thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u "
				"bytes\n", modname, *len));
			return T_REC_MORE;
		} 
		/* data doesn't fit. copy what we can, then punt the rest
		 * for next time */
		thincDbg(T_DBG_RECORD, ("%s: data doesn't fit. only have "
			"space for %u bytes, need %u\n", modname, a_len,
			obj->data_len));
		memcpy(msg, obj->data, a_len);

		rec->data = obj->data + a_len;
		rec->data_len = obj->data_len - a_len;
		*len = a_len;
		thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u "
			"bytes\n", modname, *len));
		return T_REC_MORE;
	}
}

/*********** bilevel ***********/
static int thinc_bilevelRecordCb(thincClientPtr cl, char *buf, 
	unsigned int *len, void *ptr)
{
	t_dbg_modname(modname, bitmapRecord.name)
	struct thincBitObjRec *rec = ptr;
	thincBitObj *obj = rec->obj;
	unsigned int d_len, h_len, a_len, i;
	thinc_fbBilevel *b;
	char *msg;
	register BoxPtr rects;
	register thinc_rectRegion *rp;

	thincDbg(T_DBG_RECORD, ("%s\n", modname));

	/* see if first time around, and send msg. we save all the flags, too
	 * much pain to set only those we would care about on
	 * playback. note that we assume that at the most basic, len
	 * has enough space for the message and the extra info */
	if(rec->data) {
		return thinc_bitmapRecordCbData(cl, buf, len, rec);
	}

	thincDbg(T_DBG_RECORD, ("%s: first time, recording header\n",
		modname));

	/* TODO: Add T_BUG_ON check between h_len and len */
	h_len = T_FB_BILEVEL_SIZE + obj->extra_len;

	b = (thinc_fbBilevel*) buf;
	b->x = htons(obj->x);
	b->y = htons(obj->y);
	b->width = htons(obj->width);
	b->height = htons(obj->height);
	b->fg = htonl(obj->fg);
	b->bg = htonl(obj->bg);
	b->numrects = htons(obj->numrects);

	msg = ((char*)b) + T_FB_BILEVEL_SIZE;
	memcpy(msg, obj->extra_data, obj->extra_len);
	msg += obj->extra_len;

	/* we still have to write all this ... */
	d_len = obj->data_len + (obj->numrects*T_RECTREGION_SIZE);
	/* ... and this is how much space we have available */
	a_len = *len - h_len;

	thincDbg(T_DBG_RECORD, ("%s: trying to record %u bytes, have space "
		"for %u\n", modname, d_len, a_len));

	if(a_len >= d_len) {
		/* sweetness */
		thincDbg(T_DBG_RECORD, ("%s: we fit just fine\n", modname));
		memcpy(msg, obj->data, obj->data_len);
		msg += obj->data_len;

		thincDbg(T_DBG_RECORD, ("%s: recording %d rectangles\n", 
			modname, obj->numrects));
		rp = (thinc_rectRegion*)msg;
		rects = obj->rects;

		for(i=0; i<obj->numrects; ++i, ++rects, ++rp) {
			box_to_rectp_hton(rects, rp);
		}

		*len = obj->total_len; /* wrote it all */
		thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u bytes\n", 
			modname, *len));

		return T_REC_DONE;
	} else {
		/* damn. ok, see if data fits... */
		if(a_len >= obj->data_len) {
			thincDbg(T_DBG_RECORD, ("%s: data fits. problem is "
				"damn rectangles\n", modname));
			memcpy(msg, obj->data, obj->data_len);
			msg += obj->data_len;

			/* next time, just send rects */
			rec->data = obj->data;
			rec->data_len = 0;

			*len = h_len + obj->data_len;
			thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u "
				"bytes\n", modname, *len));
			return T_REC_MORE;
		} 
		/* data doesn't fit. copy what we can, then punt the rest
		 * for next time */
		thincDbg(T_DBG_RECORD, ("%s: data doesn't fit. only have "
			"space for %u bytes, need %u\n", modname, a_len,
			obj->data_len));
		memcpy(msg, obj->data, a_len);

		rec->data = obj->data + a_len;
		rec->data_len = obj->data_len - a_len;
		*len = a_len;
		thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u "
			"bytes\n", modname, *len));
		return T_REC_MORE;
	}
}

static int thinc_bitmapRecordCbData(thincClientPtr cl, char *buf, 
	unsigned int *len, struct thincBitObjRec *rec)
{
	thincBitObj *obj = rec->obj;
	unsigned int a_len, r_len;

	if(!rec->data_len) {
		/* we're done with the data, just missing the
		 * rects. we blindly assume they all fit in :) */
		return thinc_bitmapRecordCbJustRects(cl, buf,
			len, obj);
	}
	/* still got some data to write */
	a_len = *len;
	if(a_len < rec->data_len) {
		/* and it still doesn't fit. damn */
		memcpy(buf, rec->data, a_len);

		rec->data += a_len;
		rec->data_len -= a_len;
		*len = a_len;

		return T_REC_MORE;
	}
	/* it fits */
	memcpy(buf, rec->data, rec->data_len);
	a_len -= rec->data_len;
	r_len = (obj->numrects*T_RECTREGION_SIZE);

	/* but do the rects fit? */
	if(a_len < r_len) {
		/* no :( */
		*len = rec->data_len;
		rec->data_len = 0;
		return T_REC_MORE;
	} else {
		/* in yer face! */
		buf += rec->data_len;
		thinc_bitmapRecordCbJustRects(cl, buf, len, obj);
		*len = rec->data_len + r_len;
		return T_REC_DONE;
	}
}

static int thinc_bitmapRecordCbJustRects(thincClientPtr cl, char *buf, 
	unsigned int *len, thincBitObj *obj)
{
	t_dbg_modname(modname, bitmapRecord.name)
	register BoxPtr rects;
	register thinc_rectRegion *rp;
	unsigned int i;

	thincDbg(T_DBG_RECORD, ("%s: just missing the rects\n", modname));

	rects = obj->rects;
	rp = (thinc_rectRegion*)buf;
	for(i=0; i<obj->numrects; ++i, ++rects, ++rp) {
		box_to_rectp_hton(rects, rp);
	}

	*len = (obj->numrects*T_RECTREGION_SIZE);
	thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u bytes\n", modname, 
		*len));
	return T_REC_DONE;
}
#endif /* T_RECORD */
