/* Copyright 2003, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2003, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Bitmap and bilevel handling functions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#define MASK_INIT	1U

/* the real functions. Need one for each color depth */
static int thinc_bitmap8(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg);
static int thinc_bitmap16(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg);
static int thinc_bitmap32(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg);

static int thinc_getBitmapRect1(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch);
static int thinc_getBitmapRect8(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch);
static int thinc_getBitmapRect16(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch);
static int thinc_getBitmapRect32(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch);

/* for Bilevel functions */
#define T_BGNO	1
#define T_BGYES 2
#define T_BGDONT 4

static int thinc_bitmap32Bilevel(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg,
	unsigned long *bg);
static int thinc_bitmap16Bilevel(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg,
	unsigned long *bg);
static int thinc_bitmap8Bilevel(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg,
	unsigned long *bg);

/* for eliminating transparency */
#define left_border 0
#define right_border 1
#define maybe 2

#define top_border 0
#define bottom_border 1

#define T_RET_TRANS	0

int thincInitBitmaps(thincScreenPtr pThinc)
{
	int ret = T_OK;

	switch(pThinc->bpp) {
		case 8:
			pThinc->bitmap_from_fb = thinc_bitmap8;
			pThinc->bitmap_bilevel = thinc_bitmap8Bilevel; 
			break;
		case 16:
			pThinc->bitmap_from_fb = thinc_bitmap16;
			pThinc->bitmap_bilevel = thinc_bitmap16Bilevel; 
			break;
		case 32:
			pThinc->bitmap_from_fb = thinc_bitmap32;
			pThinc->bitmap_bilevel = thinc_bitmap32Bilevel; 
			break;
		default:
			thincError("ERROR: %s: Depth %d not supported for "
				"bitmaps\n", __func__, pThinc->bpp);
			ret = T_ERROR;
			break;
	}

	return ret;
}

/* Create a bitmap (returned in bitmap) of the passed region. The region
 * only has 2 colors. A '1' is put in the bitmap if the fg color
 * is found. 
 * bitmap should have enough memory allocated for the bitmap. 
 * Returns the size of the bitmap or -1 on error.
 * TODO: X uses bitmaps which have the least significant bit in each byte as
 * the rightmost bit, so for each byte we need to process from right to
 * left. It's counterintuitive, and XAA prefers to use the other way
 * around. Maybe we should change it (need to change X client) */
int thinc_createBitmapFB(thincScreenPtr pThinc, char *bitmap, 
	xRectangle *area, unsigned long fg)
{
	t_dbg_name(func)
	int ret, off;
	char *source;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_DRAW, ("%s: depth %d\n", func, pThinc->bpp));

	/* TODO: Use generic function like t_get_fbpointer() */
	off = area->y*pThinc->widthBytes;
	off += t_pix_to_bytes(area->x, pThinc);
	source = (pThinc->theFB + off);

	t_bug_on_2(
		!pThinc->bitmap_from_fb,
		T_BUG("Bitmap functions pointers were not initialized"),
		return T_ERROR
	);
	ret = pThinc->bitmap_from_fb(pThinc, source, bitmap, area,
		pThinc->widthBytes, fg);

	thincDbg(T_DBG_DRAW, ("%s: finished %d\n", func, ret));

	stats_end(pThinc, t1, t2, glyphStats.create);
	return ret;
}

int thinc_createBitmapPixmap(thincScreenPtr pThinc, PixmapPtr pix,
	char *bitmap, xRectangle *area, unsigned long fg)
{
	const char *func = __func__;
	int ret;
	unsigned int off, width, bpp, Bpp;
	char *source;
	unsigned int stride;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	/* Some stupid apps send us invalid commands, e.g. konqueror.
	 * just ignore them :/
	 * yes, i've seen commands that say "draw text at (0,4294967279)"
	 * use konqueror and go to the bbc webpage */
	if(	(area->x > pix->drawable.width) ||
		(area->x + area->width > pix->drawable.width) ||
		(area->y > pix->drawable.height) ||
		(area->y + area->height > pix->drawable.height) ) {
		return T_OK;
	}

	stride = pix->devKind;
	width = pix->drawable.width;
	bpp = pix->drawable.bitsPerPixel;
	Bpp = (bpp == 1 ? 1 : bpp >> 3);
	off = area->y*stride;
	off += (area->x*Bpp);
	source = thinc_getPixmap(pix);
	source += off;

	thincDbg(T_DBG_DRAW, ("%s: depth %d\n", func, pix->drawable.depth));

	switch(bpp) {
		case 1:
			memcpy(bitmap, source, 
				bitmap_size(area->width, area->height));
			ret = T_OK;
			break;
		case 8:
			ret = thinc_bitmap8(pThinc, source, bitmap, area, 
				stride, fg);
			break;
		case 16:
			ret = thinc_bitmap16(pThinc, source, bitmap, area, 
				stride, fg);
			break;
		case 32:
			ret = thinc_bitmap32(pThinc, source, bitmap, area, 
				stride, fg);
			break;
		default:
			thincError("%s: Depth %d not supported for bitmaps\n",
				func, bpp);
			ret = T_ERROR;
			break;
	}

	thincDbg(T_DBG_DRAW, ("%s: finished %d\n", func, ret));

	stats_end(pThinc, t1, t2, glyphStats.create);
	return ret;
}

static int thinc_bitmap8(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg)
{
	t_dbg_name(func)
	int xe, xc; /* end,current */
	int ye, yc;
	register char *bits;
	register unsigned char mask;
	register unsigned char curr; /* current byte in bitmap */
	register CARD8* source;
	int next;
	int x = area->x, y=area->y, w=area->width, h=area->height;

	thincDbg(T_DBG_DRAW, ("%s: x %d y %d w %d h %d\n", func, x, y, w, h));

	xe = x+w;
	ye = y+h;

	next = stride - w;
	
	source = (CARD8*) from;
	bits = bitmap;

	/* loop over the region row by row */
	for(yc=y; yc<ye; yc++) {
		mask = MASK_INIT;
		curr = 0;
		/* process row */
		for(xc=x;xc<xe; xc++) {
			if(*source++ == (CARD8)fg) {
				/* found fg color, put a 1 in the bitmap */
				curr |= mask;
			}
			/* next bit */
			mask <<= 1;

			if(mask == 0) {
				/* completed one byte, store it in the 
				 * bitmap and keep going */
				*bits++ = curr;
				mask = MASK_INIT;
				curr = 0;
			}
		}

		/* chk if any partial bytes, and add them to the bitmap */
		if(mask != MASK_INIT) {
			*bits++ = curr;
		}

		source += next;
	}


	thincDbg(T_DBG_DRAW, ("%s: finished: %d\n", func, bits-bitmap));

	return bits-bitmap;
}

static int thinc_bitmap16(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg)
{
	t_dbg_name(func)
	int xe, xc; /* start,end,current */
	int ye, yc;
	register char *bits;
	register unsigned char mask;
	register unsigned char curr;
	register CARD16* source;
	int next;
	int x = area->x, y=area->y, w=area->width, h=area->height;

	thincDbg(T_DBG_DRAW, ("%s: x %d y %d w %d h %d fg %ld\n", func, 
		x, y, w,h,fg));

	xe = x+w;
	ye = y+h;

	/* stride is in bytes */
	next = (stride >> 1) - w;
	
	source = (CARD16*) from;
	bits = bitmap;

	/* loop over the region row by row */
	for(yc=y; yc<ye; yc++) {
		mask = MASK_INIT;
		curr = 0;
		/* process row */
		for(xc=x;xc<xe; xc++) {
			if(*source++ == (CARD16)fg) {
				/* found fg color, put a 1 in the bitmap */
				curr |= mask;
			}
			mask <<= 1;
			if(mask == 0) {
				/* completed one byte, store it in the 
				 * bitmap and keep going */
				*bits++ = curr;
				mask = MASK_INIT;
				curr = 0;
			}
		}

		/* chk if any partial bytes, and add them to the bitmap */
		if(mask != MASK_INIT) {
			*bits++ = curr;
		}

		source += next;
	}

	thincDbg(T_DBG_DRAW, ("%s: finished, size %d\n", func, bits-bitmap));

	return bits-bitmap;
}

static int thinc_bitmap32(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg)
{
	t_dbg_name(func)
	unsigned int xe, xc; /* end,current */
	unsigned int ye, yc;
	register char *bits;
	register unsigned char mask;
	register unsigned char curr;
	register CARD32* source;
	unsigned int next;
	unsigned int x = area->x, y=area->y, w=area->width, h=area->height;

	thincDbg(T_DBG_DRAW, ("%s: x %d y %d w %d h %d fg %#lx\n", 
		func, x, y, w, h, fg));

	xe = x+w;
	ye = y+h;
	/* stride is in bytes */
	next = (stride >> 2) - w;
	bits = bitmap;
	source = (CARD32*)from;

	/* loop over the region row by row */
	for(yc=y; yc<ye; ++yc) {
		mask = MASK_INIT;
		curr = 0;
		/* process row */
		for(xc=x;xc<xe; ++xc) {
			if(*source == fg) {
				/* found fg color, put a 1 in the bitmap */
				curr |= mask;
			} 
			++source;
			mask <<= 1;
			if(mask == 0) {
				/* completed one byte, store it in the 
				 * bitmap and keep going */
				*bits++ = curr;
				mask = MASK_INIT;
				curr = 0;
			}
		}

		/* chk if any partial bytes, and add them to the bitmap */
		if(mask != MASK_INIT) {
			*bits++ = curr;
		}

		/* next row */
		source += next;
	}

	thincDbg(T_DBG_DRAW, ("%s: finished, size %d\n", func, bits-bitmap));

	return bits-bitmap;
}

int thinc_createBitmapFBBilevel(thincScreenPtr pThinc, char *bitmap, 
	xRectangle *area, unsigned long fg, unsigned long *bg)
{
	t_dbg_name(func)
	int ret, off;
	char *source;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	/* TODO: Use generic function like t_get_fbpointer() */
	off = area->y*pThinc->widthBytes;
	off += t_pix_to_bytes(area->x,pThinc);
	source = (pThinc->theFB + off);

	thincDbg(T_DBG_DRAW, ("%s: depth %d\n", func, pThinc->bpp));

	t_bug_on_2(
		!pThinc->bitmap_bilevel,
		T_BUG("Bitmap functions pointers were not initialized"),
		return T_ERROR
	);
	ret = pThinc->bitmap_bilevel(pThinc, source, bitmap, area,
		pThinc->width, fg, bg);

	thincDbg(T_DBG_DRAW, ("%s: finished %d\n", func, ret));

	stats_end(pThinc, t1, t2, glyphStats.create);
	return ret;
}

int thinc_createBitmapPixmapBilevel(thincScreenPtr pThinc, PixmapPtr pix,
	char *bitmap, xRectangle *area, unsigned long fg, 
	unsigned long *bg)
{
	const char *func = __func__;
	int ret;
	unsigned int off, width, bpp, Bpp;
	char *source;
	unsigned int stride;

	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	/* Some stupid apps send us invalid commands, e.g. konqueror.
	 * just ignore them :/
	 * yes, i've seen commands that say "draw text at (0,4294967279)"
	 * use konqueror and go to the bbc webpage */
	if(	(area->x > pix->drawable.width) ||
		(area->x + area->width > pix->drawable.width) ||
		(area->y > pix->drawable.height) ||
		(area->y + area->height > pix->drawable.height) ) {
		return T_OK;
	}

	stride = pix->devKind;
	width = pix->drawable.width;
	bpp = pix->drawable.bitsPerPixel;
	Bpp = (bpp == 1 ? 1 : bpp >> 3);
	off = area->y*stride;
	off += (area->x*Bpp);
	source = thinc_getPixmap(pix);
	source += off;

	thincDbg(T_DBG_DRAW, ("%s: depth %d\n", func, pix->drawable.depth));

	switch(bpp) {
		case 1:
			memcpy(bitmap, source, 
				bitmap_size(area->width, area->height));
			ret = T_OK;
			break;
		case 8:
			ret = thinc_bitmap8Bilevel(pThinc, source, bitmap, 
				area, stride, fg, bg);
			break;
		case 16:
			ret = thinc_bitmap16Bilevel(pThinc, source, bitmap, 
				area, stride, fg, bg);
			break;
		case 32:
			ret = thinc_bitmap32Bilevel(pThinc, source, bitmap, 
				area, stride, fg, bg);
			break;
		default:
			thincError("%s: Depth %d not supported for bitmaps\n",
				func, bpp);
			ret = T_ERROR;
			break;
	}

	thincDbg(T_DBG_DRAW, ("%s: finished %d\n", func, ret));

	stats_end(pThinc, t1, t2, glyphStats.create);
	return ret;
}

static int thinc_bitmap8Bilevel(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg,
	unsigned long *bg_out)
{
	t_dbg_name(func)
	int xe, xc; /* end,current */
	int ye, yc;
	register char *bits;
	register unsigned char mask;
	register unsigned char curr; /* current byte in bitmap */
	register CARD8* source;
	int next;
	int x = area->x, y=area->y, w=area->width, h=area->height;
	unsigned int bg = fg;
	register unsigned char bgstatus = T_BGNO;

	thincDbg(T_DBG_DRAW, ("%s: x %d y %d w %d h %d\n", func, x, y, w, h));

	xe = x+w;
	ye = y+h;

	next = stride - w;
	
	source = (CARD8*) from;
	bits = bitmap;

	/* loop over the region row by row */
	for(yc=y; yc<ye; yc++) {
		mask = MASK_INIT;
		curr = 0;
		/* process row */
		for(xc=x;xc<xe; xc++) {
			if(*source++ == (CARD8)fg) {
				/* found fg color, put a 1 in the bitmap */
				curr |= mask;
			} else if( (bgstatus & T_BGYES) && (*source != bg) ) {
				bgstatus = T_BGDONT;
				bg = fg;
			} else if(bgstatus & T_BGNO) {
				bg = *source;
				bgstatus = T_BGYES;
			}
 
			/* next bit */
			mask <<= 1;

			if(mask == 0) {
				/* completed one byte, store it in the 
				 * bitmap and keep going */
				*bits++ = curr;
				mask = MASK_INIT;
				curr = 0;
			}
		}

		/* chk if any partial bytes, and add them to the bitmap */
		if(mask != MASK_INIT) {
			*bits++ = curr;
		}

		source += next;
	}

	*bg_out = bg;

	thincDbg(T_DBG_DRAW, ("%s: finished: %d\n", func, bits-bitmap));

	return bits-bitmap;
}

static int thinc_bitmap16Bilevel(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg,
	unsigned long *bg_out)
{
	t_dbg_name(func)
	int xe, xc; /* start,end,current */
	int ye, yc;
	register char *bits;
	register unsigned char mask;
	register unsigned char curr;
	register CARD16* source;
	int next;
	int x = area->x, y=area->y, w=area->width, h=area->height;
	unsigned int bg = fg;
	register unsigned char bgstatus = T_BGNO;

	thincDbg(T_DBG_DRAW, ("%s: x %d y %d w %d h %d fg %ld\n", func, 
		x, y, w,h,fg));

	xe = x+w;
	ye = y+h;

	/* stride is in bytes */
	next = (stride >> 1) - w;
	
	source = (CARD16*) from;
	bits = bitmap;

	/* loop over the region row by row */
	for(yc=y; yc<ye; yc++) {
		mask = MASK_INIT;
		curr = 0;
		/* process row */
		for(xc=x;xc<xe; xc++) {
			if(*source++ == (CARD16)fg) {
				/* found fg color, put a 1 in the bitmap */
				curr |= mask;
			} else if( (bgstatus & T_BGYES) && (*source != bg) ) {
				bgstatus = T_BGDONT;
				bg = fg;
			} else if(bgstatus & T_BGNO) {
				bg = *source;
				bgstatus = T_BGYES;
			}
 
			mask <<= 1;
			if(mask == 0) {
				/* completed one byte, store it in the 
				 * bitmap and keep going */
				*bits++ = curr;
				mask = MASK_INIT;
				curr = 0;
			}
		}

		/* chk if any partial bytes, and add them to the bitmap */
		if(mask != MASK_INIT) {
			*bits++ = curr;
		}

		source += next;
	}

	*bg_out = bg;

	thincDbg(T_DBG_DRAW, ("%s: finished, size %d\n", func, bits-bitmap));
	return bits-bitmap;
}

static int thinc_bitmap32Bilevel(thincScreenPtr pThinc, char *from,
	char *bitmap, xRectangle *area, int stride, unsigned long fg,
	unsigned long *bg_out)
{
	t_dbg_name(func)
	unsigned int xe, xc; /* end,current */
	unsigned int ye, yc;
	register char *bits;
	register unsigned char mask;
	register unsigned char curr;
	register CARD32* source;
	unsigned int next;
	unsigned int x = area->x, y=area->y, w=area->width, h=area->height;
	unsigned int bg = fg;
	register unsigned char bgstatus = T_BGNO;

	thincDbg(T_DBG_DRAW, ("%s: x %d y %d w %d h %d fg %#lx\n", 
		func, x, y, w, h, fg));

	xe = x+w;
	ye = y+h;
	next = (stride >> 2) - w;
	bits = bitmap;
	source = (CARD32*)from;

	/* loop over the region row by row */
	for(yc=y; yc<ye; ++yc) {
		mask = MASK_INIT;
		curr = 0;
		/* process row */
		for(xc=x;xc<xe; ++xc) {
			if(*source == fg) {
				/* found fg color, put a 1 in the bitmap */
				curr |= mask;
			} else if( (bgstatus & T_BGYES) && (*source != bg) ) {
				bgstatus = T_BGDONT;
				bg = fg;
			} else if(bgstatus & T_BGNO) {
				bg = *source;
				bgstatus = T_BGYES;
			}
			++source;
			mask <<= 1;
			if(mask == 0) {
				/* completed one byte, store it in the 
				 * bitmap and keep going */
				*bits++ = curr;
				mask = MASK_INIT;
				curr = 0;
			}
		}

		/* chk if any partial bytes, and add them to the bitmap */
		if(mask != MASK_INIT) {
			*bits++ = curr;
		}

		/* next row */
		source += next;
	}

	*bg_out = bg;
	thincDbg(T_DBG_DRAW, ("%s: finished, size %d\n", func, bits-bitmap));
	return bits-bitmap;
}

/* take out transparent parts in bitmaps */

unsigned int thinc_getBitmapRect(thincScreenPtr pThinc, PixmapPtr pixmap,
	xRectangle *area, register unsigned long fg, xRectangle *out)
{
	t_dbg_name(func)
	char *data;
	int ret;
	unsigned int size;
	unsigned int area_width, pitch, bpp, Bpp, off;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_DRAW, ("%s: getting bitmap for rectangle "
		"%u,%u=>%ux%u, fg is %#lx. pitch %d\n", func, area->x, area->y,
		area->width, area->height, fg, pixmap->devKind));

	pitch = pixmap->devKind;
	area_width = pixmap->drawable.width;
	bpp = pixmap->drawable.bitsPerPixel;
	Bpp = (bpp == 1 ? 1: bpp >> 3);
	off = area->y*pitch;
	off += (area->x*Bpp);
	data = thinc_getPixmap(pixmap);
	data += off;

	switch(bpp) {
		case 1:
			ret = thinc_getBitmapRect1(pThinc, data, area, fg,
				out, pitch);
			break;
		case 8:
			ret = thinc_getBitmapRect8(pThinc, data, area, fg, 
				out, pitch);
			break;
		case 16:
			ret = thinc_getBitmapRect16(pThinc, data, area, fg, 
				out, pitch);
			break;
		case 32:
			ret = thinc_getBitmapRect32(pThinc, data, area, fg, 
				out, pitch);
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: color depth %d "
				"not supported\n", func, bpp));
			ret = 0;
			break;
	}

	if(ret == T_RET_TRANS) {
		size = 0;
	} else {
		size = bitmap_size(out->width, out->height);
	}

	stats_end(pThinc, t1, t2, glyphStats.create);

	thincDbg(T_DBG_DRAW, ("%s: finished, size %u\n", func, size));
	return size;
}

/* TODO: Actually do what this function is supposed to do? */
static int thinc_getBitmapRect1(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch)
{
	out->x = area->x;
	out->y = area->y;
	out->width = area->width;
	out->height = area->height;
	return T_OK;
}

static int thinc_getBitmapRect8(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch)
{
	register unsigned char state1, state2;
	register unsigned int xc, x2;
	unsigned char no_xr, no_xl, no_yb;
	unsigned int yc, y2, gxl, gxr, xl, xr, yt, yb;
	register CARD8 *pix;

	x2 = area->x + area->width;
	y2 = area->y + area->height;

	/* init global borders */
	gxl = x2;
	gxr = area->x;

	state2 = top_border;
	no_yb = T_TRUE;
	yt = y2;
	yb = area->y;

	for(yc = area->y; yc < y2; ++yc) {
		pix = (CARD8*) data;
		xl = xr = area->x;
		no_xr = no_xl = T_TRUE;
		state1 = left_border;
		for(xc = area->x; xc<x2; ++xc, ++pix) {
			if(state1 == left_border) {
				if(*pix == fg) {
					state1 = right_border;
					xl = xc;
					no_xl = T_FALSE;
				}
			} else if(state1 == right_border) {
				if(*pix != fg) {
					xr = xc;
					no_xr = T_FALSE;
					state1 = maybe;
				}
			} else {
				if(*pix == fg) {
					no_xr = T_TRUE;
					state1 = right_border;
				}
			}

		}
		if(state2 == top_border) {
			if(no_xl)
				goto next;
			else {
				yt = yc;
				state2 = bottom_border;
			}
		} else if(state2 == bottom_border) {
			if(no_xl) {
				no_yb = T_FALSE;
				yb = yc;
				state2 = maybe;
				goto next;
			}
		} else {
			if(no_xl)
				goto next;
			else {
				state2 = bottom_border;
				no_yb = T_TRUE;
			}
		}

		/* non-transparent lines get here */
		if(no_xr) 
			xr = x2;
		if(xl < gxl)
			gxl = xl;
		if(xr > gxr) 
			gxr = xr;

next:
		data += pitch;
	}

	if(no_yb)
		yb = y2;

	if(gxr <= gxl  || yb <= yt)
		/* transparent bitmap, finished */
		return T_RET_TRANS;

	/* store rectangle */
	out->x = gxl;
	out->width = gxr-gxl;
	out->y = yt;
	out->height = yb-yt;

	return T_OK;
}

static int thinc_getBitmapRect16(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch)
{
	register unsigned char state1, state2;
	register unsigned int xc, x2;
	unsigned char no_xr, no_xl, no_yb;
	unsigned int yc, y2, gxl, gxr, xl, xr, yt, yb;
	register CARD16 *pix;

	x2 = area->x + area->width;
	y2 = area->y + area->height;

	/* init global borders */
	gxl = x2;
	gxr = area->x;

	state2 = top_border;
	no_yb = T_TRUE;
	yt = y2;
	yb = area->y;

	for(yc = area->y; yc < y2; ++yc) {
		pix = (CARD16*) data;
		xl = xr = area->x;
		no_xr = no_xl = T_TRUE;
		state1 = left_border;
		for(xc = area->x; xc<x2; ++xc, ++pix) {
			if(state1 == left_border) {
				if(*pix == fg) {
					state1 = right_border;
					xl = xc;
					no_xl = T_FALSE;
				}
			} else if(state1 == right_border) {
				if(*pix != fg) {
					xr = xc;
					no_xr = T_FALSE;
					state1 = maybe;
				}
			} else {
				if(*pix == fg) {
					no_xr = T_TRUE;
					state1 = right_border;
				}
			}

		}
		if(state2 == top_border) {
			if(no_xl)
				goto next;
			else {
				yt = yc;
				state2 = bottom_border;
			}
		} else if(state2 == bottom_border) {
			if(no_xl) {
				no_yb = T_FALSE;
				yb = yc;
				state2 = maybe;
				goto next;
			}
		} else {
			if(no_xl)
				goto next;
			else {
				state2 = bottom_border;
				no_yb = T_TRUE;
			}
		}

		/* non-transparent lines get here */
		if(no_xr) 
			xr = x2;
		if(xl < gxl)
			gxl = xl;
		if(xr > gxr) 
			gxr = xr;

next:
		data += pitch;
	}

	if(no_yb)
		yb = y2;

	if(gxr <= gxl  || yb <= yt)
		/* transparent bitmap, finished */
		return T_RET_TRANS;

	/* store rectangle */
	out->x = gxl;
	out->width = gxr-gxl;
	out->y = yt;
	out->height = yb-yt;

	return T_OK;
}

static int thinc_getBitmapRect32(thincScreenPtr pThinc, char *data,
	xRectangle *area, register unsigned long fg, xRectangle *out,
	unsigned int pitch)
{
	register unsigned char state1, state2;
	register unsigned int xc, x2;
	unsigned char no_xr, no_xl, no_yb;
	unsigned int yc, y2, gxl, gxr, xl, xr, yt, yb;
	register CARD32 *pix;

	x2 = area->x + area->width;
	y2 = area->y + area->height;

	/* init global borders */
	gxl = x2;
	gxr = area->x;

	state2 = top_border;
	no_yb = T_TRUE;
	yt = y2;
	yb = area->y;

	for(yc = area->y; yc < y2; ++yc) {
		pix = (CARD32*) data;
		xl = xr = area->x;
		no_xr = no_xl = T_TRUE;
		state1 = left_border;
		for(xc = area->x; xc<x2; ++xc, ++pix) {
			if(state1 == left_border) {
				if(*pix == fg) {
					state1 = right_border;
					xl = xc;
					no_xl = T_FALSE;
				}
			} else if(state1 == right_border) {
				if(*pix != fg) {
					xr = xc;
					no_xr = T_FALSE;
					state1 = maybe;
				}
			} else {
				if(*pix == fg) {
					no_xr = T_TRUE;
					state1 = right_border;
				}
			}

		}
		if(state2 == top_border) {
			if(no_xl)
				goto next;
			else {
				yt = yc;
				state2 = bottom_border;
			}
		} else if(state2 == bottom_border) {
			if(no_xl) {
				no_yb = T_FALSE;
				yb = yc;
				state2 = maybe;
				goto next;
			}
		} else {
			if(no_xl)
				goto next;
			else {
				state2 = bottom_border;
				no_yb = T_TRUE;
			}
		}

		/* non-transparent lines get here */
		if(no_xr) 
			xr = x2;
		if(xl < gxl)
			gxl = xl;
		if(xr > gxr) 
			gxr = xr;

next:
		data += pitch;
	}

	if(no_yb)
		yb = y2;

	if(gxr <= gxl  || yb <= yt)
		/* transparent bitmap, finished */
		return T_RET_TRANS;

	/* store rectangle */
	out->x = gxl;
	out->width = gxr-gxl;
	out->y = yt;
	out->height = yb-yt;

	return T_OK;
}
