/* Copyright 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for Buffering of Messages
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/* TODO: Add checking for command buffering option and call Flush() directly
 * if it's disabled */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include <stdarg.h>
#include <sys/poll.h>

static int thincBufferBatchEndCl(thincScreenPtr pThinc, thincClientPtr cl);
static int thincBufferBatchStartCl(thincScreenPtr pThinc, thincClientPtr cl,
	RegionPtr reg);
static int thincBufferCmdForReal(thincScreenPtr pThinc, WindowPtr pWin,
	thincCmdPtr cmd, int docopy);
static int thincBufferNewCmdReal(thincScreenPtr pThinc, WindowPtr pWin,
	int type, RegionPtr reg, va_list args);
static int thincBufferCmdClient(thincScreenPtr pThinc, thincClientPtr cl, 
	thincCmdPtr cmd);
static int thincSendPing(thincClientPtr cl);

static int thincBufferCmdClientNotBatch(thincScreenPtr pThinc, 
	thincClientPtr cl, thincCmdPtr cmd);
static int thincBufferCmdClientBatch(thincScreenPtr pThinc, thincClientPtr cl,
	thincCmdPtr cmd);

static int thincBufferCopy(thincClientPtr cl, thincCmdPtr cmd);

/* for thincback */
static void thincDrawStatsReg(thincScreenPtr pThinc, RegionPtr reg);
static void thincDrawStats(thincScreenPtr pThinc, BoxPtr box);

/* call this when you're going to buffer a number of commands at once
 * (for example, ExecutePixmapQueue does this). The idea is that when doing
 * this we have no need to check for merge and overwrite on every command
 * that's buffered, since that has already been checked in the source.
 * What do we do:
 * - first set up some vars to let BufferCmd know that this is happening
 * - then use the passed region to overwrite all commands in the buffer
 *   at once.
 *
 * Finally, don't forget to call thincBufferEnd() when you're done, or the
 * buffer will think that none of the commands need to be overwritten or
 * merged :)
 */
int thincBufferBatchStart(thincScreenPtr pThinc, RegionPtr reg)
{
	t_dbg_name(func)
	thincClientPtr cl;
	int res;

	thincDbg(T_DBG_CLIENT, ("%s: Preparing buffer for batch buffering\n",
		func));
	call_for_all(thincBufferBatchStartCl, (pThinc, cl, reg));

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thincBufferBatchStartCl(thincScreenPtr pThinc, thincClientPtr cl,
	RegionPtr reg)
{
	t_dbg_name(func)
	cl->buffer_batch = 1;

	if(thincOverwriteCmds(pThinc, &cl->cmdBuffer, reg) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't overwrite "
			"commands in client buffer\n", func));
		return T_ERROR;
	}

	return T_OK;
}

int thincBufferBatchEnd(thincScreenPtr pThinc)
{
	thincClientPtr cl;
	int res;
	call_for_all(thincBufferBatchEndCl, (pThinc, cl));
	return T_OK;
}

/* this really can't fail, but call_for_all expects something to be returned.
 * boooo (is it ok to complain about my own code??) */
static int thincBufferBatchEndCl(thincScreenPtr pThinc, thincClientPtr cl)
{
	cl->buffer_batch = 0;
	return T_OK;
}

/* Buffer a command for all connected clients. Only thing it needs to do
 * is clip the command for each client's clip region. Called by the
 * Execute() function for the command. 
 * Doesn't really fail (if buffering the command fails we just kill
 * the client :) */
/* TODO: Since we make a copy of the command, then allow the Execute()
 * functions to pass down the clip and deltas, even though it doesn't
 * really correspond to the semantics of this function */
int thincBufferCmd(thincScreenPtr pThinc, WindowPtr pWindow, thincCmdPtr cmd)
{
	t_dbg_name(func)

	thincDbg(T_DBG_CLIENT, ("%s: buffering command %#lx\n", func,
		cmd->id));
	if(cmd->type != T_FB_CURSOR) {
		thincDrawStats(pThinc, &cmd->area);
	}
	if(pThinc->clientCount == 0) {
		thincDbg(T_DBG_CLIENT, ("%s: Nobody connected, no need to "
			"buffer anything\n", func));
		return T_OK;
	}

	thincBufferCmdForReal(pThinc, pWindow, cmd, T_TRUE);
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

/* same as above, but doesn't require the buffer to make a copy of the
 * command if it doesn't have to. will return T_COPIED if it had to make
 * a copy anyway */
int thincBufferCmdNoCopy(thincScreenPtr pThinc, WindowPtr pWindow, 
	thincCmdPtr cmd)
{
	if(cmd->type != T_FB_CURSOR) {
		thincDrawStats(pThinc, &cmd->area);
	}
	/*  XXX: temporary hack. we return T_COPIED so that caller
	 *  destroys it since we didn't do anything with it */
	if(pThinc->clientCount == 0) 
		return T_COPIED;

	thincBufferCmdForReal(pThinc, pWindow, cmd, T_FALSE);

	if(pThinc->clientCount > 1)
		return T_COPIED;
	return T_OK;
}

int thincBufferNewCmd(thincScreenPtr pThinc, WindowPtr pWin, int type, 
	int x, int y, int w, int h, ...)
{
	RegionRec reg;
	BoxRec box;
	va_list args;
	int ret;

	box.x1 = x;
	box.y1 = y;
	box.x2 = x+w;
	box.y2 = y+h;

	va_start(args, h);
	REGION_INIT(pThinc->pScreen, &reg, &box, 1);
	ret = thincBufferNewCmdReal(pThinc, pWin, type, &reg, args);
	REGION_UNINIT(pThinc->pScreen, &reg);
	va_end(args);

	return ret;
}

int thincBufferNewCmdReg(thincScreenPtr pThinc, WindowPtr pWin, int type, 
	RegionPtr reg, ...)
{
	va_list args;
	int ret;

	va_start(args, reg);
	ret = thincBufferNewCmdReal(pThinc, pWin, type, reg, args);
	va_end(args);

	return ret;
}

static int thincBufferNewCmdReal(thincScreenPtr pThinc, WindowPtr pWin,
	int type, RegionPtr reg, va_list args)
{
	t_dbg_name(func)
	thincCmdPtr cmd;

	thincDbg(T_DBG_CLIENT, ("%s: buffering command of type %d\n", func,
		type));
	print_reg(T_DBG_CLIENT, pThinc->pScreen, reg, "region of new command:");

	if(type != T_FB_CURSOR) {
		thincDrawStatsReg(pThinc, reg);
	}
	/* check that somebody's connected before we do all this */
	if(pThinc->clientCount == 0) {
		thincDbg(T_DBG_CLIENT, ("%s: Nobody connected, no need to "
			"buffer anything\n", func));
		return T_OK;
	}

	if(region_is_empty(reg)) {
		thincDbg(T_DBG_CLIENT, ("%s: Command has an empty region. "
			"finished\n", func));
		return T_OK;
	}

	/* create new command */
	thincDbg(T_DBG_CLIENT, ("%s: creating new command\n", func));
	cmd = thincCreateCmd(pThinc, type, reg, args);
	if(!cmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create new "
			"command\n", func));
		return T_ERROR;
	}

	/* now buffer it in each client's queue. We cheat here and
	 * optimize things so that if there's only 1 client connected
	 * we don't make a copy of the command and call the buffering
	 * function directly. If there's more than 1 client connected
	 * we do need to make a copy for each client and then destroy
	 * ours */
	thincDbg(T_DBG_CLIENT, ("%s: Have %d clients connected, do we need "
		"to copy?\n", func, pThinc->clientCount));
	if(pThinc->clientCount == 1) {
		thincBufferCmdForReal(pThinc, pWin, cmd, T_FALSE);
	} else {
		thincBufferCmdForReal(pThinc, pWin, cmd, T_TRUE);
		thincDestroyCmd(pThinc, cmd);
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

/* This is the main buffering function, where we make all the decisions.
 * Nothing should go around us 
 * IMPORTANT: It assumes that there's clients where we should buffer
 * stuff. check for this before calling this function or you'll die */
static int thincBufferCmdForReal(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, int docopy)
{
	t_dbg_name(func)
	thincClientPtr cl;
	int res;

	thincDbg(T_DBG_CLIENT, ("%s: Buffering command %#lx coming from "
		"window\n", func, cmd->id));

	/* check if we're buffering commands */
	if(!pThinc->doCmdBuffer) {
		/* nope. flush it on all clients */
		call_for_all(thincFlushCmd, (pThinc, cl, cmd));
		return T_OK;
	}

	if(docopy) {
		/* call_for_all handles closing the client if an error
		 * occurs */
		call_for_all(thincBufferCmdClient, (pThinc, cl, cmd));
	} else {
		/* normally here because there's only one client */
		cl = t_first_client(pThinc);
		res = cl->buffer(cl, cmd);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer "
				"command in client's queue\n", func));
			cl->close_cl(cl);
		}
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

/* create a copy of the command for this client, buffer it. If that
 * fails, destroy the copy and let caller handle the rest. If everything
 * is okay, just return. The copy will be destroyed when the command
 * is flushed or purged */
static int thincBufferCmdClient(thincScreenPtr pThinc, thincClientPtr cl, 
	thincCmdPtr cmd)
{
	int ret;
	thincCmdPtr cmd2 = thincCopyCommand(pThinc, cmd);
	if(!cmd2) 
		return T_ERROR;

	ret = cl->buffer(cl, cmd2);
	if(ret == T_ERROR)
		thincDestroyCmd(pThinc, cmd2);

	return ret;
}

/* if the command is COPY we need to treat it especially: commands that
 * are before it in the queue, and draw to the source region, shouldn't be
 * overwritten by subsequent commands or the copy will be erroneous
 * (for example, in xterm when a command makes the output scroll, the
 * following happens:
 * => some bilevel cmds are generated for the initial part of the output
 * => a copy is made to scroll up these commands.
 * => a fill and some other bilevels are generated for the rest of
 * the output. 
 * The last commands draw to the same region as the first ones. However,
 * the first ones shouldn't be overwritten since their output is now
 * somewhere else as part of the copy. Note that changing these commands
 * output to reflect their final destination won't work since the
 * copy has to be made anyway: any additional data that's already drawn
 * and is not overwritten by the new commands needs to be copied too.
 */
#if 0
int thincBufferCmdClientReal(thincClientPtr cl, thincCmdPtr cmd)
{
	const char *func = __func__;
	RegionPtr clip;
	RegionPtr reg;
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: buffering command %#lx in client %d\n",
		func, cmd->id, cl->sock));

	thincDbg(T_DBG_CLIENT, ("%s: clipping command to client's viewport\n",
		func));

	clip = &cl->clClip;

	/* if copy need to treat specially */
	cmd = thincModifyCommand(pThinc, cmd, T_CMD_CLIP, clip);
	if(cmd == (thincCmdPtr) T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't clip command "
			"to client's clipping region\n", func));
		return T_ERROR;
	} else if(!cmd) {
		thincDbg(T_DBG_CLIENT, ("%s: Command draws outside of the "
			"client's clip region. Nothing to buffer\n", func));
		return T_OK;
	}

	thincDbg(T_DBG_CLIENT, ("%s: clipped command, checking if it "
		"overwrites any other commands\n", func));

	/* if COPY, mark any commands in the source region as untouchable.
	 * Do this before overwritting them with the copy's dst region */
	if((!t_list_is_empty(&cl->cmdBuffer)) && (cmd->type == T_FB_COPY) ) {
		if(thincCmdSaveCopy(pThinc, &cl->cmdBuffer, cmd) != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't fix "
				"commands to account for new COPY\n", func));
			return T_ERROR;
		}
	}

	/* If the new command is not opaque, then no commands will be 
	 * overwritten. For example, a GLYPH only writes in parts of the 
	 * region it affects. The rest is transparent */
	if(cmd->flags & T_CMD_TRANS) {
		thincDbg(T_DBG_CLIENT, ("%s: new command is transparent, no "
			"need to check for overwriting\n", func));
	} else {
		thincCmdGetRegion(pThinc, cmd, &reg);
		if(!reg) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get "
				"command's draw region\n", func));
			return T_ERROR;
		}
		if(thincOverwriteCmds(pThinc,&cl->cmdBuffer,reg) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't overwrite "
				"commands already in the queue\n", func));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_CLIENT, ("%s: ok, ready to add it to the queue. "
		"Trying to merge it first\n", func));
	/* ok, now add it to the buffer */
	if(!thincTryMergeCmd(pThinc, &cl->cmdBuffer, cmd)) {
		/* bummer.. oh, well, just add it to the list */
		thincDbg(T_DBG_CLIENT, ("%s: Couldn't merge it, adding it "
			"to the tail of the queue\n", func));
		if(thincAddCommand(cl, cmd) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add "
				"command to client's buffer\n", func));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}
#else
int thincBufferCmdClientReal(thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)
	RegionPtr clip;
	thincScreenPtr pThinc = cl->parent;
	int err;

	thincDbg(T_DBG_CLIENT, ("%s: buffering command %#lx in client %s\n",
		func, cmd->id, cl->id));

	clip = &cl->clClip;

	thincDbg(T_DBG_CLIENT, ("%s: clipping command to client's viewport\n",
		func));

	cmd = thincModifyCommand(pThinc, cmd, T_CMD_CLIP, clip);
	if(cmd == (thincCmdPtr) T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't clip command "
			"to client's clipping region\n", func));
		return T_ERROR;
	} else if(!cmd) {
		thincDbg(T_DBG_CLIENT, ("%s: Command draws outside of the "
			"client's clip region. Nothing to buffer\n", func));
		return T_OK;
	}

	thincDbg(T_DBG_CLIENT, ("%s: clipped command, checking if it "
		"overwrites any other commands\n", func));

#if 0
	/* if COPY, mark any commands in the source region as untouchable.
	 * Do this before overwritting them with the copy's dst region */
	if((!t_list_is_empty(&cl->cmdBuffer)) && (cmd->type == T_FB_COPY) ) {
		if(thincFlushCmdAll(pThinc) == T_PART) {
			thincError("PART PART PART %#lx\n", cmd->id);
		}
		thincFlushBufferAll(pThinc);
		if(thincCmdSaveCopy(pThinc, &cl->cmdBuffer, cmd) != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't fix "
				"commands to account for new COPY\n", func));
			return T_ERROR;
		}
	}
#else
	/* special treatment */
	if(cmd->type == T_FB_COPY) {
		err = thincBufferCopy(cl, cmd);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer "
				"COPY command\n", func));
			return T_ERROR;
		}
	}
#endif

	if(cl->buffer_batch) {
		err = thincBufferCmdClientBatch(pThinc, cl, cmd);
	} else {
		err = thincBufferCmdClientNotBatch(pThinc, cl, cmd);
	}
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer command\n",
			func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thincBufferCopy(thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)
	int err;
	thincScreenPtr pThinc = cl->parent;
	thincCmdPtr other;

	thincDbg(T_DBG_CLIENT, ("%s: %#lx\n", func, cmd->id));

	if(t_list_is_empty(&cl->cmdBuffer)) {
		thincDbg(T_DBG_CLIENT, ("%s: command buffer is empty, "
			"finished\n", func));
		return T_OK;
	}

	thincDbg(T_DBG_CLIENT, ("%s: checking if we can merge new copy "
		"with an existing one\n", func));

	other = t_list_first_entry(&cl->cmdBuffer, thincCmdRec, list);
	if(other->type == T_FB_COPY) {
		thincDbg(T_DBG_CLIENT, ("%s: have outstanding COPY command "
			"%#lx, trying to merge them\n", func, other->id));
		err = thincCopyMerge(pThinc, other, cmd);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't merge "
				"COPY commands\n", func));
			return T_ERROR;
		} else if(err == T_TRUE) {
			thincDbg(T_DBG_CLIENT, ("%s: merged them!\n", func));
			thincDeleteCmdAll(pThinc, other);
			t_list_len_dec(&cl->cmdBuffer);
		} 
	}

	thincDbg(T_DBG_CLIENT, ("%s: flushing buffers\n", func));

	err = thincFlushCmdAll(pThinc);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush commands "
			"from buffer before buffering COPY\n", func));
		return T_ERROR;
	} else if(err == T_PART) {
		/* FIXME: If we can't clean up the buffer we're a little
		 * bit screwed i think. but let's see if this helps.
		 * We also need to make sure that the scheduler doesn't
		 * put it in front of any of the saved commands. ugh */
		if(thincCmdSaveCopy(pThinc, &cl->cmdBuffer, cmd) != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't fix "
				"commands to account for new COPY\n", func));
			return T_ERROR;
		}
	} else {
		err = thincFlushBufferAll(pThinc);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush "
				"commands "
				"from buffer before buffering COPY\n", func));
			return T_ERROR;
		}
	}
	
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thincBufferCmdClientBatch(thincScreenPtr pThinc, thincClientPtr cl,
	thincCmdPtr cmd) 
{
	return thincAddCommand(cl, cmd);
}

static int thincBufferCmdClientNotBatch(thincScreenPtr pThinc, 
	thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)
	RegionPtr reg;

	thincDbg(T_DBG_CLIENT, ("%s: first thing: try to merge the command\n",
		func));

	if(thincTryMergeCmd(pThinc, &cl->cmdBuffer, cmd)) {
		/* all done! */
		thincDbg(T_DBG_CLIENT, ("%s: merged new command, i'm done, "
			"finished, the end, finito, dat's it, bye bye\n", 
			func));
		return T_OK;
	}

	thincDbg(T_DBG_CLIENT, ("%s: Couldn't merge it, adding it "
		"to the tail of the queue then\n", func));

	/* If the new command is not opaque, then no commands will be 
	 * overwritten. For example, a GLYPH only writes in parts of the 
	 * region it affects. The rest is transparent */
	if(cmd->flags & T_CMD_TRANS) {
		thincDbg(T_DBG_CLIENT, ("%s: new command is transparent "
			"(type %d), no need to check for overwriting\n", func,
			cmd->type));
	} else {
		thincCmdGetRegion(pThinc, cmd, &reg);
		if(!reg) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get "
				"command's draw region\n", func));
			return T_ERROR;
		}
		if(thincOverwriteCmds(pThinc,&cl->cmdBuffer,reg) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't overwrite "
				"commands already in the queue\n", func));
			return T_ERROR;
		}
	}

	/* ok, now add it to the buffer */
	if(thincAddCommand(cl, cmd) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add "
			"command to client's buffer\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}
#endif

/* flushes command buffer of all connected clients. On error, clients are
 * disconnected. */
int thincFlushCmdAll(thincScreenPtr pThinc)
{
	thincClientPtr cl, t;
	int res, part = T_FALSE;
	unsigned int now;

	gettime_ms(now);

	t_list_for_each_entry(cl, t, thincClientRec, &pThinc->clients, link) {
		if(!t_cl_connected(cl)) {
			continue; 
		} 
#ifdef T_RECORD
		if(now - cl->last_flush < pThinc->rate_ms) {
			/* so we set timer */
			part = T_TRUE;
			continue;
		}
#endif
		/* function call */ 
		res = thincFlushCmdClient(pThinc, cl);
#ifdef T_RECORD
		cl->last_flush = now;
#endif
		if(res == T_ERROR) 
			cl->close_cl(cl); 
		else if(res == T_PART) {
			part = T_TRUE;
			break;
		}
	}

	if(part)
		return T_PART;
	else
		return T_OK;
}

/* Checks if we can write data to this client. If yes, tries to flush
 * as much data as possible. If not, it just returns. */
int thincFlushCmdClient(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_ANNOY, ("%s: flushing client %s's buffer\n", func,
		cl->id));

	err = cl->flush(pThinc, cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush client's "
			"buffer\n", func));
		return T_ERROR;
	} else if(err == T_PART) {
		/* can't write to this client now */
		thincDbg(T_DBG_ANNOY, ("%s: Client is busy, can't flush its "
			"buffer. finished\n", func));
		return T_PART;
	}

	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));

	return T_OK;
}

int thincFlushCmds(thincScreenPtr pThinc, thincClientPtr cl)
{
	int err = thincDoPoll(cl->sock, POLLOUT, 0);
	if(err == T_ERROR) {
		return T_ERROR;
	} else if(err == T_TIMEOUT) {
		/* can't write to this client now */
		return T_PART;
	}
	if(t_list_is_empty(&cl->cmdBuffer)) 
		return T_OK;

	return thincFlushListBuffer(pThinc, cl, &cl->cmdBuffer);
}

int thincFlushListBuffer(thincScreenPtr pThinc, thincClientPtr cl, 
	thincListPtr buffer)
{
	t_dbg_name(func)
	register thincCmdPtr cmd, next;
	int err;

	stats_dbg(("Flush:Buffer:Client[%s]: len %u max_len %u\n",
		cl->id, t_list_len(buffer), t_list_maxlen(buffer)));
	t_list_for_each_entry(cmd, next, thincCmdRec, buffer, list) {
		thincDbg(T_DBG_ANNOY, ("%s: flushing command %#lx\n", 
			func, cmd->id));

		err = thincFlushCmd(pThinc, cl, cmd);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush "
				"command %#lx from client %s's queue\n", 
				func, cmd->id, cl->id));
			return T_ERROR;
		} else if(err == T_DONE) {
			/* command flushed completely, take it out */
			thincDeleteCmdList(pThinc, cmd);
			t_list_len_dec(buffer);
			continue;
		} else if(err == T_PART) {
			/* if command not completely flushed, then we can't
			 * write any more data. get out */
			thincDbg(T_DBG_ANNOY, ("%s: command not completely "
				"flushed. returning\n", func));
			break;
		} else { /* huh? */
			T_BUG_ARG("Unknown return value %d from Flush()",
				err);
			thincDeleteCmdList(pThinc, cmd);
			t_list_len_dec(buffer);
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));
	return T_OK;
}

int thincFlushListSched(thincScreenPtr pThinc, thincClientPtr cl, 
	thincListPtr queue)
{
	t_dbg_name(func)
	register thincCmdPtr cmd, next;
	int err;

	t_list_for_each_entry(cmd, next, thincCmdRec, queue, sched) {
		thincDbg(T_DBG_ANNOY, ("%s: flushing command[%d] %#lx\n", 
			func, cmd->type, cmd->id));

		err = thincFlushCmd(pThinc, cl, cmd);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush "
				"command %#lx from client %s's queue\n", 
				func, cmd->id, cl->id));
			return T_ERROR;
		} else if(err == T_DONE) {
			/* command flushed completely, take it out */
			thincDeleteCmdSched(pThinc, cmd);
			t_list_len_dec(queue);
			t_list_len_dec(&cl->cmdBuffer);
			continue;
		} else if(err == T_PART) {
			/* if command not completely flushed, then we can't
			 * write any more data. get out */
			thincDbg(T_DBG_ANNOY, ("%s: command not completely "
				"flushed. returning\n", func));
			break;
		} else { /* huh? */
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Unknown return "
				"value %d from Flush() function\n", func,
				err));
			thincDeleteCmdSched(pThinc, cmd);
			t_list_len_dec(queue);
			t_list_len_dec(&cl->cmdBuffer);
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));
	return T_OK;
}

int thincFlushCmd(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
	int index = thincCmd_idx(cmd->type);
	cmdFlushFunc flush = pThinc->cmdHandlerList[index]->cmdFlush;

	return (*flush)(pThinc, cl, cmd);
}

/* Important, important: the ping functions assume there's nothing buffered
 * and waiting to be sent. There's no point in buffering pings, that who
 * knows when they will be sent. Check the buffers before calling us! */
int thincSendPingAll(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincClientPtr cl;
	int res;

	thincDbg(T_DBG_ANNOY, ("%s: sending ping to clients\n", func));
	call_for_all(thincSendPing, (cl));
	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));
	return T_OK;
}

static int thincSendPing(thincClientPtr cl)
{
	if(thincSendMsg(cl, T_SPING, T_SPING_SIZE, 0) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send PING to "
			"client\n", __func__));
		return T_ERROR;
	} else 
		return T_OK;
}

OsTimerPtr thincInitATimer(thincScreenPtr pThinc)
{
	OsTimerPtr timer = TimerSet(NULL, 0, 0, NULL, NULL);
	if(!timer)
		return NULL;
	else 
		return timer;
}

OsTimerPtr thincSetATimer(thincScreenPtr pThinc, OsTimerPtr timer, int flags, 
	int millis, OsTimerCallback func, pointer arg)
{
	timer = TimerSet(timer, flags, millis, func, arg);
	if(!timer)
		return NULL;
	else
		return timer;
}

int thincInitTimer(thincScreenPtr pThinc)
{
	pThinc->timer = thincInitATimer(pThinc);
	if(!pThinc->timer)
		return T_ERROR;
	else
		return T_OK;
}

int thincSetTimer(thincScreenPtr pThinc, int flags, int millis, 
	OsTimerCallback func, pointer arg)
{
	pThinc->timer = thincSetATimer(pThinc, pThinc->timer, flags, 
		millis, func, arg);
	if(!pThinc->timer)
		return T_ERROR;
	else
		return T_OK;
}

#ifdef T_NODRAW
/* Buffer function for client0. the idea is to make this operation as
 * simple as possible. So for example, we don't check for overlaps among
 * commands. Instead, we wait until the system is idle (e.g. when
 * SaveScreen is called) and come clean things up (or when a new client gets 
 * connected).
 * As a result, it doesn't seem necessary for client0 to do scheduling, 
 * so just add it directly to the command queue */
/* FIXME: Copies don't work with this model */
int thincClient0Buffer(thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)

	thincDbg(T_DBG_CLIENT, ("%s: adding command to client0's buffer\n",
		func));
	return thincAddCmdList(&cl->cmdBuffer, cmd);
}

/* Flush function for client0. Don't do anything, as we need to
 * preserve all commands relevant to the state of the fb */
int thincClient0Flush(thincScreenPtr pThinc, thincClientPtr cl)
{
	/* if nobody connected, then let's purge the buffer, to minimize
	 * startup time for new clients */
	if(pThinc->clientCount == 1) {
		return thincClient0Purge(pThinc);
	} else 
		return T_OK;
}

/* flush client0's buffer into cl's socket ... kinky */
int thincClient0SendFB(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	int err;
	thincClientPtr cl0 = pThinc->client0;
	thincCmdPtr cmd;

	thincDbg(T_DBG_CLIENT, ("%s: purging client0's buffer before "
		"sending it to new client\n", func));

	err = thincClient0Purge(pThinc);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't purge client0's "
			"buffer, we're screwed\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: sending client0's buffer to client "
		"%s\n", func, cl->id));

	t_list_for_each_entry_fast(cmd, thincCmdRec, &cl0->cmdBuffer, list) {
again:
		thincDbg(T_DBG_CLIENT, ("%s: flushing command %#lx\n", func,
			cmd->id));
		err = thincFlushCmd(pThinc, cl, cmd);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush "
				"initial command %#lx\n", func, cmd->id));
			return T_ERROR;
		} else if(err == T_PART) {
			/* FIXME: We should punt for now and try later
			 * as this will block the server */
			thincDbg(T_DBG_CLIENT, ("%s: command not completely "
				"flushed, trying again\n", func));
			goto again;
		} 
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}
#endif	/* T_NODRAW */

/* we allow to pass just a box as a convenience function for callers that
 * have no easy access to the region (for example if you get passed a cmd
 * you don't really know the region unless you ask for it). The drawback
 * of this is that box may not be the precise area where a command
 * is drawing. we dont care. */
static void thincDrawStats(thincScreenPtr pThinc, BoxPtr box)
{
	RegionRec reg;

	REGION_INIT(pThinc->pScreen, &reg, box, 1);
	thincDrawStatsReg(pThinc, &reg);
	REGION_UNINIT(pThinc->pScreen, &reg);
}

/* this is more precise, as we only take into account the actual rectangles,
 * not the extents of the region. funky */
static void thincDrawStatsReg(thincScreenPtr pThinc, RegionPtr reg)
{
#if 0
	pThinc->display_count++;
	REGION_UNION(pThinc->pScreen, &pThinc->display_change,
		&pThinc->display_change, reg);
#ifdef T_RECORD
	if(pThinc->doRecord) {
		thincRecordPtr record = t_recordPtr(pThinc); 
		record->display_count++;
		REGION_UNION(pThinc->pScreen, &record->display_change,
				&record->display_change, reg);
	}
#endif
#else
	thincIncDisplayCount(pThinc);
	thincAddDisplayChange(pThinc, reg);
#endif
}
