/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Client handling functions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <zlib.h>

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#include "xf86.h"
#include "xf86_OSproc.h"

/* local functions */
static thincClientPtr thincNewClient(thincScreenPtr pThinc);
static int thincHandleClientNormal(thincClientPtr cl, int msg_type);

/* connection functions */
static int thincNewClientv2(thincScreenPtr pThinc, thincClientPtr cl);
static int thincClientConnected(thincClientPtr cl);
static int thincZendInitialFB(thincClientPtr cl);

/* Handle new thincClient connections */
int thincNewConnection(thincScreenPtr pThinc)
{
	thincClientPtr cl;
	t_dbg_name(func)
	BoxRec clip;

	thincDbg(T_DBG_CLIENT, ("%s\n", func));

	cl = thincNewClient(pThinc);
	if(!cl) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize new "
			"client\n", func));
		return T_ERROR;
	}

	/* add to list */
	t_list_add_head(&cl->link, &pThinc->clients);
	thincAddSocket(pThinc, cl->sock);

	/* initialize clipping region */
	/* TODO: Support different clipping region per client */
	clip.x1 = 0;
	clip.y1 = 0;
	clip.x2 = clip.x1 + pThinc->width;
	clip.y2 = clip.y1 + pThinc->height;
	REGION_INIT(pThinc->pScreen, &cl->clClip, &clip, 1);
	cl->clipIsSub = T_FALSE;

	/* init the read buffer */
	cl->rBufStart = NULL;
	cl->rBufSize = 0;
	memset(cl->rBuffer, 0, T_READ_BUFLEN);

	/* init the write buffer */
	t_list_init(&cl->cmdBuffer);
	cl->buffer_batch = 0;

	/* init scheduler */
	if(pThinc->doSched) {
		if(thincSchedInit(cl) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"initialize client's command scheduler. "
				"New client rejected\n", func));
			cl->close_cl(cl);
			return T_ERROR;
		}
		cl->last = 0;
		cl->flush = thincFlushScheduler;
	} else
		cl->flush = thincFlushCmds;

	if(pThinc->doWriteBuffer) {
		cl->msgBufMaxSize = T_WRITE_BUFLEN;
		cl->msgBufSize = 0;
		cl->msgBufP = cl->msgBuffer;
		memset(cl->msgBuffer, 0, T_WRITE_BUFLEN);
	} else {
		cl->msgBufMaxSize = -1;
		cl->msgBufSize = 0;
		cl->msgBufP = NULL;
	}

	/* init resize */
	cl->resized = T_CL_RESZ_NONE;
	if(pThinc->allowResize) {
		cl->curr_width = pThinc->width;
		cl->curr_height = pThinc->height;
		cl->factor_x = cl->factor_y = 1.0;
		/* TODO: Eventually base this on the client's bpp */
		switch(pThinc->bpp) {
			case 8:
				cl->resize = thinc_resize8;
				break;
			case 16:
				cl->resize = thinc_resize16;
				break;
			case 24:
			case 32:
				cl->resize = thinc_resize32;
				break;
			default:
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Color "
					"depth %d not supported for "
					"resizing\n", func, pThinc->bpp));
				cl->close_cl(cl);
				return T_ERROR;
		}
	} else {
		cl->curr_width = cl->curr_height = 0;
		cl->factor_x = cl->factor_y = 0.0;
		cl->resize = NULL;
	}

	/* init cache */
	if(pThinc->useRawCache) {
		cl->raw_cache = thinc_cache_init(pThinc->raw_cache_bits);
		if(!cl->raw_cache) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"initialize client's image cache\n", func));
			cl->close_cl(cl);
			return T_ERROR;
		}
	}
	if(pThinc->useBitCache) {
		cl->bit_cache = thinc_cache_init(pThinc->bit_cache_bits);
		if(!cl->bit_cache) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"initialize client's bitmap cache\n", func));
			cl->close_cl(cl);
			return T_ERROR;
		}
	}
	if(pThinc->usePixCache) {
		cl->pix_cache = thinc_cache_init(pThinc->pix_cache_bits);
		if(!cl->pix_cache) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"initialize client's pixmap cache\n", func));
			cl->close_cl(cl);
			return T_ERROR;
		}
	}

	/* initialize message stacks */
	cl->rawStack = NULL;
	cl->bitmapStack = NULL;
	cl->pixmapStack = NULL;
	cl->fillStack = NULL;
	cl->copyStack = NULL;
	if(thinc_rawInitStack(cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize RAW "
			"message stack\n", func));
		cl->close_cl(cl);
		return T_ERROR;
	}
	if(thinc_bitmapInitStack(cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"BITMAP message stack\n", func));
		cl->close_cl(cl);
		return T_ERROR;
	}
	if(thinc_pixmapInitStack(cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"PIXMAP message stack\n", func));
		cl->close_cl(cl);
		return T_ERROR;
	}
	if(thinc_fillInitStack(cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"SFILL message stack\n", func));
		cl->close_cl(cl);
		return T_ERROR;
	}

	if(thinc_copyInitStack(cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"COPY message stack\n", func));
		cl->close_cl(cl);
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static thincClientPtr thincNewClient(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincClientPtr cl;
	int sock, one, err;
	struct sockaddr_in addr;
	socklen_t addr_len;

	thincDbg(T_DBG_CLIENT, ("%s: processing new client connection\n",
		func));

	/* do accept first so that the client doesn't get stuck
	 * waiting if we fail at something else */
	addr_len = sizeof(addr);
	sock = accept(pThinc->listenSocket, (struct sockaddr*)&addr, 
		&addr_len);
	if (sock == -1) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't accept() new "
			"client: %s\n", func, strerror(errno)));
		return NULL;
	}

	thincDbg(T_DBG_CLIENT, ("%s: got new client from %s:%d\n", func,
		inet_ntoa(addr.sin_addr), addr.sin_port));

	/* set it non-blocking */
	if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't set non-blocking "
			"flag: %s\n", func, strerror(errno)));
		tcpThincClose(sock);
		return NULL;
	}

	/* disable nagle's algorithm if we're doing write buffering */
	if(pThinc->doWriteBuffer) {
		one = 1;
		if(setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&one, 
			sizeof(one)) < 0) {
			thincWarn("Warning: couldn't disable Nagle's "
				"algorithm. Your performance may suffer\n");
		}
	}

	cl = calloc(1, sizeof(thincClientRec));
	if (!cl) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create new "
			"connection: %s\n", func, strerror(errno)));
		tcpThincClose(sock);
		return NULL;
	}

	cl->sock = sock;
	cl->parent = pThinc;
	cl->print_msg = T_TRUE;

	strncpy(cl->id, inet_ntoa(addr.sin_addr), T_CLID_LEN);
	cl->id[T_CLID_LEN-1] = '\0';	/* paranoia! */

	err = thincNewClientv2(pThinc, cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"new client\n", func));
		tcpThincClose(cl->sock);
		free(cl);
		return NULL;
	}

	cl->buffer = thincBufferCmdClientReal;
	cl->close_cl = thincCloseClient;
	cl->send_msg = thincSendMsg;

	thincDbg(T_DBG_CLIENT, ("%s: finished. new client is %s\n",
		func, cl->id));

	return cl;
}

/* add socket to select(2) list */
void thincAddSocket(thincScreenPtr pThinc, int sock)
{
	FD_SET(sock, &(pThinc->fds));
	if (sock >= pThinc->fdmax)
		pThinc->fdmax = sock + 1;
	AddEnabledDevice(sock);
}

void thincRemoveSocket(thincScreenPtr pThinc, int sock)
{
	RemoveEnabledDevice(sock);
	FD_CLR(sock, &pThinc->fds);
}

int thincHaveNewClients(thincScreenPtr pThinc, fd_set *socks)
{
	t_dbg_name(func)

	if (FD_ISSET(pThinc->listenSocket, socks)) {
		thincDbg(T_DBG_ANNOY, ("%s: got a new connection\n", func));
		/* have new connection, handle it */
		if (!thincNewConnection(pThinc)) {
			/* not fatal */
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't accept "
				"new client connection\n", func));
			return T_ERROR;
		}
		return T_YES;
	} else {
		thincDbg(T_DBG_ANNOY, ("%s: finished no new connections\n", 
			func));
		return T_NO;
	}
}

int thincCheckClients(thincScreenPtr pThinc, fd_set *socks, int num)
{
	t_dbg_name(func)
	thincClientPtr cl, t;
	int err;

	thincDbg(T_DBG_ANNOY, ("%s: %d clients to check\n", func, num));

	t_list_for_each_entry(cl, t, thincClientRec, &pThinc->clients, link) {
		thincDbg(T_DBG_ANNOY, ("%s: checking client %s\n", func, 
			cl->id));
		if (FD_ISSET(cl->sock, socks)) {
			thincDbg(T_DBG_ANNOY, ("%s: handling client "
				"%s\n", func, cl->id));
			--num;

			err = thincHandleClient(cl);
			if(err == T_CLOSED || err == T_ERROR) {
				if(cl->print_msg) {
					thincError("Client %s has closed its "
						"connection. au revoir!\n", 
						cl->id);
				}
				cl->close_cl(cl);
			}
			if(!num)
				break;
		}
	}
	/* shouldn't happen */
	t_bug_on_2(
		num,
		T_BUG_ARG("select() says we have more active clients than we "
			"could find: %d", num),
		return T_ERROR
	);

	return T_OK;
}

int thincHandleClient(thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_header h;
	int res;

	thincDbg(T_DBG_CLIENT, ("%s: Handling client %s\n",func, cl->id));

	do {
		if(t_cl_connected(cl)) {
			res = thincReadHeader(cl, &h);
			if(res == T_ERROR || res == T_CLOSED) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"read message header from client\n", 
					func));
				return T_ERROR;
			} 
			thincDbg(T_DBG_CLIENT, ("%s: got message %d from "
				"client %s\n", func, h.type, cl->id));
			res = thincHandleClientNormal(cl, h.type);
			if(res == T_ERROR || res == T_CLOSED) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"handle connected client %s\n", func,
					cl->id));
				return T_ERROR;
			}
		} else {
			res = thinc_msgStackStepExec(cl, cl->connectStack,
				&cl->sent);
			if(res == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"handle connecting client %s\n",
					func, cl->id));
				return T_ERROR;
			}
			if(t_cl_connected(cl)) {
				thincDbg(T_DBG_CLIENT, ("%s: client %s fully "
					"connected\n", func, cl->id));
				if(thincClientConnected(cl) == T_ERROR)
					return T_ERROR;
			}
		}
	} while(!thincClientRBufferEmpty(cl));

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

/* FIXME: Use thinc_compress API for this */
static int thincZendInitialFB(thincClientPtr cl)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = cl->parent;
	char *zbuff, *zcurr, *curr, *end;
	int buf_size, res, diff, total;
	unsigned long zlen;
	thinc_fbZData *zline;

	thincDbg(T_DBG_CLIENT, ("%s: Sending compressed framebuffer\n", func));

	/* send a line at a time */
	buf_size = pThinc->widthBytes;

	/* allocate buffer. Leave enough space in front to put the
	 * compression header */
	total = buf_size + T_FB_ZDATA_SIZE;
	/* use ALLOCATE_LOCAL */
	zbuff = malloc(total);
	if(!zbuff) {
		thincDbg(T_DBG_ERROR, ("%s: Error allocating memory for "
			"compressed framebuffer\n", func));
		return T_ERROR;
	}

	zline = (thinc_fbZData*) zbuff;
	zcurr = zbuff + T_FB_ZDATA_SIZE;

	/* all good, send fb */
	curr = pThinc->theFB;
	end = curr + (buf_size*pThinc->height);
	diff = abs(end-curr);

	zline->usize = htonl(buf_size);

	while(diff) {
		/* compress the data */
		zlen = buf_size;
		res = compress((Bytef*)zcurr, &zlen, (Bytef*)curr, buf_size);
		if(res != Z_OK) {
			thincDbg(T_DBG_ERROR, ("%s: Couldn't compress "
				"framebuffer line: %d\n", func, res));
			free(zbuff);
			return T_ERROR;
		}

		zline->zsize = htonl(zlen);
		total = zlen + T_FB_ZDATA_SIZE;
		res = thincWriteData(cl, zbuff, total);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("%s: Error sending compressed "
				"scanline\n", func));
			free(zbuff);
			return T_ERROR;
		} 
		curr += buf_size;
		diff -= buf_size;
	}

	free(zbuff);
	return T_OK;
}

/*
 * Functions called from here should return
 * -1 on error
 *  0 if client closed the connection
 *  1 if everything good
 * if returning 0 or -1 the client shouldn't be closed. Either this function
 * or the caller will take care of that. closing the client and returning
 * 1 may be ok for now. However if you need to close the client by yourself,
 * it may be necessary to rethink this model, instead of doing things
 * behind our back.
 */
static int thincHandleClientNormal(thincClientPtr cl, int msg_type)
{
	t_dbg_name(func)
	int res = T_OK;

	thincDbg(T_DBG_CLIENT, ("%s: Handling active client %s, msg %d\n", 
		func, cl->id, msg_type));

	/* chk message type */
	switch(msg_type) {
		/* event messages */
		case T_EV_MOTION:
			thincDbg(T_DBG_CLIENT, ("%s: Got motion event from "
				"client\n", func));
			res = thincHandleMotionEv(cl);
			break;
		case T_EV_KEYB:
			thincDbg(T_DBG_CLIENT, ("%s: Got keyboard event from "
				"client\n", func));
			res = thincHandleKeybEv(cl);
			break;
		case T_EV_BUTTON:
			thincDbg(T_DBG_CLIENT, ("%s: Got button event from "
				"client\n", func));
			res = thincHandleButtonEv(cl);
			break;
		/* video messages */
		case T_VIDEO_STARTOK:
			thincDbg(T_DBG_CLIENT, ("%s: Got Video reply message "
				"from client\n", func));
			res = thincHandleVidStartOK(cl);
			break;
		/* resize messages */
		case T_RESZ_EVENT:
			thincDbg(T_DBG_CLIENT, ("%s: Got resize event message "
				"from client\n", func));
			res = thincHandleReszEvent(cl);
			break;
		case T_RESZ_UPDATE:
			thincDbg(T_DBG_CLIENT, ("%s: Got resize update request"
				" from client\n", func));
			res = thincHandleReszUpdate(cl);
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: invalid client "
				"message %d received\n", func, msg_type));
			return T_ERROR;
	}

	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't process client's "
			"message\n", func));
		return T_ERROR;
	} else if(res == T_CLOSED) {
		thincDbg(T_DBG_CLIENT, ("%s: Connection closed while "
			"processing client's message\n", func));
		return T_CLOSED;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));

	return T_OK;
}

/* close thincClient connection, delete from list, delete from fd_set 
 * and free memory */
void thincCloseClient(thincClientPtr cl)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: Closing client %s\n", func, cl->id));

	REGION_UNINIT(pThinc->pScreen, &cl->clClip);
	if(cl->resized == T_CL_RESZ_LESS)
		pThinc->resizeCount--;

	t_list_del_el(&cl->link);

	if(!t_list_is_empty(&cl->cmdBuffer)) {
		stats_log(("Close:Buffer:Client[%s]: len %u max_len %u\n",
			cl->id, t_list_len(&cl->cmdBuffer),
			t_list_maxlen(&cl->cmdBuffer)));
		thincEmptyCmdListBuffer(pThinc, &cl->cmdBuffer);
	}
	if(pThinc->doSched) {
		thincSchedDestroy(cl);
	}
	if(pThinc->useRawCache && cl->raw_cache) {
		thinc_cache_destroy(cl->raw_cache);
	}
	if(pThinc->useBitCache && cl->bit_cache) {
		thinc_cache_destroy(cl->bit_cache);
	}
	if(pThinc->usePixCache && cl->pix_cache) {
		thinc_cache_destroy(cl->pix_cache);
	}

#ifdef T_USEXVIDEO
	if(pThinc->useVideo) 
		/* assumes clientCount hasn't been decreased */
		thincVidCloseClient(pThinc, cl);
#endif

	/* destroy message stacks */
	if(cl->rawStack)
		thinc_rawDestroyStack(cl);
	if(cl->bitmapStack)
		thinc_bitmapDestroyStack(cl);
	if(cl->pixmapStack)
		thinc_pixmapDestroyStack(cl);
	if(cl->fillStack)
		thinc_fillDestroyStack(cl);
	if(cl->copyStack)
		thinc_copyDestroyStack(cl);

	thincRemoveSocket(pThinc, cl->sock);
	/* TODO: Close SSL connection */
	tcpThincClose(cl->sock);

	if(t_cl_connected(cl)) {
		--pThinc->clientCount;
#ifdef T_RECORD
		if(pThinc->doRecord) {
			thincRecordFlush(pThinc);
			thincRecordClientDisable(pThinc);
		}
#endif
	}

	t_list_inv(&cl->link);
	cl->parent = NULL;
	free(cl);


#if 0
	if(!pThinc->clients)
		/* no one connected - enable VT switching */
		xf86EnableVTSwitch(TRUE);
#endif

#ifdef T_NODRAW
	/* FIXME: If we're in close screen, no need to do this */
	if(thincClient0Purge(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't purge client0's "
			"buffer at client exit. we're probably screwed\n",
			func));
	}
#endif

	/* done. */
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
}

#ifdef T_NODRAW

#ifdef T_RECORD
#error "record and nodraw dont like each other"
#endif

/* existentialist THINC: client 0 is the client that's not really a client.  */
int thincClient0Init(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincClientPtr cl;
	BoxRec clip;

	thincDbg(T_DBG_CLIENT, ("%s: initializing client 0\n", func));

	/* so we only have to initialize what's not 0 */
	cl = calloc(1, sizeof(thincClientRec));
	if(!cl) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"client 0\n", func));
		return T_ERROR;
	}

	cl->sock = -1;
	strncpy(cl->id, "CLIENT 0", 8);
	cl->id[T_CLID_LEN-1] = '\0'; /* paranoia! */

	cl->buffer = thincClient0Buffer;
	cl->flush = thincClient0Flush;
	cl->close_cl = thincClient0Close;

	t_list_init(&cl->cmdBuffer);
	cl->buffer_batch = 0;

	/* TODO: Support different clipping region per client */
	clip.x1 = 0;
	clip.y1 = 0;
	clip.x2 = clip.x1 + pThinc->width;
	clip.y2 = clip.y1 + pThinc->height;
	REGION_INIT(pThinc->pScreen, &cl->clClip, &clip, 1);
	cl->clipIsSub = T_FALSE;
	cl->resized = T_CL_RESZ_NONE;

	cl->connected = T_TRUE;
	cl->parent = pThinc;
	t_list_add_head(&cl->link, &pThinc->clients);
	pThinc->client0 = cl;
	pThinc->clientCount = 1;

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

void thincClient0Close(thincClientPtr cl)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = cl->parent;
	
	thincDbg(T_DBG_CLIENT, ("%s: Closing client 0\n", func));

	REGION_UNINIT(pThinc->pScreen, &cl->clClip);
	t_list_del_el(&cl->link);
	thincEmptyCmdListBuffer(pThinc, &cl->cmdBuffer);
	cl->parent = NULL;
	pThinc->client0 = NULL;
	--pThinc->clientCount;

	free(cl);

	/* in case we close it before we're supposed to */
	pThinc->client0 = NULL;

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
}

/* every once in a while we want to purge stuff out of the queue to make
 * sure only the most relevant things are still in it. this function
 * takes care that.
 * How it works:
 * - traverse client0's buffer from the back.
 * - for each command:
 *   - get its region.
 *   - call OverwriteCmds(cmdbuffer, reg)
 *   - delete command from the cmdbuffer and add it to the head of a new
 *   list ('clean')
 * - move all commands in 'clean' back to the cmd buffer
 *
 * If this functions returns T_ERROR, it's probably a good idea to
 * exit the server as we may have no notion of what the framebuffer
 * looks like.
 */
/* an error here is pretty bad as we may have lost the state of
 * the framebuffer. make sure callers do something drastic */
/* TODO: Put some kind of marker so we don't purge all the time if
 * something hasn't happened */
int thincClient0Purge(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincClientPtr cl0 = pThinc->client0;
	thincListRec clean;
	thincCmdPtr curr;
	RegionPtr reg = NULL;

	instr_add("CLIENT0_PURGE", TI_START);

	thincDbg(T_DBG_CLIENT, ("%s: purging client0's buffer\n", func));

	t_bug_on_2(
		!cl0,
		T_BUG("Trying to purge non-existent client0"),
		return T_ERROR
	);

	t_list_init(&clean);

	/* must do it like this as we have no control over what
	 * OverwriteCommands() does to the list */
	while(!t_list_is_empty(&cl0->cmdBuffer)) {
		curr = t_list_last_entry(&cl0->cmdBuffer, thincCmdRec, list);
		thincDbg(T_DBG_CLIENT, ("%s: processing command %#lx "
			"(type %d)\n", func, curr->id, curr->type));

		/* take it out of the list or it will overwrite itself! */
		t_list_del_el(cmd_list(curr));

		/* if command is transparent, it doesn't overwrite anybody */
		if(curr->flags & T_CMD_TRANS) 
			goto add;

		/* get its region */
		thincCmdGetRegion(pThinc, curr, &reg);

		t_bug_on_2(	/* paranoia will kill us one day */
			!reg,
			T_BUG_ARG("Couldn't get region for command %#lx\n",
				curr->id),
			return T_ERROR
		);

		thincDbg(T_DBG_CLIENT, ("%s: got command's region, "
			"overwriting\n", func));

		if(thincOverwriteCmds(pThinc,&cl0->cmdBuffer,reg) == T_ERROR) {
			T_BUG_ARG_2("Couldn't overwrite commands in client0's "
				"buffer after command %#lx (type %d)\n", 
				curr->id, curr->type);
			return T_ERROR;
		}

add:
		/* add it to new list */
		t_list_add_head(cmd_list(curr), &clean);
	}

	thincDbg(T_DBG_CLIENT, ("%s: almost there, moving relevant commands "
		"back to the buffer\n", func));

	t_bug_on_2(
		!t_list_is_empty(&cl0->cmdBuffer),
		T_BUG("After purging, client0's buffer is not empty!"),
		return T_ERROR
	);

	/* since cmdbuffer is empty, this is ok */
	t_list_move(&cl0->cmdBuffer, &clean);

	t_bug_on_2(
		!t_list_is_empty(&clean),
		T_BUG("After purging, client0's buffer is not empty!"),
		return T_ERROR
	);

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	instr_add("CLIENT0_PURGE", TI_END);
	return T_OK;
}
#endif /* T_NODRAW */

/* XXX: temp hack */
static const char *thinc_version_str = "THINC_0.2";
static const unsigned int thinc_version_len = 10; /* with \0 */

static int thincV2InitConnStack(thincClientPtr cl);
static void thincV2UninitConnStack(thincClientPtr cl);

/* some useful macros */
#define v2_addmodule(cl, mod) do { \
	err = thinc_msgStepAddUsingModule((cl)->connectStack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add v2 connect " \
			"module %s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
} while(0)

#define v2_readHeader(err, cl, h) do {		\
	err = thincReadHeader(cl, h);		\
	if(err == T_ERROR || err == T_CLOSED) {	\
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read message " \
			"from client\n", func));	\
		return T_ERROR;			\
	}					\
} while(0)

#define v2_readMsg(err, cl, m, sz) do {		\
	err = thincReadAnyMsg(cl, m, sz);	\
	if(err == T_ERROR || err == T_CLOSED) {	\
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read message " \
			"from client\n", func));	\
		return T_ERROR;			\
	}					\
} while(0)

#define v2_readData(err, cl, data, size) do {		\
	err = thincReadData(cl, data, size);		\
	if(err == T_ERROR || err == T_CLOSED) {	\
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read message " \
			"from client\n", func));	\
		return T_ERROR;			\
	}					\
} while(0)


#define v2_poll(err, cl) do {			\
	err = thincDoPoll(cl->sock, POLLIN, 0);	\
	if(err == T_ERROR) {			\
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read message " \
			"from client\n", func));	\
		return T_ERROR;			\
	} else if(err == T_TIMEOUT) {		\
		return T_WOULDBLOCK;		\
	}					\
} while(0)

/* the modules for the connection */
static int thinc_v2Version(thincClientPtr cl, void *handle);
static int thinc_v2SecCaps(thincClientPtr cl, void *handle);
static int thinc_v2SSL(thincClientPtr cl, void *handle);
static int thinc_v2Auth(thincClientPtr cl, void *handle);
static int thinc_v2Request(thincClientPtr cl, void *handle);
static int thinc_v2Done(thincClientPtr cl, void *handle);

/* request funcs */
static int thinc_v2ProcessRequest(thincClientPtr cl, int h_length);
static int thinc_v2ReplyUnknown(thincClientPtr cl);
static int thinc_v2ReplyNotOk(thincClientPtr cl);
static int thinc_v2ReplyOk(thincClientPtr cl);
static int thinc_v2SendSimpleReply(thincClientPtr cl, int reply);

static int thinc_v2ReqFBInfo(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqCursor(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqEncoder(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqFBData(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqCacheSz(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqVideo(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqNoVideo(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqVideoServFmts(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqVideoClientFmts(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqAppSharing(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqAppList(thincClientPtr cl, thinc_reqGen *req);
static int thinc_v2ReqAppGet(thincClientPtr cl, thinc_reqGen *req);

/* helper funcs */
static int thinc_v2ReadNotOk(thincClientPtr cl);
static int thinc_v2SendNotOk(thincClientPtr cl, const char *reason);
static int thinc_v2SendOK(thincClientPtr cl);
static int thinc_v2ReadVersionStr(thincClientPtr cl);
static int thinc_v2ReadSecCaps(thincClientPtr cl);
static int thinc_v2ReadAuth(thincClientPtr cl, struct thinc_user *info);
static int thinc_v2ReadRequestHeader(thincClientPtr cl, thinc_header *h);

#ifdef T_APPSHARE
static int thincSendAppDesc(thincClientPtr cl, thincAppPtr app);
#endif	/* T_APPSHARE */

/* priorities */
enum {
	T_V2CONN_VERSION,
	T_V2CONN_SEC_CAPS,
	T_V2CONN_SSL,
	T_V2CONN_AUTH,
	T_V2CONN_REQ,
	T_V2CONN_DONE
} thincV2ConnPrio;

thincDefModule(v2Version, "V2_CONN_VERSION", 16, T_V2CONN_VERSION);
thincDefModule(v2SecCaps, "V2_CONN_SECCAPS", 16, T_V2CONN_SEC_CAPS);
thincDefModule(v2SSL	, "V2_CONN_SSL",     12, T_V2CONN_SSL);
thincDefModule(v2Auth	, "V2_CONN_AUTH",    13, T_V2CONN_AUTH);
thincDefModule(v2Request, "V2_CONN_REQUEST", 16, T_V2CONN_REQ);
thincDefModule(v2Done	, "V2_CONN_DONE",    13, T_V2CONN_DONE);

static int thincNewClientv2(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)

	thincDbg(T_DBG_CLIENT, ("%s: got new v2 client\n", func));

	cl->doAuth = 0;
	cl->doVideo = 0;
	/* TODO: Make cache optional. if client doesn't ask about it
	 * assume it doesn't support it */

	if(thincV2InitConnStack(cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't init connection "
			"stack\n", func));
		return T_ERROR;
	}

	/* initialize basic stuff */
	cl->read = thincReadDataSocket;
	cl->write = thincWriteDataSocket;

	cl->sent = T_FALSE;

	/* get the ball rolling */
	if(thinc_msgStackStepExec(cl, cl->connectStack, &cl->sent) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't execute "
			"connection stack\n", func));
		thincV2UninitConnStack(cl);
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thincV2InitConnStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_CLIENT, ("%s: initializing connection stack\n", func));

	cl->connectStack = thinc_msgStackStepCreate();
	if(!cl->connectStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create connection "
			"message stack\n", func));
		return T_ERROR;
	}

	v2_addmodule(cl, &v2Version);
	v2_addmodule(cl, &v2SecCaps);
	v2_addmodule(cl, &v2SSL);
	v2_addmodule(cl, &v2Auth);
	v2_addmodule(cl, &v2Request);
	v2_addmodule(cl, &v2Done);

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static void thincV2UninitConnStack(thincClientPtr cl)
{
	thinc_msgStackStepDestroy(cl->connectStack);
}

static int thinc_v2Version(thincClientPtr cl, void *handle)
{
	t_dbg_modname(modname, v2Version.name)
	int err;
	int* sent = handle;

	thincDbg(T_DBG_CLIENT, ("%s\n", modname));

	if(!*sent) {
		thincDbg(T_DBG_CLIENT, ("%s: sending version string to new "
			"client\n", modname));
		err = thincWriteData(cl, thinc_version_str, thinc_version_len);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send server "
				"version string", modname));
			return T_ERROR;
		}
		*sent = T_TRUE;
	} 

	err = thinc_v2ReadVersionStr(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client "
			"version string\n", modname));
		/* for future connections */
		*sent = T_FALSE;
		return T_ERROR;
	} else if(err == T_WOULDBLOCK) {
		thincDbg(T_DBG_CLIENT, ("%s: client hasn't sent its version "
			"string, must wait\n", modname));
		return T_MSG_NOTDONE;
	}
	/* for future connections */
	*sent = T_FALSE;

	thincDbg(T_DBG_CLIENT, ("%s: got client version string, finished\n", 
		modname));
	return T_MSG_DONE;
}

static int thinc_v2SecCaps(thincClientPtr cl, void *handle)
{
	t_dbg_modname(modname, v2SecCaps.name)
	int err;
	thinc_serverSecCaps *s;
	thinc_sessionSecCaps *ss;
	int *sent = handle;

	thincDbg(T_DBG_CLIENT, ("%s: exchanging security caps\n", modname));

	if(*sent)
		goto read_it;
	s = (thinc_serverSecCaps*)t_msgBuffer;
	s->caps = 0;
	if(cl->parent->beSecure) {
		s->caps |= T_CAPS_SEC_ENC;
	}
	if(cl->parent->useAuth) {
		s->caps |= T_CAPS_SEC_AUTH;
	}

	thincDbg(T_DBG_CLIENT, ("%s: sending server sec caps %#x\n", modname,
		s->caps));

	s->caps = htonl(s->caps);
	err = thincSendMsg(cl, T_SERV_SEC_CAPS, T_SERV_SEC_CAPS_SIZE, 0);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send server "
			"security caps\n", modname));
		return T_ERROR;
	}
	*sent = T_TRUE;

read_it:
	thincDbg(T_DBG_CLIENT, ("%s: reading client sec caps\n", modname));
	err = thinc_v2ReadSecCaps(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client "
			"security caps\n", modname));
		*sent = T_FALSE;
		return T_ERROR;
	} else if(err == T_WOULDBLOCK) {
		thincDbg(T_DBG_CLIENT, ("%s: Client sec caps not available "
			"yet\n", modname));
		return T_MSG_NOTDONE;
	}

	*sent = T_FALSE;

	/* send the agreed upon session caps */
	ss = (thinc_sessionSecCaps*)t_msgBuffer;
	ss->caps = 0;
	if(cl->isSecure)
		ss->caps |= T_CAPS_SEC_ENC;
	if(cl->doAuth)
		ss->caps |= T_CAPS_SEC_AUTH;

	ss->caps = htonl(ss->caps);

	err = thincSendMsg(cl, T_SESS_SEC_CAPS, T_SESS_SEC_CAPS_SIZE, 0);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send session "
			"security caps\n", modname));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", modname));
	return T_MSG_DONE;
}

static int thinc_v2SSL(thincClientPtr cl, void *handle)
{
	t_dbg_modname(modname, v2SSL.name)
	int err;
	int *retry = handle;

	if(!cl->isSecure)
		return T_MSG_DONE;

	thincDbg(T_DBG_CLIENT, ("%s: Doing SSL handshake\n", modname));
	if(!*retry) {
		cl->ssl = NULL;
	}
	err = thinc_ssl_accept(cl->sock, cl->parent->ctx, &cl->ssl);
	if( (err == T_SSL_CLOSED) || (err == T_SSL_ERROR ) ) {
		thincError("Error: Couldn't initialize secure connection with "
			"client %s\n", cl->id);
		cl->print_msg = T_FALSE;
		return T_ERROR;
	} else if(err == T_SSL_WOULDBLOCK) {
		thincDbg(T_DBG_CLIENT, ("%s: SSL handshake didn't "
			"finish\n", modname));
		*retry = T_TRUE;
		return T_MSG_NOTDONE;
	}

	cl->read = thincReadDataSecure;
	cl->write = thincWriteDataSecure;

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", modname));
	return T_MSG_DONE;
}

static int thinc_v2Auth(thincClientPtr cl, void *handle)
{
	t_dbg_modname(modname, v2Auth.name)
	int err;
	struct thinc_user info;

	if(!cl->doAuth)
		return T_MSG_DONE;

	thincDbg(T_DBG_CLIENT, ("%s: doing authentication\n", modname));

	t_bug_on_2(
		(!cl->parent->owner || cl->parent->owner_len == 0),
		T_BUG("Trying to authenticate user but session owner hasn't "
			"been initialiazed!!!"),
		return T_ERROR
	);

	err = thinc_v2ReadAuth(cl, &info);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read "
			"authentication information\n", modname));
		thinc_v2SendNotOk(cl, "Authentication failed");
		return T_ERROR;
	} else if(err == T_WOULDBLOCK) {
		thincDbg(T_DBG_CLIENT, ("%s: auth info not yet here\n",
			modname));
		return T_MSG_NOTDONE;
	}

	thincDbg(T_DBG_CLIENT, ("%s: got username %s (%d chars). "
		"authenticating...\n", modname, info.user, info.user_len));

	err = thinc_authenticate(&info, cl->parent->owner, 
		cl->parent->owner_len);
	if(err == T_AUTH_NOMATCH) {
		/* user is not owner */
		thincError("Authentication failed: User '%s' does not match "
			"session owner '%s'\n", info.user, cl->parent->owner);
		cl->print_msg = T_FALSE;
		thinc_v2SendNotOk(cl, "You're not the owner of this session");
		goto error;
	} else if(err == T_AUTH_ERROR) {
		/* user couldn't be authenticated */
		thincError("Could not authenticate user %s\n", info.user);
		cl->print_msg = T_FALSE;
		thinc_v2SendNotOk(cl, "Authentication failed");
		goto error;
	}
	/* don't need this anymore */
	thinc_free_user_info(&info);

	err = thinc_v2SendOK(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send auth OK "
			"message to client %s\n", modname, cl->id));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", modname));
	return T_MSG_DONE;

error:
	thinc_free_user_info(&info);
	return T_ERROR;
}

static int thinc_v2Request(thincClientPtr cl, void *handle)
{
	t_dbg_modname(modname, v2Request.name)
	int err;
	thinc_header h;
	
	thincDbg(T_DBG_CLIENT, ("%s: reading requests from client\n", modname));

	err = thinc_v2ReadRequestHeader(cl, &h);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read request "
			"header\n", modname));
	} else if(err == T_WOULDBLOCK) {
		thincDbg(T_DBG_CLIENT, ("%s: client request not yet here\n",
			modname));
		return T_MSG_NOTDONE;
	}

	switch (h.type) {
		case T_CLIENT_REQ:
			thincDbg(T_DBG_CLIENT, ("%s: got another client "
				"request\n", modname));
			err = thinc_v2ProcessRequest(cl, h.length);
			if(err == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("%s: Couldn't process "
					"client request\n", modname));
				return T_ERROR;
			}
			break;
		case T_CLIENT_DONE:
			thincDbg(T_DBG_CLIENT, ("%s: client is finished "
				"with handshake\n", modname));
			return T_MSG_DONE;
		case T_CLIENT_NOTOK:
			thincDbg(T_DBG_CLIENT, ("%s: client rejected the "
				"connection\n",	modname));
			thinc_v2ReadNotOk(cl);
			return T_ERROR;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: unknown message %d "
			"received\n", modname, h.type));
			return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", modname));
	return T_MSG_NOTDONE;
}

static int thinc_v2Done(thincClientPtr cl, void *handle)
{
	t_dbg_modname(modname, v2Request.name)

	thincDbg(T_DBG_CLIENT, ("%s: client %s is connected\n", modname, 
		cl->id));
	
#ifdef T_NODRAW
	/* now send the fb (as collection of commands) to the client */
	if(thincClient0SendFB(cl->parent, cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send framebuffer "
			"to new client\n", func));
		return T_ERROR;
	}
#endif	/* T_NODRAW */

	cl->connected = T_TRUE;

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", modname));
	return T_MSG_DONE;
}

static int thincClientConnected(thincClientPtr cl)
{
	thincV2UninitConnStack(cl); 
	cl->parent->clientCount++;

#ifdef T_USEXVIDEO
	/* tell video the client is connected to give it a change to
	 * initialize any active video streams */
	if(cl->parent->useVideo) {
		if(thincClientConnectedVideo(cl) == T_ERROR) {
			/* awesome */
			return T_ERROR;
		}
	}
#endif

#ifdef T_RECORD
	/* _SHOULD_ be called after clientCount has been increased */
	if(cl->parent->doRecord) {
		if(thincRecordClientEnable(cl->parent) == T_ERROR)
			return T_ERROR;
		gettime_ms(cl->last_flush);
	}
#endif
	return T_OK;
}

/**************** request handling functions ***********/
static int thinc_v2ProcessRequest(thincClientPtr cl, int h_length)
{
	t_dbg_name(func)
	int err, which;
	thinc_reqGen req;

	thincDbg(T_DBG_CLIENT, ("%s: reading client's request, %d bytes\n", 
		func, h_length));

	v2_readMsg(err, cl, &req, h_length);

	which = t_req_type(&req) = t_req_read_type(&req);
	thincDbg(T_DBG_CLIENT, ("%s: got request %d\n", func, which));
	
	switch (which) {
		case T_REQ_FBINFO:
			err = thinc_v2ReqFBInfo(cl, &req);
			break;
		case T_REQ_CURSOR:
			err = thinc_v2ReqCursor(cl, &req);
			break;
		case T_REQ_ENCODER:
			err = thinc_v2ReqEncoder(cl, &req);
			break;
		case T_REQ_FBDATA:
			err = thinc_v2ReqFBData(cl, &req);
			break;
		case T_REQ_CACHESZ:
			err = thinc_v2ReqCacheSz(cl, &req);
			break;
		case T_REQ_VIDEO:
			err = thinc_v2ReqVideo(cl, &req);
			break;
		case T_REQ_NOVIDEO:
			err = thinc_v2ReqNoVideo(cl, &req);
			break;
		case T_REQ_VIDEO_SERV_FMTS:
			err = thinc_v2ReqVideoServFmts(cl, &req);
			break;
		case T_REQ_VIDEO_CLIENT_FMTS:
			err = thinc_v2ReqVideoClientFmts(cl, &req);
			break;
		case T_REQ_KEEPALIVE:
			err = thinc_v2ReplyUnknown(cl);
			break;
		case T_REQ_APP_SHARING:
			err = thinc_v2ReqAppSharing(cl, &req);
			break;
		case T_REQ_APP_LIST:
			err = thinc_v2ReqAppList(cl, &req);
			break;
		case T_REQ_APP_GET:
			err = thinc_v2ReqAppGet(cl, &req);
			break;
		default:
			thincDbg(T_DBG_CLIENT, ("%s: Unknown request %d "
				"received\n", func, which));
			err = thinc_v2ReplyUnknown(cl);
			break;
	}
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't process client "
			"request %d\n", func, which));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2ReqFBInfo(thincClientPtr cl, thinc_reqGen *req)
{
	t_dbg_name(func)
	int err;
	thinc_replyFBInfo *fb;
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: sending fb info\n", func));

	v2_newReply(fb, thinc_replyFBInfo, T_REQ_FBINFO);
	fb->byteOrder = thincEndian();
	fb->mbz = 0;
	fb->depth = pThinc->depth;
	fb->bpp = pThinc->bpp;
	fb->width = htons(pThinc->width);
	fb->height = htons(pThinc->height);

	v2_sendReply(err, cl, t_reply_size(T_REPLY_FBINFO));

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2ReqCursor(thincClientPtr cl, thinc_reqGen *req)
{
	t_dbg_name(func)
	int err;
	thinc_replyCursor *c;
	thincScreenPtr pThinc = cl->parent;
	thincCursorPtr cursor = pThinc->cursor;

	if(!pThinc->useHWCursor) {
		return thinc_v2ReplyNotOk(cl);
	}

	thincDbg(T_DBG_CLIENT, ("%s: sending cursor info\n", func));

	v2_newReply(c, thinc_replyCursor, T_REQ_CURSOR);
	c->flags = 0;
	if(cursor->argb) {
		c->flags |= T_CURSOR_ARGB;
	}
	c->flags = htons(c->flags);
	c->size = htons(cursor->size);
	c->fg = htonl(cursor->fg);
	c->bg = htonl(cursor->bg);
	c->width = cursor->width;
	c->height = cursor->height;
	c->xhot = cursor->xhot;
	c->yhot = cursor->yhot;
	c->x = htons(cursor->x);
	c->y = htons(cursor->y);

	v2_sendReply(err, cl, t_reply_size(T_REPLY_CURSOR));

	err = thincWriteData(cl, pThinc->cursor->cursor, 
		pThinc->cursor->size);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send initial "
			"cursor\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2ReqEncoder(thincClientPtr cl, thinc_reqGen *req)
{
	t_dbg_name(func)
	int err;
	thinc_replyEncoder *e;
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: sending encoder info\n", func));

	v2_newReply(e, thinc_replyEncoder, T_REQ_ENCODER);
	if(pThinc->compress) {
#ifdef T_USEPNG
		e->which = T_ENCODER_PNG;
#else
		e->which = T_ENCODER_ZLIB;
#endif	/* PNG */
	} else {
		e->which = T_ENCODER_NONE;
	}

	v2_sendReply(err, cl, t_reply_size(T_REPLY_ENCODER));

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

#ifdef T_NODRAW
static int thinc_v2ReqFBData(thincClientPtr cl, thinc_reqGen *req)
{
	return thinc_v2ReplyNotOk(cl);
}
#else
static int thinc_v2ReqFBData(thincClientPtr cl, thinc_reqGen *req)
{
	t_dbg_name(func)
	int err;
	thinc_replyFBData *d;
	char *curr, *end;
	int buf_size, diff;
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: sending fb data\n", func));

	v2_newReply(d, thinc_replyFBData, T_REQ_FBDATA);

	d->flags = 0;
	if(pThinc->compress) {
		d->flags |= T_FB_COMPRESSED;
	} 
	d->flags = htonl(d->flags);

	v2_sendReply(err, cl, t_reply_size(T_REPLY_FBDATA));
	if(pThinc->compress) {
		if(thincZendInitialFB(cl) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"initial framebuffer\n", func));
			return T_ERROR;
		}
	} else {
		/* send a line at a time */
		buf_size = pThinc->widthBytes;

		/* all good, send fb */
		curr = pThinc->theFB;
		end = curr + (buf_size*pThinc->height);
		diff = abs(end-curr);
		while(diff) {
			err = thincWriteData(cl, curr, buf_size);
			if(err == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Error "
					"sending initial framebuffer\n", 
					func));
				return T_ERROR;
			} 
			curr += err;
			diff -= err;
		}
	}


	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}
#endif	/* T_NODRAW */

static int thinc_v2ReqCacheSz(thincClientPtr cl, thinc_reqGen *req)
{
	t_dbg_name(func)
	int err;
	thinc_replyCacheSz *c;
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: sending fb data\n", func));

	v2_newReply(c, thinc_replyCacheSz, T_REQ_CACHESZ);

	if(pThinc->useRawCache)
		c->img = pThinc->raw_cache_bits;
	else
		c->img = 0;
	if(pThinc->useBitCache)
		c->bit = pThinc->bit_cache_bits;
	else
		c->bit = 0;
	if(pThinc->usePixCache)
		c->pix = pThinc->pix_cache_bits;
	else
		c->pix = 0;

	c->reserved = 0;

	thincDbg(T_DBG_CLIENT, ("%s: sending sizes %d, %d, %d\n", func,
		c->img, c->bit, c->pix));
	
	v2_sendReply(err, cl, t_reply_size(T_REPLY_CACHESZ));

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2ReqVideo(thincClientPtr cl, thinc_reqGen *req)
{
	t_dbg_name(func)
	int err;
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: telling client if we have video or not\n",
		func));

	if(pThinc->useVideo && pThinc->thincVid) {
		err = thinc_v2ReplyOk(cl);
	} else {
		err = thinc_v2ReplyNotOk(cl);
	}
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send reply to "
			"video request\n", func));
		return T_ERROR;
	} else {
		thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
		return T_OK;
	}
}

static int thinc_v2ReqNoVideo(thincClientPtr cl, thinc_reqGen *req)
{
	return thincVideoDisable(cl);
}

static int thinc_v2ReqVideoServFmts(thincClientPtr cl, thinc_reqGen *req)
{
	return thincVInitNewClient(cl);
}

static int thinc_v2ReqVideoClientFmts(thincClientPtr cl, thinc_reqGen *req)
{
	thinc_reqVideoClientFmts fmts = req->video_client_fmts;

	if(thincVInitClientFmts(cl, &fmts.fmts) == T_OK) {
		thinc_v2ReplyOk(cl);
		return T_OK;
	} else {
		thinc_v2SendNotOk(cl, "Cannot process your video formats");
		return T_ERROR;
	}
}

static int thinc_v2ReqAppSharing(thincClientPtr cl, thinc_reqGen *req)
{
#ifdef T_APPSHARE
	return thinc_v2ReplyOk(cl);
#else
	return thinc_v2ReplyNotOk(cl);
#endif
}

static int thinc_v2ReqAppList(thincClientPtr cl, thinc_reqGen *req)
{
#ifdef T_APPSHARE
	t_dbg_name(func)
	int err, i;
	thinc_replyAppList *a;
	thincScreenPtr pThinc = cl->parent;
	thincAppPtr app;

	thincDbg(T_DBG_CLIENT, ("%s: sending application list\n", func));

	v2_newReply(a, thinc_replyAppList, T_REQ_APP_LIST);

	a->count = pThinc->app_count;
	a->mbz = 0;

	v2_sendReply(err, cl, t_reply_size(T_REPLY_APP_LIST));
	
	thincDbg(T_DBG_CLIENT, ("%s: sending %d apps\n", func, 
		pThinc->app_count));
	app = pThinc->apps;
	for(i=0; i<pThinc->app_count; ++i, ++app) {
		err = thincSendAppDesc(cl, app);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s; Couldn't send "
				"app %#x\n", func, app->id));
			return T_ERROR;
		}
	}
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
#else
	return T_ERROR;	/* shouldn't happen */
#endif	/* T_APPSHARE */
}

#ifdef T_APPSHARE
static int thincSendAppDesc(thincClientPtr cl, thincAppPtr app)
{
	t_dbg_name(func)
	thinc_appDesc d;
	int err;

	thincDbg(T_DBG_CLIENT, ("%s: sending app %#x (%s)\n", func, app->id,
		app->name));

	d.id = app->id;
	d.name_len = app->name_len;
	d.win_count = htons(app->count);

	err = thincWriteData(cl, &d, sizeof(d));
	if(err == T_ERROR) {
		return T_ERROR;
	}
	if(app->name_len) {
		err = thincWriteData(cl, app->name, app->name_len);
		if(err == T_ERROR) {
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}
#endif /* T_APPSHARE */

static int thinc_v2ReqAppGet(thincClientPtr cl, thinc_reqGen *req)
{
#ifdef T_APPSHARE
	return T_ERROR;
#else
	return T_ERROR;	/* shouldn't happen */
#endif	/* T_APPSHARE */
}

static int thinc_v2ReplyOk(thincClientPtr cl)
{
	return thinc_v2SendSimpleReply(cl, T_REPLY_OK);
}

static int thinc_v2ReplyNotOk(thincClientPtr cl)
{
	return thinc_v2SendSimpleReply(cl, T_REPLY_NOTOK);
}

static int thinc_v2ReplyUnknown(thincClientPtr cl)
{
	return thinc_v2SendSimpleReply(cl, T_REPLY_UNKNOWN);
}

static int thinc_v2SendSimpleReply(thincClientPtr cl, int reply)
{
	t_dbg_name(func)
	int err;
	thinc_serverReply* r;
	
	thincDbg(T_DBG_CLIENT, ("%s: sending reply %d\n", func, reply));

	r = (thinc_serverReply*)t_msgBuffer;
	r->reply = htons(reply);
	r->unused = 0;

	err = thincSendMsg(cl, T_SERV_REPLY, T_SERV_REPLY_SIZE, 0);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send server "
			"reply\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

int thinc_v2SendReply(thincClientPtr cl, int reply_size)
{
	return thincSendMsg(cl, T_SERV_REPLY, reply_size, 0);
}

/**************** helper functions ********************/
static int thinc_v2ReadVersionStr(thincClientPtr cl)
{
	t_dbg_name(func)
	int err, len=0, max=31;
	char *c, *ver;

	thincDbg(T_DBG_CLIENT, ("%s: reading client's version string\n", 
		func));

	ver = malloc(max+1);
	if(!ver)
		return T_ERROR;
	c = ver;

	while(1) {
		err = thincReadDataNoblock(cl, c, 1);
		if(err == T_ERROR || err == T_CLOSED) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read "
				"data from clientn", func));
			free(ver);
			return T_ERROR;
		} else if(err == T_WOULDBLOCK) {
			/* TODO: Support T_WOULDBLOCK here i.e. support 
			 * multiple calls for same version string */
			free(ver);
			thincDbg(T_DBG_CLIENT, ("%s: Version string not "
				"here yet (len %d). finished\n", func, len));
			return T_WOULDBLOCK;
		}
		if(!*c)
			break;
		++c;
#if 0
		if(++len == max) {
			max <<= 2;
			ver = realloc(ver, max);
			if(!ver)
				return T_ERROR;
			c = &ver[len];
		}
#else
		/* ok, that's a long as we'll let you go */
		if(++len == max) {
			*c = '\0';
			break;
		}
#endif
	}

	thincDbg(T_DBG_CLIENT, ("%s: client version (%d len): '%s'\n", func, 
		len, ver));
	/* TODO: check the version */
	free(ver);

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2ReadSecCaps(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;
	thinc_header h;
	thinc_clientSecCaps c;

	thincDbg(T_DBG_CLIENT, ("%s: reading client sec caps\n", func));

	v2_poll(err, cl);

	v2_readHeader(err, cl, &h);

	thincDbg(T_DBG_CLIENT, ("%s: got message of type %d\n", func, h.type));
	if(h.type == T_CLIENT_NOTOK) {
		thincDbg(T_DBG_CLIENT, ("%s: client rejected our sec caps\n",
			func));
		thinc_v2ReadNotOk(cl);
		return T_ERROR;
	} else if(h.type != T_CLIENT_SEC_CAPS) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: unknown message %d "
			"received. Expected sec caps (%d)\n", func, h.type,
			T_CLIENT_SEC_CAPS));
		return T_ERROR;
	}

	v2_readMsg(err, cl, &c, T_CLIENT_SEC_CAPS_SIZE);

	c.caps = ntohl(c.caps);
	thincDbg(T_DBG_CLIENT, ("%s: got client caps: %#x\n", func, c.caps));

	/* NOTES:
	 * If we support ssl, but the client doesn't it's a tough call, but
	 * for now, we let the connection go on. However, if we do 
	 * authentication and the client doesn't, we have to deny the
	 * connection or we would let just anybody in. */
	if((c.caps & T_CAPS_SEC_ENC) && (cl->parent->beSecure)) {
		thincDbg(T_DBG_CLIENT, ("%s: using secure channel\n", func));
		cl->isSecure = 1;
	} else {
		thincDbg(T_DBG_CLIENT, ("%s: using insecure channel\n", func));
		thincMsg("Warning: Using insecure channel for client %s\n", 
			cl->id);
		cl->isSecure = 0;
	}
	if(cl->parent->useAuth) {
		if(!(c.caps & T_CAPS_SEC_AUTH)) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Client does not "
				"support authentication. Can't allow it to "
				"connect\n", func));
			thinc_v2SendNotOk(cl, "Cannot connect without "
				"authenticating");
			thincError("Error: Rejecting client without "
				"authentication support %s\n", cl->id);
			cl->print_msg = T_FALSE;
			return T_ERROR;
		} else {
			thincDbg(T_DBG_CLIENT, ("%s: using authentication\n", 
				func));
			cl->doAuth = 1;
		}
	} else {
		thincDbg(T_DBG_CLIENT, ("%s: not doing authentication\n", 
			func));
		cl->doAuth = 0;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2ReadNotOk(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;
	thinc_clientNotOk n;
	char *reason;

	thincDbg(T_DBG_CLIENT, ("%s: reading client NOTOK\n", func));
	
	v2_readMsg(err, cl, &n, T_CLIENT_NOTOK_SIZE);

	n.reason_len = ntohl(n.reason_len);
	if(!n.reason_len)
		return T_OK;

	reason = malloc(n.reason_len+1);
	if(!reason)
		return T_ERROR;

	v2_readData(err, cl, reason, n.reason_len);
	reason[n.reason_len] = '\0';

	thincInfo("Client rejected the connection: %s\n", reason);

	free(reason);
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2SendNotOk(thincClientPtr cl, const char *reason)
{
	t_dbg_name(func)
	int err, len;
	thinc_serverNotOk *n;

	thincDbg(T_DBG_CLIENT, ("%s: sending NOTOK to client %s: '%s'\n", func,
		cl->id, reason));

	len = strlen(reason);

	n = (thinc_serverNotOk*)t_msgBuffer;
	n->reason_len = htonl(len);
	err = thincSendMsg(cl, T_SERV_NOTOK, T_SERV_NOTOK_SIZE, 0);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send server "
			"NOTOK\n", func));
		return T_ERROR;
	}
	err = thincWriteData(cl, reason, len);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send server "
			"NOTOK\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_v2SendOK(thincClientPtr cl)
{
	return thincSendMsg(cl, T_SERV_OK, T_SERV_OK_SIZE, 0);
}

static int thinc_v2ReadAuth(thincClientPtr cl, struct thinc_user *info)
{
	t_dbg_name(func)
	int err;
	thinc_header h;
	thinc_cAuth auth;

	thincDbg(T_DBG_CLIENT, ("%s: reading authentication information\n",
		func));

	v2_poll(err, cl);

	v2_readHeader(err, cl, &h);
	if(h.type == T_CLIENT_NOTOK) {
		thincDbg(T_DBG_CLIENT, ("%s: client rejected connection here\n",
			func));
		thinc_v2ReadNotOk(cl);
	} else if(h.type != T_CAUTH) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: unknown message %d "
			"received. Expected auth (%d)\n", func, h.type,
			T_CAUTH));
		return T_ERROR;
	}

	v2_readMsg(err, cl, &auth, T_CAUTH_SIZE);

	thincDbg(T_DBG_CLIENT, ("%s: got CAUTH message, reading "
		"authentication info\n", func));

	/* cool, get the authentication info. need +1 for the \0 */
	info->user_len = ntohs(auth.user_len);
	info->pass_len = ntohs(auth.pass_len);

	info->user = malloc(info->user_len+1);
	info->passwd = malloc(info->pass_len+1);
	if(!info->user || !info->passwd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"to authenticate user\n", func));
		free(info->user);
		free(info->passwd);
		return T_ERROR;
	}

	err = thincReadData(cl, info->user, info->user_len);
	if(err == T_ERROR || err == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client's "
			"username\n", func));
		goto error;
	}
	err = thincReadData(cl, info->passwd, info->pass_len);
	if(err == T_ERROR || err == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client's "
			"username\n", func));
		goto error;
	}

	info->user[info->user_len] = '\0';
	info->passwd[info->pass_len] = '\0';

	return T_OK;

error:
	thinc_free_user_info(info);
	return T_ERROR;
}

static int thinc_v2ReadRequestHeader(thincClientPtr cl, thinc_header *h)
{
	t_dbg_name(func)
	int err;

	v2_poll(err, cl);
	v2_readHeader(err, cl, h);

	return err;
}
