/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Command functions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include <stdarg.h>

RegionRec thincNullReg = { {0,0,0,0}, NULL };
BoxRec thincNullBox = {0, 0, 0, 0};

/* local functions */
static thincListPtr thincGetListRegion(thincScreenPtr pThinc, 
	thincPixmapPtr pPriv, RegionPtr pReg, int deltax, int deltay);
static void thincDestroyCmdWrap(thincScreenPtr pThinc, void *cmd);

/*************** <play with cmdInfos> *******************/
/* cmdInfo is a structure with pointers to functions that know how to handle
 * a particular type of command. 
 * namelen cannot be larger than T_MAXCMD_NAME */
thincCmdInfoPtr thincAllocCmdInfo(int cmdId, char* cmdName, int namelen, 
	cmdCreateFunc create, cmdDestroyFunc destroy, cmdCopyFunc copy,
	cmdModifyFunc modify, cmdExecuteFunc execute, 
	cmdGetInfoFunc getinfo, cmdFlushFunc flush)
{
	t_dbg_name(func)
	thincCmdInfoPtr cmd;

	thincDbg(T_DBG_CMD, ("%s: allocating cmdRec for command %d\n", 
		func, cmdId));

	/* silently truncate cmd name if too long */
	if(namelen > T_MAXCMD_NAME) {
		namelen = T_MAXCMD_NAME;
		thincDbg(T_DBG_CMD, ("%s: command name too long (%d), "
			"truncating\n", func, namelen));
	}

	cmd = malloc(sizeof(thincCmdInfoRec));
	if(!cmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate cmdRec\n",
			func));
		return NULL;
	}

	cmd->cmdId = cmdId;
	cmd->cmdCreate = create;
	cmd->cmdDestroy = destroy;
	cmd->cmdCopy = copy;
	cmd->cmdModify = modify;
	cmd->cmdGetInfo = getinfo;
	cmd->cmdExecute = execute;
	cmd->cmdFlush = flush;

	memcpy(cmd->cmdName, cmdName, namelen);
	cmd->cmdName[namelen] = '\0';

	thincDbg(T_DBG_CMD, ("%s: finished\n", func)); 

	return cmd;
}

void thincDestroyCmdInfo(thincCmdInfoPtr cmd)
{
	if(!cmd)
		return;
	free(cmd);
}
/*************** </play with cmdInfos> *******************/

/****************** <Queue a command> ************************/
/* What to do:
 * -> Create() the command structure.
 * -> Check all existing commands for overwriting, and Modify() as
 *  necessary.
 * -> Add the command to the end of the queue
 */
int thincQueueCommand(thincScreenPtr pThinc, PixmapPtr pPix, int type,
	RegionPtr reg, ...)
{
	t_dbg_name(func)
	thincPixmapPtr pPriv = THINCPIXPtr(pPix);
	thincCmdPtr cmd;
	va_list data;
	int ret;
	stats_var2(sa, ea)
	stats_var2(sf, ef)

	thincDbg(T_DBG_CMD, ("%s: Queueing command of type %d in pixmap "
		"%#lx\n", func, type, t_pixID(pPix)));
	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "Region of new command");

	if(region_is_empty(reg)) {
		thincDbg(T_DBG_CMD, ("%s: Command has an "
			"empty region. finished\n", func));
		return T_OK;
	}

	/* create the new command */
	thincDbg(T_DBG_CMD, ("%s: creating new command\n", func));
	stats_start(pThinc, sa);
	va_start(data, reg);
	cmd = thincCreateCmd(pThinc, type, reg, data);
	va_end(data);
	stats_end_simple(pThinc, ea);
	if(!cmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create new "
			"command\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CMD, ("%s: Got new command. queueing it\n", func));

	ret = thincQueueThisCommand(pThinc, pPriv, cmd, reg);
	if(ret == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't queue new "
			"command\n", func));
		thincDestroyCmd(pThinc, cmd);
		return T_ERROR;
	} else {
		if(ret == T_MERGED) {
			thincDbg(T_DBG_CMD, ("%s: command was merged "
				"destroying our copy\n", func));
			stats_start(pThinc, sf);
			thincDestroyCmd(pThinc, cmd);
			stats_end(pThinc, sf, ef, merge_time_free);
			stats_add(pThinc, merge_time_alloc, ea-sa);
		}
		thincDbg(T_DBG_CMD, ("%s: finished\n", func));
		return T_OK;
	}
}

/* TODO:
 * Having to pass the region ptr is nasty. Should we move the
 * RegionRec that all commands have in their private struct to the
 * common part? 
 * for now, RegionPtr may be NULL, in which case we create a region
 * out of the cmd->area BoxRec
 */
#if 1
int thincQueueThisCommand(thincScreenPtr pThinc, thincPixmapPtr pPriv,
	thincCmdPtr cmd, RegionPtr reg)
{
	t_dbg_name(func)
	RegionRec rgn;
	unsigned char local_rgn = T_FALSE;
	thincDbg(T_DBG_CMD, ("%s: queueing command %#lx, type %d\n", func,
		cmd->id, cmd->type));

	if(!reg) {
		REGION_INIT(pThinc->pScreen, &rgn, &cmd->area, 1);
		reg = &rgn;
		local_rgn = T_TRUE;
	}

	/* If the new command is not opaque, then no commands will be 
	 * overwritten. For example, a GLYPH only writes in parts of the 
	 * region it affects. The rest is transparent */
	if(cmd->flags & T_CMD_TRANS) {
		thincDbg(T_DBG_CMD, ("%s: new command is a GLYPH, no "
			"need to check for overwriting\n", func));
		stats_inc(pThinc, overwrite_trans);
	} else {
		if(thincOverwriteCmds(pThinc, &pPriv->list, reg) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't overwrite "
				"commands already in the queue\n", func));
			goto error;
		}
	}

	/* Can we merge it with the last command in the list? This is
	 * extremely useful for broken programs like mozilla that for
	 * some reason like to put images on screen one line at a time.
	 * Retarded. */
	/* TODO: Maybe we should do this earlier in the game? so we don't
	 * have to destroy the thing now... */
	if(thincTryMergeCmd(pThinc, &pPriv->list, cmd)) {
		/* woohoo!! */
		thincDbg(T_DBG_CMD, ("%s: merged new command into the "
			"tail. no need to add it\n", func));
		if(local_rgn)
			REGION_UNINIT(pThinc->pScreen, reg);
		return T_MERGED;
	} else {
		/* oh well, add the command to the end of the list then */
		if(thincAddCmdList(&pPriv->list, cmd) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add new "
				"command to queue\n", func));
			goto error;
		}
	}


	if(local_rgn)
		REGION_UNINIT(pThinc->pScreen, reg);
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return T_OK;

error:
	if(local_rgn)
		REGION_UNINIT(pThinc->pScreen, reg);
	return T_ERROR;
}
#else
int thincQueueThisCommand(thincScreenPtr pThinc, thincPixmapPtr pPriv,
	thincCmdPtr cmd, RegionPtr reg)
{
	const char *func = __func__;
	RegionRec rgn;
	unsigned char local_rgn = T_FALSE;
	thincDbg(T_DBG_CMD, ("%s: queueing command %#lx\n", func,
		cmd->id));

	if(!reg) {
		REGION_INIT(pThinc->pScreen, &rgn, &cmd->area, 1);
		reg = &rgn;
		local_rgn = T_TRUE;
	}

	/* Can we merge it with the last command in the list? This is
	 * extremely useful for broken programs like mozilla that for
	 * some reason like to put images on screen one line at a time.
	 * Retarded. */
	/* TODO: Maybe we should do this earlier in the game? so we don't
	 * have to destroy the thing now... */
	if(thincTryMergeCmd(pThinc, &pPriv->list, cmd)) {
		/* woohoo!! */
		thincDbg(T_DBG_CMD, ("%s: merged new command into the "
			"tail. no need to add it\n", func));
		if(local_rgn)
			REGION_UNINIT(pThinc->pScreen, reg);
		return T_MERGED;
	}

	/* If the new command is not opaque, then no commands will be 
	 * overwritten. For example, a GLYPH only writes in parts of the 
	 * region it affects. The rest is transparent */
	if(cmd->flags & T_CMD_TRANS) {
		thincDbg(T_DBG_CMD, ("%s: new command is a GLYPH, no "
			"need to check for overwriting\n", func));
		stats_inc(pThinc, overwrite_trans);
	} else {
		if(thincOverwriteCmds(pThinc, &pPriv->list, reg) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't overwrite "
				"commands already in the queue\n", func));
			goto error;
		}
	}

	/* oh well, add the command to the end of the list then */
	if(!thincAddCmdList(&pPriv->list, cmd)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add new "
					"command to queue\n", func));
		goto error;
	}

	if(local_rgn)
		REGION_UNINIT(pThinc->pScreen, reg);
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return T_OK;

error:
	if(local_rgn)
		REGION_UNINIT(pThinc->pScreen, reg);
	return T_ERROR;
}
#endif


/****************** </Queue a command> ************************/

/****************** Copy commands from one pixmap to another **************/
/* need to:
 * - create a list of which src commands need to be copied. clip and
 * translate them. (Modify())
 * - find out which dst commands are being overwritten. clip them
 * (Modify()) and Destroy() if necessary.
 * - Add the src commands to the end of the list.
 */
int thincCopyPixmapQueue(thincScreenPtr pThinc, PixmapPtr srcpix, 
	PixmapPtr dstpix, int srcx, int srcy, int w, int h, int dstx, 
	int dsty)
{
	t_dbg_name(func)
	thincPixmapPtr pSrc, pDst;
	thincListPtr src_cmds;
	RegionRec srcReg, dstReg;
	int deltax, deltay;
	int res;

	thincDbg(T_DBG_CMD, ("%s: copying from %#lx to %#lx\n", func,
		t_pixID(srcpix), t_pixID(dstpix)));

	pSrc = THINCPIXPtr(srcpix);
	if(t_list_is_empty(&pSrc->list)) {
		thincDbg(T_DBG_CMD, ("%s: no commands in source queue, "
			"finished\n", func));
		return 0;
	}

	pDst = THINCPIXPtr(dstpix);

	/* translation */
	deltax = dstx - srcx;
	deltay = dsty - srcy;

	/* src reg */
	coords_to_reg(pThinc->pScreen, &srcReg, srcx, srcy, w, h);

	print_reg(T_DBG_CMD, pThinc->pScreen, &srcReg, "source region:");

	/* get source command list */
	src_cmds = thincGetListRegion(pThinc, pSrc, &srcReg, deltax,
		deltay);
	REGION_UNINIT(pThinc->pScreen, &srcReg);
	if(!src_cmds) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get list of "
			"source commands\n", func));
		return T_ERROR;
	}

	if(t_list_is_empty(src_cmds)) {
		thincDbg(T_DBG_CMD, ("%s: couldn't find any source "
			"commands to copy\n", func));
		return 0;
	}

	stats_dbg(("Copy:Src:Pixmap[%#lx]: len %u max_len %u\n",
		t_pixID(srcpix), t_list_len(&pSrc->list),
		t_list_maxlen(&pSrc->list)));
	stats_dbg(("Copy:Cmds:Pixmap[%#lx]: len %u max_len %u\n", 
		t_pixID(srcpix), t_list_len(src_cmds), 
		t_list_maxlen(src_cmds)));
	stats_dbg(("Copy:Dst:Pixmap[%#lx]: len %u max_len %u\n",
		t_pixID(srcpix), t_list_len(&pDst->list),
		t_list_maxlen(&pDst->list)));

	/* now mess with the destination commands */
	coords_to_reg(pThinc->pScreen, &dstReg, dstx, dsty, w, h);

	print_reg(T_DBG_CMD, pThinc->pScreen, &dstReg, 
		"destination region:");

	res = thincOverwriteCmds(pThinc, &pDst->list, &dstReg);
	REGION_UNINIT(pThinc->pScreen, &dstReg);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't overwrite "
			"commands in destination queue\n", func));
		thincDestroyCmdListBuffer(pThinc, src_cmds);
		return T_ERROR;
	}

	/* now add the commands in the source list to the destination
	 * pixmap */
	/* If there's only one command to move, try to merge it instead */
	if(t_list_one_el(src_cmds)) {
		if(!thincTryMergeCmd(pThinc, &pDst->list, 
			(t_list_first_entry(src_cmds, thincCmdRec, list)) ) ) {
			/* no luck */
			t_list_move(&pDst->list, src_cmds);
		}
	} else {
		t_list_move(&pDst->list, src_cmds);
	}

	/* t_list_move should make src_cmds empty so this will just free it */
	thincDestroyCmdListBuffer(pThinc, src_cmds);

	/* the end! */
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return T_OK;
}

/* Returns a list of the subset of commands in pPriv which draw inside
 * pReg. Returns NULL on error */
static thincListPtr thincGetListRegion(thincScreenPtr pThinc, 
	thincPixmapPtr pPriv, RegionPtr pReg, int deltax, int deltay)
{
	t_dbg_name(func)
	thincListPtr list;
	register thincCmdPtr cmd, next, newc, tmpc;
	BoxPtr currArea, ext;
	
	thincDbg(T_DBG_CMD, ("%s: pixmap %#lx\n", func, pPriv->id));
	print_reg(T_DBG_CMD, pThinc->pScreen, pReg, 
		"looking for commands which draw in: ");

	t_list_alloc(list);
	if(!list) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for command list\n", func));
		return NULL;
	}

	ext = REGION_EXTENTS(pThinc->pScreen, pReg);
	t_list_for_each_entry(cmd, next, thincCmdRec, &pPriv->list, list) {
		thincDbg(T_DBG_CMD, ("%s: checking command %d with %#lx\n",
			func, cmd->type, cmd->id));

		currArea = &cmd->area;
		print_box(T_DBG_CMD, currArea, 
			"current command's draw area:");
		if(!t_intersectBox(currArea, ext) ) {
			/* sweet, next! */
			thincDbg(T_DBG_CMD, ("%s: current command doesn't "
				"draw where we're interested. next\n",
				func));
			continue;
		}

		/* ok, get a copy */
		newc = thincCopyCommand(pThinc, cmd);
		if(!newc) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get "
				"copy of command. Ignoring it\n", func));
			continue;
		}

		/* now need to clip and translate it. Do this before
		 * adding to the list in case it gets completely clipped
		 * out */
		tmpc = thincModifyCommand(pThinc, newc, 
			(T_CMD_CLIP | T_CMD_MOVE), pReg, deltax, deltay);
		if(tmpc == (thincCmdPtr)T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't modify "
				"command. Ignoring it\n", func));
			thincDestroyCmd(pThinc, newc);
			continue;
		} else if(tmpc == NULL) {
			thincDbg(T_DBG_CMD, ("%s: Command to be copied "
				"got clipped out\n", func));
			thincDestroyCmd(pThinc, newc);
			continue;
		}
		newc = tmpc;
		
		/* cool, add it to the list */
		if(thincTryMergeCmd(pThinc, list, newc)) {
			thincDbg(T_DBG_CMD, ("%s: merged command to copy, "
				"no need to add it to the list\n", func));
			thincDestroyCmd(pThinc, newc);
			continue;
		} else {
			thincDbg(T_DBG_CMD, ("%s: Couldn't merge copy, "
				"adding it to the list\n", func));
			if(!thincAddCmdList(list, newc)) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"add command to list. Ignoring it\n", 
					func));
				thincDestroyCmd(pThinc, newc);
				continue;
			}
		}
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return list;
}

int thincOverwriteCmds(thincScreenPtr pThinc, thincListPtr list,
	RegionPtr pReg)
{
	t_dbg_name(func)

	thincCmdPtr cmd, next;
	RegionRec clip;
#ifdef SLOW
	RegionRec currReg;
#endif
	BoxPtr currArea, exts= NULL;
	int modret, inreg, modop, chk_eq = T_FALSE;
	int over_count = 0;
	stats_var2(start, end);

#ifdef T_ENABLE_DEBUG_OPTIONS
	if(pThinc->allowOverlap)
		return T_OK;
#endif

	stats_start(pThinc, start);
	stats_inc(pThinc, overwrite_try);
	stats_add(pThinc, overwrite_try_cmds, t_list_len(list));

	thincDbg(T_DBG_CMD, ("%s: Overwriting commands.\n", func));

	print_reg(T_DBG_CMD, pThinc->pScreen, pReg, "overwrite region:");

	if(REGION_NUM_RECTS(pReg) == 1) {
		chk_eq = T_TRUE;
		exts = REGION_EXTENTS(pThinc->pScreen, pReg);
	}

	t_list_for_each_entry(cmd, next, thincCmdRec, list, list) {
		thincDbg(T_DBG_CMD, ("%s: id %#lx, type %d\n",
			func, cmd->id, cmd->type));

		if(cmd->flags & T_CMD_NO_OVER) {
			/* "can't touch this" */
			thincDbg(T_DBG_CMD, ("%s: found untouchable "
				"command. skipping\n", func));
			stats_inc(pThinc, overwrite_noover);
			continue;
		}

		currArea = &cmd->area;

		print_box(T_DBG_CMD, currArea, "command's area:");

		if(chk_eq && box_equal(exts, currArea)) {
			thincDbg(T_DBG_CMD, ("%s: command's area is same "
				"as new one. you're outa hier!\n", func));
			/* XXX: This is kind of sketchy, but needed. In
			 * theory, we should just delete the command from
			 * the actual list we're passed. but:
			 * 1. we don't know if that list is cmd->list or
			 * cmd->sched
			 * 2. sometimes we need to delete it from both.
			 */
			thincDeleteCmdAll(pThinc, cmd);
			t_list_len_dec(list);
			stats_inc(pThinc, overwrite_complete);
			++over_count;
			continue;
		}
		/*
		 * this checks if currArea is inside, outside or miti-miti
		 * of pReg.
		 * then:
		 * -> if currArea is inside pReg, then the command is
		 *  completely clipped out and we should just delete it.
		 * -> if currArea is outside pReg, then we don't have
		 *  to do anything with this dud so just go to next
		 *  command.
		 * -> if there's some overlap but no clear winner, then
		 *  need to clip the command with currReg - pReg
		 */
		inreg = RECT_IN_REGION(pThinc->pScreen, pReg, currArea);
		if(inreg == rgnOUT) {
			/* sweet, next */
			thincDbg(T_DBG_CMD, ("%s: current command is not "
				"overwritten by region, ignoring\n", func));
			continue;
		} else if(inreg == rgnIN) {
			thincDbg(T_DBG_CMD, ("%s: current command "
				"completely obscured by new command. "
				"Deleting from list\n", func));
			/* XXX: This is kind of sketchy, but needed. In
			 * theory, we should just delete the command from
			 * the actual list we're passed. but:
			 * 1. we don't know if that list is cmd->list or
			 * cmd->sched
			 * 2. sometimes we need to delete it from both.
			 */
			thincDeleteCmdAll(pThinc, cmd);
			t_list_len_dec(list);
			stats_inc(pThinc, overwrite_complete);
			++over_count;
			continue;
		}

		/* inreg == rgnPART */

		/* there's overlap. Need to modify this command */
		thincDbg(T_DBG_CMD, ("%s: Command conficts with region. "
			"Modifying\n", func));

#ifdef SLOW
		REGION_INIT(pThinc->pScreen, &clip, NullBox, 0);
		REGION_INIT(pThinc->pScreen, &currReg, currArea, 1);

		REGION_SUBTRACT(pThinc->pScreen, &clip, &currReg, pReg);

		REGION_UNINIT(pThinc->pScreen, &currReg);
#else
		REGION_INIT(pThinc->pScreen, &clip, currArea, 1);
		REGION_SUBTRACT(pThinc->pScreen, &clip, &clip, pReg);
#endif
		print_reg(T_DBG_CMD, pThinc->pScreen, &clip,
			"clipping command with: ");

		/* do this before, since we may destroy the cmd afterwards.
		 * It may count errors too, but damn, can't be too picky with
		 * these numbers. and besides, there shouldn't be any errors :) */
		stats_inc(pThinc, overwrite_type[cmd->type]);

		modop = T_CMD_OVER;
		modret = thincModifyCmdInList(pThinc, cmd, modop, &clip);
		REGION_UNINIT(pThinc->pScreen, &clip);
		/* don't care about the other return values since we
		 * don't do anything else. Note that if modret is 0,
		 * curr is no longer valid, so be careful if you add
		 * code before the end of the loop */
		if(modret == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't modify "
				"command in queue. Destroying it\n", func));
			/* XXX: This is kind of sketchy, but needed. In
			 * theory, we should just delete the command from
			 * the actual list we're passed. but:
			 * 1. we don't know if that list is cmd->list or
			 * cmd->sched
			 * 2. sometimes we need to delete it from both
			 * (e.g. when overwriting on the client buffer,
			 * and the command is already scheduled)
			 */
			thincDeleteCmdAll(pThinc, cmd);
			t_list_len_dec(list);
			continue;
		} else if(modret == T_CLIPPED) {
			/* Modify() doesn't have enough information to
			 * do this so here we are */
			t_list_len_dec(list);
			stats_inc(pThinc, overwrite_complete);
		}
		++over_count;
	}

	if(over_count) {
		stats_dbg(("%s: had %d commands, overwrote %d "
			"of them\n", func, t_list_len(list), over_count));
		stats_inc(pThinc, overwrite_any);
		stats_add(pThinc, overwrite_any_cmds, t_list_len(list));
		stats_add(pThinc, overwrites, over_count);
	}
	
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, start, end, overwrite_time);
	return T_OK;
}

thincCmdPtr thincCopyCommand(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	thincCmdPtr c;
	int index = thincCmd_idx(cmd->type);
	cmdCopyFunc copy = pThinc->cmdHandlerList[index]->cmdCopy;
	c = (*copy)(pThinc, cmd);
	if(!c)
		return NULL;
	t_list_init(cmd_sched(c));
	return c;
}

/****************** </Copy commands from one pixmap to another> ************/

/* for each command in the queue:
 * -> check if it affects the area of interest
 * -> if yes, call its execute handler.
 * -> if no, go to next
 */
int thincExecPixmapQueue(thincScreenPtr pThinc, WindowPtr dst, PixmapPtr pix, 
	int srcx, int srcy, int w, int h, int dstx, int dsty, 
	RegionPtr dstReg)
{
	t_dbg_name(func)
	thincPixmapPtr pPriv;
	thincCmdPtr cmd;
	BoxPtr currArea;
	RegionRec srcReg;
	cmdExecuteFunc execute;
	int deltax, deltay, index, count = 0;

	thincDbg(T_DBG_CMD, ("%s: Executing commands for pixmap %#lx\n",
		func, t_pixID(pix)));

	pPriv = THINCPIXPtr(pix);

	if(t_list_is_empty(&pPriv->list)) {
		thincDbg(T_DBG_CMD, ("%s: No commands in queue. finished\n",
			func));
		return 0;
	}

	stats_dbg(("Exec:Pixmap[%#lx]: len %u max_len %u\n",
		t_pixID(pix), t_list_len(&pPriv->list),
		t_list_maxlen(&pPriv->list)));

	/* tell buffer we're gonna be batch-ing. we only do it if we're
	 * exec'ing more than one element tho */
	if(!t_list_one_el(&pPriv->list)) {
		if(thincBufferBatchStart(pThinc, dstReg) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't prepare "
				"client buffer for batch operation\n", func));
			return T_ERROR;
		}
	}

	deltax = dstx - srcx;
	deltay = dsty - srcy;

	coords_to_reg(pThinc->pScreen, &srcReg, srcx, srcy, w, h);
	print_reg(T_DBG_CMD, pThinc->pScreen, &srcReg, "source region:");

	t_list_for_each_entry_fast(cmd, thincCmdRec, &pPriv->list, list) {
		thincDbg(T_DBG_CMD, ("%s: checking command %d with "
			"id %#lx\n", func, cmd->type, cmd->id));

		currArea = &cmd->area;

		print_box(T_DBG_CMD, currArea, "command's drawing area:");

		if(!t_intersectBox(currArea, REGION_EXTENTS(pThinc->pScreen, 
			&srcReg)) ) {
			thincDbg(T_DBG_CMD, ("%s: command doesn't affect "
				"drawing region, ignoring\n", func));
			continue;
		}

		/* ok, found one, exec it */
		index = thincCmd_idx(cmd->type);
		execute = pThinc->cmdHandlerList[index]->cmdExecute;

		(*execute)(pThinc, dst, cmd, dstReg, deltax, deltay);
		++count;
	}

	/* seriously, this can't fail */
	if(!t_list_one_el(&pPriv->list)) {
		thincBufferBatchEnd(pThinc);
	}

	thincDbg(T_DBG_CMD, ("%s: finished. Executed %d commands\n", func,
		count));

	return count;
}

/*************** <do something to a command> ********************/
thincCmdPtr thincCreateCmd(thincScreenPtr pThinc, int type, RegionPtr reg, 
	va_list args)
{
	int index;
	cmdCreateFunc create;
	thincCmdPtr cmd;

	index = thincCmd_idx(type);
	create = pThinc->cmdHandlerList[index]->cmdCreate;
	cmd = (*create)(pThinc, reg, args);
	if(!cmd)
		return NULL;

	t_list_init(&cmd->list);
	t_list_init(&cmd->sched);
	cmd->refcnt = 0;

	return cmd;
}

/* get an id for a command */
unsigned long thinc_getCmdID()
{
	static unsigned long id = 0xFFFFFFFF;

	return --id;
}

/* FIXME: Do we need this (we don't even use it) */
int thincCommandInvalid(thincCmdPtr cmd)
{
	return (cmd->flags & T_CMD_INVALID);
}

static void thincDestroyCmdWrap(thincScreenPtr pThinc, void *cmd)
{
	thincDeleteCmdAll(pThinc, (thincCmdPtr)cmd);
}

/* Note that this one doesn't take the command out of any list
 * Use thincDeleteCmd() to take a command out of a command list and destroy
 * it */
void thincDestroyCmd(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	int index;
	cmdDestroyFunc destroy;

	/* FIXME: I'm not sure we really need this */
	if(cmd->refcnt > 1) {
		/* can't destroy but let everybody now this guy should
		 * die...soon */
		cmd->refcnt--;
		cmd->flags |= T_CMD_INVALID;
		return;
	}

	index = thincCmd_idx(cmd->type);
	destroy = pThinc->cmdHandlerList[index]->cmdDestroy;

	(*destroy)(pThinc, cmd);
}

/* returns modified command. Semantics are the same as the actual
 * modification functions */
thincCmdPtr thincModifyCommand(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, ...)
{
	thincCmdPtr mcmd;
	va_list args;
	va_start(args, mod);
	mcmd = thincModifyCommandV(pThinc, cmd, mod, args);
	va_end(args);
	return mcmd;
}

thincCmdPtr thincModifyCommandV(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	int index = thincCmd_idx(cmd->type);
	cmdModifyFunc modify = pThinc->cmdHandlerList[index]->cmdModify;
	return (*modify)(pThinc, cmd, mod, args);
}

/*
 * Modify a command or a command in a list
 * returns:
 * T_ERROR on error
 * T_CLIPPED if command was destroyed
 * T_OK if everything the same and modification was successful
 *
 * mod is modification (see thincPixmap.h)
 *
 * If command is clipped out, it is deleted from the list (unless list 
 * is NULL)
 */
int thincModifyCmdInList(thincScreenPtr pThinc, thincCmdPtr cmd, int mod, ...)
{
	va_list args;
	thincCmdPtr mcmd;
	va_start(args, mod);
	mcmd = thincModifyCommandV(pThinc, cmd, mod, args);
	va_end(args);

	if(mcmd == (thincCmdPtr)T_ERROR)
		return T_ERROR;
	else if(mcmd == NULL) {
#if 0
		/* command completely clipped out. delete from list. this
		 * may not free its memory but it will mark it as invalid
		 * so nobody else uses it */
		if(!t_list_is_empty(cmd_list(cmd))) {
			thincDeleteCmdList(pThinc, cmd);
		}
		if(!t_list_is_empty(cmd_sched(cmd))) {
			thincDeleteCmdSched(pThinc, cmd);
		}
#else
		thincDeleteCmdAll(pThinc, cmd);
#endif
		return T_CLIPPED;
	} else {
		/* need to revalidate pointers in any lists the command may
		 * be.
		 * mcmd->prev->next = mcmd;
		 * mcmd->next->prev = mcmd;
		 */
		if(!t_list_is_empty(cmd_list(mcmd))) {
			t_list_next(t_list_prev(cmd_list(mcmd) )) = 
				cmd_list(mcmd);
			t_list_prev(t_list_next(cmd_list(mcmd) )) = 
				cmd_list(mcmd);
		}
		if(!t_list_is_empty(cmd_sched(mcmd))) {
			t_list_next(t_list_prev(cmd_sched(mcmd) )) = 
				cmd_sched(mcmd);
			t_list_prev(t_list_next(cmd_sched(mcmd) )) = 
				cmd_sched(mcmd);
		}
		return T_OK;
	}
}

/* tries to merge the cmd with the last command in the list. For now just
 * for raw updates. returns T_TRUE or T_FALSE */
/* TODO: Make it generic, not just for raw updates */
int thincTryMergeCmd(thincScreenPtr pThinc, thincListPtr list, 
	thincCmdPtr cmd)
{
	t_dbg_name(func)
	thincCmdPtr last;
	int merged;

	stats_var2(start, end);

	stats_start(pThinc, start);
	stats_inc(pThinc, merge_try);

	thincDbg(T_DBG_CMD, ("%s: trying to merge command %#lx of type %d\n", 
		func, cmd->id, cmd->type));

	/* FIXME */
	if( (cmd->type != T_FB_RUPDATE) && (cmd->type != T_FB_SFILL) ) {
		thincDbg(T_DBG_CMD, ("%s: sorry, only RAW and SFILL updates "
			"supported for now. finished\n", func));
		stats_end(pThinc, start, end, merge_time);
		return T_FALSE;
	}

	/* anything to merge? */
	if(t_list_is_empty(list)) {
		thincDbg(T_DBG_CMD, ("%s: finished. list empty, nothing "
			"to merge with\n", func));
		stats_end(pThinc, start, end, merge_time);
		return T_FALSE;
	}

	last = t_list_last_entry(list, thincCmdRec, list);

	thincDbg(T_DBG_CMD, ("%s: trying to merge with %#lx\n", func,
		last->id));

	/* same type? */
	if(last->type != cmd->type) {
		thincDbg(T_DBG_CMD, ("%s: commands have different type. "
			"can't merge. finished\n", func));
		stats_end(pThinc, start, end, merge_time);
		return T_FALSE;
	}

	last = thincModifyCommand(pThinc, last, T_CMD_MERGE, cmd, &merged);
	/* hopefully gcc will make this chunk of code dissapear when stats
	 * are disabled */
	if(merged) {
		stats_inc(pThinc, merges);
		if(cmd->type == T_FB_RUPDATE) {
			stats_inc(pThinc, merge_raw);
		} else if(cmd->type == T_FB_SFILL) {
			stats_inc(pThinc, merge_sfill);
		}
	}

	thincDbg(T_DBG_CMD, ("%s: Finished\n", func));

	stats_end(pThinc, start, end, merge_time);
	return merged;
}

/*************** </do something to a command> ********************/

/************** <get information from commands> *******************/
void thincCmdGetRegion(thincScreenPtr pThinc, thincCmdPtr cmd, RegionRec **reg)
{
	int index = thincCmd_idx(cmd->type);
	cmdGetInfoFunc getinfo = pThinc->cmdHandlerList[index]->cmdGetInfo;

	(*getinfo)(pThinc, cmd, T_CMD_REGION, reg);
}

void thincCmdGetBitmap(thincScreenPtr pThinc, thincCmdPtr cmd, 
	char **bitmap, int *bitmap_size)
{
	int index = thincCmd_idx(cmd->type);
	cmdGetInfoFunc getinfo = pThinc->cmdHandlerList[index]->cmdGetInfo;

	(*getinfo)(pThinc, cmd, T_CMD_BITMAP, bitmap, bitmap_size);
}

unsigned int thincCmdGetSize(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	unsigned int size, index = thincCmd_idx(cmd->type);
	cmdGetInfoFunc getinfo = pThinc->cmdHandlerList[index]->cmdGetInfo;

	(*getinfo)(pThinc, cmd, T_CMD_SIZE, &size);

	return size;
}

/************** </get information from commands> *******************/

/****************** <do something with a list of commands> *****************/

/* adds cmd to tail of list */
int thincAddCmdList(thincListPtr list, thincCmdPtr cmd)
{
	t_list_add_tail(cmd_list(cmd), list);
	cmd->refcnt++;
	return T_OK;
}

int thincAddCmdSched(thincListPtr list, thincCmdPtr cmd)
{
	t_list_add_tail(cmd_sched(cmd), list);
	cmd->refcnt++;
	return T_OK;
}

void thincDestroyCmdListBuffer(thincScreenPtr pThinc, thincListPtr buffer)
{
	thincEmptyCmdListBuffer(pThinc, buffer);
	free(buffer);
}

void thincDestroyCmdListSched(thincScreenPtr pThinc, thincListPtr queue)
{
	thincEmptyCmdListSched(pThinc, queue);
	free(queue);
}

void thincEmptyCmdListBuffer(thincScreenPtr pThinc, thincListPtr buffer)
{
	thincCmdPtr cmd;
	t_list_empty(cmd, thincCmdRec, buffer, list, thincDestroyCmdWrap, 
		pThinc);
}

void thincEmptyCmdListSched(thincScreenPtr pThinc, thincListPtr queue)
{
	thincCmdPtr cmd;
	t_list_empty(cmd, thincCmdRec, queue, sched, thincDestroyCmdWrap, 
		pThinc);
}

/* Use this function to take a command out of a linked list and destroy it.
 * Use DestroyCmd() if you only want to free its memory, but note that
 * any outside references to it in a list will still be around.
 * At the end, the command will at least be:
 * -> marked as invalid
 * -> its buffer pointer = NULL
 * -> its refcnt decreased by 1
 * If conditions are appropriate, the command will be destroyed.
 */
void thincDeleteCmdList(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	if(!t_list_is_empty(&cmd->list))
		t_list_del_el(&cmd->list);
	thincDestroyCmd(pThinc, cmd);
}

void thincDeleteCmdAll(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	if(!t_list_is_empty(&cmd->list)) {
		t_list_del_el(&cmd->list);
	}
	if(!t_list_is_empty(&cmd->sched)) {
		t_list_del_el(&cmd->sched);
	}

	thincDestroyCmd(pThinc, cmd);
}

/* deletes a command from one of the scheduler queues. It also takes it out
 * from the command buffer. Useful when flushing the queues, so that we
 * don't have to go and traverse the command buffer figuring out which
 * commands were flushed */
void thincDeleteCmdSched(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_list_del_el(&cmd->sched);
	cmd->refcnt--;

	if(!t_list_is_empty(&cmd->list)) {
		t_list_del_el(&cmd->list);
		cmd->refcnt--;
	}

	thincDestroyCmd(pThinc, cmd);
}
/****************** </do something with a list of commands> *****************/

/****************** <merge two boxes> ***********************/
/* returns T_TRUE if it merged. And it returns one of 
 * T_MERGE_TOP/_BOTTOM/_RIGHT/_LEFT to let you know where it merged.
 * if merged, rect will be changed to reflect the merge operation.
 * where is optional. */
int thincMergeBoxes(BoxPtr rect, BoxPtr mrect, int *where)
{
	t_dbg_name(func)
	int w = T_MERGE_TOP, ret;

	print_box(T_DBG_CMD, mrect, "merging this")
	print_box(T_DBG_CMD, rect, "into this");


	/* compare coords. we know they can't be all equal */
	if((mrect->x1 == rect->x1) && (mrect->x2 == rect->x2)) {
		/* can plug them horizontally */
		if(mrect->y2 == rect->y1) {
			/* woohoo!! cmd is on top of last!! */
			thincDbg(T_DBG_CMD, ("%s: got a merge! new command "
				"is on top of last command in the list\n",
				func));
			rect->y1 = mrect->y1;
			ret = T_TRUE;
			w = T_MERGE_TOP;
		} else if(rect->y2 == mrect->y1) {
			/* woohoo!! last is on top of cmd!! */
			thincDbg(T_DBG_CMD, ("%s: got a merge! new command "
				"is below last command in the list\n",
				func));
			rect->y2 = mrect->y2;
			ret = T_TRUE;
			w = T_MERGE_BOTTOM;
		} else {
			thincDbg(T_DBG_CMD, ("%s: almost but not quite. y's "
				"didn't match. sorry\n", func));
			ret = T_FALSE;
		}
	} else if((mrect->y1 == rect->y1) && (mrect->y2 == rect->y2)) {
		/* can plug them vertically */
		if(rect->x2 == mrect->x1) {
			/* woohoo!! cmd is to the rigth of last!! */
			thincDbg(T_DBG_CMD, ("%s: got a merge! new command "
				"is to the right of last command in the list\n",
				func));
			rect->x2 = mrect->x2;
			ret = T_TRUE;
			w = T_MERGE_RIGHT;
		} else if(mrect->x2 == rect->x1) {
			/* woohoo!! cmd is to the left of last!! */
			thincDbg(T_DBG_CMD, ("%s: got a merge! new command "
				"is to the left of last command in the list\n",
				func));
			rect->x1 = mrect->x1;
			w = T_MERGE_LEFT;
			ret = T_TRUE;
		} else {
			thincDbg(T_DBG_CMD, ("%s: almost but not quite. x's "
				"didn't match. sorry\n", func));
			ret = T_FALSE;
		}
	} else {
		thincDbg(T_DBG_CMD, ("%s: coordinates don't match. sorry, "
			"can't merge\n", func));
		ret = T_FALSE;
	}

	if(where && ret)
		*where = w;

	return ret;
}
