/* Copyright 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for offscreen manipulation of BITMAP (glyph/bilevel) commands
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

/* local funcs */
static thincCmdPtr thincGlyphCreateReal(thincScreenPtr pThinc, RegionPtr reg, 
	char *bits, int x, int y, int w, int h, unsigned long color,
	unsigned short flags);
static thincCmdPtr thincBilevelCreateReal(thincScreenPtr pThinc, RegionPtr reg, 
	char *bits, int x, int y, int w, int h, unsigned long fg,
	unsigned long bg, unsigned short flags);
static unsigned int thincGlyphSize(struct glyphCmd *glyph);

/* Glyph command is stored:
 * ------------------------------
 * | CmdRec | struct   |  data  |
 * |        | glyphCmd |        |
 * ------------------------------
 *
 * this function just extracts the arguments from the va_list and passes it
 * to the real creation function (below)
 */
thincCmdPtr thincGlyphCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data)
{
	/* the args to the real function */
	char *bits;
	int x, y, w, h;
	unsigned long color;
	thincCmdPtr cmd;
	unsigned short flags;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	bits = va_arg(data, char*);
	x = va_arg(data, int);
	y = va_arg(data, int);
	w = va_arg(data, int);
	h = va_arg(data, int);
	color = va_arg(data, unsigned long);
	flags = (unsigned short) va_arg(data, unsigned int);

	cmd = thincGlyphCreateReal(pThinc,pReg, bits, x, y, w, h, color, flags);

	stats_end(pThinc, t1, t2, glyphStats.create);
	return cmd;
}

/* if bits is not NULL, the bitmap needs to be stored with the command. In
 * this case, x, y, w, h are the actual coordinates of the bitmap inside
 * the bitmap :) */
static thincCmdPtr thincGlyphCreateReal(thincScreenPtr pThinc, RegionPtr pReg, 
	char *bits, int x, int y, int w, int h, unsigned long color,
	unsigned short flags)
{
	t_dbg_name(func)

	struct glyphCmd *glyph;
	thincCmdPtr cmd;
	char *thecmd;
	char *bitscopy;
	BoxPtr ext;
	int totalsize, cmdsize, glyphsize;
	xRectangle rect;

	thincDbg(T_DBG_CMD, ("%s: creating glyph command with color %#lx. "
		"have %ld rectangles\n", func, color, REGION_NUM_RECTS(pReg)));
	print_reg(T_DBG_CMD, pThinc->pScreen, pReg, "region for glyph:");

	if(!bits) {
		ext = REGION_EXTENTS(pThinc->pScreen, pReg);
		box_to_xrect(ext, rect);
		x = rect.x;
		y = rect.y;
		w = rect.width;
		h = rect.height;
	}
	glyphsize = bitmap_size(w,h);

	cmdsize = sizeof(struct glyphCmd);
	cmdsize += glyphsize;

	totalsize = cmdsize + T_CMDREC_SIZE;
	thincDbg(T_DBG_CMD, ("%s: the size of the glyph "
		"is %d bytes, the size of the command is %d bytes. Need to "
		"allocate %d bytes\n", func, glyphsize, cmdsize,
		totalsize));

	thecmd = malloc(totalsize);
	if(!thecmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for GLYPH command\n", func));
		return NULL;
	}
	
	/* pack the command */
	cmd = (thincCmdPtr)thecmd;
	cmd->type = T_FB_GLYPH;
	cmd->flags = T_CMD_TRANS | T_CMD_DEPS_NEEDS;
	cmd->id = thinc_getCmdID();
	thincDbg(T_DBG_CMD, ("%s: new command %#lx has flags %d\n", func, cmd->id,
		cmd->flags));
	ext = REGION_EXTENTS(pThinc->pScreen, pReg);
	copy_boxp(&cmd->area, ext);
	cmd->cmdSize = cmdsize;

	if(flags & T_BIT_FILL) {
		thincDbg(T_DBG_DRAW, ("%s: got actual glyph fill\n", func));
		cmd->flags |= T_BIT_FILL;
	}

	/* fill glyph struct */
	glyph = cmd_data(cmd, struct glyphCmd*);
	glyph->color = color;
	glyph->size = glyphsize;
	copy_boxp(&glyph->orig_area, ext);
	glyph->reg = thincNullReg;
	glyph->sched_sz = 0;

	if(!REGION_COPY(pThinc->pScreen, &glyph->reg, pReg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy GLYPH "
			"command's region\n", func));
		free(thecmd);
		return NULL;
	}

	glyph->x = x;
	glyph->y = y;
	glyph->w = w;
	glyph->h = h;
	/* copy bitmap or create it */
	bitscopy = (char*) (glyph+1);
	if(bits) 
		memcpy(bitscopy, bits, glyphsize);
	else {
		if(thinc_createBitmapFB(pThinc, bitscopy, &rect, color)
			== T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create "
				"bitmap for GLYPH command\n", func));
			free(thecmd);
			return NULL;
		}
	}

	/* no mol */
	thincDbg(T_DBG_CMD, ("%s: finished. command's id %#lx\n", func,
		cmd->id));

	return cmd;
}

void thincGlyphDestroy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	char *glyphcmd = (char*)cmd;
	struct glyphCmd *glyph;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: Destroying command %#lx\n", func,
		cmd->id));

	glyph = cmd_data(cmd, struct glyphCmd*);
	REGION_UNINIT(pThinc->pScreen, &glyph->reg);

	free(glyphcmd);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, glyphStats.destroy);
}

thincCmdPtr thincGlyphCopy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)

	char *newcmd, *oldcmd;
	int totalsize;

	struct glyphCmd *nglyph, *oglyph;
	thincCmdPtr ncmd;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: copying command of size %d\n", func,
		cmd->cmdSize));

	totalsize = cmd->cmdSize + T_CMDREC_SIZE;

	newcmd = malloc(totalsize);
	if(!newcmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for copy of command\n", func));
		return NULL;
	}

	oldcmd = (char*)cmd;
	memcpy(newcmd, oldcmd, totalsize);

	/* copy the region */
	ncmd = (thincCmdPtr) newcmd;
	nglyph = cmd_data(ncmd, struct glyphCmd*);
	oglyph = cmd_data(cmd, struct glyphCmd*);
	nglyph->reg = thincNullReg;

	if(!REGION_COPY(pThinc->pScreen, &nglyph->reg, &oglyph->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy region "
			"of GLYPH command\n", func));
		free(newcmd);
		return NULL;
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, glyphStats.copy);
	return (thincCmdPtr)newcmd;
}

thincCmdPtr thincGlyphModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	t_dbg_name(func)

	struct glyphCmd *glyph;
	RegionPtr reg;

	RegionPtr clip;
	int deltax, deltay;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: modifying command %#lx, modifications "
		"%d\n", func, cmd->id, mod));

	/* retrieve the command */
	glyph = cmd_data(cmd, struct glyphCmd*);
	reg = &glyph->reg;

	if(mod & (T_CMD_CLIP | T_CMD_OVER)) {
		/* clip it */
		clip = va_arg(args, RegionPtr);
		REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
		if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
			thincDbg(T_DBG_CMD, ("%s: command has been "
				"completely obscured. finished\n", func));
			return NULL;
		}

		thincDbg(T_DBG_CMD, ("%s: clipped the region, it now "
			"has %ld rectangles\n", func, REGION_NUM_RECTS(reg)));
		print_reg(T_DBG_CMD, pThinc->pScreen, reg, 
			"clipped region:");
	}

	if(mod & T_CMD_MOVE) {
		/* translate */
		BoxPtr orig_area;
		deltax = va_arg(args, int);
		deltay = va_arg(args, int);
		REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);

		orig_area = &glyph->orig_area;
		box_translate(orig_area, deltax, deltay);

		/* also have to move the bitmap (openoffice depends on this) */
		glyph->x += deltax;
		glyph->y += deltay;
		
		print_reg(T_DBG_CMD, pThinc->pScreen, reg, 
			"translated region:");
	}

	if(mod & T_CMD_SCHEDSZ) {
		unsigned int sz = thincGlyphSize(glyph);
		unsigned int sched_sz = va_arg(args, unsigned int);
		if(sched_sz > sz)
			glyph->sched_sz = sched_sz;
	}

	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, glyphStats.modify);

	return cmd;
}

void thincGlyphGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...)
{
	va_list args;
	struct glyphCmd *glyph = cmd_data(cmd, struct glyphCmd*);
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	va_start(args, info);

	if(info & T_CMD_REGION) {
		RegionRec **reg = va_arg(args, RegionRec**);
		*reg = &glyph->reg;
	}

	if(info & T_CMD_SIZE) {
		unsigned int *size = va_arg(args, unsigned int*);
		*size = thincGlyphSize(glyph);
	}

	if(info & T_CMD_BITMAP) {
		char **bitmap = va_arg(args, char**);
		int *bit_size = va_arg(args, int*);

		if(!glyph->size) {
			*bitmap = NULL;
			*bit_size = 0;
		} else {
			*bitmap = (char*) (glyph+1);
			*bit_size = glyph->size;
		}
	}

	va_end(args);
	stats_end(pThinc, t1, t2, glyphStats.info);
}

/* size of this command is:
 * header size + T_FB_GLYPH_SIZE + bitmap size + size of rects */
static unsigned int thincGlyphSize(struct glyphCmd *glyph)
{
	unsigned int total;

	if(glyph->sched_sz)
		return glyph->sched_sz;
	
	total = T_FB_GLYPH_SIZE + THINC_HEADERSIZE;
	total += REGION_NUM_RECTS(&glyph->reg)*T_RECTREGION_SIZE;

	if(glyph->size)
		total += glyph->size;
	else
		total += bitmap_size(glyph->w, glyph->h);

	return total;
}

void thincGlyphExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay)
{
	t_dbg_name(func)

	struct glyphCmd *glyph;
	char *bits;
	RegionRec bak = thincNullReg;
	RegionPtr reg;
	BoxPtr orig_area;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: deltax %d, deltay %d. clip has %ld rects\n", 
		func, deltax, deltay, REGION_NUM_RECTS(clip)));
	print_reg(T_DBG_CMD, pThinc->pScreen, clip, "clipping with:");

	glyph = cmd_data(cmd, struct glyphCmd*);

	thincDbg(T_DBG_CMD, ("%s: got glyph command, have "
		"bitmap of size %d\n", func, glyph->size));

	if(glyph->size)
		bits = (char*) (glyph+1);
	else
		bits = NULL;

	if(!REGION_COPY(pThinc->pScreen, &bak, &glyph->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create copy of "
			"GLYPH region\n", func));
		return;
	}

	reg = &glyph->reg;
	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "glyph region:");

	REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
	REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, reg)) {
		thincDbg(T_DBG_CMD, ("%s: finished. Queued command is "
			"completely obscured\n", func));
		REGION_UNINIT(pThinc->pScreen, reg);
		copy_reg(glyph->reg, bak);
		return;
	}

	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "after clipping:");
	orig_area = &glyph->orig_area;
	box_translate(orig_area, deltax, deltay);

	/* also have to move the bitmap (openoffice depends on this) */
	thincDbg(T_DBG_CMD, ("%s: moving glyph from %d,%d by %d,%d\n", func,
		glyph->x, glyph->y, deltax, deltay));
	glyph->x += deltax;
	glyph->y += deltay;

	/* send the command */
#if 0
	if(bits) {
		thincDbg(T_DBG_CMD, ("%s: have bitmap, sending update\n",
			func));

		thincSendFBGlyphAllRegion(pThinc, &reg, bits, glyph->x,
			glyph->y, glyph->w, glyph->h, glyph->color);
	} else {
		thincDbg(T_DBG_CMD, ("%s: Don't have glyph. Converting "
			"region to glyph and sending update\n", func));
		thincSendFBGlyphAll(pThinc, &reg, glyph->color);
	}
	REGION_UNINIT(pThinc->pScreen, &reg);
#else
	/* update this, or overwrite() will be useless! */
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincBufferCmd(pThinc, pWindow, cmd);

	/* clean up */
	REGION_UNINIT(pThinc->pScreen, reg);
	copy_reg(glyph->reg, bak);
	box_translate(orig_area, -deltax, -deltay);
	glyph->x -= deltax;
	glyph->y -= deltay;
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, &glyph->reg)));
#endif
	
	thincDbg(T_DBG_CMD, ("%s; finished\n", func));
	stats_end(pThinc, t1, t2, glyphStats.exec);
}

/* Bilevel command is stored:
 * ----------------------------------------------
 * | CmdRec | struct     | bitmap | Region Data |
 * |        | bilevelCmd |        |             |
 * ----------------------------------------------
 *
 * bitmap is present only if bilevelCmd->size > 0
 * RegionData is present only if bilevelCmd->numrects > 0
 *
 * this function just extracts the arguments from the va_list and passes it
 * to the real creation function (below)
 */
thincCmdPtr thincBilevelCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data)
{
	/* the args to the real function */
	char *bits;
	int x, y, w, h;
	unsigned long fg, bg;
	thincCmdPtr cmd;
	unsigned short flags;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	bits = va_arg(data, char*);
	x = va_arg(data, int);
	y = va_arg(data, int);
	w = va_arg(data, int);
	h = va_arg(data, int);
	fg = va_arg(data, unsigned long);
	bg = va_arg(data, unsigned long);
	flags = (unsigned short) va_arg(data, unsigned int);

	cmd = thincBilevelCreateReal(pThinc, pReg, bits, x, y, w, h, fg,
		bg, flags);
	stats_end(pThinc, t1, t2, bilevelStats.create);
	return cmd;
}

/* if bits is not NULL, the bitmap needs to be stored with the command. In
 * this case, x, y, w, h are the actual coordinates of the bitmap inside
 * the bitmap :) */
static thincCmdPtr thincBilevelCreateReal(thincScreenPtr pThinc, 
	RegionPtr pReg, char *bits, int x, int y, int w, int h, 
	unsigned long fg, unsigned long bg, unsigned short flags)
{
	t_dbg_name(func)

	struct bilevelCmd *bilevel;
	thincCmdPtr cmd;
	char *thecmd;
	char *bitscopy;
	BoxPtr ext;
	int totalsize, cmdsize, bilevelsize;
	xRectangle rect;

	thincDbg(T_DBG_CMD, ("%s: creating bilevel command with fg %ld "
		"and bg %ld\n",	func, fg, bg));
	print_reg(T_DBG_CMD, pThinc->pScreen, pReg, "region for bilevel:");

	if(!bits) {
		ext = REGION_EXTENTS(pThinc->pScreen, pReg);
		box_to_xrect(ext, rect);
		x = rect.x;
		y = rect.y;
		w = rect.width;
		h = rect.height;
	}
	bilevelsize = bitmap_size(w,h);

	cmdsize = sizeof(struct bilevelCmd);
	cmdsize += bilevelsize;

	totalsize = cmdsize + T_CMDREC_SIZE;
	thincDbg(T_DBG_CMD, ("%s: the size of the bilevel "
		"is %d bytes, the size of the command is %d bytes. Need to "
		"allocate %d bytes\n", func, bilevelsize, cmdsize,
		totalsize));

	thecmd = malloc(totalsize);
	if(!thecmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for GLYPH command\n", func));
		return NULL;
	}
	
	/* pack the command */
	cmd = (thincCmdPtr)thecmd;
	cmd->type = T_FB_BILEVEL;
	cmd->flags = 0;
	cmd->id = thinc_getCmdID();
	ext = REGION_EXTENTS(pThinc->pScreen, pReg);
	copy_boxp(&cmd->area, ext);
	cmd->cmdSize = cmdsize;

	/* fill bilevel struct */
	bilevel = cmd_data(cmd, struct bilevelCmd*);
	bilevel->fg = fg;
	bilevel->bg = bg;
	bilevel->size = bilevelsize;
	copy_boxp(&bilevel->orig_area, ext);
	bilevel->reg = thincNullReg;
	if(!REGION_COPY(pThinc->pScreen, &bilevel->reg, pReg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy BILEVEL "
			"command's region\n", func));
		free(thecmd);
		return NULL;
	}

	bilevel->x = x;
	bilevel->y = y;
	bilevel->w = w;
	bilevel->h = h;

	if(flags & T_BIT_FILL) {
		thincDbg(T_DBG_DRAW, ("%s: got actual bilevel fill\n", func));
		cmd->flags |= T_BIT_FILL;
	}

	/* copy or create bitmap */
	bitscopy = (char*) (bilevel+1);
	if(bits) {
		memcpy(bitscopy, bits, bilevelsize);
	} else {
		if(thinc_createBitmapFB(pThinc, bitscopy, &rect, fg) 
			== T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create "
				"bitmap for BILEVEL command\n", func));
			free(thecmd);
			return NULL;
		}
	}

	/* no mol */
	thincDbg(T_DBG_CMD, ("%s: finished. command's id %#lx\n", func,
		cmd->id));

	return cmd;
}

void thincBilevelDestroy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	char *blevelcmd = (char*)cmd;
	struct bilevelCmd *bilevel;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s\n", func));

	bilevel = cmd_data(cmd, struct bilevelCmd*);
	REGION_UNINIT(pThinc->pScreen, &bilevel->reg);
	
	free(blevelcmd);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, bilevelStats.destroy);
}

thincCmdPtr thincBilevelCopy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)

	char *newcmd, *oldcmd;
	int totalsize;
	struct bilevelCmd *nbilevel, *obilevel;
	thincCmdPtr ncmd;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: copying command of size %d\n", func,
		cmd->cmdSize));

	totalsize = cmd->cmdSize + T_CMDREC_SIZE;

	newcmd = malloc(totalsize);
	if(!newcmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for copy of command\n", func));
		return NULL;
	}

	oldcmd = (char*)cmd;
	memcpy(newcmd, oldcmd, totalsize);

	/* copy the region */
	ncmd = (thincCmdPtr) newcmd;
	nbilevel = cmd_data(ncmd, struct bilevelCmd*);
	obilevel = cmd_data(cmd, struct bilevelCmd*);
	nbilevel->reg = thincNullReg;

	if(!REGION_COPY(pThinc->pScreen, &nbilevel->reg, &obilevel->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy region "
			"of BILEVEL command\n", func));
		free(newcmd);
		return NULL;
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, bilevelStats.copy);

	return (thincCmdPtr)newcmd;
}

thincCmdPtr thincBilevelModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	t_dbg_name(func)

	struct bilevelCmd *bilevel;
	RegionPtr reg;
	RegionPtr clip;
	int deltax, deltay;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: modifying command %#lx, modifications "
		"%d\n", func, cmd->id, mod));

	/* retrieve the command */
	bilevel = cmd_data(cmd, struct bilevelCmd*);
	reg = &bilevel->reg;

	if(mod & T_CMD_CLIP) {
		/* clip it */
		clip = va_arg(args, RegionPtr);
		REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
		if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
			thincDbg(T_DBG_CMD, ("%s: command has been "
				"completely obscured. finished\n", func));
			return NULL;
		}

		thincDbg(T_DBG_CMD, ("%s: clipped the region, it now "
			"has %ld rectangles\n", func, REGION_NUM_RECTS(reg)));
		print_reg(T_DBG_CMD, pThinc->pScreen, reg, 
			"clipped region:");
	}

	if(mod & T_CMD_MOVE) {
		/* translate */
		BoxPtr orig_area;
		deltax = va_arg(args, int);
		deltay = va_arg(args, int);
		REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
		
		orig_area = &bilevel->orig_area;
		box_translate(orig_area, deltax, deltay);

		/* also have to move the bitmap (openoffice depends on this) */
		bilevel->x += deltax;
		bilevel->y += deltay;
	
		print_reg(T_DBG_CMD, pThinc->pScreen, reg, 
			"translated region:");
	}

	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, bilevelStats.modify);
	return cmd;
}

void thincBilevelGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...)
{
	va_list args;
	struct bilevelCmd *bilevel = cmd_data(cmd, struct bilevelCmd*);
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	va_start(args, info);

	if(info & T_CMD_REGION) {
		RegionRec **reg = va_arg(args, RegionRec**);
		*reg = &bilevel->reg;
	}

	if(info & T_CMD_SIZE) {
		/* size of this command is:
		 * header size + T_FB_BILEVEL_SIZE + bitmap size + size rects
		 */
		unsigned int *size = va_arg(args, unsigned int*);
		unsigned int total = T_FB_GLYPH_SIZE + THINC_HEADERSIZE;
		total += REGION_NUM_RECTS(&bilevel->reg)*T_RECTREGION_SIZE;

		if(bilevel->size)
			total += bilevel->size;
		else
			total += bitmap_size(bilevel->w, bilevel->h);

		*size = total;
	}

	if(info & T_CMD_BITMAP) {
		char **bitmap = va_arg(args, char**);
		int *bit_size = va_arg(args, int*);

		if(!bilevel->size) {
			*bitmap = NULL;
			*bit_size = 0;
		} else {
			*bitmap = (char*) (bilevel+1);
			*bit_size = bilevel->size;
		}
	}



	va_end(args);
	stats_end(pThinc, t1, t2, bilevelStats.info);
}

void thincBilevelExecute(thincScreenPtr pThinc, WindowPtr pWindow, 
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay)
{
	t_dbg_name(func)

	struct bilevelCmd *bilevel;
#if 0
	char *bits;
#endif
	RegionRec bak = thincNullReg;
	RegionPtr reg;
	BoxPtr orig_area;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: deltax %d, deltay %d\n", func, deltax,
		deltay));
	print_reg(T_DBG_CMD, pThinc->pScreen, clip, "clipping with:");

	bilevel = cmd_data(cmd, struct bilevelCmd*);

	thincDbg(T_DBG_CMD, ("%s: got bilevel command, have "
		"bitmap of size %d\n", func, bilevel->size));

#if 0
	if(bilevel->size)
		bits = (char*) (bilevel+1);
	else
		bits = NULL;
#endif

	if(!REGION_COPY(pThinc->pScreen, &bak, &bilevel->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create copy of "
			"BILEVEL region\n", func));
		return;
	}

	reg = &bilevel->reg;
	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "bilevel region:");

	REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
	REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, reg)) {
		thincDbg(T_DBG_CMD, ("%s: finished. Queued command is "
			"completely obscured\n", func));
		REGION_UNINIT(pThinc->pScreen, reg);
		copy_reg(bilevel->reg, bak);
		return;
	}

	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "after clipping:");

	orig_area = &bilevel->orig_area;
	box_translate(orig_area, deltax, deltay);

	/* also have to move the bitmap (openoffice depends on this) */
	bilevel->x += deltax;
	bilevel->y += deltay;
	
	/* send the command */
#if 0
	if(bits) {
		thincDbg(T_DBG_CMD, ("%s: have bitmap, sending update\n",
			func));

		thincSendFBBilevelAllRegion(pThinc, &reg, bits, bilevel->x,
			bilevel->y, bilevel->w, bilevel->h, bilevel->fg,
			bilevel->bg);
	} else {
		thincDbg(T_DBG_CMD, ("%s: Don't have bilevel. Converting "
			"region to bilevel and sending update\n", func));
		thincSendFBBilevelAll(pThinc, &reg, bilevel->fg,
			bilevel->bg);
	}
	
	REGION_UNINIT(pThinc->pScreen, &reg);
#else
	/* update this, or overwrite() will be useless! */
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincBufferCmd(pThinc, pWindow, cmd);

	/* clean up */
	REGION_UNINIT(pThinc->pScreen, reg);
	copy_reg(bilevel->reg, bak);
	box_translate(orig_area, -deltax, -deltay);
	bilevel->x -= deltax;
	bilevel->y -= deltay;
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, &bilevel->reg)));
#endif

	thincDbg(T_DBG_CMD, ("%s; finished\n", func));
	stats_end(pThinc, t1, t2, bilevelStats.exec);
}
