/* Copyright 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for COPY message
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include <netinet/in.h>

/* Sends copy commands to copy region copyReg (dx,dy) pixels away. 
 * The works:
 * - Need 2 regions, source and destination.
 * - The source is the passed region, clipped to the client's viewport.
 * - The destination is the passed region, translated by (dx,dy), and clipped
 *   to the client's viewport.
 * Since source and destination must be equal except for the translation,
 * we need to:
 * 1. clip source
 * 2. clip destination
 * 2. intersect source with destination. The resulting region is the
 * part that we may copy. We extract its constituent rectangles and send
 * a copy command for each of those.
 *
 * Changes the passed region.
 *
 * TODO:
 * If the copy involves a part of the screen not visible to the client
 * but visible to the normal X, then we need to generate the corresponding
 * fb updates. Only case where this doesn't need to be done is when the
 * client has the same clipping area as the normal fb 
 * (ie pThinc->x,y,width,height)
 * For this, we may assume that the fb has been updated, and just send
 * FBRUpdates() for the offscreen parts.
 */
/* the methods */
#ifdef T_GETSTATS
static int thinc_copyPreStats(thincClientPtr, void *);
static int thinc_copyPostStats(thincClientPtr, void *);
#endif
static int thinc_copyCompute(thincClientPtr cl, void *msg);
static int thinc_copySend(thincClientPtr cl, void *msg);
#ifdef T_RECORD
static int thinc_copyRecord(thincClientPtr cl, void *msg);
static int thinc_copyRecordCb(thincClientPtr cl, char *buf, unsigned int *len,
	void *ptr);
#endif

/* the priorities, current order:
 * PRESTATS
 * COMPUTE
 * SEND
 * RECORD
 * POSTSTATS
 */
#ifdef T_GETSTATS
#define T_COPYPRIO_PRESTATS	T_PRIO_MIN
#define T_COPYPRIO_POSTSTATS	T_PRIO_MAX
#endif
#define T_COPYPRIO_SEND		T_PRIO_SEND
#define T_COPYPRIO_COMPUTE	T_COPYPRIO_SEND-10
#ifdef T_RECORD
#define T_COPYPRIO_RECORD	T_COPYPRIO_SEND+1
#endif

/* the static modules. From here, we make copies to each client's stack */
#ifdef T_GETSTATS
thincDefModule(copyPreStats, "COPY_PRESTATS", 14, T_COPYPRIO_PRESTATS);
thincDefModule(copyPostStats, "COPY_POSTSTATS", 15, T_COPYPRIO_POSTSTATS);
#endif
thincDefModule(copyCompute, "COPY_COMPUTE", 13, T_COPYPRIO_COMPUTE);
thincDefModule(copySend, "COPY_SEND", 10, T_COPYPRIO_SEND);
#ifdef T_RECORD
thincDefModule(copyRecord, "COPY_RECORD", 12, T_COPYPRIO_RECORD);
#endif

typedef struct thincCopyObj_S {
	int		dx:16;
	int		dy:16;
	unsigned int	total_len:32;
	unsigned int	numrects:16;
	unsigned int	numcopies:16;
	BoxPtr		rects;
	thinc_fbCopy*	copies;	/* numrects of them */
} thincCopyObj;

#define copy_addmodule(cl, mod) { \
	err = thinc_msgAddUsingModule((cl), (cl)->copyStack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add COPY module"\
			" %s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}

int thinc_copyInitStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_CLIENT, ("%s: creating COPY message stack\n", func));

	cl->copyStack = thinc_msgStackCreate(cl);
	if(!cl->copyStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create COPY "
			"message stack\n", func));
		return T_ERROR;
	}

	/* add the modules */
#ifdef T_GETSTATS
	copy_addmodule(cl, &copyPreStats);
	copy_addmodule(cl, &copyPostStats);
#endif

	copy_addmodule(cl, &copyCompute);
	copy_addmodule(cl, &copySend);

#if 0
#ifdef T_RECORD
	/* only for first client */
	if(cl->parent->doRecord) {
		thincScreenPtr pThinc = cl->parent;
		thincRecordPtr record = t_recordPtr(pThinc);
		/* this mess because when we get called, clientCount hasn't
		 * been set */
		if( 
		    ((record->cl_enabled) && (pThinc->clientCount == 1) ) ||
		    (pThinc->clientCount == 0)
		) {
			copy_addmodule(cl, &copyRecord);
		}
	}
#endif
#endif

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

void thinc_copyDestroyStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->copyStack);
	cl->copyStack = NULL;
}

int thincCopyFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct copyCmd *copy = cmd_data(cmd, struct copyCmd*);
	int ret;
	thincCopyObj obj;
	static thinc_fbCopy* copies = NULL;
	static unsigned int num_copies=0;

	thincDbg(T_DBG_PROTO, ("%s: flushing command %#lx\n", func, cmd->id));

	REGION_TRANSLATE(pThinc->pScreen, &copy->reg, -copy->deltax,
		-copy->deltay);

	obj.dx = copy->deltax;
	obj.dy = copy->deltay;
	obj.numrects = REGION_NUM_RECTS(&copy->reg);
	obj.rects = REGION_RECTS(&copy->reg);
	obj.total_len = 0;

	if(num_copies < obj.numrects) {
		copies = realloc(copies, sizeof(*copies)*obj.numrects);
		if(!copies) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: no memory\n", func));
			return T_ERROR;
		}
		num_copies = obj.numrects;
	}
	obj.copies = copies;
	obj.numcopies = obj.numrects;

	ret = thinc_msgStackExec(cl, cl->copyStack, &obj);

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return ret;
}

static int thinc_copyCompute(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, copyCompute.name)
	thinc_fbCopy *c;
	int numrects, i,j, rectsinband;
	BoxPtr rects, curr, band;
	int xsign, ysign, count = 0;
	thincCopyObj *obj = msg;
	int deltax, deltay;

	thincDbg(T_DBG_PROTO, ("%s: computing copy with deltas x: %d, y: %d\n", 
		modname, obj->dx, obj->dy));

	numrects = obj->numrects;
	rects = obj->rects;

	deltax = obj->dx;
	deltay = obj->dy;

	/* We need to send the rectangles in the correct order, ie. in such
	 * a way that we won't overwrite the source data for a subsequent
	 * copy.
	 * We know the rectangles are YX-banded (ordered by Y then by X)
	 * therefore, suppose following is region to be copied (Rx represents
	 * the rectangle, with x being its position in a YX-banded region):
	 *
	 *  ------------------
	 *  |       R1       |
	 *  ------------------
	 *  | R |    |  R  |
	 *  | 2 |    |  3  |
	 *  |   |    |     |
	 *  ------------------
	 *  |  R4  |    | R5 |
	 *  |      |    |    |
	 *  ------------------
	 *
	 *  Possible movements:
	 *  To NW (dx < 0, dy < 0): Normal order: R1, R2, R3, R4, R5
	 *  To NE (dx > 0, dy < 0): Bands in normal order, but per band order
	 *  inversed: R1, R3, R2, R5, R4
	 *  To SE (dx > 0, dy > 0): Completely inverse order:R5, R4, R3, R2, R1
	 *  To SW (dx < 0, dy > 0): Bands inverse order, but per band order
	 *  normal: R4, R5, R2, R3, R1
	 */
	/* first figure out which case we have. If it's only 1 rectangle,
	 * then trivial case. Note that normal (increasing) order is when
	 * both deltas are negative, that's why the logic seems reversed */
	curr = rects;
	if(numrects == 1)
		xsign = ysign = 1;
	else {
		if(deltax <= 0)
			xsign = 1;
		else
			xsign = -1;

		if(deltay <= 0)
			ysign = 1;
		else {
			/* if deltay > 0, need to start from the last band 
			 * and go up */
			curr = rects + (numrects-1);
			ysign = -1;
		}
	}

	c = obj->copies;

	/* if both deltas have same sign then it's trivial case (note that
	 * in case of completely inverse order, curr is already pointing to
	 * the last rectangle) */
	if(xsign == ysign) {
		for(i=0; i<numrects; i++, curr+=ysign) {
			c->sx = curr->x1;
			c->sy = curr->y1;
			c->dx = curr->x1+deltax;
			c->dy = curr->y1+deltay;
			/* FIXME: should use box_width/_height */
			c->width = curr->x2-curr->x1;
			c->height = curr->y2-curr->y1;

			thincDbg(T_DBG_PROTO, ("%s: computed copy "
				"(%d,%d)->(%d,%d), %d,%d\n", modname, curr->x1, 
				curr->y1, curr->x1+deltax, curr->y1+deltay, 
				curr->x2 - curr->x1, curr->y2 - curr->y1));
			++count;
			++c;
		}
	} else {
		/* "complicated" case:
		 * - dx > 0, dy <= 0 -> NE -> per band order reversed
		 *   xsign -1, ysign 1. curr points to 1st rectangle.
		 *   
		 * - dx <= 0, dy > 0 -> SW -> per band order normal.
		 *   xsign 1, ysing -1. curr points to last rectangle.
		 *
		 * Inverse or normal band order has been taken care of
		 * by the sign assignment (making curr point to first
		 * or last rectangle in the region). Therefore, we just
		 * need to send the rectangles in each band in the correct
		 * order. Idea is then:
		 * - Get all the rectangles in a band (by comparing y1)
		 * - Send them in correct order (using xsign to go left or
		 *   right)
		 * - Go to next band (using ysign)
		 *
		 * We don't need to distinguish between the two cases because:
		 * - when the band order must be reversed, we start on the
		 *   first rectangle. Therefore, after finding all the
		 *   rectangles in a band we are at the last rectangle in
		 *   the band, which is the one we need to send first.
		 * - when the band order is normal, we start on the last
		 *   rectangle. Therefore, after finding all the rectangles
		 *   in a band we are at the first rectangle in the band,
		 *   which is the one we need to send first.
		 */
		band = curr;
		for(i=0; i<numrects; i++) {
			rectsinband = 1;
			/* get all the rectangles in a band, by comparing the
			 * next rectangle's (in correct order terms) y1 coord 
			 * with current's */
			while((i<numrects-1) && band[ysign].y1 == band->y1) {
				band += ysign;
				rectsinband++, i++;
			}
			/* found a band, send it in correct order */
			curr = band;
			for(j=0; j<rectsinband; j++, curr += xsign) {
				thincDbg(T_DBG_PROTO, ("%s: computed rectangle: "
					"(%d,%d)x(%d,%d)\n", modname,
					curr->x1, curr->y1, curr->x2, 
					curr->y2));
				c->sx = curr->x1;
				c->sy = curr->y1;
				c->dx = curr->x1+deltax;
				c->dy = curr->y1+deltay;
				/* FIXME: should use box_width/_height */
				c->width = curr->x2-curr->x1;
				c->height = curr->y2-curr->y1;

				++count;
				++c;
			}
			/* get next band */
			band += ysign;
		}
	}

	obj->numcopies = count;

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}
static int thinc_copySend(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, copySend.name)
	thinc_fbCopy *from, *to;
	int i;
	int res;
	int dobuffer = T_TRUE;
	thincCopyObj *obj = msg;

	thincDbg(T_DBG_PROTO, ("%s: sending copy\n", modname));

	/* XXX testing testing testing XXX */
	if(obj->numcopies != obj->numrects) {
		thincError("sending more (or less?) copies (%u) than "
			"rectangles %u. WTF?\n", obj->numcopies, 
			obj->numrects);
	}
	/* XXX testing testing testing XXX */

	from = obj->copies;
	for(i=0; i<obj->numcopies; ++i, ++from) {
		/* TODO: this could probably be optimized so we don't
		 * try and try every time */
		to = (thinc_fbCopy*)thincBufferMsg(cl, T_FB_COPY, 0,
			T_FB_COPY_SIZE, 0);
		if(!to) {
			/* revert to normal writes */
			thincDbg(T_DBG_PROTO, ("%s: Couldn't buffer "
				"COPY message\n", modname));
			dobuffer = T_FALSE;
			thincFlushBuffer(cl);
			to = (thinc_fbCopy*)t_msgBuffer;
		}

		to->sx = htons(from->sx);
		to->sy = htons(from->sy);
		to->dx = htons(from->dx);
		to->dy = htons(from->dy);
		to->width = htons(from->width);
		to->height = htons(from->height);

		thincDbg(T_DBG_PROTO, ("%s: sending copy "
			"(%d,%d)->(%d,%d), %d,%d\n", modname, from->sx, 
			from->sy, from->sy, from->dy, 
			from->width, from->height));
		if(!dobuffer) {
			res = thincSendMsg(cl, T_FB_COPY, T_FB_COPY_SIZE, 0);
			if(res == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"send COPY message\n", modname));
				return T_ERROR;
			}
		}
	}

	obj->total_len = obj->numcopies*(T_FB_COPY_SIZE+THINC_HEADERSIZE);
	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}

/******************************** recording stuff ****************************/
#ifdef T_RECORD
int thinc_copyInitRecordStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_RECORD, ("%s: creating COPY message stack for "
		"record command\n", func));

	cl->copyStack = thinc_msgStackCreate(cl);
	if(!cl->copyStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create COPY "
			"message stack\n", func));
		return T_ERROR;
	}

	copy_addmodule(cl, &copyCompute);
	copy_addmodule(cl, &copyRecord);

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

void thinc_copyDestroyRecordStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->copyStack);
	cl->copyStack = NULL;
}

static int thinc_copyRecord(thincClientPtr cl, void *m)
{
	t_dbg_modname(modname, copyRecord.name)
	thinc_fbCopy *curr, *orig;
	int i, err;
	thincCopyObj *obj = m;

	thincDbg(T_DBG_RECORD, ("%s: recording %d copies\n", modname,
		obj->numcopies));

	orig = obj->copies;
	curr = obj->copies;

	/* kinda tricky, because each rect is a separate command. we overload
	 * obj->copies so it points to the copy the callback should send.
	 */
	for(i=0; i<obj->numcopies; ++i, ++curr) {
		obj->copies = curr;
		err = thincRecord(cl, T_FB_COPY_SIZE,
			T_FB_COPY, 0, thinc_copyRecordCb, m);
	}

	obj->copies = orig;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

/* since we record so little data, we blindly assume that len has more than
 * enough space for us */
static int thinc_copyRecordCb(thincClientPtr cl, char *buf, unsigned int *len,
	void *ptr)
{
	t_dbg_modname(modname, copyRecord.name)
	thinc_fbCopy *from, *to;
	thincCopyObj *obj = ptr;

	thincDbg(T_DBG_RECORD, ("%s: sending _a_ copy\n", modname));

	to = (thinc_fbCopy*)buf;
	from = obj->copies;

	to->sx = htons(from->sx);
	to->sy = htons(from->sy);
	to->dx = htons(from->dx);
	to->dy = htons(from->dy);
	to->width = htons(from->width);
	to->height = htons(from->height);

	*len = T_FB_COPY_SIZE;

	thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u bytes\n", modname,
		*len));
	return T_REC_DONE;
}
#endif /* T_RECORD */

#ifdef T_GETSTATS
static unsigned long t_start;	/* needed for statistics */

static int thinc_copyPreStats(thincClientPtr cl, void *msg)
{
	stats_start_nostr(t_start);
	stats_init_simple(writeTime);
	stats_init_simple(blockTime);

	return T_OK;
}

static int thinc_copyPostStats(thincClientPtr cl, void *msg)
{
	unsigned long t_end;
	thincCopyObj *obj = msg;

	/* TODO: Maybe stats should be per client?? */
	stats_end(cl->parent, t_start, t_end, copyStats.flush);
	stats_inc(cl->parent, copyStats.count);
	stats_add(cl->parent, copyStats.write, writeTime);
	stats_add(cl->parent, copyStats.block, blockTime);
	stats_add(cl->parent, copyStats.size, obj->total_len);
	return T_OK;
}
#endif /* T_GETSTATS */

#if 0
static int thincSendFBCopy(thincScreenPtr pThinc, thincClientPtr cl, 
	RegionPtr copyReg, int dx, int dy);

int thincCopyFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
	struct copyCmd *copy = cmd_data(cmd, struct copyCmd*);
	int ret;
	stats_var2(t1, t2);

	stats_init_simple(writeTime);
	stats_init_simple(blockTime);
	stats_start(pThinc, t1);

	REGION_TRANSLATE(pThinc->pScreen, &copy->reg, -copy->deltax,
		-copy->deltay);

	ret = thincSendFBCopy(pThinc, cl, &copy->reg, copy->deltax, 
		copy->deltay);
	stats_end(pThinc, t1, t2, copyStats.flush);
	stats_add(pThinc, copyStats.write, writeTime);
	stats_add(pThinc, copyStats.block, blockTime);

	return ret;
}

static int thincSendFBCopy(thincScreenPtr pThinc, thincClientPtr cl, 
	RegionPtr copyReg, int deltax, int deltay)
{
	t_dbg_name(func)
	thinc_fbCopy *c;
	int numrects, i,j, rectsinband;
	BoxPtr rects, curr, band;
	int xsign, ysign, res, count = 0;
	int dobuffer = T_TRUE;

	thincDbg(T_DBG_PROTO, ("%s: sending copy with deltas x: %d, y: %d\n", 
		func, deltax, deltay));

#if 0
	/* clip source region */
	REGION_INTERSECT(pThinc->pScreen, copyReg, &cl->clClip, copyReg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, copyReg)) {
		thincDbg(T_DBG_PROTO, ("%s: finished. Tried to send copy "
			"with offscreen source\n", func));
		return T_OK;
	}

	/* do the translation that the copy requires and clip again */
	REGION_TRANSLATE(pThinc->pScreen, copyReg, deltax, deltay);
	REGION_INTERSECT(pThinc->pScreen, copyReg, &cl->clClip, copyReg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, copyReg)) {
		thincDbg(T_DBG_PROTO, ("%s: finished. Tried to send copy to "
			"offscreen destination\n", func));
		return T_OK;
	}

	/* ok, copyReg now has the area that the client has and to which we can 
	 * copy. Translate back and send it */
	REGION_TRANSLATE(pThinc->pScreen, copyReg, -deltax, -deltay);
#endif

	numrects = REGION_NUM_RECTS(copyReg);
	rects = REGION_RECTS(copyReg);

	/* We need to send the rectangles in the correct order, ie. in such
	 * a way that we won't overwrite the source data for a subsequent
	 * copy.
	 * We know the rectangles are YX-banded (ordered by Y then by X)
	 * therefore, suppose following is region to be copied (Rx represents
	 * the rectangle, with x being its position in a YX-banded region):
	 *
	 *  ------------------
	 *  |       R1       |
	 *  ------------------
	 *  | R |    |  R  |
	 *  | 2 |    |  3  |
	 *  |   |    |     |
	 *  ------------------
	 *  |  R4  |    | R5 |
	 *  |      |    |    |
	 *  ------------------
	 *
	 *  Possible movements:
	 *  To NW (dx < 0, dy < 0): Normal order: R1, R2, R3, R4, R5
	 *  To NE (dx > 0, dy < 0): Bands in normal order, but per band order
	 *  inversed: R1, R3, R2, R5, R4
	 *  To SE (dx > 0, dy > 0): Completely inverse order:R5, R4, R3, R2, R1
	 *  To SW (dx < 0, dy > 0): Bands inverse order, but per band order
	 *  normal: R4, R5, R2, R3, R1
	 */
	/* first figure out which case we have. If it's only 1 rectangle,
	 * then trivial case. Note that normal (increasing) order is when
	 * both deltas are negative, that's why the logic seems reversed */
	curr = rects;
	if(numrects == 1)
		xsign = ysign = 1;
	else {
		if(deltax <= 0)
			xsign = 1;
		else
			xsign = -1;

		if(deltay <= 0)
			ysign = 1;
		else {
			/* if deltay > 0, need to start from the last band 
			 * and go up */
			curr = rects + (numrects-1);
			ysign = -1;
		}
	}

	/* if both deltas have same sign then it's trivial case (note that
	 * in case of completely inverse order, curr is already pointing to
	 * the last rectangle) */
	if(xsign == ysign) {
		for(i=0; i<numrects; i++, curr+=ysign) {
			c = (thinc_fbCopy*)thincBufferMsg(cl, T_FB_COPY, 0,
				T_FB_COPY_SIZE, 0);
			if(!c) {
				/* revert to normal writes */
				thincDbg(T_DBG_PROTO, ("%s: Couldn't buffer "
					"COPY message\n", func));
				dobuffer = T_FALSE;
				thincFlushBuffer(cl);
				c = (thinc_fbCopy*)t_msgBuffer;
			}

			c->sx = htons(curr->x1);
			c->sy = htons(curr->y1);
			c->dx = htons(curr->x1+deltax);
			c->dy = htons(curr->y1+deltay);
			c->width = htons(curr->x2-curr->x1);
			c->height = htons(curr->y2-curr->y1);

			thincDbg(T_DBG_PROTO, ("%s: sending copy "
				"(%d,%d)->(%d,%d), %d,%d\n", func, curr->x1, 
				curr->y1, curr->x1+deltax, curr->y1+deltay, 
				curr->x2 - curr->x1, curr->y2 - curr->y1));
			if(!dobuffer) {
				res = thincSendMsg(cl, T_FB_COPY, 
					T_FB_COPY_SIZE, 0);
				if(res == T_ERROR) {
					thincError("ERROR: Couldn't send "
						"framebuffer copy message\n");
					return T_ERROR;
				}
			}
			++count;
		}
	} else {
		/* "complicated" case:
		 * - dx > 0, dy <= 0 -> NE -> per band order reversed
		 *   xsign -1, ysign 1. curr points to 1st rectangle.
		 *   
		 * - dx <= 0, dy > 0 -> SW -> per band order normal.
		 *   xsign 1, ysing -1. curr points to last rectangle.
		 *
		 * Inverse or normal band order has been taken care of
		 * by the sign assignment (making curr point to first
		 * or last rectangle in the region). Therefore, we just
		 * need to send the rectangles in each band in the correct
		 * order. Idea is then:
		 * - Get all the rectangles in a band (by comparing y1)
		 * - Send them in correct order (using xsign to go left or
		 *   right)
		 * - Go to next band (using ysign)
		 *
		 * We don't need to distinguish between the two cases because:
		 * - when the band order must be reversed, we start on the
		 *   first rectangle. Therefore, after finding all the
		 *   rectangles in a band we are at the last rectangle in
		 *   the band, which is the one we need to send first.
		 * - when the band order is normal, we start on the last
		 *   rectangle. Therefore, after finding all the rectangles
		 *   in a band we are at the first rectangle in the band,
		 *   which is the one we need to send first.
		 */
		band = curr;
		for(i=0; i<numrects; i++) {
			rectsinband = 1;
			/* get all the rectangles in a band, by comparing the
			 * next rectangle's (in correct order terms) y1 coord 
			 * with current's */
			while((i<numrects-1) && band[ysign].y1 == band->y1) {
				band += ysign;
				rectsinband++, i++;
			}
			/* found a band, send it in correct order */
			curr = band;
			for(j=0; j<rectsinband; j++, curr += xsign) {
				c = (thinc_fbCopy*)thincBufferMsg(cl, 
					T_FB_COPY, 0, T_FB_COPY_SIZE, 0);
				if(!c) {
					/* revert to normal writes */
					thincDbg(T_DBG_PROTO, 
						("%s: Couldn't buffer COPY" 
						" message\n", func));
					dobuffer = T_FALSE;
					thincFlushBuffer(cl);
					c = (thinc_fbCopy*)t_msgBuffer;
				}

				thincDbg(T_DBG_PROTO, ("%s: sending rectangle: "
					"(%d,%d)x(%d,%d)\n", func,
					curr->x1, curr->y1, curr->x2, 
					curr->y2));
				c->sx = htons(curr->x1);
				c->sy = htons(curr->y1);
				c->dx = htons(curr->x1+deltax);
				c->dy = htons(curr->y1+deltay);
				c->width = htons(curr->x2-curr->x1);
				c->height = htons(curr->y2-curr->y1);

				thincDbg(T_DBG_PROTO, ("%s: sending copy "
					"(%d,%d)->(%d,%d), %d,%d\n", func, 
					curr->x1, curr->y1, curr->x1+deltax, 
					curr->y1+deltay, curr->x2 - curr->x1,
					curr->y2 - curr->y1));

				if(!dobuffer) {
					res = thincSendMsg(cl, T_FB_COPY, 
						T_FB_COPY_SIZE, 0);
					if(res == T_ERROR) {
						thincError("ERROR: "
							"Couldn't send "
							"framebuffer copy "
							"message\n");
						return T_ERROR;
					}
				}
				++count;
			}
			/* get next band */
			band += ysign;
		}
	}

	stats_add(pThinc, copyStats.count, count);
	stats_add(pThinc, copyStats.size, 
		(count*(T_FB_COPY_SIZE+THINC_HEADERSIZE)));

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));

	return T_OK;
}
#endif

/* Command functions */
/* pReg is source region. We store the destination */
thincCmdPtr thincCopyCreate(thincScreenPtr pThinc, RegionPtr pReg,
	va_list data)
{
	t_dbg_name(func)
	BoxPtr ext;
	int cmdsize, totalsize;

	thincCmdPtr cmd;
	struct copyCmd *copy;
	char *copycmd;

	int deltax, deltay;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: size of copyCmd: %d bytes\n", func,
		sizeof(struct copyCmd)));

	cmdsize = sizeof(struct copyCmd);

	deltax = va_arg(data, int);
	deltay = va_arg(data, int);

	totalsize = cmdsize + T_CMDREC_SIZE;

	thincDbg(T_DBG_CMD, ("%s: command size %d, total size %d bytes\n", 
		func, cmdsize, totalsize));
	print_reg(T_DBG_CMD, pThinc->pScreen, pReg, "queueing copy:");

	copycmd  = malloc(totalsize);
	if(!copycmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
			"memory for COPY command\n", func));
		return NULL;
	}

	/* fill the command structure */
	cmd = (thincCmdPtr) copycmd;
	cmd->type = T_FB_COPY;
	cmd->flags = 0;
	cmd->id = thinc_getCmdID();
	cmd->cmdSize = cmdsize;

	copy = cmd_data(cmd, struct copyCmd*);
	copy->deltax = deltax;
	copy->deltay = deltay;

	REGION_INIT(pThinc->pScreen, &copy->reg, &thincNullBox, 
		REGION_NUM_RECTS(pReg)); 
	if(!REGION_COPY(pThinc->pScreen, &copy->reg, pReg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy COPY "
			"command's region\n", func));
		free(copycmd);
		return NULL;
	}
	REGION_TRANSLATE(pThinc->pScreen, &copy->reg, deltax, deltay);
	ext = REGION_EXTENTS(pThinc->pScreen, &copy->reg);
	copy_boxp(&cmd->area, ext);

	/* the end */
	thincDbg(T_DBG_CMD, ("%s: finished, command's id %#lx\n", func,
		cmd->id));
	stats_end(pThinc, t1, t2, copyStats.create);
	return cmd;
}

void thincCopyDestroy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	char *copycmd = (char*)cmd;
	struct copyCmd *copy;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: destroying command %#lx\n", func, 
		cmd->id));

	copy = cmd_data(cmd, struct copyCmd*);
	REGION_UNINIT(pThinc->pScreen, &copy->reg);

	free(copycmd);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, copyStats.destroy);
}

thincCmdPtr thincCopyCopy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)

	char *cnewcmd, *coldcmd;
	int totalsize;

	struct copyCmd *ncopy, *ocopy;
	thincCmdPtr ncmd;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: Copying command %#lx of size %d bytes\n",
		func, cmd->id, cmd->cmdSize));

	totalsize = cmd->cmdSize + T_CMDREC_SIZE;

	cnewcmd = malloc(totalsize);
	if(!cnewcmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for copy of COPY command\n", func));
		return NULL;
	}

	/* easier to just memcpy */
	coldcmd = (char*)cmd;
	memcpy(cnewcmd, coldcmd, totalsize);

	/* now copy the region */
	ncmd = (thincCmdPtr) cnewcmd;
	ncopy = cmd_data(ncmd, struct copyCmd*);
	ocopy = cmd_data(cmd, struct copyCmd*); 
	ncopy->reg = thincNullReg;
	if(!REGION_COPY(pThinc->pScreen, &ncopy->reg, &ocopy->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy region "
			"of COPY command\n", func));
		free(cnewcmd);
		return NULL;
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, copyStats.copy);
	return ncmd;
}

/* In case of clipping only do it in the destination region, not the
 * source (normally clipping is done for overwritting) */
thincCmdPtr thincCopyModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	t_dbg_name(func)

	struct copyCmd *copy;
	RegionPtr reg;
	int deltax, deltay;
	RegionPtr clip;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: modifying cmd %#lx. operations: %d\n",
		func, cmd->id, mod));

	copy = cmd_data(cmd, struct copyCmd*);
	reg = &copy->reg;

	/* clip? this should be passed when clipping against the
	 * client's clip, for example, since in this case we need to
	 * clip both the src and dst regions (we can't ask the client
	 * to copy something it doesn't have). On the other hand,
	 * if you just want to overwrite the command's output, pass
	 * T_CMD_OVER, since in that case we only want to overwrite
	 * the dst region (it'd be erronous to clip the source since we
	 * don't draw there). */
	if(mod & T_CMD_CLIP) {
		clip = va_arg(args, RegionPtr);

		/* clip source */
		REGION_TRANSLATE(pThinc->pScreen, reg, -copy->deltax,
			-copy->deltay);
		REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
		if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
			thincDbg(T_DBG_CMD, 
				("%s: command has been completely "
				"obscured. finished\n", func));
			return NULL;
		}

		/* clip destination */
		REGION_TRANSLATE(pThinc->pScreen, reg, copy->deltax,
			copy->deltay);
		REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
		if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
			thincDbg(T_DBG_CMD, 
				("%s: command has been completely "
				"obscured. finished\n", func));
			return NULL;
		}

		thincDbg(T_DBG_CMD, ("%s: clipped the command, it now "
			"has %ld rectangles\n", func, REGION_NUM_RECTS(reg)));
		print_reg(T_DBG_CMD,pThinc->pScreen,reg,"clipped region:");
	}

	if(mod & T_CMD_OVER) {
		/* just clip dst */
		clip = va_arg(args, RegionPtr);

		REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
		if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
			thincDbg(T_DBG_CMD, 
				("%s: command has been completely "
				"obscured. finished\n", func));
			return NULL;
		}

		thincDbg(T_DBG_CMD, ("%s: clipped the command, it now "
			"has %ld rectangles\n", func, REGION_NUM_RECTS(reg)));
		print_reg(T_DBG_CMD,pThinc->pScreen,reg,"clipped region:");
	}

	/* translate? */
	if(mod & T_CMD_MOVE) {
		deltax = va_arg(args, int);
		deltay = va_arg(args, int);
		REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
		
		print_reg(T_DBG_CMD, pThinc->pScreen, reg,
			"translated region:");
	}

	/* update the bounding box */
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	/* finished */
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, copyStats.modify);
	return cmd;
}

void thincCopyGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...)
{
	va_list args;
        struct copyCmd *copy = cmd_data(cmd, struct copyCmd*);
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	va_start(args, info);

	if(info & T_CMD_REGION) {
		RegionRec **reg = va_arg(args, RegionRec**);
		*reg = &copy->reg;
	}

	if(info & T_CMD_SIZE) {
		/* size of this command is:
		 * number of rects * (header size + T_FB_COPY_SIZE)
		 */
		int *size = va_arg(args, int*);
		int numrects = REGION_NUM_RECTS(&copy->reg);
		int total = numrects * (THINC_HEADERSIZE * T_FB_COPY_SIZE);

		*size = total;
	}

	va_end(args);
	stats_end(pThinc, t1, t2, copyStats.info);
}

/* This function's never called, since we don't store copies in offscreen
 * pixmaps */
void thincCopyExecute(thincScreenPtr pThinc, WindowPtr pWin, thincCmdPtr cmd, 
	RegionPtr clip, int deltax, int deltay)
{
	T_BUG("I was called!! What's going on??"); 
}

/* marks those commands in list which write to cmd's source region as
 * untouchable */
int thincCmdSaveCopy(thincScreenPtr pThinc, thincListPtr list, 
	thincCmdPtr copy_cmd)
{
	t_dbg_name(func)
	RegionPtr src;
	register BoxPtr src_box, cmd_box;
	register thincCmdPtr cmd;
	struct copyCmd *copy = cmd_data(copy_cmd, struct copyCmd*);
	int ret = T_OK;

	if(t_list_is_empty(list)) {
		thincDbg(T_DBG_CMD, ("%s: no commands to check. finished\n",
			func));
		return T_OK;
	}

	src = &copy->reg;
	REGION_TRANSLATE(pThinc->pScreen, src, -copy->deltax, -copy->deltay);
	src_box = REGION_EXTENTS(pThinc->pScreen, src);

	t_list_for_each_entry_fast(cmd, thincCmdRec, list, list) {
		/* we're conservative and only check if the extents
		 * intersect. Faster and chances are pretty good that
		 * the actual region will intersect */
		cmd_box = &cmd->area;

		if(t_intersectBox(src_box, cmd_box)) {
			thincDbg(T_DBG_CMD, ("%s: marking command %#lx as "
				"untouchable\n", func, cmd->id));
			cmd->flags |= T_CMD_NO_OVER;

			/* The command needs to save the data it's using
			 * to draw, in case the source region is overwritten
			 * and that data becomes obsolete */
			if(thincModifyCmdInList(pThinc, cmd, T_CMD_SAVE)
				== T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Command "
					"couldn't save its draw data\n",
					func));
				ret = T_ERROR;
				goto end;
			}
		}
	}

end:
	/* fix things up */
	REGION_TRANSLATE(pThinc->pScreen, src, copy->deltax, copy->deltay);

	if(ret == T_OK) {
		thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	}

	return ret;
}

static int thincCopyCanMerge(thincScreenPtr pThinc, struct copyCmd *a,
	struct copyCmd *b);

/* try to merge one with two (so two will be combination of doing copy one
 * and then copy two. if return T_TRUE, you can safely destroy one */
int thincCopyMerge(thincScreenPtr pThinc, thincCmdPtr one, thincCmdPtr two)
{
	t_dbg_name(func)
	struct copyCmd *a, *b;
	BoxRec bm, orig;
	int can;

	thincDbg(T_DBG_CMD, ("%s: one %#lx, two %#lx\n", func, one->id, 
		two->id));

	a = cmd_data(one, struct copyCmd*);
	b = cmd_data(two, struct copyCmd*);

	can = thincCopyCanMerge(pThinc, a, b);
	if(can == T_NO) {
		thincDbg(T_DBG_CMD, ("%s: can't merge 'em. finished\n", func));
		return T_FALSE;
	}

	copy_boxp(&bm, REGION_EXTENTS(pThinc->pScreen, &b->reg));
	copy_box(orig, bm);

	REGION_UNINIT(pThinc->pScreen, &b->reg);

	if(a->deltay < 0) {
		/* going up. sum cuz it's negative and we want to substract...
		 * i think... */
		bm.y2 += a->deltay;
	} else {
		/* going down */
		bm.y1 += a->deltay;
	}
	if(bm.y2 < bm.y1) {
		thincDbg(T_DBG_CMD, ("%s: something weird has happened. "
			"can't merge them\n", func));
		/* undo */
		REGION_INIT(pThinc->pScreen, &b->reg, &orig, 1);
		return T_FALSE;
	}

	REGION_INIT(pThinc->pScreen, &b->reg, &bm, 1);
	copy_box(two->area, bm);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return T_TRUE;
}

/* determines if we can merge them. for now just vertical */
static int thincCopyCanMerge(thincScreenPtr pThinc, struct copyCmd *a,
	struct copyCmd *b)
{
	t_dbg_name(func)
	RegionPtr ar, br;
	BoxPtr ab, bb;

	thincDbg(T_DBG_CMD, ("%s: can we merge them?\n", func));

	ar = &a->reg;
	br = &b->reg;

	/* XXX just vertical copies please */
	if(a->deltax || b->deltax) {
		thincDbg(T_DBG_CMD, ("%s: moving hz. can't merge. finished\n",
			func));
		return T_NO;
	}

	/* check they're going in same direction */
	if(	( (a->deltay > 0) && (b->deltay < 0) ) ||
		( (a->deltay < 0) && (b->deltay > 0) )
	  ) {
		thincDbg(T_DBG_CMD, ("%s: going in opposite directions, "
			"can't merge. finished\n", func));
		return T_NO;
	}

	/* XXX just simple regions please */
	if( 	(REGION_NUM_RECTS(ar) != 1) ||
		(REGION_NUM_RECTS(br) != 1)
	) {
		thincDbg(T_DBG_CMD, ("%s: commands not simple enough to merge "
			"finished\n", func));
		return T_NO;
	}

	ab = REGION_EXTENTS(pThinc->pScreen, ar);
	bb = REGION_EXTENTS(pThinc->pScreen, br);

#if 0
	/* since we store the destination, we need to go through all this
	 * scheisse */
	box_translate(ab, 0, -a->deltay);
	box_translate(bb, 0, -b->deltay);
#endif

	print_box(T_DBG_CMD, ab, "a src");
	print_box(T_DBG_CMD, bb, "b src");

	/* XXX for now just vertical merge is allowed */
	if((ab->x1 != bb->x1) || (ab->x2 != bb->x2)) {
		thincDbg(T_DBG_CMD, ("%s: copy boxes do not match x, "
			"finished\n", func));
#if 0
		/* undo */
		box_translate(ab, 0, a->deltay);
		box_translate(bb, 0, b->deltay);
#endif
		return T_NO;
	}

#if 0
	/* undo */
	box_translate(ab, 0, a->deltay);
	box_translate(bb, 0, b->deltay);
#endif

	thincDbg(T_DBG_CMD, ("%s: yes we can. finished\n", func));
	return T_YES;
}
