/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for hw cursor handling.
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */
/*
 * docs: Xserver/hw/xfree86/ramdac/CURSOR.NOTES
 */

/* TODO:
 * What happens if the driver specifies its own hardware cursor?
 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include "xorg/xf86Cursor.h"
#include "cursorstr.h"	/* needed for realizeCursor */

#include <netinet/in.h>

#define T_CURSOR_MAX_W	64
#define T_CURSOR_MAX_H	64

#define t_cursor(pThinc) \
	thincCursorPtr cursor = (pThinc)->cursor

#define t_cursor_pre(pScreen) \
	thincScreenPtr pThinc = THINCPtr(pScreen);	\
	t_cursor(pThinc)

#define t_cursor_pre_scrn(pScrn) \
	t_cursor_pre((pScrn)->pScreen)

/* use noret for void */
#define t_cursor_chk(ret)			\
	if(!pThinc->useHWCursor || !cursor)	\
		return ret;

#define t_cursor_sz(cursor)					\
	( (cursor)->argb ? 					\
	  ((cursor)->width << 2) * (cursor)->height : 		\
	  2 * bitmap_size((cursor)->width, (cursor)->height)	\
	)

#define t_cursor_maxsz(cursor)	\
	( ((cursor->info)->MaxWidth << 2) * (cursor->info->MaxHeight) )

/**** required ****/
static void thincShowCursor(ScrnInfoPtr pScreen);
static void thincHideCursor(ScrnInfoPtr pScreen);
static void thincShowHideCursor(thincScreenPtr pThinc, int showhide);
/* set the cursor position to x,y. x or y may be < 0 -> cursor image is
 * partially offscreen. It's up to the driver to trap for this and deal
 * with the situation */
static void thincSetCursorPosition(ScrnInfoPtr pScreen, int x, int y);
/* set the cursor's fg and bg colors. 
 * In 8bpp -> fg,bg are indices into current colormap unless the
 * flag HARDWARE_CURSOR_TRUECOLOR_AT_8BPP is set. 
 * fg,bg -> 8-8-8 RGB format */
static void thincSetCursorColors(ScrnInfoPtr pScreen, int bg, int fg);
/* hw cursor bits computed by RealizeCursor will be passed to the
 * driver when the cursor shape needs to be changed */
static void thincLoadCursorImage(ScrnInfoPtr pScreen, unsigned char *bits);
#ifdef ARGB_CURSOR
static Bool thincUseHWCursorARGB(ScreenPtr, CursorPtr);
static void thincLoadCursorARGB(ScrnInfoPtr, CursorPtr);
#endif /* ARGB_CURSOR */

/**** optional ****/
/* If not provided one will be given based on the Flags passed by the
 * driver. Function must be provided if the cursor format is not
 * supported by the ramdac module */
/*
static char* thincRealizeCursor(xf86CursorInfoPtr, CursorPtr);
 */

/* If the driver is unable to use a hardware cursor for reasons
 * other than the cursor being larger than the maximum specified
 * in the MaxWidth or MaxHeight field below, it can supply the
 * UseHWCursor function.  If UseHWCursor is provided by the driver,
 * it will be called whenever the cursor shape changes or the video
 * mode changes.  This is useful for when the hardware cursor cannot
 * be used in interlaced or doublescan modes.
 *
 * Yeah, yeah, bladabladablada. we just use it to get the hostspot info 
 * every time the cursor image is changed. We smart dude.
 */
static Bool thincUseHWCursor(ScreenPtr, CursorPtr);

/* Thinc-Command structure for cursor commands:
 * => All cursor commands are wrapped by struct cursorCmd
 * => Cursor commands are scheduled as real time. Flag: T_CMD_RT
 * => Cursor commands are unmodifiable (for now. Eventually we may allow
 * modification by their own type). Flag: T_CMD_NO_OVER
 * => Cursor commands do not overwrite other commands. Flag: T_CMD_TRANS
 * => There's no cursor create/execute function.
 * => New commands are created in-situ by each function called by X, buffered 
 * using thincBufferCmd and then destroyed (easiest path).
 */

struct cursorCmd {
	int which;
};

/* which */
#define T_CMD_CURSOR_CHANGE	1
#define T_CMD_CURSOR_SHOW	2
#define T_CMD_CURSOR_MOVE	3
#define T_CMD_CURSOR_COLOR	4

static thincCmdPtr thincCursorCreateGen(int which);
static int thincCursorBufferGen(thincScreenPtr pThinc, int which);
static int thincCursorFlush_color(thincScreenPtr pThinc, thincClientPtr cl);
static int thincCursorFlush_move(thincScreenPtr pThinc, thincClientPtr cl);
static int thincCursorFlush_sh(thincScreenPtr pThinc, thincClientPtr cl); 
static int thincCursorFlush_change(thincScreenPtr pThinc, thincClientPtr cl); 

/* Initialize our Hardware Cursor. Insipired on mga_hwcurs.c and 
 * ramdac/CURSOR.NOTES */
int THINCHWCursorInit(ScreenPtr pScreen)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = THINCPtr(pScreen);
	thincCursorPtr cursor;
	xf86CursorInfoPtr infoPtr;

	thincDbg(T_DBG_INPUT, ("%s: initializing hw cursor\n", func));

	pThinc->cursor = malloc(sizeof(*pThinc->cursor));
	if(!pThinc->cursor) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for cursor\n", func));
		return T_ERROR;
	}
	cursor = pThinc->cursor;

	infoPtr = xf86CreateCursorInfoRec(); 
	if(!infoPtr) 
		return T_ERROR;

	thincDbg(T_DBG_INPUT, ("%s: initializing CursorInfo struct\n", func));

	cursor->info = infoPtr;
	/* we start assuming it's not ARGB */
	cursor->argb = T_FALSE;

	/* width and height are in pixels. The largest sized cursor that can 
	 * be a hw cursor. Will fall * back to a software cursor if the cursor
	 * is bigger than this. */
	cursor->width = infoPtr->MaxWidth = T_CURSOR_MAX_W;
	cursor->height = infoPtr->MaxHeight = T_CURSOR_MAX_H; 
	cursor->size = t_cursor_sz(cursor);

	thincDbg(T_DBG_INPUT, ("%s: cursor is %dx%d, %d bytes. max size "
		"%d bytes\n", func, cursor->width, cursor->height, 
		cursor->size, t_cursor_maxsz(cursor)));
	/* allocate memory for the cursor. we allocate the maximum possible
	 * so there's no need to realloc later on */
	cursor->cursor = calloc(1, t_cursor_maxsz(cursor));
	if(!cursor->cursor) {
		xf86DestroyCursorInfoRec(infoPtr);
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for cursor\n", func));
		return T_ERROR;
	}

	/* these are only initialization values. They will 
	 * be filled with appropriate values by the cursor
	 * functions */
	cursor->x = 0;
	cursor->y = 0;
	cursor->xhot = 0;
	cursor->yhot = 0;
	cursor->fg = 0;
	cursor->bg = 0xFFFFFF;
	cursor->on = T_SHOW;

	/* FLAGS:
	 * 
	 * Color related:
	 * HARDWARE_CURSOR_TRUECOLOR_AT_8BPP:
	 *  Always pass fg/bg in RGB even in 8bpp (if not set, in 8bpp
	 *  fg,bg would be indices into the colormap)
	 *  
	 * Data loading:
	 * HARDWARE_CURSOR_SHOW_TRANSPARENT
	 *  Don't call HideCursor. Instead pass a completely transparent
	 *  cursor to LoadCursorImage
	 * HARDWARE_CURSOR_UPDATE_UNHIDDEN
	 *  Prevents a call to HideCursor just before LoadCursorImage is
	 *  called.
	 *  
	 * Data packing:
	 * Hardware cursor data t_dbg_name(func)
	 * The mask is a bitmap indicating which parts of the cursor are 
	 * transparent and which parts are drawn.  The source is a bitmap
	 * indicating which parts of the non-transparent portion of the the
	 * cursor should be painted in the foreground color and which should
	 * be painted in the background color.
	 * HARDWARE_CURSOR_INVERT_MASK
	 *  By default, set bits indicate the opaque part of the mask bitmap
	 *  and clear bits indicate the transparent part. This flag will 
	 *  invert the mask.
	 * HARDWARE_CURSOR_SWAP_SOURCE_AND_MASK
	 *  By defalt the source is stored first and then the mask. Set the
	 *  flag to reverse this order.
	 * HARDWARE_CURSOR_AND_SOURCE_WITH_MASK
	 *  This flag will have the module logical AND the source with the 
	 *  mask to make sure there are no source bits set if the 
	 *  corresponding mask bits aren't set.
	 * HARDWARE_CURSOR_BIT_ORDER_MSBFIRST
	 *  By default, it is assumed that the least significant bit in each 
	 *  byte corresponds to the leftmost pixel on the screen. Set flag
	 *  to reverse this order.
	 * HARDWARE_CURSOR_NIBBLE_SWAPPED
	 *  To enable byte swapping of the hw cursor.
	 *
	 * source-mask interleaving:
	 * By default source and mask data are inlined (source first). 
	 * Another option is to pack them interleaved, ie. X number of 
	 * source bits followed by X mask bits.
	 * HARDWARE_CURSOR_SOURCE_MASK_NOT_INTERLEAVED
	 *  Default
	 * HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_1        
	 * HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_8        
	 * HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_16       
	 * HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_32       
	 * HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_64
	 * For interleaved cursors.
	 */
	infoPtr->Flags = HARDWARE_CURSOR_TRUECOLOR_AT_8BPP |
			 HARDWARE_CURSOR_UPDATE_UNHIDDEN |
			 HARDWARE_CURSOR_SOURCE_MASK_NOT_INTERLEAVED |
			 HARDWARE_CURSOR_AND_SOURCE_WITH_MASK |
			 /* our current clients can't hide the cursor */
			 HARDWARE_CURSOR_SHOW_TRANSPARENT;

	/* the cursor handling funcs. */
	infoPtr->SetCursorColors = thincSetCursorColors;
	infoPtr->SetCursorPosition = thincSetCursorPosition;
	infoPtr->LoadCursorImage = thincLoadCursorImage;
	infoPtr->HideCursor = thincHideCursor;
	infoPtr->ShowCursor = thincShowCursor;
	/* we use UseHWCursor to get the hotspot info. RealizeCursor is not
	 * good enough, because it's only called when a new cursor is loaded,
	 * not on every change */
	infoPtr->UseHWCursor = thincUseHWCursor;
#ifdef ARGB_CURSOR
	infoPtr->UseHWCursorARGB = thincUseHWCursorARGB;
	infoPtr->LoadCursorARGB = thincLoadCursorARGB;
#endif

	thincDbg(T_DBG_INPUT, ("%s: registering cursor with xf86\n", func));

	if(!xf86InitCursor(pScreen, infoPtr)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize the "
			"cursor\n", func));
		xf86DestroyCursorInfoRec(cursor->info);
		free(cursor->cursor);
		free(cursor);
		pThinc->cursor = NULL;
		return T_ERROR;
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
	return T_OK;
}

void THINCHWCursorUninit(thincScreenPtr pThinc)
{
	t_cursor(pThinc);
	if(!cursor)
		return;

	xf86DestroyCursorInfoRec(cursor->info);
	free(cursor->cursor);
	free(cursor);
	pThinc->cursor = NULL;
}

/* This just steals the hot spot info about the cursor and then call's
 * the default realize cursor function */
static Bool thincUseHWCursor(ScreenPtr pScreen, CursorPtr pCursor)
{
	t_dbg_name(func)
	t_cursor_pre(pScreen);

	t_cursor_chk(FALSE);

	if( (pCursor->bits->width > cursor->width) || 
	    (pCursor->bits->height > cursor->height) ) 
	{
		thincDbg(T_DBG_INPUT, ("%s: cursor is too big (%dx%d)\n",
			func, pCursor->bits->width, 
			pCursor->bits->height));
		return FALSE;	/* use x's bool */
	}

	cursor->xhot = pCursor->bits->xhot;
	cursor->yhot = pCursor->bits->yhot;

	thincDbg(T_DBG_INPUT, ("%s: xhot %d, yhot %d\n", func, 
		cursor->xhot, cursor->yhot));
	thincDbg(T_DBG_INPUT, ("%s: cursor is %dx%d\n", func,
		pCursor->bits->width, pCursor->bits->height));

	return T_TRUE;
}

/* Send T_CURSOR_CHANGE message */
static void thincLoadCursorImage(ScrnInfoPtr pScrn, unsigned char *bits)
{
	t_dbg_name(func)
	int res;
	t_cursor_pre_scrn(pScrn);
	
	/* just in case */
	t_cursor_chk(noret);

	thincDbg(T_DBG_INPUT, ("%s: xhot %d yhot %d\n", func, 
		cursor->xhot, cursor->yhot));

	/* clean things up first */
	memset(cursor->cursor, 0, cursor->size);

#ifdef ARGB_CURSOR
	if(cursor->argb) {
		thincDbg(T_DBG_INPUT, ("%s: changing from ARGB to normal "
			"cursor\n", func));
		cursor->argb = T_FALSE;
		cursor->size = t_cursor_sz(cursor);
	}
#endif

	/* copy it to our local storage */
	memcpy(cursor->cursor, bits, cursor->size);

	res = thincCursorBufferGen(pThinc, T_CMD_CURSOR_CHANGE);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer cursor "
			"command\n", func));
		return; 
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}

#ifdef ARGB_CURSOR
/* based on radeon argb cursor (ati/radeon_cursor.c) */
static Bool thincUseHWCursorARGB(ScreenPtr pScreen, CursorPtr pCursor)
{
	t_dbg_name(func)
	t_cursor_pre(pScreen);

	t_cursor_chk(FALSE);

	if( (pCursor->bits->width > cursor->width) || 
	    (pCursor->bits->height > cursor->height) ) 
	{
		thincDbg(T_DBG_INPUT, ("%s: cursor is too big (%dx%d)\n",
			func, pCursor->bits->width, 
			pCursor->bits->height));
		return FALSE;	/* use x's bool */
	}

	cursor->xhot = pCursor->bits->xhot;
	cursor->yhot = pCursor->bits->yhot;

	thincDbg(T_DBG_INPUT, ("%s: xhot %d, yhot %d\n", func, 
		cursor->xhot, cursor->yhot));
	thincDbg(T_DBG_INPUT, ("%s: cursor is %dx%d\n", func,
		pCursor->bits->width, pCursor->bits->height));

	return T_TRUE;
}

static void thincLoadCursorARGB(ScrnInfoPtr pScrn, CursorPtr pCursor)
{
	t_dbg_name(func)
	t_cursor_pre_scrn(pScrn);
	unsigned long *from, *to, *next;
	int x, y, w, h, err;

	thincDbg(T_DBG_INPUT, ("%s: cursor is %dx%d\n", func,
		pCursor->bits->width, pCursor->bits->height));

	/* clean things up first */
	memset(cursor->cursor, 0, cursor->size);

	if(!cursor->argb) {
		thincDbg(T_DBG_INPUT, ("%s: changing from normal to ARGB "
			"cursor\n", func));
		cursor->argb = T_TRUE;
		cursor->size = t_cursor_sz(cursor);
	}

	w = MIN(pCursor->bits->width, cursor->width);
	h = MIN(pCursor->bits->height, cursor->height);
	from = pCursor->bits->argb;
	to = (unsigned long*)cursor->cursor;
	for(y=0; y< h; ++y) {
		next = from + pCursor->bits->width;
		for(x=0; x<w; ++x)
			*to++ = *from++;

		/* now go to next line */
		for(; x<cursor->width; ++x, ++to);

		from = next;
	}

	err = thincCursorBufferGen(pThinc, T_CMD_CURSOR_CHANGE);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer "
			"cursor command\n", func));
		return;
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}
#endif	/* ARGB_CURSOR */

/* Send T_CURSOR_SHOW message */
static void thincShowCursor(ScrnInfoPtr pScrn)
{
	thincShowHideCursor(THINCPtr(pScrn->pScreen), T_SHOW);
}

/* Send T_CURSOR_HIDE message */
static void thincHideCursor(ScrnInfoPtr pScrn)
{
	thincShowHideCursor(THINCPtr(pScrn->pScreen), T_HIDE);
}

static void thincShowHideCursor(thincScreenPtr pThinc, int showhide)
{
	t_dbg_name(func)
	int res;
	t_cursor(pThinc);

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	cursor->on = showhide;
	res = thincCursorBufferGen(pThinc, T_CMD_CURSOR_SHOW);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer cursor "
			"show/hide command\n", func));
		return;
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}

/* Send T_CURSOR_COLOR message. Only for core cursors, not ARGB */
static void thincSetCursorColors(ScrnInfoPtr pScrn, int bg, int fg)
{
	t_dbg_name(func)
	t_cursor_pre_scrn(pScrn);
	int res;

	cursor->fg = fg;
	cursor->bg = bg;

	thincDbg(T_DBG_INPUT, ("%s: fg %ld, bg %ld\n", func, cursor->fg,
		cursor->bg));

	res = thincCursorBufferGen(pThinc, T_CMD_CURSOR_COLOR);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer cursor "
			"command\n", func));
		return; 
	}
	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}

/* Send T_CURSOR_MOVE message */
static void thincSetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
{
	t_dbg_name(func)
	t_cursor_pre_scrn(pScrn);
	int res;

	thincDbg(T_DBG_INPUT, ("%s: x %d, y %d\n", func, x, y));

	cursor->x = x;
	cursor->y = y;

	/* if counter is 0, this position was not generated as
	 * a response to a mouse movement, and need to send it to all 
	 * clients. Otherwise, decrease counter and send to all clients
	 * except one that sent last event (checked at flush time) */
	if(pThinc->moveCount == 0) {
		thincDbg(T_DBG_INPUT, ("%s: Got app move event. Sending to "
			"all clients\n", func));
		pThinc->last_mouse_client = NULL;
	} else { /* normal event, update the counter  */
		pThinc->moveCount--;
		thincDbg(T_DBG_INPUT, ("%s: normal move event. have %d "
			"left\n", func,	pThinc->moveCount));
	}
	res = thincCursorBufferGen(pThinc, T_CMD_CURSOR_MOVE);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer "
			"cursor position command\n", func));
		return;
	}
	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}

/**** command functions ****/

/* generic create function called by 'which'-specific create functions. Do not
 * confuse with proper cursor create function (which is not defined) */
static thincCmdPtr thincCursorCreateGen(int which)
{
	t_dbg_name(func)
	thincCmdPtr cmd;
	struct cursorCmd *cursor;
	int cmdsize;

	cmdsize = sizeof(struct cursorCmd) + T_CMDREC_SIZE;
	cmd = malloc(cmdsize);
	if(!cmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for new cursor command\n", func));
		return NULL;
	}

	cmd->type = T_FB_CURSOR;
	cmd->flags = T_CMD_TRANS | T_CMD_NO_OVER | T_CMD_RT;
	cmd->id = thinc_getCmdID();
	cmd->area = thincNullBox;
	cmd->cmdSize = sizeof(struct cursorCmd);
	cmd->refcnt = 0;
	t_list_init(&cmd->list);
	t_list_init(&cmd->sched);

	cursor = cmd_data(cmd, struct cursorCmd*);
	cursor->which = which;

	return cmd;
}

static int thincCursorBufferGen(thincScreenPtr pThinc, int which)
{
	t_dbg_name(func)
	thincCmdPtr cmd;
	int err;

	cmd = thincCursorCreateGen(which);
	if(!cmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create cursor "
			"command\n", func));
		return T_ERROR;
	}

	/* NULL -> pWindow: Used to find duplicates, so o.k. to pass null */
	err = thincBufferCmd(pThinc, NULL, cmd);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer cursor "
			"command\n", func));
		thincCursorDestroy(pThinc, cmd);
		return T_ERROR;
	}

	/* we're finished, get rid of it. It's this or having a separate
	 * create function :/ */
	thincCursorDestroy(pThinc, cmd);
	return T_OK;
}


void thincCursorDestroy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	free(cmd);
}

thincCmdPtr thincCursorCopy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	thincCmdPtr cmd2;
	int size;

	size = cmd->cmdSize + T_CMDREC_SIZE;
	cmd2 = malloc(size);
	if(!cmd2) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory for "
			"copy of cursor command\n", func));
		return NULL;
	}
	memcpy(cmd2, cmd, size);
	return cmd2;
}

int thincCursorFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct cursorCmd *cursor;
	int err;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	cursor = cmd_data(cmd, struct cursorCmd*);
	thincDbg(T_DBG_INPUT, ("%s: got cursor command %d\n", func,
		cursor->which));
	
	switch(cursor->which) {
		case T_CMD_CURSOR_CHANGE:
			err = thincCursorFlush_change(pThinc, cl);
			break;
		case T_CMD_CURSOR_SHOW:
			err = thincCursorFlush_sh(pThinc, cl);
			break;
		case T_CMD_CURSOR_MOVE:
			err = thincCursorFlush_move(pThinc, cl);
			break;
		case T_CMD_CURSOR_COLOR:
			err = thincCursorFlush_color(pThinc, cl);
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Unknown cursor "
				"message %d to flush\n", func, cursor->which));
			return T_ERROR;
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
	return err;
}

/* TODO: Purge other cursor change messages from queue. Alternatively, allow
 * a new cursor change msg to purge old ones (see TODO in flags in 
 * thincCommand.h) */
static int thincCursorFlush_change(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_cursorChange *c;
	int res, flags;
	char *curr;
	t_cursor(pThinc);

	thincDbg(T_DBG_INPUT, ("%s: sending CURSOR CHANGE message to client\n",
		func));

	flags = 0;
	if(cursor->argb)
		flags |= T_CURSOR_ARGB;

	c = thincBufferMsg(cl, T_CURSOR_CHANGE, flags, T_CURSOR_CHANGE_SIZE,
		cursor->size);
	if(!c) {
		/* send it to the clients */
		c = (thinc_cursorChange*)t_msgBuffer;
		c->xhot = cursor->xhot;
		c->yhot = cursor->yhot;
		c->size = htons(cursor->size);
		res = thincSendMsg(cl, T_CURSOR_CHANGE, T_CURSOR_CHANGE_SIZE, 
			flags);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send cursor "
				"change command\n", func));
			return T_ERROR;
		}
		/* send the new cursor. First source, then mask */
		res = thincWriteData(cl, cursor->cursor, cursor->size);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"cursor change command\n", func));
			return T_ERROR;
		}
	} else {
		c->xhot = cursor->xhot;
		c->yhot = cursor->yhot;
		c->size = htons(cursor->size);
		curr = ((char*)c) + T_CURSOR_CHANGE_SIZE;

		memcpy(curr, cursor->cursor, cursor->size);

		thincDbg(T_DBG_INPUT, ("%s: buffered cursor change "
			"message\n", func));
	}
	thincDbg(T_DBG_INPUT, ("%s: sent CURSOR_CHANGE message to "
		"client %d,%d\n", func, c->xhot, c->yhot));
	return T_OK;
}

static int thincCursorFlush_sh(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_cursorShowHide *s;
	t_cursor(pThinc);
	int res;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	s = thincBufferMsg(cl, T_CURSOR_SHOWHIDE, 0,T_CURSOR_SHOWHIDE_SIZE, 
		0);
	if(!s) {
		s = (thinc_cursorShowHide*)t_msgBuffer;
		s->showhide = cursor->on;
		
		res = thincSendMsg(cl, T_CURSOR_SHOWHIDE, 
			T_CURSOR_SHOWHIDE_SIZE, 0);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"cursor message to client\n", func));
			return T_ERROR;
		}
	} else {
		s->showhide = cursor->on;
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
	return T_OK;
}

static int thincCursorFlush_move(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	int res;
	thinc_cursorMove *m;
	t_cursor(pThinc);

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	if (cl == pThinc->last_mouse_client) {
		thincDbg(T_DBG_INPUT, ("%s: client is one that sent "
			"mouse event. finished\n", func));
		return T_OK;
	}

	m = thincBufferMsg(cl, T_CURSOR_MOVE, 0, T_CURSOR_MOVE_SIZE, 0);
	if(!m) {
		m = (thinc_cursorMove*)t_msgBuffer;
		m->x = ntohs(cursor->x);
		m->y = ntohs(cursor->y);
				
		res = thincSendMsg(cl, T_CURSOR_MOVE, 
			T_CURSOR_MOVE_SIZE, 0);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"cursor message to client\n", func));
			return T_ERROR;
		}
	} else {
		m->x = ntohs(cursor->x);
		m->y = ntohs(cursor->y);
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
	return T_OK;
}

static int thincCursorFlush_color(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	int res;
	thinc_cursorColor *c;
	t_cursor(pThinc);

	thincDbg(T_DBG_INPUT, ("%s\n", func));
	c  = thincBufferMsg(cl, T_CURSOR_COLOR, 0,
		T_CURSOR_COLOR_SIZE, 0);
	if(!c) {
		c = (thinc_cursorColor*)t_msgBuffer;
		c->fg = htonl(cursor->fg);
		c->bg = htonl(cursor->bg);
		
		res = thincSendMsg(cl, T_CURSOR_COLOR, T_CURSOR_COLOR_SIZE, 
			0);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"cursor message to client\n", func));
			return res;
		}
	} else {
		c->fg = htonl(cursor->fg);
		c->bg = htonl(cursor->bg);
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
	return T_OK;
}

/* TODO */
void thincCursorGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...)
{
	T_BUG_ARG("Cursor GetInfo function called. write me!! They want %d",
		info);
}

/* unmodifiable. don't do anything */
thincCmdPtr thincCursorModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	return cmd;
}

/* these are required but not used */

/* create is embedded on each of the functions above. This one shouldn't be
 * used */
thincCmdPtr thincCursorCreate(thincScreenPtr pThinc, RegionPtr pRegion, 
	va_list args)
{
	T_BUG("Cursor Create function called. why??");
	return NULL;
}

/* not used. not stored in offscreen pixmaps */
void thincCursorExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay)
{
	T_BUG("I was called!! What's going on??");
}
