/* Copyright 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for Solid and Pixmap Fill messages
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include <netinet/in.h>

/* local vars */
#ifdef T_GETSTATS
static unsigned long t_start;	/* needed for statistics */
#endif

/********************* Solid Fills ***************************/
/* the methods */
#ifdef T_GETSTATS
static int thinc_fillPreStats(thincClientPtr, void *);
static int thinc_fillPostStats(thincClientPtr, void *);
#endif
static int thinc_fillSend(thincClientPtr cl, void *msg);
#ifdef T_RECORD
static int thinc_fillRecord(thincClientPtr cl, void *msg);
static int thinc_fillRecordCb(thincClientPtr cl, char *buf, unsigned int *len,
	void *ptr);
#endif

/* the priorities, current order:
 * PRESTATS
 * SEND
 * RECORD
 * POSTSTATS
 */
#ifdef T_GETSTATS
#define T_FILLPRIO_PRESTATS	T_PRIO_MIN
#define T_FILLPRIO_POSTSTATS	T_PRIO_MAX
#endif
#define T_FILLPRIO_SEND		T_PRIO_SEND
#ifdef T_RECORD
#define T_FILLPRIO_RECORD	T_FILLPRIO_SEND+1
#endif

/* the static modules. From here, we make copies to each client's stack */
#ifdef T_GETSTATS
thincDefModule(fillPreStats, "FILL_PRESTATS", 14, T_FILLPRIO_PRESTATS);
thincDefModule(fillPostStats, "FILL_POSTSTATS", 15, T_FILLPRIO_POSTSTATS);
#endif
thincDefModule(fillSend, "FILL_SEND", 10, T_FILLPRIO_SEND);
#ifdef T_RECORD
thincDefModule(fillRecord, "FILL_RECORD", 12, T_FILLPRIO_RECORD);
#endif

typedef struct thincFillObj_S {
	unsigned long	color:32;
	BoxPtr		rects;
	unsigned int	numrects:16;

	unsigned int	total_len:32;
} thincFillObj;

#define fill_addmodule(cl, mod) { \
	err = thinc_msgAddUsingModule((cl), (cl)->fillStack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add SFILL module"\
			" %s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}

int thinc_fillInitStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_CLIENT, ("%s: creating SFILL message stack\n", func));

	cl->fillStack = thinc_msgStackCreate(cl);
	if(!cl->fillStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create SFILL "
			"message stack\n", func));
		return T_ERROR;
	}

#ifdef T_GETSTATS
	fill_addmodule(cl, &fillPreStats);
	fill_addmodule(cl, &fillPostStats);
#endif

	/* add the modules */
	fill_addmodule(cl, &fillSend);

#if 0
#ifdef T_RECORD
	/* only for first client */
	if(cl->parent->doRecord) {
		thincScreenPtr pThinc = cl->parent;
		thincRecordPtr record = t_recordPtr(pThinc);
		/* this mess because when we get called, clientCount hasn't
		 * been set */
		if( 
		    ((record->cl_enabled) && (pThinc->clientCount == 1) ) ||
		    (pThinc->clientCount == 0)
		) {
			fill_addmodule(cl, &fillRecord);
		}
	}
#endif
#endif

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

void thinc_fillDestroyStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->fillStack);
	cl->fillStack = NULL;
}

#if 0
static int thincSendFBSolidFill(thincScreenPtr pThinc, thincClientPtr cl,
	RegionPtr reg, unsigned long pixel);

int thincSFillFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
	struct fillCmd *fill = cmd_data(cmd, struct fillCmd*);
	int ret;
	stats_var2(t1, t2);

	stats_init_simple(writeTime);
	stats_init_simple(blockTime);
	stats_start(pThinc, t1);

	ret = thincSendFBSolidFill(pThinc, cl, &fill->reg, fill->color);

	stats_end(pThinc, t1, t2, sfillStats.flush);
	stats_add(pThinc, sfillStats.write, writeTime);
	stats_add(pThinc, sfillStats.block, blockTime);
	return ret;
}

/* changes the passed region */
static int thincSendFBSolidFill(thincScreenPtr pThinc, thincClientPtr cl,
	RegionPtr reg, unsigned long pixel)
{
	const char *func = __func__;
	thinc_fbSFill *s;
	thinc_rectRegion *rp;	/* used when buffering */
	thinc_rectRegion r;	/* used when sending directly */
	int numrects, i, res, datasize;
	register BoxPtr rects;
	int nobuffer = T_FALSE;

#ifndef DISABLE_NORAW
	thincDbg(T_DBG_PROTO, ("%s: sending fill with color %#lx\n", func, 
		pixel));

#if 0
	REGION_INTERSECT(pThinc->pScreen, reg, reg, &cl->clClip);
	if(!REGION_NOTEMPTY(pThinc->pScreen, reg)) {
		thincDbg(T_DBG_PROTO, ("%s: finished. Tried to send fill for "
			"offscreen region\n",
			func));
		return T_OK;
	}
#endif

	/* check how big our message is going to be */
	numrects = REGION_NUM_RECTS(reg);

	datasize = numrects * T_RECTREGION_SIZE;

	thincDbg(T_DBG_PROTO, ("%s: sending fill for %d rectangles. Total "
		"size %d\n", func, numrects, datasize));

	s = (thinc_fbSFill*)thincBufferMsg(cl, T_FB_SFILL, 0, T_FB_SFILL_SIZE,
		datasize);
	if(!s) {
		/* do normal writes */
		thincDbg(T_DBG_PROTO, ("%s: Couldn't buffer SFILL message\n", 
			func));
		thincFlushBuffer(cl);
		s = (thinc_fbSFill*)t_msgBuffer;
		nobuffer = T_TRUE;
	}

	/* build the message */
	s->pixel = htonl(pixel);
	s->numrects = htons(numrects);

	if(nobuffer) {
		res = thincSendMsg(cl, T_FB_SFILL, T_FB_SFILL_SIZE, 0);
		if(res == T_ERROR) {
			thincError("ERROR: %s: Couldn't send SFILL message\n",
				func);
			return T_ERROR;
		}
		rects = REGION_RECTS(reg);

		for(i=0; i<numrects; i++, rects++) {
			box_to_rect_hton(rects, r);

			res = thincWriteData(cl, &r, T_RECTREGION_SIZE);
			if(res == T_ERROR) {
				thincError("ERROR: %s: Couldn't send SFILL "
					"message\n", func);
				return T_ERROR;
			}
		}
	} else {
		/* if we use s+1 we'll include the padding in the message :/ */
		rp = (thinc_rectRegion*) (((char*)s) + T_FB_SFILL_SIZE);
		rects = REGION_RECTS(reg);
		
		for(i=0; i<numrects; ++i, ++rects, ++rp) {
			print_box(T_DBG_PROTO, rects, "sending rectangle:");
			box_to_rectp_hton(rects, rp);
			print_rect(T_DBG_PROTO, rp, "sent rectangle:");
		}
	}

	stats_inc(&pThinc->sfillStats, count);
	stats_add(pThinc, sfillStats.size, 
		datasize + T_FB_SFILL_SIZE + THINC_HEADERSIZE);
	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
#endif

	return T_OK;
}
#endif
int thincSFillFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct fillCmd *fill = cmd_data(cmd, struct fillCmd*);
	int ret;
	thincFillObj obj;

	thincDbg(T_DBG_PROTO, ("%s: flushing command %#lx\n", func, cmd->id));

	obj.rects = REGION_RECTS(&fill->reg);
	obj.numrects = REGION_NUM_RECTS(&fill->reg);
	obj.color = fill->color;
	obj.total_len = 0;

	ret = thinc_msgStackExec(cl, cl->fillStack, &obj);

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return ret;
}

#ifdef T_GETSTATS
static int thinc_fillPreStats(thincClientPtr cl, void *msg)
{
	stats_start_nostr(t_start);
	stats_init_simple(writeTime);
	stats_init_simple(blockTime);

	return T_OK;
}

static int thinc_fillPostStats(thincClientPtr cl, void *msg)
{
	unsigned long t_end;
	thincFillObj *obj = msg;

	/* TODO: Maybe stats should be per client?? */
	stats_end(cl->parent, t_start, t_end, sfillStats.flush);
	stats_inc(cl->parent, sfillStats.count);
	stats_add(cl->parent, sfillStats.write, writeTime);
	stats_add(cl->parent, sfillStats.block, blockTime);
	stats_add(cl->parent, sfillStats.size, obj->total_len);
	return T_OK;
}
#endif /* T_GETSTATS */

static int thinc_fillSend(thincClientPtr cl, void *m)
{
	t_dbg_modname(modname, fillSend.name)
	thincFillObj *obj = m;
	thinc_fbSFill *s;
	thinc_rectRegion *rp;	/* used when buffering */
	thinc_rectRegion r;	/* used when sending directly */
	unsigned int numrects, i, datasize;
	int res;
	register BoxPtr rects;
	int nobuffer = T_FALSE;

	thincDbg(T_DBG_PROTO, ("%s: sending fill with color %#lx\n", modname, 
		obj->color));

	/* check how big our message is going to be */
	numrects = obj->numrects;
	datasize = numrects * T_RECTREGION_SIZE;

	thincDbg(T_DBG_PROTO, ("%s: sending fill for %d rectangles. Total "
		"size %d\n", modname, numrects, datasize));

	s = (thinc_fbSFill*)thincBufferMsg(cl, T_FB_SFILL, 0, T_FB_SFILL_SIZE,
		datasize);
	if(!s) {
		/* do normal writes */
		thincDbg(T_DBG_PROTO, ("%s: Couldn't buffer SFILL message\n", 
			modname));
		thincFlushBuffer(cl);
		s = (thinc_fbSFill*)t_msgBuffer;
		nobuffer = T_TRUE;
	}

	/* build the message */
	s->pixel = htonl(obj->color);
	s->numrects = htons(numrects);

	if(nobuffer) {
		res = thincSendMsg(cl, T_FB_SFILL, T_FB_SFILL_SIZE, 0);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"SFILL message\n", modname));
			return T_ERROR;
		}
		rects = obj->rects;

		for(i=0; i<numrects; i++, rects++) {
			box_to_rect_hton(rects, r);

			res = thincWriteData(cl, &r, T_RECTREGION_SIZE);
			if(res == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"send SFILL message\n", modname));
				return T_ERROR;
			}
		}
	} else {
		/* if we use s+1 we'll include the padding in the message :/ */
		rp = (thinc_rectRegion*) (((char*)s) + T_FB_SFILL_SIZE);
		rects = obj->rects;
		
		for(i=0; i<numrects; ++i, ++rects, ++rp) {
			box_to_rectp_hton(rects, rp);
		}
	}

	obj->total_len = datasize + T_FB_SFILL_SIZE + THINC_HEADERSIZE;
	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));

	return T_OK;
}

/******************************** recording stuff ****************************/
#ifdef T_RECORD
int thinc_fillInitRecordStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_RECORD, ("%s: creating SFILL message stack for "
		"record command\n", func));

	cl->fillStack = thinc_msgStackCreate(cl);
	if(!cl->fillStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create SFILL "
			"message stack\n", func));
		return T_ERROR;
	}
	
	fill_addmodule(cl, &fillRecord);

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

void thinc_fillDestroyRecordStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->fillStack);
	cl->fillStack = NULL;
}

static int thinc_fillRecord(thincClientPtr cl, void *m)
{
	t_dbg_modname(modname, fillRecord.name)
	thincFillObj *obj = m;
	unsigned int numrects, datasize, o_total;
	int err;

	thincDbg(T_DBG_RECORD, ("%s: recording fill with color %#lx\n", 
		modname, obj->color));

	/* check how big our message is going to be */
	numrects = obj->numrects;
	datasize = numrects * T_RECTREGION_SIZE;

	o_total = obj->total_len;
	obj->total_len = datasize + T_FB_SFILL_SIZE;

	thincDbg(T_DBG_RECORD, ("%s: recording fill for %d rectangles. Total "
		"size %d\n", modname, numrects, obj->total_len));

	err = thincRecord(cl, obj->total_len, T_FB_SFILL, 0, 
		thinc_fillRecordCb, m);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't record SFILL "
			"command\n", modname));
		return T_ERROR;
	}

	obj->total_len = o_total;
	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

/* since we record so little data, we blindly assume that len has more than
 * enough space for us */
static int thinc_fillRecordCb(thincClientPtr cl, char *buf, unsigned int *len,
	void *ptr)
{
	t_dbg_modname(modname, fillRecord.name)
	thincFillObj *obj = ptr;
	thinc_fbSFill *s;
	thinc_rectRegion *rp;	/* used when buffering */
	unsigned int numrects, i;
	register BoxPtr rects;

	thincDbg(T_DBG_RECORD, ("%s: recording fill with color %#lx\n", 
		modname, obj->color));

	/* FIXME:
	 * about:plugins in firefox triggers this.
	 */
	if(*len < obj->total_len) {
		T_BUG_ARG_2("Can't record solid fill. need %d bytes, only "
			"have %d bytes available", obj->total_len, *len);
		return T_ERROR;
	}

	s = (thinc_fbSFill*)buf;
	numrects = obj->numrects;

	/* build the message */
	s->pixel = htonl(obj->color);
	s->numrects = htons(numrects);

	/* if we use s+1 we'll include the padding in the message :/ */
	rp = (thinc_rectRegion*) (((char*)s) + T_FB_SFILL_SIZE);
	rects = obj->rects;
	
	for(i=0; i<numrects; ++i, ++rects, ++rp) {
		box_to_rectp_hton(rects, rp);
	}

	*len = obj->total_len;

	thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u bytes\n", modname,
		*len));
	return T_REC_DONE;
}
#endif /* T_RECORD */

/**************************** Pixmap Fills **********************************/

/*********************** <OLD> ****************************/
#if 0
static int thincSendFBPixmapFill(thincScreenPtr pThinc, thincClientPtr cl,
	RegionPtr reg, char *pixmap, int x, int y, int w, int h);
#endif

#if 0
#define WRITEPIXTOFILE
#endif

#ifdef WRITEPIXTOFILE
/* needed for WriteFile module */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

int thinc_writePixmapToFile(thincClientPtr cl, char *pix,
	unsigned int w, unsigned int h);
#endif /* WRITEPIXTOFILE */

/************************** </OLD> *************************/

/* the methods */
#ifdef T_GETSTATS
static int thinc_pixPreStats(thincClientPtr, void *);
static int thinc_pixPostStats(thincClientPtr, void *);
#endif
static int thinc_pixmapSend(thincClientPtr cl, void *msg);
static int thinc_pixmapResize(thincClientPtr cl, void *msg);
static int thinc_pixmapCompress(thincClientPtr cl, void *msg);
static int thinc_pixmapChkCache(thincClientPtr cl, void *msg);
static int thinc_pixmapAddCache(thincClientPtr cl, void *msg);
#ifdef WRITEPIXTOFILE
static int thinc_pixmapWriteFile(thincClientPtr cl, void *msg);
#endif
#ifdef T_RECORD
static int thinc_pixmapRecord(thincClientPtr cl, void *msg);
static int thinc_pixmapRecordCb(thincClientPtr cl, char *buf, unsigned int *len,
	void *ptr);
#endif

/* the priorities. current order:
 * PRESTATS
 * RESIZE
 * CHKCACHE
 * COMPRESS
 * SEND
 * RECORD
 * ADDCACHE
 * POSTSTATS
 */
#ifdef T_GETSTATS
#define T_PIXPRIO_PRESTATS	T_PRIO_MIN
#define T_PIXPRIO_POSTSTATS	T_PRIO_MAX
#endif
#define T_PIXPRIO_SEND		T_PRIO_SEND
#define T_PIXPRIO_RESIZE	T_PIXPRIO_SEND-100
#define T_PIXPRIO_COMPRESS	T_PIXPRIO_SEND-10
#define T_PIXPRIO_CHKCACHE	T_PIXPRIO_SEND-50
#define T_PIXPRIO_ADDCACHE	T_PIXPRIO_SEND+10
#ifdef WRITEPIXTOFILE
/* change to modify where in the pipeline you want to save the data */
#define T_PIXPRIO_WRITEFILE	T_PIXPRIO_SEND-60
#endif
#ifdef T_RECORD
#define T_PIXPRIO_RECORD	T_PIXPRIO_SEND+1
#endif

/* the static modules. From here, we make copies to each client's stack */
#ifdef T_GETSTATS
thincDefModule(pixPreStats, "PIX_PRESTATS", 13, T_PIXPRIO_PRESTATS);
thincDefModule(pixPostStats, "PIX_POSTSTATS", 14, T_PIXPRIO_POSTSTATS);
#endif
thincDefModule(pixmapSend, "PIX_SEND", 9, T_PIXPRIO_SEND);
thincDefModule(pixmapResize, "PIX_RESIZE", 11, T_PIXPRIO_RESIZE);
thincDefModule(pixmapCompress, "PIX_COMPRESS", 13, T_PIXPRIO_COMPRESS);
thincDefModule(pixmapChkCache, "PIX_CHKCACHE", 13, T_PIXPRIO_CHKCACHE);
thincDefModule(pixmapAddCache, "PIX_ADDCACHE", 13, T_PIXPRIO_ADDCACHE);
#ifdef WRITEPIXTOFILE
thincDefModule(pixmapWriteFile, "PIX_WRITEFILE", 14, T_PIXPRIO_WRITEFILE);
#endif
#ifdef T_RECORD
thincDefModule(pixmapRecord, "PIX_RECORD", 11, T_PIXPRIO_RECORD);
#endif

/* the pixmap object */
#define T_PIX_EXTRALEN	32

typedef struct thincPixObj_S {
	unsigned int x:16;
	unsigned int y:16;
	unsigned int width:16;
	unsigned int height:16;

	/* Basically how much total data was sent as a result of this pixmap 
	 * object */
	unsigned int total_len:32;

	/* the pixmap data. Always stored contiguously */
	char *data;
	char *orig_data;
	unsigned int data_len:32;
	unsigned int data_width:32;	/* width of one line of the pixmap */

	/* the rectangles to fill. may be NULL */
	BoxPtr rects;
	unsigned int numrects:16;

	/* extra headers for the message */
	unsigned int flags:8;		/* message flags */
	unsigned int extra_len:8;	/* current length <= T_PIX_EXTRALEN */
	char *extra;		/* current end */
	char extra_data[T_PIX_EXTRALEN];

	unsigned int hash:32;
} thincPixObj;

/**** useful macros ****/
/* checks that we haven't overflowed the extra data buffer. Shouldn't happen
 * as long as extra messages and the buffer size are kept in sync. Should
 * be called every time something is added to the extra buffer */
#define checkExtraBuffer(obj) \
	if((obj)->extra_len > T_PIX_EXTRALEN) { \
		thincError("BUG: %s: %d: %s(): PIX update extra headers " \
			"overflowed static data buffer in PIXMAP message " \
			"object! Current lenght is %u but buffer's size is " \
			"%u\n", __FILE__, __LINE__, modname, obj->extra_len, \
			T_PIX_EXTRALEN); \
		return T_ERROR; \
	}

#define pixmap_addmodule(cl, mod) { \
	err = thinc_msgAddUsingModule((cl), (cl)->pixmapStack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add PIXMAP module"\
			" %s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}

/* initializes and adds appropriate modules to the PIXMAP message stack */
int thinc_pixmapInitStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;
	
	thincDbg(T_DBG_CLIENT, ("%s: creating PIXMAP message stack\n", func));
	
	cl->pixmapStack = thinc_msgStackCreate(cl);
	if(!cl->pixmapStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create PIXMAP "
			"message stack\n", func));
		return T_ERROR;
	}

#ifdef T_GETSTATS
	pixmap_addmodule(cl, &pixPreStats);
	pixmap_addmodule(cl, &pixPostStats);
#endif

	/* add the modules */
	pixmap_addmodule(cl, &pixmapSend);
#if 0	/* not for now */
	pixmap_addmodule(cl, &pixmapCompress);
#endif

	if(cl->parent->usePixCache) {
		/* cache needs the data contiguously */
		pixmap_addmodule(cl, &pixmapChkCache);
		pixmap_addmodule(cl, &pixmapAddCache);
	}

	/* only for debugging */
#ifdef WRITEPIXTOFILE
	pixmap_addmodule(cl, &pixmapWriteFile);
#endif	/* WRITEPIXTOFILE */

#if 0
#ifdef T_RECORD
	/* only for first client */
	if(cl->parent->doRecord) {
		thincScreenPtr pThinc = cl->parent;
		thincRecordPtr record = t_recordPtr(pThinc);
		/* this mess because when we get called, clientCount hasn't
		 * been set */
		if( 
		    ((record->cl_enabled) && (pThinc->clientCount == 1) ) ||
		    (pThinc->clientCount == 0)
		) {
			pixmap_addmodule(cl, &pixmapRecord);
		}
	}
#endif
#endif

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

void thinc_pixmapDestroyStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->pixmapStack);
	cl->pixmapStack = NULL;
}

int thinc_pixStackClResized(thincClientPtr cl, int prevstate)
{
	t_dbg_name(func)
	int err;

	if(!cl->parent->allowResize)
		return T_OK;

	if(cl->resized == T_CL_RESZ_LESS) {
		if(prevstate != T_CL_RESZ_LESS)
			pixmap_addmodule(cl, &pixmapResize);
	} else {
		if(prevstate == T_CL_RESZ_LESS) {
			thinc_msgDelUsingModule(cl, cl->pixmapStack, 
				&pixmapResize);
		}
	}
	return T_OK;
}

/************** flush a pixmap fill update ******************/
int thincPFillFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct pixmapFillCmd* pixmap = cmd_data(cmd, struct pixmapFillCmd*);
	char *pixdata = (char*) (pixmap+1);
	int ret;
#if 0
	stats_var2(t1, t2);

	stats_init_simple(writeTime);
	stats_init_simple(blockTime);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_PROTO, ("%s: flushing command %#lx\n", func, cmd->id));

	ret = thincSendFBPixmapFill(pThinc, cl, &pixmap->reg, pixdata,
		pixmap->x, pixmap->y, pixmap->w, pixmap->h);

	stats_end(pThinc, t1, t2, pfillStats.flush);
	stats_add(pThinc, pfillStats.write, writeTime);
	stats_add(pThinc, pfillStats.block, blockTime);


#else
	thincPixObj obj;

	thincDbg(T_DBG_PROTO, ("%s: flushing command %#lx\n", func, cmd->id));

	obj.x = pixmap->x;
	obj.y = pixmap->y;
	obj.width = pixmap->w;
	obj.height = pixmap->h;

	obj.data = pixdata;
	obj.orig_data = pixdata;
	obj.data_width = t_pix_to_bytes(obj.width, pThinc);
	obj.data_len = obj.data_width * obj.height;

	thincDbg(T_DBG_PROTO, ("%s: pixmap %dx%d has size %d bytes\n", func,
		obj.width, obj.height, obj.data_len));

	obj.rects = REGION_RECTS(&pixmap->reg);
	obj.numrects = REGION_NUM_RECTS(&pixmap->reg);

	obj.flags = 0;
	obj.extra_len = 0;
	obj.extra = obj.extra_data;

	obj.hash = 0;

	obj.total_len = 0;

	ret = thinc_msgStackExec(cl, cl->pixmapStack, &obj);
#endif

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));

	return ret;
}

#if 0
static int thincSendFBPixmapFill(thincScreenPtr pThinc, thincClientPtr cl,
	RegionPtr reg, char *pixmap, int x, int y, int w, int h)
{
	t_dbg_name(func)
	thinc_fbPFill *p;
	int numrects, i, res;
	register BoxPtr rects;
	thinc_rectRegion r;
	register thinc_rectRegion *rp;
	int size, total_size;
	int dobuffer;
	/* initialize b to NULL so compiler doesn't complain */
	char *b=NULL;

	static char *rszpix = NULL;
	static unsigned int rszpix_sz = 0;
	unsigned int rx, ry, rw, rh;
	char *data;
	unsigned int flags;

#ifndef DISABLE_NORAW
	thincDbg(T_DBG_PROTO, ("%s: x %d y %d w %d, h %d. %ld rectangles\n", 
		func, x, y, w, h, REGION_NUM_RECTS(reg)));
	print_reg(T_DBG_PROTO, pThinc->pScreen, reg, "update: ");

	/* first clip the update to the client's viewport */
	REGION_INTERSECT(pThinc->pScreen, reg, reg, &cl->clClip);
	if(!REGION_NOTEMPTY(pThinc->pScreen, reg)) {
		thincDbg(T_DBG_PROTO, ("%s: finished. Tried to send fill for "
			"offscreen region\n", func));
		return T_OK;
	}

	/* Extract number of rectangles */
	numrects = REGION_NUM_RECTS(reg);

	thincDbg(T_DBG_PROTO, ("%s: Sending Pixmap fill command w: %d, h: %d, "
		"n: %d\n", func, w, h, numrects));

	/* if client is resized, resize the pixmap for it */
	if(cl->parent->allowResize && (cl->resized == T_CL_RESZ_LESS) ) {
		rw = CEILING(w * cl->factor_x);
		rh = CEILING(h * cl->factor_y);
		rx = CEILING(x * cl->factor_x);
		ry = CEILING(y * cl->factor_x);

		thincDbg(T_DBG_PROTO, ("%s: resizing pixmap to %ux%u\n", func, 
			rw, rh));
		size = rh*t_pix_to_bytes(rw,pThinc);
		if(size > rszpix_sz) {
			rszpix = realloc(rszpix, size);
			if(!rszpix) {
				/* don't treat it as an error, just punt */
				thincDbg(T_DBG_PROTO, ("%s: Couldn't allocate "
					"memory for resized pixmap, punting\n",
					func));
				goto normal;
			}
			rszpix_sz = size;
		}

#if 0
		res = thinc_resize32(pixmap, rszpix, w, h, rw, rh, w, 
			t_pix_to_bytes(rw,pThinc));
#else
		res = cl->resize(pixmap, rszpix, w, h, rw, rh, 
			t_pix_to_bytes(w,pThinc), 
			t_pix_to_bytes(rw,pThinc));
#endif
		if(res == T_ERROR) {
			/* don't cry, just punt */
			thincDbg(T_DBG_PROTO, ("%s: Oh well, couldn't resize "
				"pixmap, punting\n", func));
			goto normal;
		}

		thincDbg(T_DBG_PROTO, ("%s: got resized pixmap of size %u\n", 
			func, size));
		flags = T_FB_PFILL_RESIZED;
		x = rx;
		y = ry;
		w = rw;
		h = rh;
		data = rszpix;
	} else {
normal:
		/* calculate total size of message and see if we can buffer 
		 * it */
		size = h*t_pix_to_bytes(w,pThinc);
		flags = 0;
		data = pixmap;
	}

	total_size = size;		/* the pixmap */
	total_size += (numrects*T_RECTREGION_SIZE); /* the rectangles */

	if(total_size > T_MAXDATA(T_FB_PFILL)) {
		/* nope, flush buffers and send data the normal way */
		thincDbg(T_DBG_PROTO, ("%s: PFILL message too big (%d) to "
			"buffer\n", func, total_size));
		thincFlushBuffer(cl);
		dobuffer = T_FALSE;
		p = (thinc_fbPFill*)t_msgBuffer;
	} else {
		/* yes! buffer it */
		thincDbg(T_DBG_PROTO, ("%s: buffering a PFILL message of "
			"size %d\n", func, total_size));
		p = (thinc_fbPFill*)thincBufferMsg(cl, T_FB_PFILL, flags,
			T_FB_PFILL_SIZE, total_size);
		if(!p) {
			thincDbg(T_DBG_PROTO, ("%s: couldn't buffer, reverting "
				"to normal writes\n", func));
			thincFlushBuffer(cl);
			p = (thinc_fbPFill*)t_msgBuffer;
			dobuffer = T_FALSE;
		} else {
			dobuffer = T_TRUE;
		}
	}

	p->x = htons(x);
	p->y = htons(y);
	p->width = htons(w);
	p->height = htons(h);
	p->numrects = htons(numrects);

	if(dobuffer) {
		b = ((char*)p)+T_FB_PFILL_SIZE;
	} else {
		res = thincSendMsg(cl, T_FB_PFILL, T_FB_PFILL_SIZE, flags);
		if(res == T_ERROR) {
			thincError("Error sending pixmap fill message\n");
			return T_ERROR;
		}
	} 
	thincDbg(T_DBG_PROTO, ("%s: Sent message\n", func));

	/* send the pixmap */
	if(dobuffer) {
		memcpy(b, data, size);
		b += size;
	} else {
		res = thincWriteData(cl, data, size);
		if(res == T_ERROR) {
			thincError("Error sending pixmap for pixmap fill "
				"commandn\n");
			return T_ERROR;
		}
	}
	thincDbg(T_DBG_PROTO, ("%s: Sent pixmap\n", func));
	
	/* send the rectangles */
	rects = REGION_RECTS(reg);

	if(dobuffer) {
		rp = (thinc_rectRegion*)b;
		for(i=0; i<numrects; i++, rects++, rp++) {
			print_box(T_DBG_PROTO, rects, "sending rectangle:");
			box_to_rectp_hton(rects, rp);
		}
	} else {
		for(i=0; i<numrects; i++, rects++) {
			box_to_rect_hton(rects, r);
			thincDbg(T_DBG_PROTO, ("%s: Sending rectangle x:%d, "
				"y:%d, w:%d, h:%d\n", func, r.x, r.y, 
				r.width, r.height));

			res = thincWriteData(cl, &r, T_RECTREGION_SIZE);
			if(res == T_ERROR) {
				thincError("Error sending rectangle for pixmap "
					"fill command (rectangle %d)\n", i);
				return res;
			}
		}
	}

#ifdef WRITEPIXTOFILE
	thinc_writePixmapToFile(cl, pixmap, w, h);
#endif

	stats_inc(&pThinc->pfillStats, count);
	stats_add(pThinc, pfillStats.size, 
		total_size + T_FB_PFILL_SIZE + THINC_HEADERSIZE);

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
#endif
	return T_OK;
}
#endif

#ifdef WRITEPIXTOFILE
/* Writes the update to a local file. useful for debugging. */
#define do_write(fd, ptr, len) \
	err = write(fd, ptr, len); \
	if(err == -1) { \
		thincDbg(T_DBG_ERROR, ("%s: Couldn't write PIXMAP update to " \
			"file: %s\n", modname, strerror(errno))); \
		return T_ERROR; \
	}

/* TODO: Add config options for this?? */
int thinc_writePixmapToFile(thincClientPtr cl, char *pixmap,
	unsigned int w, unsigned int h)
{
	const char *modname = "writepixmaptofile";
	static int count = 0;
	unsigned int a, i;
	char fn[255];
	int err, dump; /* fd */
	const char *base = "/tmp/pix/pix";
	register char *curr;
	const int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

	/* Write a header?? */
	const unsigned char write_header = T_TRUE;

	int data_width, pitch;

	data_width = pitch = t_pix_to_bytes(w,cl->parent);

	strncpy(fn, base, 12);
	sprintf(fn+12, "%d", count++);

	thincError("writing file %s\n", fn);

	dump = creat(fn, mode);
	if(dump == -1) {
		thincDbg(T_DBG_ERROR, ("%s: Couldn't open raw file %s. "
			"count %d: %s\n", modname,fn,count,strerror(errno)));
		return T_ERROR;
	}

	if(write_header) {
		a = w;
		a = htonl(a);
		do_write(dump, &a, 4);
		a = h;
		a = htonl(a);
		do_write(dump, &a, 4);
		a = cl->parent->Bpp;
		a = htonl(a);
		do_write(dump, &a, 4);
		a= 0;	/* pad */
		do_write(dump, &a, 4);
	}

	curr = pixmap;
	for(i=0; i<h; ++i) {
		do_write(dump, curr, data_width);
		curr += pitch;
	}

	close(dump);
	return T_OK;
}
#endif	/* WRITEPIXTOFILE */

static int thinc_pixmapSend(thincClientPtr cl, void *m)
{
	t_dbg_modname(modname, pixmapSend.name)
	thincPixObj *obj = m;
	unsigned int len, dobuffer;
	thinc_fbPFill *p;
	int err, i;
	char *msg;
	register BoxPtr rects;
	register thinc_rectRegion *rp;
	thinc_rectRegion r;

	thincDbg(T_DBG_PROTO, ("%s: x %d y %d w %d, h %d. %u rectangles\n", 
		modname, obj->x, obj->y, obj->width, obj->height, 
		obj->numrects));

	len = obj->extra_len + obj->data_len + 
		(obj->numrects*T_RECTREGION_SIZE);
	obj->total_len = THINC_HEADERSIZE + T_FB_PFILL_SIZE + len;

	thincDbg(T_DBG_PROTO, ("%s: update's total length is %d bytes\n", 
		modname, obj->total_len));

	/* see if we can buffer it */
	if(len > T_MAXDATA(T_FB_PFILL)) {
		/* nope, flush buffers and send data the normal way */
		thincDbg(T_DBG_PROTO, ("%s: PFILL message too big (%d) to "
			"buffer\n", modname, len));
		thincFlushBuffer(cl);
		dobuffer = T_FALSE;
		p = (thinc_fbPFill*)t_msgBuffer;
	} else {
		/* yes! buffer it */
		thincDbg(T_DBG_PROTO, ("%s: buffering a PFILL message of "
			"size %d\n", modname, len));
		p = (thinc_fbPFill*)thincBufferMsg(cl, T_FB_PFILL, obj->flags,
			T_FB_PFILL_SIZE, len);
		if(!p) {
			thincDbg(T_DBG_PROTO, ("%s: couldn't buffer, reverting "
				"to normal writes\n", modname));
			thincFlushBuffer(cl);
			p = (thinc_fbPFill*)t_msgBuffer;
			dobuffer = T_FALSE;
		} else {
			dobuffer = T_TRUE;
		}
	}

	p->x = htons(obj->x);
	p->y = htons(obj->y);
	p->width = htons(obj->width);
	p->height = htons(obj->height);
	p->numrects = htons(obj->numrects);

	rects = obj->rects;

	if(dobuffer) {
		thincDbg(T_DBG_PROTO, ("%s: buffering header\n", modname));
		msg = ((char*)p) + T_FB_PFILL_SIZE;
		memcpy(msg, obj->extra_data, obj->extra_len);
		msg += obj->extra_len;
		/* chk if cached */
		if(!obj->data_len)
			goto rects1;
		thincDbg(T_DBG_PROTO, ("%s: buffering pixmap data (%d "
			"bytes)\n", modname, obj->data_len));
		memcpy(msg, obj->data, obj->data_len);
		msg+=obj->data_len;
rects1:
		thincDbg(T_DBG_PROTO, ("%s: buffering %d rectangles\n", 
			modname, obj->numrects));
		rp = (thinc_rectRegion*)msg;
		for(i=0; i<obj->numrects; ++i, ++rects, ++rp) {
			box_to_rectp_hton(rects, rp);
		}
	} else {
		thincDbg(T_DBG_PROTO, ("%s: sending header\n", modname));
		err = thincSendMsg(cl, T_FB_PFILL,T_FB_PFILL_SIZE,obj->flags);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"PFILL update\n", modname));
			return T_ERROR;
		}
		err = thincWriteData(cl, obj->extra_data, obj->extra_len);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"PFILL update\n", modname));
			return T_ERROR;
		}
		/* chk if cached */
		if(!obj->data_len)
			goto rects2;

		thincDbg(T_DBG_PROTO, ("%s: sending pixmap data (%d bytes)\n",
			modname, obj->data_len));
		err = thincWriteData(cl, obj->data, obj->data_len);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"PFILL update\n", modname));
			return T_ERROR;
		}
rects2:
		rp = &r;
		thincDbg(T_DBG_PROTO, ("%s: sending %d rectangles\n", modname,
			obj->numrects));
		for(i=0; i<obj->numrects; ++i, ++rects) {
			box_to_rect_hton(rects, r);
			err = thincWriteData(cl, rp, T_RECTREGION_SIZE);
			if(err == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"send PFILL update\n", modname));
				return T_ERROR;
			}
		}
	}

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_pixmapResize(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, pixmapResize.name)
	thincPixObj *obj = msg;
	static char *rdata = NULL; /* the resized data */
	static unsigned int rdata_sz = 0;
	unsigned int rw, rh, rx, ry, rsz, rw_bytes;
	int err;

	thincDbg(T_DBG_PROTO, ("%s: resizing pixmap fill update\n",
		modname));

	rx = obj->x * cl->factor_x;
	ry = obj->y * cl->factor_y;
	rw = obj->width * cl->factor_x;
	rh = obj->height * cl->factor_y;

	/* XXX: Nasty fix, but it seems to come up with quite nice results :) */
	if(rw == 0) rw = 1;
	if(rh == 0) rh = 1;

	thincDbg(T_DBG_PROTO, ("%s: pixmap resized from %dx%d to %dx%d\n",
		modname, obj->width, obj->height, rw, rh));

	rw_bytes = t_pix_to_bytes(rw, cl->parent);
	rsz = rw_bytes * rh;
	if(rdata_sz < rsz) {
		rdata = realloc(rdata, rsz);
		rdata_sz = rsz;
		if(!rdata) {
			thincDbg(T_DBG_PROTO, ("%s: Couldn't allocate "
				"memory for resized pixmap data. punting. "
				"finished\n", 
				modname));
			return T_OK;
		}
	}

	thincDbg(T_DBG_PROTO, ("%s: resizing pixmap data\n", modname));

	err = cl->resize(obj->data, rdata, obj->width, obj->height, rw, rh,
		obj->data_width, rw_bytes);
	if(err == T_ERROR) {
		/* don't cry, just punt */
		thincDbg(T_DBG_PROTO, ("%s: Couldn't resize pixmap data "
			"punting. finished\n", modname));
		return T_OK;
	}

	obj->flags |= T_FB_PFILL_RESIZED;
	obj->x = rx;
	obj->y = ry;
	obj->width = rw;
	obj->height = rh;

	obj->data = rdata;
	obj->data_width = rw_bytes;
	obj->data_len = rsz;

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_pixmapCompress(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, pixmapCompress.name)
#if 0
	thincPixObj *obj = msg;
#endif

	thincDbg(T_DBG_PROTO, ("%s: %d\n", modname, obj->x));

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_pixmapChkCache(thincClientPtr cl, void *msg)
{
	const char *modname = pixmapChkCache.name;
	thincPixObj *obj = msg;
	int err;
	unsigned long hash;

	thincDbg(T_DBG_PROTO, ("%s: checking if pixmap is in cache\n", 
		modname));

	err = thinc_cache_query_data(cl->pix_cache, obj->data, 
		obj->data_len, &hash);
	obj->hash = hash;
	if(err == T_NO) {
		thinc_cacheAdd *a;
		
		thincDbg(T_DBG_PROTO, ("%s: pixmap is not in cache :*( "
			"telling client to add it, hash is %u\n",
			modname, obj->hash));
		obj->extra_len += T_FB_CACHE_ADD_SIZE;
		checkExtraBuffer(obj);
		a = (thinc_cacheAdd*) obj->extra;
		a->cache_id = htonl(obj->hash);
		obj->extra += T_FB_CACHE_ADD_SIZE;
		obj->flags |= T_FB_PIXMAP_ADDCACHE;
	} else { 
		thinc_cacheData *c;
		thincDbg(T_DBG_PROTO, ("%s: sweeet, found cached data. hash "
			"is %u\n", modname, obj->hash));
		obj->extra_len += T_FB_CACHE_DATA_SIZE;
		checkExtraBuffer(obj);
		c = (thinc_cacheData*)obj->extra;
		c->cache_id = htonl(obj->hash);
		obj->extra += T_FB_CACHE_DATA_SIZE;
		obj->flags |= T_FB_PIXMAP_CACHED;

		/* XXX: don't fuck with the data_width so record has any chance
		 * of recovering from this */
		obj->data = NULL;
		obj->data_len = 0;
	}
	
	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_pixmapAddCache(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, pixmapAddCache.name)
	thincPixObj *obj = msg;

	thincDbg(T_DBG_PROTO, ("%s: possibly adding PFILL udpate to cache\n", 
		modname));

	if(obj->flags & T_FB_PIXMAP_CACHED) {
		thincDbg(T_DBG_PROTO, ("%s: udpate already cached. finished\n",
			modname));
		return T_OK;
	} else {
		int err;
		err = thinc_cache_add_hash(cl->pix_cache, obj->hash, NULL);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add "
				"PFILL update to cache\n", modname));
			return T_ERROR;
		} else {
			thincDbg(T_DBG_PROTO, ("%s: Added update to cache with "
				"hash %u, finished\n", modname, obj->hash));
			return T_OK;
		}
	}
}

#ifdef T_GETSTATS
static int thinc_pixPreStats(thincClientPtr cl, void *msg)
{
	stats_start_nostr(t_start);
	stats_init_simple(writeTime);
	stats_init_simple(blockTime);

	return T_OK;
}

static int thinc_pixPostStats(thincClientPtr cl, void *msg)
{
	unsigned long t_end;
	thincPixObj *obj = msg;

	/* TODO: Maybe stats should be per client?? */
	stats_end(cl->parent, t_start, t_end, pfillStats.flush);
	stats_inc(cl->parent, pfillStats.count);
	stats_add(cl->parent, pfillStats.write, writeTime);
	stats_add(cl->parent, pfillStats.block, blockTime);
	stats_add(cl->parent, pfillStats.size, obj->total_len);
	return T_OK;
}
#endif /* T_GETSTATS */


#ifdef WRITEPIXTOFILE
static int thinc_pixmapWriteFile(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, pixmapWriteFile.name)
	thincPixObj *obj = msg;

	thincDbg(T_DBG_PROTO, ("%s: %d\n", modname, obj->x));

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}
#endif /* WRITEPIXTOFILE */

#ifdef T_RECORD
int thinc_pixmapInitRecordStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_RECORD, ("%s: creating PIXMAP message stack for "
		"record command\n", func));

	cl->pixmapStack = thinc_msgStackCreate(cl);
	if(!cl->pixmapStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create PIXMAP "
			"message stack\n", func));
		return T_ERROR;
	}
	
	pixmap_addmodule(cl, &pixmapRecord);

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

void thinc_pixmapDestroyRecordStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->pixmapStack);
	cl->pixmapStack = NULL;
}

/* wrapper obj in case we need to call callback more than once */
struct thincPixObjRec {
	thincPixObj*	obj;
	char*		data;	/* next data to write */
	unsigned int	data_len;	/* how much we have yet to write.
					if 0, then only rects are missing
					(which we assume to fit all in
					buffer) */
};

/* somewhat tricky as we need to deal with caching. but leaving caching
 * allowed as we may be able to use it in the future */
static int thinc_pixmapRecord(thincClientPtr cl, void *m)
{
	t_dbg_modname(modname, pixmapRecord.name)
	thincPixObj *obj = m;
	unsigned int o_total = 0;
	int err;
	char cached = T_FALSE;
	struct thincPixObjRec o_rec;

	thincDbg(T_DBG_RECORD, ("%s: \n", modname));

	/* if we're the recorder client, then total_len hasn't been computed.
	 * merde */
	o_total = obj->total_len;
	if(!obj->total_len) {
		obj->total_len = THINC_HEADERSIZE + T_FB_PFILL_SIZE + 
			(obj->numrects*T_RECTREGION_SIZE) + obj->data_len +
			obj->extra_len;
	} else if(!obj->data_len) {
		/* it's cached. fix things up */ 
		thincDbg(T_DBG_RECORD, ("%s: it's cached. need to get it "
			"and fix some things up\n", modname));
		cached = T_TRUE;
		obj->data = obj->orig_data;
		obj->data_len = obj->data_width * obj->height;
		thincDbg(T_DBG_RECORD, ("%s: recovered cached update. data "
			"length is %u\n", modname, obj->data_len));
		obj->total_len += obj->data_len;
	}
	obj->total_len -= THINC_HEADERSIZE;
	
	thincDbg(T_DBG_RECORD, ("%s: update's total length is %d bytes\n", 
		modname, obj->total_len));

	o_rec.obj = obj;
	o_rec.data = NULL;

	err = thincRecord(cl, obj->total_len, T_FB_PFILL, obj->flags,
		thinc_pixmapRecordCb, &o_rec);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't record PIXMAP "
			"command\n", modname));
		return T_ERROR;
	}

	if(cached) {
		/* revert our changes */
		obj->data = NULL;
		obj->data_len = 0;
	}
	obj->total_len = o_total;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

/* such a mess... */
static int thinc_pixmapRecordCbData(thincClientPtr cl, char *buf, 
	unsigned int *len, struct thincPixObjRec *rec);
static int thinc_pixmapRecordCbJustRects(thincClientPtr cl, char *buf, 
	unsigned int *len, thincPixObj *obj);

static int thinc_pixmapRecordCb(thincClientPtr cl, char *buf, unsigned int *len,
	void *ptr)
{
	t_dbg_modname(modname, pixmapRecord.name)
	struct thincPixObjRec *rec = ptr;
	thincPixObj *obj = rec->obj;
	unsigned int d_len, h_len, a_len, i;
	thinc_fbPFill *p;
	char *msg;
	register BoxPtr rects;
	register thinc_rectRegion *rp;

	thincDbg(T_DBG_RECORD, ("%s\n", modname));

	/* see if first time around, and send msg. we save all the flags, too
	 * much pain to set only those we would care about on
	 * playback. note that we assume that at the most basic, len
	 * has enough space for the message and the extra info */
	if(rec->data) {
		return thinc_pixmapRecordCbData(cl, buf, len, rec);
	}

	thincDbg(T_DBG_RECORD, ("%s: first time, recording header\n",
		modname));

	/* TODO: Add T_BUG_ON check between h_len and len */
	h_len = T_FB_PFILL_SIZE + obj->extra_len;

	p = (thinc_fbPFill*) buf;
	p->x = htons(obj->x);
	p->y = htons(obj->y);
	p->width = htons(obj->width);
	p->height = htons(obj->height);
	p->numrects = htons(obj->numrects);

	msg = ((char*)p) + T_FB_PFILL_SIZE;
	memcpy(msg, obj->extra_data, obj->extra_len);
	msg += obj->extra_len;

	/* we still have to write all this ... */
	d_len = obj->data_len + (obj->numrects*T_RECTREGION_SIZE);
	/* ... and this is how much space we have available */
	a_len = *len - h_len;

	thincDbg(T_DBG_RECORD, ("%s: trying to record %u bytes, have space "
		"for %u\n", modname, d_len, a_len));

	if(a_len >= d_len) {
		/* sweetness */
		thincDbg(T_DBG_RECORD, ("%s: we fit just fine\n", modname));
		memcpy(msg, obj->data, obj->data_len);
		msg += obj->data_len;

		thincDbg(T_DBG_RECORD, ("%s: recording %d rectangles\n", 
			modname, obj->numrects));
		rp = (thinc_rectRegion*)msg;
		rects = obj->rects;

		for(i=0; i<obj->numrects; ++i, ++rects, ++rp) {
			box_to_rectp_hton(rects, rp);
		}

		*len = obj->total_len; /* wrote it all */
		thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u bytes\n", 
			modname, *len));

		return T_REC_DONE;
	} else {
		/* damn. ok, see if data fits... */
		if(a_len >= obj->data_len) {
			thincDbg(T_DBG_RECORD, ("%s: data fits. problem is "
				"damn rectangles\n", modname));
			memcpy(msg, obj->data, obj->data_len);
			msg += obj->data_len;

			/* next time, just send rects */
			rec->data = obj->data;
			rec->data_len = 0;

			*len = h_len + obj->data_len;
			thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u "
				"bytes\n", modname, *len));
			return T_REC_MORE;
		} 
		/* data doesn't fit. copy what we can, then punt the rest
		 * for next time */
		thincDbg(T_DBG_RECORD, ("%s: data doesn't fit. only have "
			"space for %u bytes, need %u\n", modname, a_len,
			obj->data_len));
		memcpy(msg, obj->data, a_len);

		rec->data = obj->data + a_len;
		rec->data_len = obj->data_len - a_len;
		*len = a_len;
		thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u "
			"bytes\n", modname, *len));
		return T_REC_MORE;
	}
}

static int thinc_pixmapRecordCbData(thincClientPtr cl, char *buf, 
	unsigned int *len, struct thincPixObjRec *rec)
{
	t_dbg_modname(modname, pixmapRecord.name)
	thincPixObj *obj = rec->obj;
	unsigned int a_len, r_len;

	thincDbg(T_DBG_RECORD, ("%s: data_len %u\n", modname, rec->data_len));

	if(!rec->data_len) {
		thincDbg(T_DBG_RECORD, ("%s: no data left, just rects\n",
			modname));
		/* we're done with the data, just missing the
		 * rects. we blindly assume they all fit in :) */
		return thinc_pixmapRecordCbJustRects(cl, buf,
			len, obj);
	}
	/* still got some data to write */
	a_len = *len;
	if(a_len < rec->data_len) {
		thincDbg(T_DBG_RECORD, ("%s: still doesn't fit. have space "
			"for %u, need %u\n", modname, a_len, rec->data_len));
		/* and it still doesn't fit. damn */
		memcpy(buf, rec->data, a_len);

		rec->data += a_len;
		rec->data_len -= a_len;
		*len = a_len;

		thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u "
			"bytes\n", modname, *len));
		return T_REC_MORE;
	}
	/* it fits */
	memcpy(buf, rec->data, rec->data_len);
	a_len -= rec->data_len;
	r_len = (obj->numrects*T_RECTREGION_SIZE);

	thincDbg(T_DBG_RECORD, ("%s: done with data. have space for %u, "
		"need %u for rects\n", modname, a_len, r_len));

	/* but do the rects fit? */
	if(a_len < r_len) {
		/* no :( */
		*len = rec->data_len;
		rec->data = NULL;
		rec->data_len = 0;
		thincDbg(T_DBG_RECORD, ("%s: finished, but rects still "
			"missing (%u bytes)\n", modname, r_len));
		return T_REC_MORE;
	} else {
		/* in yer face! */
		buf += rec->data_len;
		thinc_pixmapRecordCbJustRects(cl, buf, len, obj);
		*len = rec->data_len + r_len;

		thincDbg(T_DBG_RECORD, ("%s: finished, wrote %u bytes\n",
			modname, *len));
		return T_REC_DONE;
	}
}

static int thinc_pixmapRecordCbJustRects(thincClientPtr cl, char *buf, 
	unsigned int *len, thincPixObj *obj)
{
	t_dbg_modname(modname, pixmapRecord.name)
	register BoxPtr rects;
	register thinc_rectRegion *rp;
	unsigned int i;

	thincDbg(T_DBG_RECORD, ("%s: just missing the rects\n", modname));

	rects = obj->rects;
	rp = (thinc_rectRegion*)buf;
	for(i=0; i<obj->numrects; ++i, ++rects, ++rp) {
		box_to_rectp_hton(rects, rp);
	}

	*len = (obj->numrects*T_RECTREGION_SIZE);
	thincDbg(T_DBG_RECORD, ("%s: finished, recorded %u bytes\n", modname, 
		*len));
	return T_REC_DONE;
}
#endif /* T_RECORD */
