/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Wrapper functions for drawing (wrap GC ops)
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/* TODO
 * See which functions that look like we cant do anything can be improved
 * by using mi() so that it calls our Spans functions.
 * Candidates -> *Arc()
 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include "mi.h"
#include "migc.h"
#include "fb.h"

/*** local functions ***/

/* send a Copy between two windows on the screen */
static void thincCopyArea_onScreen(thincScreenPtr pThinc, GCPtr pGC,
	WindowPtr srcW, WindowPtr dstW, int srcx, int srcy, int w, int h, 
	int dstx, int dsty);

/* helper functions */
static int thincFillRegion_Solid(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, unsigned long color);
static int thincFillRegion_Pixmap(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, GCPtr pGC);
static int thincFillRegion_Glyph(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, GCPtr pGC);
static int thincFillRegion_Bilevel(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, GCPtr pGC);

/******************* GC FUNCS ****************/
/* Make the ones not necessary equal NULL here and equal to driver ops 
 * upon GC creation */
GCOps thincGCOps = {
	thincFillSpans,
	thincSetSpans,
	thincPutImage,
	thincCopyArea,
	thincCopyPlane,
	thincPolyPoint,
	thincPolylines,
	thincPolySegment,
	thincPolyRectangle,
	thincPolyArc,
	thincFillPolygon,
	thincPolyFillRect,
	thincPolyFillArc,
	thincPolyText8,
	thincPolyText16,
	thincImageText8,
	thincImageText16,
	thincImageGlyphBlt,
	thincPolyGlyphBlt,
	thincPushPixels,
#ifdef NEED_LINEHELPER	/* deprecated op */
	NULL
#endif
};

GCFuncs thincGCFuncs = {
	thincValidateGC,
	miChangeGC,
	miCopyGC,
	miDestroyGC,
	miChangeClip,
	miDestroyClip,
	miCopyClip,
};

/* The GC includes two vectors of procedure pointers, the first 
 * operate on the GC itself and are called GC funcs.
 * The second, called GC ops, contains the functions that carry out 
 * the fundamental graphic operations such as drawing lines, polygons, 
 * arcs, text, and copying bitmaps.
 * NOTE: For any particular GC, the func vector must remain constant, 
 * the ops can vary
 */
Bool thincCreateGC(GCPtr pGC)
{
	t_dbg_name(func)
	ScreenPtr pScreen = pGC->pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);
	thincGCPtr pThincGC = THINCGCPtr(pGC);
	Bool res;
	instr_add_extra("CREATEGC", TI_START);

	thincDbg(T_DBG_DRAW, ("%s\n", func));

	thinc_unwrap(pThinc, pScreen, CreateGC);
	instr_add_extra("CREATEGC", TI_DRAW_START);
	res = (*pScreen->CreateGC)(pGC);
	instr_add_extra("CREATEGC", TI_DRAW_END);
	thinc_wrap(pThinc, pScreen, CreateGC, thincCreateGC);

	pThincGC->driver_ops = pGC->ops;
	pGC->ops = &thincGCOps;

	pThincGC->driver_funcs = pGC->funcs;
	pGC->funcs = &thincGCFuncs;
	
	/* misc init */
	pThincGC->flags = gc_none;

	/* legacy op */
#ifdef NEED_LINEHELPER
	pGC->ops->LineHelper = pThincGC->driver_ops->LineHelper;
#endif

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	instr_add_extra("CREATEGC", TI_END);
	return res;
}

/* 
 * Validate the GC before drawing
 * Things to do:
 * -> If we're drawing to somewhere we don't care about (currently 
 *  !(DRAWABLE_WINDOW || DRAWABLE_CMD) ) then just put the normal GC
 *  ops in the GC and off we go
 */

#define call_validate(pGC) \
	pGC->funcs = pThincGC->driver_funcs; \
	(*pGC->funcs->ValidateGC)(pGC, changes, pDraw); \
	pGC->funcs = &thincGCFuncs;

void thincValidateGC(GCPtr pGC, unsigned long changes, DrawablePtr pDraw)
{
	t_dbg_name(func)
	int drawToDst, isOff, gcmarked = T_FALSE;
	thincGCPtr pThincGC = THINCGCPtr(pGC);
	thincScreenPtr pThinc = THINCPtr(pGC->pScreen);
	int depthmask;
	instr_add_extra("VALIDATEGC", TI_START);

	thincDbg(T_DBG_DRAW, ("%s\n", func));
	thincDrawable_print(pDraw, "");

	drawToDst = is_onscreen(pDraw);
	isOff = is_offscreen(pDraw);

#ifdef T_APPSHARE
	if(drawToDst) {
		WindowPtr pWin = (WindowPtr)pDraw;
		if(thincWindowIsShared(pThinc, pWin)) {
			thincMsg("Window %#lx:%#lx is shared\n", 
				pDraw->id, t_winID(pWin));
		} else {
			thincMsg("Window %#lx:%#lx is not shared\n",
				pDraw->id, t_winID(pWin));
		}
	}
#endif	/* T_APPSHARE */

	/* chk if we're interested in this drawable */
	if(!drawToDst && !isOff) {
		thincDbg(T_DBG_DRAW, ("%s: Drawing to a boring drawable, "
			"ignoring\n", func));
		/* we don't care about this dud, off you go! */
		goto lame;
	} else if(isOff && !pThinc->useOffscreen) {
		thincDbg(T_DBG_DRAW, ("%s: Drawing to a pixmap, and offscreen "
			"pixmaps are disabled\n", func));
		goto lame;
	}

	/* chk if there are any weird settings and mark the gc as a weirdo */

	/* alu */
	if(pGC->alu != GXcopy) {
		thincDbg(T_DBG_DRAW, ("%s: Unsupported ALU %d, marking GC\n",
			func, pGC->alu));
		pThincGC->flags |= gc_weirdo;
		gcmarked = T_TRUE;
	}
	/* planemask */
	depthmask = FbFullMask(pDraw->depth);
	if((pGC->planemask & depthmask) != depthmask) {
		/* weeeiiirdooo */
		thincDbg(T_DBG_DRAW, ("%s: Unsupported planemask, marking GC\n",
			func));
		pThincGC->flags |= gc_weirdo;
		gcmarked = T_TRUE;
	}

	if(!gcmarked)
		gc_clearflag(pThincGC, gc_weirdo);

	/* call driver's validate func */
	instr_add_extra("VALIDATEGC", TI_DRAW_START);
	call_validate(pGC);
	instr_add_extra("VALIDATEGC", TI_DRAW_END);

	/* make sure our functions are called */
	pGC->ops = &thincGCOps;
	
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	instr_add_extra("VALIDATEGC", TI_END);
	return;

lame:	/* lame drawing ops that we don't care about end up here */
	pGC->ops = pThincGC->driver_ops;

	call_validate(pGC);

	thincDbg(T_DBG_DRAW, ("%s: lame op. finished\n", func));
}

/* Destroy returned region with REGION_DESTROY */
int thincGC_clipUpdate(GCPtr pGC, RegionPtr reg, int x, int y, int w, int h)
{
	t_dbg_name(func)
	BoxRec b;
	instr_add_extra("CLIPUPDATE", TI_START);

	b.x1 = x;
	b.y1 = y;
	b.x2 = x+w;
	b.y2 = y+h;

	thincDbg(T_DBG_DRAW, ("%s: clipping update (%d, %d) (%d, %d)\n", 
		func, b.x1, b.y1, b.x2, b.y2));

	REGION_INIT(pGC->pScreen, reg, &b, 1);
	
	if(thincGC_clipUpdateRegion(pGC, reg) == T_ERROR) {
		return T_ERROR;
	} else {
		thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
		instr_add_extra("CLIPUPDATE", TI_END);
		return T_OK;
	}
}

/* returns 0 on success, -1 on error */
int thincGC_clipUpdateRegion(GCPtr pGC, RegionPtr update)
{
	t_dbg_name(func)
	RegionPtr clip = pGC->pCompositeClip;
	instr_add_extra("CLIPUPDATEREG", TI_START);

	if(!clip) {
		thincDbg(T_DBG_ERROR, 
			("ERROR: %s: Couldn't find a CompositeClip\n", func));
		return T_ERROR;
	
	}

	thincDbg(T_DBG_DRAW, ("%s. GC clip has %ld rectangles\n", func, 
		REGION_NUM_RECTS(clip)));
	print_reg(T_DBG_DRAW, pGC->pScreen, clip, "GC clip:");

	/* clip the update */
	if(!REGION_INTERSECT(pGC->pScreen, update, clip, update)) {
		/* Intersect failed for some reason */
		thincDbg(T_DBG_ERROR, ("ERROR: %s: REGION_INTERSECT() "
			"failed\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	instr_add_extra("CLIPUPDATEREG", TI_END);
	return T_OK;
}

void thincPutImage( DrawablePtr dst, GCPtr pGC, int depth, int x, int y, 
	int w, int h, int leftpad, int format, char *pBits)
{
	t_dbg_name(func)
	RegionRec updateRec;
	int isOff;
	int res;
	thincGC_pre(pGC);

	instr_add("PUTIMAGE", TI_START);

	thincDrawable_print(dst, "");
	thincDbg(T_DBG_DRAW, ("%s (%d,%d) -> (%d, %d). depth %d, leftpad %d "
		"format %d, fg %ld bg %ld\n", func, x, y, w, h, depth, 
		leftpad, format, pGC->fgPixel, pGC->bgPixel));

	stats_inc(pThinc, countPutImage);

	/* call driver's func */
	instr_add("PUTIMAGE", TI_DRAW_START);
	(*pGC->ops->PutImage)(dst, pGC, depth, x, y, w, h, leftpad, format, 
		pBits);
	instr_add("PUTIMAGE", TI_DRAW_END);

	/* clip the update */
	x+=dst->x;
	y+=dst->y;

	thincDbg(T_DBG_DRAW, ("%s: sending update (%d,%d) -> (%d,%d)\n", 
		func, x,y,w,h));

	call_clipUpdate(pGC, &updateRec, x, y, w, h, noret);

	/* chk what kind of drawable we're dealing with */
	isOff = is_offscreen(dst);

	if(isOff) {
		res = thincQueueCommand(pThinc, (PixmapPtr)dst,
			T_FB_RUPDATE, &updateRec, T_RAW_VA_NONE);
	} else {
		res = thincBufferNewCmdReg(pThinc, (WindowPtr)dst, 
			T_FB_RUPDATE, &updateRec, T_RAW_VA_NONE);
	}

	/* Only thing we can do is print an error, since DIX doesn't
	 * expect us to return anything */
	if(res != T_OK) {
		thincWarn("THINC: Couldn't process PutImage command\n");
	}

	REGION_UNINIT(pScreen, &updateRec);

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("PUTIMAGE", TI_END);
}

#ifdef T_USEMITSHM
/* from fbShmPutImage and miShmPutImage. We do some additional 
 * thinc-specific things, and we call our functions directly */
/* Assumes offscreen pixmaps are enabled. If they're not, there's no
 * point in calling this function, just use mi's version */
void thincShmPutImage(DrawablePtr dst, GCPtr pGC, int depth,
	unsigned int format, int w, int h, int sx, int sy, int sw, int sh, 
	int dx, int dy, char *data)
{
	t_dbg_name(func)

	thincScreenPtr pThinc = THINCPtr(pGC->pScreen);

	PixmapPtr pmap;
	GCPtr putGC;
	PixmapPtr pPixmap;
	RegionRec update;
	instr_add("SHMPUTIMAGE", TI_START);

	thincDbg(T_DBG_DRAW, ("%s: Putting SHM Image at x: %d, y: %d "
		"(sx: %d, sy: %d) w: %d, h: %d (sw %d, sh %d), format %d\n", 
		func, dx, dy, sx, sy, w, h, sw, sh, format));
	thincDrawable_print(dst, "");

	stats_inc(pThinc, countPutImage);
	
	if ((format == ZPixmap) || (depth == 1))
	{
		thincDbg(T_DBG_DRAW, ("%s: using scratch pixmap to transfer "
			"the shm image\n", func));
		instr_add("SHMPUTIMAGE", TI_ALLOC_START);
		pPixmap = GetScratchPixmapHeader(dst->pScreen, w, h, depth,
			BitsPerPixel(depth), PixmapBytePad(w, depth), 
			(pointer)data);
		if (!pPixmap)
			return;
		instr_add("SHMPUTIMAGE", TI_ALLOC_END);
		/* Insert a fake RAW Update so that the copy is done
		 * correctly */
		thincDbg(T_DBG_DRAW, ("%s: injecting fake RAW "
			"update to scratch pixmap\n", func));
		/* the part of the image that will be actually put */
		coords_to_reg(pThinc->pScreen, &update, 0, 0, w, h);
		if(thincQueueCommand(pThinc, pPixmap, T_FB_RUPDATE, 
			&update, T_RAW_VA_NONE) == T_ERROR) {
			thincWarn("THINC: Couldn't process "
				"SHMem PutImage command\n");
		}
		REGION_UNINIT(dst->pScreen, &update);
		
		instr_add("SHMPUTIMAGE", TI_DRAW_START);
		if (format == XYBitmap)
			thincCopyPlane((DrawablePtr)pPixmap, 
				dst, pGC, sx, sy, sw, sh, dx, dy, 1L);
		else
			thincCopyArea((DrawablePtr)pPixmap, dst, 
				pGC, sx, sy, sw, sh, dx, dy);
		instr_add("SHMPUTIMAGE", TI_DRAW_END);

		thincDbg(T_DBG_DRAW, ("%s: done. freeing scratch pixmap\n",
			func));
		thincEmptyPixmapQueue(pPixmap);
		FreeScratchPixmapHeader(pPixmap);
	} else {
		/* For this case PutImage() is called so we don't need to
		 * do anything */
		putGC = GetScratchGC(depth, dst->pScreen);
		if (!putGC)
			return;
		instr_add("SHMPUTIMAGE", TI_ALLOC_START);
		pmap = thincCreatePixmap(dst->pScreen, sw, sh, 
			depth);
		if (!pmap)
		{
			FreeScratchGC(putGC);
			return;
		}
		instr_add("SHMPUTIMAGE", TI_ALLOC_END);
		ValidateGC((DrawablePtr)pmap, putGC);
		instr_add("SHMPUTIMAGE", TI_DRAW_START);
		thincPutImage((DrawablePtr)pmap, putGC, depth, 
			-sx, -sy, w, h, 0, 
			( (format == XYPixmap) ? XYPixmap : ZPixmap), data);
		instr_add("SHMPUTIMAGE", TI_DRAW_END);
		FreeScratchGC(putGC);
		instr_add("SHMPUTIMAGE", TI_DRAW_START);
		if (format == XYBitmap)
			thincCopyPlane((DrawablePtr)pmap, dst, 
				pGC, 0, 0, sw, sh, dx, dy, 1L);
		else
			thincCopyArea((DrawablePtr)pmap, dst, 
				pGC, 0, 0, sw, sh, dx, dy);
		instr_add("SHMPUTIMAGE", TI_DRAW_END);
		thincDestroyPixmap(pmap);
	}

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("SHMPUTIMAGE", TI_END);
}
#endif	/* T_USEMITSHM */

RegionPtr thincCopyArea(DrawablePtr src, DrawablePtr dst, GCPtr pGC,
	int srcx, int srcy, int w, int h, int dstx, int dsty)
{
	t_dbg_name(func)
	RegionPtr res;
	RegionRec dstReg;
	int dstIsOff, srcIsOff, srcIsW, dstIsW;
	int ret;
	WindowPtr srcW, dstW;
	thincGC_pre(pGC);
	instr_add("COPYAREA", TI_START);

	thincDrawable_print(src, "SRC: ");
	thincDrawable_print(dst, "DST: ");

	thincDbg(T_DBG_DRAW, ("%s: src (%d, %d), dst (%d, %d), size (%dx%d). "
		"alu %d\n", func, srcx, srcy, dstx+dst->x, dsty+dst->y, w, 
		h, pGC->alu));

	stats_inc(pThinc, countCopyArea);

	/* call driver's func */
#if 0
	res = (*pGC->ops->CopyArea)(src, dst, pGC, srcx, srcy, 
		w, h, dstx, dsty);
#endif

	/* possibilities:
	 * both are offscreen pixmaps:
	 * -> thincCopyPixmapQueue()
	 * src offscreen pixmap, dst onscreen
	 * -> thincExecPixmapQueue()
	 * src onscreen, dst onscreen
	 * -> thincSendFBCopy()
	 * the rest: 
	 * -> do nothing
	 *
	 * a drawable could be:
	 * 1. A pixmap => is_offscreen() == true
	 * 2. An undrawable window
	 * 3. A drawable unrealized window
	 * 4. A drawable realized window => is_onscreen() == true
	 * 
	 * we only know what to do for 1 and 4. Undrawable windows shouldn't 
	 * be drawn to, and unrealized windows appear to be in the ToDo
	 * 
	 * TODO:
	 * if pixmaps are the same, shouldn't queue a Copy?
	 * TODO:
	 * What about unrealized windows? should we implement offscreen
	 * commands for those too?
	 */
	srcIsOff = is_offscreen(src);
	dstIsOff = is_offscreen(dst);
	srcIsW = is_onscreen(src);
	dstIsW = is_onscreen(dst);

	if(srcIsOff && dstIsW) {
		instr_add("COPYAREA", TI_DRAW_START);
		res = (*pGC->ops->CopyArea)(src, dst, pGC, srcx, srcy, w, h, 
			dstx, dsty);
		instr_add("COPYAREA", TI_DRAW_END);
		if(pThinc->useOffscreen) {
			thincDbg(T_DBG_DRAW, ("%s: source is offscreen pixmap, "
				"dst is window. Playing back commands\n", 
				func));
		} else {
			thincDbg(T_DBG_DRAW, ("%s: source is offscreen pixmap, "
				"destination is a window, sending RAW update\n",
				func));
		}

		thincDbg(T_DBG_DRAW, ("%s: ClipType %d,  subWindowMode %d\n",
			func, pGC->clientClipType, pGC->subWindowMode));

		dstx += dst->x;
		dsty += dst->y;

		dstW = (WindowPtr)dst;

		/* this is the region to be copied */
		coords_to_reg(pThinc->pScreen, &dstReg, dstx, dsty, w, h);

		thincGC_clipUpdateRegion(pGC, &dstReg);
		
		/* clip against the destination window */
		REGION_INTERSECT(pThinc->pScreen, &dstReg, &(dstW->borderClip),
			&dstReg);
		if(!REGION_NOTEMPTY(pThinc->pScreen, &dstReg)) {
			thincDbg(T_DBG_DRAW, ("%s: finished. Destination "
				"region is completely obscured\n", func));
			thincGC_post(pGC);
			return res;
		}

		print_reg(T_DBG_DRAW, pThinc->pScreen, &dstReg, "final:");

		if(pThinc->useOffscreen) {
			/* playback stored commands. Returns number of
			 * played back commands. If 0, we need to issue
			 * a RAW update */
			instr_add("COPYAREA", TI_EXEC_QUEUE_START);
			ret = thincExecPixmapQueue(pThinc, (WindowPtr)dst,
				(PixmapPtr)src, srcx, srcy, w, h, dstx, dsty, 
				&dstReg);
			instr_add("COPYAREA", TI_EXEC_QUEUE_END);
			if(ret == T_ERROR) {
				/* is this correct behavior?? :) */
				thincDbg(T_DBG_ERROR, 
					("ERROR: %s: Couldn't play back "
					"stored offscreen commands\n", func));
				REGION_DESTROY(pThinc->pScreen, res);
				thincGC_post(pGC);
				return NULL;
#ifdef T_ENABLE_DEBUG_OPTIONS
			} else if((ret == 0) && pThinc->emptyQueuesOK) {
				thincWarn("THINC: No commands stored "
					"in source pixmap, sending RAW "
					"update\n");
				thincBufferNewCmdReg(pThinc, (WindowPtr)dst,
					T_FB_RUPDATE, &dstReg, T_RAW_VA_NONE);
#else
			} else if(ret == 0) {
				thincError("THINC: No commands stored in "
					"source pixmap, FIX ME!\n");
#endif
			} else {
				thincDbg(T_DBG_DRAW, ("%s: played back %d "
					"commands\n", func, ret));
			}
		} else {
			/* send RAW update */
			instr_add("COPYAREA", TI_BUFFER_START);
			thincBufferNewCmdReg(pThinc, (WindowPtr)dst, 
				T_FB_RUPDATE, &dstReg, T_RAW_VA_NONE);
			instr_add("COPYAREA", TI_BUFFER_END);
		}
	} else if(srcIsW && dstIsW) {
		/* issue a copy. both drawables are WindowPtr
	 	 * must clip to dst's clipping region */
		/* FIXME: Don't we need to clip the region here too?? */
		thincDbg(T_DBG_DRAW, ("%s: both drawables are windows. Sending "
			"normal copy\n", func));

		srcW = (WindowPtr)src;
		dstW = (WindowPtr)dst;

		thincCopyArea_onScreen(pThinc, pGC, srcW, dstW, srcx, srcy, 
			w, h, dstx, dsty);
		instr_add("COPYAREA", TI_DRAW_START);
		res = (*pGC->ops->CopyArea)(src, dst, 
			pGC, srcx, srcy, w, h, dstx, dsty);
		instr_add("COPYAREA", TI_DRAW_END);
	} else if(srcIsOff && dstIsOff) {
		instr_add("COPYAREA", TI_DRAW_START);
		res = (*pGC->ops->CopyArea)(src, dst, 
			pGC, srcx, srcy, w, h, dstx, dsty);
		instr_add("COPYAREA", TI_DRAW_END);

		if(!pThinc->useOffscreen)
			goto end;

		thincDbg(T_DBG_DRAW, ("%s: Both pixmaps are offscreen. Copying "
			"commands between their queues\n", func));
		instr_add("COPYAREA", TI_COPY_QUEUE_START);
		ret = thincCopyPixmapQueue(pThinc, (PixmapPtr)src, 
			(PixmapPtr)dst, srcx, srcy, w, h, dstx, dsty);
		instr_add("COPYAREA", TI_COPY_QUEUE_END);
		if(ret == T_ERROR) {
			/* if we can't nobody can!!! */
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy "
				"command queue to destination pixmap\n", func));
			REGION_DESTROY(pThinc->pScreen, res);
			thincGC_post(pGC);
			return NULL;
#ifdef T_ENABLE_DEBUG_OPTIONS
		} else if((ret == 0) && pThinc->emptyQueuesOK) {
			/* bummer, don't know anything about the src pixmap, 
			 * only thing to do is inject a RAW update in the dst 
			 * pixmap. */
			thincWarn("THINC: No commands stored in "
				"source pixmap, injecting RAW update\n");

			coords_to_reg(pThinc->pScreen, &dstReg, dstx, dsty,
				w, h);
			thincGC_clipUpdateRegion(pGC, &dstReg);

			thincQueueCommand(pThinc, (PixmapPtr)dst, 
				T_FB_RUPDATE, &dstReg, T_RAW_VA_NONE);
#else
		} else if(ret == 0) {
			thincError("THINC: No commands stored in "
				"source pixmap, fix me!!\n");
#endif
		} else {
			thincDbg(T_DBG_DRAW, ("%s: copied %d commands\n",
				func, ret));
		}
	} else {
		/* the weird case. do nothing */
		instr_add("COPYAREA", TI_DRAW_START);
		res = (*pGC->ops->CopyArea)(src, dst, 
			pGC, srcx, srcy, w, h, dstx, dsty);
		instr_add("COPYAREA", TI_DRAW_END);

		thincDbg(T_DBG_DRAW, ("%s: Weird case, src off %d, on %d, "
			"dst off %d, on %d. Not doing anything\n", func,
			srcIsOff, srcIsW, dstIsOff, dstIsW));
	}
end:

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	instr_add("COPYAREA", TI_END);
	return res;
}

/* Issue a Copy between two windows on the screen */
static void thincCopyArea_onScreen(thincScreenPtr pThinc, GCPtr pGC,
	WindowPtr srcW, WindowPtr dstW, int srcx, int srcy, int w, int h, 
	int dstx, int dsty)
{
	t_dbg_name(func)
	int deltax, deltay;
	RegionRec copyReg;

	instr_add("COPYAREA_ONSCR", TI_START);
	thincDbg(T_DBG_DRAW, ("%s\n", func));

	/* compute actual screen coordinates */
	srcx += srcW->drawable.x;
	srcy += srcW->drawable.y;
	dstx += dstW->drawable.x;
	dsty += dstW->drawable.y;

	deltax = dstx-srcx;
	deltay = dsty-srcy;

	thincDbg(T_DBG_DRAW, ("%s: src: %d,%d dst: %d,%d\n",
		func, srcx, srcy, dstx, dsty));

	/* initialize the region to be copied with destination coords */
#if 0
	coords_to_reg(pThinc->pScreen, &copyReg, dstx, dsty, w, h);

	thincGC_clipUpdateRegion(pGC, &copyReg);

	/* clip against the destination window. We do this first
	 * because SendFBCopy() receives the source region, so
	 * we save 1 TRANSLATE */
	REGION_INTERSECT(pThinc->pScreen, &copyReg, &dstW->borderClip,
		&copyReg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, &copyReg)) {
		thincDbg(T_DBG_DRAW, ("%s: finished. Destination "
			"region is completely obscured\n", func));
		return;
	}

	/* translate to the source region */
	REGION_TRANSLATE(pThinc->pScreen, &copyReg, -deltax, -deltay);

	/* clip so we know what's available */
	REGION_INTERSECT(pThinc->pScreen, &copyReg, &srcW->borderClip,
		&copyReg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, &copyReg)) {
		thincDbg(T_DBG_DRAW, ("%s: finished. Source region is "
			"completely obscured\n", func));
		return;
	}

#else
	coords_to_reg(pThinc->pScreen, &copyReg, dstx, dsty, w, h);

	thincGC_clipUpdateRegion(pGC, &copyReg);

	/* translate to the source region */
	REGION_TRANSLATE(pThinc->pScreen, &copyReg, -deltax, -deltay);
#endif
	/* ok, have the region to copy, send the command */
	if(pThinc->useCopy) {
		thincBufferNewCmdReg(pThinc, dstW, T_FB_COPY, 
			&copyReg, deltax, deltay);
	} else {
		thincBufferNewCmdReg(pThinc, dstW, T_FB_RUPDATE, &copyReg,
			T_RAW_VA_NONE);
	}

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("COPYAREA_ONSCR", TI_END);
}

/* FIXME: Problematic function for offscreen pixmaps. We can't do the same as
 * in CopyArea because some of the commands may (and probably will) affect 
 * planes outside of the one being copied. The solution is to insert a fake
 * RAW update in the destination pixmap that covers the region which is being
 * copied. Ugly but no way around, at least for now.
 * In addition, in the visible case the best we can do is just issue a RAW
 * update since the client doesn't have this capability.
 * Fortunately this function doesn't seem to be used very often...
 */
RegionPtr thincCopyPlane(DrawablePtr src, DrawablePtr dst, GCPtr pGC, 
	int srcx, int srcy, int w, int h, int dstx, int dsty, 
	unsigned long plane)
{
	t_dbg_name(func)
	RegionRec updateRec;
	RegionPtr res;
	int x, y;
	int isOff;
	int ret;

	thincGC_pre(pGC);

	instr_add("COPYPLANE", TI_START);
	thincDrawable_print(dst, "");

	stats_inc(pThinc, countCopyPlane);

	/* call driver's func */
	instr_add("COPYPLANE", TI_DRAW_START);
	res = (*pGC->ops->CopyPlane)(src, dst, pGC, srcx, srcy, 
		w, h, dstx, dsty, plane);
	instr_add("COPYPLANE", TI_DRAW_END);

	x = dstx+dst->x;
	y = dsty+dst->y;

	thincDbg(T_DBG_DRAW, ("%s: sending update (%d,%d)->(%d,%d)\n", 
		func, x, y, w, h));

	call_clipUpdate(pGC, &updateRec, x, y, w, h, res);

	/* chk what kind of drawable we're dealing with */
	isOff = is_offscreen(dst);
	/* We're fcuked :*( */
		thincWarn("THINC: CopyPlane called, injecting fake RAW "
			"update\n");
	punt_to_raw(pThinc, isOff, ret, dst, &updateRec);
	if(ret != T_OK) {
		thincWarn("THINC: Couldn't process CopyPlane command\n");
	}

	REGION_UNINIT(pScreen, &updateRec);

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("COPYPLANE", TI_END);
	return res;
}

/* 
 * draws outlines of rectangles for each rectangle in the array
 * Update is region composed of the 4 borders of each rectangle
 */
void thincPolyRectangle( DrawablePtr dst, GCPtr pGC, int n, xRectangle *pRect)
{
	t_dbg_name(func)
	RegionPtr reg;
	int add, lw;
	int i;
	static xRectangle *rects=NULL;
	register xRectangle *s, *d;
	static int prev_n;
	int isOff;
	int res = T_OK;

	thincGC_pre(pGC);

	instr_add("POLYRECT", TI_START);
	thincDrawable_print(dst, "");
	thincDbg(T_DBG_DRAW, ("%s: Drawing %d rectangles, ALU %d\n",func, n,
		pGC->alu));
	
	stats_inc(pThinc, countPolyRectangle);

	/* how much to add to each border to make sure we cover the whole
	 * outline */
	lw = pGC->lineWidth;
	if(lw == 0) 
		lw = 1;
	add = lw/2;

	thincDbg(T_DBG_DRAW, ("%s: linewidth %d, add %d\n", func,lw,add));

	/* we need 4 rectangles for each outline that needs to be drawn.
	 * In addition we allocate a little more space to reduce unnecessary
	 * allocations in the future */
	instr_add("POLYRECT", TI_ALLOC_START);
	if(!rects)
		rects = calloc(4*(n+2), sizeof(xRectangle));
	else if(prev_n < n)
		rects = realloc(rects, 4*(n+2)*sizeof(xRectangle));
	if(!rects) {
		goto error;
	}
	instr_add("POLYRECT", TI_ALLOC_END);
	prev_n = n+2;

	s = pRect;
	d = rects;
	/* the line is centered around the coordinates, so 1/2 linewidth
	 * is projected outwards and 1/2 linewidth is projected inwards 
	 * of the rectangle's original box. For some case we add 1 in case
	 * the linewidth is 1 */
	for(i=0; i<n; i++, d+=4, s++) {
		thincDbg(T_DBG_DRAW, ("%s: rectangle %d: x:%d y:%d w:%d "
			"h:%d\n", func, i, s->x, s->y, s->width, s->height));
		/* top border */
		d[0].x = s->x-add;
		d[0].y = s->y-add;
		d[0].width = s->width + (add ? 2*add : 1);
		d[0].height = (add ? 2*add : 1);
		/* left border */
		d[1].x = s->x - add;
		d[1].y = s->y - add;
		d[1].width = (add ? 2*add : 1);
		d[1].height = s->height + (add ? 2*add : 1);
		/* bottom border */
		d[2].x = s->x - add;
		d[2].y = s->y + s->height - add;
		d[2].width = s->width + (add ? 2*add : 1);
		d[2].height = (add ? 2*add : 1);
		/* right border */
		d[3].x = s->x + s->width - add;
		d[3].y = s->y - add;
		d[3].width = (add ? 2*add : 1);
		d[3].height = s->height+ (add ? 2*add : 1);
	}
	reg = RECTS_TO_REGION(pScreen, n*4, rects, CT_NONE);
	REGION_TRANSLATE(pScreen, reg, dst->x, dst->y);

	/* call driver's func */
	instr_add("POLYRECT", TI_DRAW_START);
	(*pGC->ops->PolyRectangle)(dst, pGC, n, pRect);
	instr_add("POLYRECT", TI_DRAW_END);
	
	/* clip the region */
	call_clipUpdateRegion(pGC, reg, noret);

	/* figure out what kind of drawable this is */
	isOff = is_offscreen(dst);

	/* any weird GC settings? */
	if(pThincGC->flags & gc_weirdo) {
		thincDbg(T_DBG_DRAW, ("%s: weird settings on drawing, bailing "
			"out\n", func));
		punt_to_raw(pThinc, isOff, res, dst, reg);
	/* ok, can we send a solid fill update? */
	} else if(pThinc->useSFill && pGC->lineStyle == LineSolid) {
		/* woohoo!! */
		thincDbg(T_DBG_DRAW, ("%s: Outline is solid, sending Fill\n", 
			func));
		if(isOff) 
			res = thincQueueCommand(pThinc, (PixmapPtr)dst, 
				T_FB_SFILL, reg, pGC->fgPixel);
		else
			res = thincBufferNewCmdReg(pThinc, (WindowPtr)dst,
				T_FB_SFILL, reg, pGC->fgPixel);
	} else {
		/* oh well, screw it, we're going normal update then */
		punt_to_raw(pThinc, isOff, res, dst, reg);
	}

	if(res != T_OK) {
		thincWarn("THINC: Couldn't process PolyRectangle command\n");
	}

	REGION_DESTROY(pScreen, reg);

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	instr_add("POLYRECT", TI_END);
	return;

error:
	/* call driver's func even on error */
	instr_add("POLYRECT", TI_DRAW_START);
	(*pGC->ops->PolyRectangle)(dst, pGC, n, pRect);
	instr_add("POLYRECT", TI_DRAW_END);
	thincGC_post(pGC);
	thincDbg(T_DBG_ERROR, ("%s: Error occurred, called dirver's function "
		"anyway\n", func));
	instr_add("POLYRECT", TI_END);
}

/* 
 * Update is region enclosing the rectangles
 * Depending on pGC->FillStyle any of Solid, Pixmap, Glyph or Bilevel fill 
 * may be used to send the update.
 */
void thincPolyFillRect( DrawablePtr dst, GCPtr pGC, int n, xRectangle *pRect)
{
	t_dbg_name(func)
	RegionPtr pRegion;
	int isOff;
	int res = T_OK;
	thincGC_pre(pGC);

	instr_add("POLYFILLRECT", TI_START);
	thincDrawable_print(dst, "");
	thincDbg(T_DBG_DRAW, ("%s: %d rectangles, FillStyle %d\n", func, n,
		pGC->fillStyle));

	stats_inc(pThinc, countPolyFillRect);

	if(n <=0) {
		instr_add("POLYFILLRECT", TI_DRAW_START);
		(*pGC->ops->PolyFillRect)(dst, pGC, n, pRect);
		instr_add("POLYFILLRECT", TI_DRAW_END);
		thincGC_post(pGC);
		thincDbg(T_DBG_DRAW, ("%s: not sending update\n", func));
		return;
	}

	/* what kind of drawable is this? */
	isOff = is_offscreen(dst);

	/* Collapse all the rectangles into one region. Done here in case 
	 * the driver's function changes things */
	pRegion = RECTS_TO_REGION(pScreen, n, pRect, CT_NONE);

	/* translate them to screen coordinates */
	REGION_TRANSLATE(pScreen, pRegion, dst->x, dst->y);

	/* call driver's func */
	instr_add("POLYFILLRECT", TI_DRAW_START);
	(*pGC->ops->PolyFillRect)(dst, pGC, n, pRect);
	instr_add("POLYFILLRECT", TI_DRAW_END);
	/* clip the region */
	call_clipUpdateRegion(pGC, pRegion, noret);

	if(region_is_empty(pRegion)) {
		thincDbg(T_DBG_DRAW, ("%s: draw region is empty, nomol\n",
			func));
		goto end;
	}

	thincDbg(T_DBG_DRAW, ("%s: clipped region, have %ld rectangles\n", 
		func, REGION_NUM_RECTS(pRegion)));

	/* Figure out how to send the update. Lots of options here */
	if(pThincGC->flags & gc_weirdo) {
		/* nothing to do here */
		thincDbg(T_DBG_DRAW, ("%s: Unsupported GC settings, bailing "
			"out\n", func));
		punt_to_raw(pThinc, isOff, res, dst, pRegion);
	/* can we do a solid fill? */
	} else if(pThinc->useSFill && pGC->fillStyle == FillSolid) {
		thincDbg(T_DBG_DRAW, ("%s: Using Solid Fill for update. "
			"color is %#lx\n", func, pGC->fgPixel));
		res = thincFillRegion_Solid(pThinc, dst, pRegion, isOff,
			pGC->fgPixel);
	/* can we do a pixmap fill then? */
	} else if(pThinc->usePFill && pGC->fillStyle == FillTiled) {
		thincDbg(T_DBG_DRAW, ("%s: Using Pixmap Fill for update\n", 
			func));
		res = thincFillRegion_Pixmap(pThinc, dst, pRegion,isOff,pGC);
	/* ok, try glyph fill (stipple @ pGC->stipple) */
	} else if(pThinc->useGlyph && pGC->fillStyle == FillStippled) {
		thincDbg(T_DBG_DRAW, ("%s: Using Glyph Fill for update\n", 
			func));
		res = thincFillRegion_Glyph(pThinc, dst, pRegion,isOff,pGC);
	/* ok, try bilevel fill then (stipple @ pGC->stipple) */
	} else if(pThinc->useBilevel && pGC->fillStyle==FillOpaqueStippled) {
		thincDbg(T_DBG_DRAW, ("%s: Using Bilevel Fill for update\n", 
			func));
		res = thincFillRegion_Bilevel(pThinc,dst,pRegion,isOff,pGC);
	/* pppff */
	} else {
		thincDbg(T_DBG_DRAW, ("%s: Couldn't optimize update, sending "
			"it RAW\n", func));
		punt_to_raw(pThinc, isOff, res, dst, pRegion);
	}

	if(res != T_OK) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't process FillRect "
			"command\n", func));
		REGION_DESTROY(pThinc->pScreen, pRegion);
		thincGC_post(pGC);
		return;
	}

end:
	REGION_DESTROY(pScreen, pRegion);
	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("POLYFILLRECT", TI_END);
}

static int thincFillRegion_Solid(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, unsigned long color)
{
	int res = T_OK;
	instr_add("POLYFILLRECT_SOLID", TI_START);
	
	if(isOff) {
		res = thincQueueCommand(pThinc, (PixmapPtr)dst, 
			T_FB_SFILL, pRegion, color);
	} else {
		res = thincBufferNewCmdReg(pThinc, (WindowPtr)dst, T_FB_SFILL, 
			pRegion, color);
	}

	instr_add("POLYFILLRECT_SOLID", TI_END);
	return res;
}

static int thincFillRegion_Pixmap(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, GCPtr pGC)
{
	t_dbg_name(func)
	int px, py, pw, ph, pitch, psz;
	BoxPtr ext;
	PixmapPtr pDst = (PixmapPtr)dst;
	int res = T_OK;
	char *bits;
	instr_add("POLYFILLRECT_PIX", TI_START);
	
	thincDbg(T_DBG_DRAW, ("%s: drawable coords %d,%d\n", func, dst->x,
		dst->y));

	if(pGC->tileIsPixel && pThinc->useSFill) {
		thincDbg(T_DBG_DRAW, ("%s: pixmap is actually just a color, "
			"sending SFILL instead\n", func));
		instr_add("POLYFILLRECT_PIX", TI_END);
		return thincFillRegion_Solid(pThinc, dst, pRegion, isOff,
			pGC->tile.pixel);
	}

	/* Pixmap at pGC->tile.pixmap. Coords are relative to drawable so
	 * adjust as needed */
	px = pGC->patOrg.x + dst->x;
	py = pGC->patOrg.y + dst->y;
	pw = pGC->tile.pixmap->drawable.width;
	ph = pGC->tile.pixmap->drawable.height;
	psz = pw*ph;
	pitch = pGC->tile.pixmap->devKind;
	ext = REGION_EXTENTS(pThinc->pScreen, pRegion);

	thincDbg(T_DBG_DRAW, ("%s: pixmap dimensions: px %d, py %d, pw %d "
		"ph %d, pitch %d\n", func, px, py, pw, ph, pitch));
	thincDbg(T_DBG_DRAW, ("%s: drawing to: x1 %d y1 %d x2 %d y2 %d\n",
		func, ext->x1, ext->y1, ext->x2, ext->y2));

	/* is it worth sending the pixmap fill? */
	if(psz > T_PIXMAP_MAXS) {
		/* TODO: Maybe in this case we should do RAW + COPY? */
		thincDbg(T_DBG_DRAW, ("%s: pixmap too big, sending "
			"normal update\n", func));
		punt_to_raw(pThinc, isOff, res, dst, pRegion);
	} else if( box_size_pix(ext) <= psz) {
		thincDbg(T_DBG_DRAW, ("%s: region too small %dx%d "
			"sending normal update\n", func, 
			box_width(ext), box_height(ext)));
		punt_to_raw(pThinc, isOff, res, dst, pRegion);
	} else {
		/* woohooo!! */
		bits = thinc_getPixmap(pGC->tile.pixmap);
		if(isOff) {
			res = thincQueueCommand(pThinc, pDst, T_FB_PFILL,
				pRegion, bits, px, py, pw, ph);
		} else {
			res = thincBufferNewCmdReg(pThinc, (WindowPtr)dst, 
				T_FB_PFILL, pRegion, bits, px, py, pw, ph);
		}
	}

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("POLYFILLRECT_PIX", TI_END);
	return res;
}

static int thincFillRegion_Glyph(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, GCPtr pGC)
{
	t_dbg_name(func)
	/* stipple dimensions */
	int sx, sy, sw, sh;
	int res = T_OK;
	char *bits;
	char *from, *to;
	unsigned int i, bw, size;
	PixmapPtr pDst = (PixmapPtr)dst;
	instr_add("POLYFILLRECT_GLYPH", TI_START);

	thincDbg(T_DBG_DRAW, ("%s: %d %d\n", func, dst->x, dst->y));

	thincDrawable_print((DrawablePtr)pGC->stipple, "stipple");

	sx = pGC->patOrg.x + dst->x;
	sy = pGC->patOrg.y + dst->y;
	sw = pGC->stipple->drawable.width;
	sh = pGC->stipple->drawable.height;

	thincDbg(T_DBG_DRAW, ("%s: bitmap coords: x %d y %d w %d h %d "
		"(%d,%d)\n", func, sx, sy, sw, sh, pGC->stipple->drawable.x,
		pGC->stipple->drawable.y));
	
	if(bitmap_size(sw,sh) > T_MAX_BITMAP_SIZE) {
		/* FIXME: Should be able to break it up and do something 
		 * about this */
		thincDbg(T_DBG_DRAW, ("%s: bitmap too big, sending normal "
			"update\n", func));
		punt_to_raw(pThinc, isOff, res, dst, pRegion);
	} else {
		/* alright! */
		size = bitmap_size(sw, sh);
		instr_add("POLYFILLRECT_GLYPH", TI_ALLOC_START);
		bits = malloc(size);
		if(!bits) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
				"memory for bitmap\n", func));
			return T_ERROR;
		}
		instr_add("POLYFILLRECT_GLYPH", TI_ALLOC_END);

		bw = bitmap_width(sw);
		to = bits;
		from = thinc_getPixmap(pGC->stipple);
		for(i=0; i<sh; ++i) {
			memcpy(to, from, bw);
			to += bw;
			from += pGC->stipple->devKind;
		}
		if(isOff)
			res = thincQueueCommand(pThinc, pDst, T_FB_GLYPH,
				pRegion, bits, sx, sy, sw, sh, pGC->fgPixel,
				T_BIT_FILL);
		else {
			res = thincBufferNewCmdReg(pThinc, (WindowPtr)dst, 
				T_FB_GLYPH, pRegion, 
				bits, sx, sy, sw, sh, pGC->fgPixel, T_BIT_FILL);
		}
		free(bits);
	}

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	instr_add("POLYFILLRECT_GLYPH", TI_END);
	return res;
}

static int thincFillRegion_Bilevel(thincScreenPtr pThinc, DrawablePtr dst,
	RegionPtr pRegion, int isOff, GCPtr pGC)
{
	t_dbg_name(func)
	/* bilevel dimensions */
	int sx, sy, sw, sh;
	int res = T_OK;
	char *bits;
	char *from, *to;
	unsigned int i, bw, size;
	PixmapPtr pDst = (PixmapPtr)dst;

	instr_add("POLYFILLRECT_BILEVEL", TI_START);
	thincDbg(T_DBG_DRAW, ("%s\n", func));
	
	/* FIXME: For some reason, these values get screwed up from time
	   to time (end up being set to things like 6667). */
#if 1
	sx = pGC->patOrg.x + dst->x;
	sy = pGC->patOrg.y + dst->y;
#else
	sx = 0;
	sy = 0;
#endif
	sw = pGC->stipple->drawable.width;
	sh = pGC->stipple->drawable.height;

	thincDbg(T_DBG_DRAW, ("%s: sx %d, sy %d\n", func, sx, sy));
	
	if(bitmap_size(sw,sh) > T_MAX_BITMAP_SIZE) {
		thincDbg(T_DBG_DRAW, ("%s: bitmap too big, sending normal "
			"update\n", func));
		punt_to_raw(pThinc, isOff, res, dst, pRegion);
	} else {
		/* alright! */
		size = bitmap_size(sw, sh);
		instr_add("POLYFILLRECT_BILEVEL", TI_ALLOC_START);
		bits = malloc(size);
		if(!bits) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
				"memory for bitmap\n", func));
			return T_ERROR;
		}
		instr_add("POLYFILLRECT_BILEVEL", TI_ALLOC_END);

		bw = bitmap_width(sw);
		to = bits;
		from = thinc_getPixmap(pGC->stipple);
		for(i=0; i<sh; ++i) {
			memcpy(to, from, bw);
			to += bw;
			from += pGC->stipple->devKind;
		}
		if(isOff) 
			res = thincQueueCommand(pThinc, pDst, T_FB_BILEVEL,
				pRegion, bits, sx, sy, sw, sh, pGC->fgPixel, 
				pGC->bgPixel, T_BIT_FILL);
		else {
			res = thincBufferNewCmdReg(pThinc, (WindowPtr)dst, 
				T_FB_BILEVEL, 
				pRegion, bits, sx, sy, sw, sh, pGC->fgPixel, 
				pGC->bgPixel, T_BIT_FILL);
		}
		free(bits);
	}
	
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	instr_add("POLYFILLRECT_BILEVEL", TI_END);
	return res;
}

/* 
 * Update is a box which encloses all the points drawn 
 */
void thincPolyPoint(DrawablePtr dst, GCPtr pGC, int mode, int n, 
	DDXPointPtr pPoint)
{
	t_dbg_name(func)
	int xmax, ymax, xmin, ymin;
	int i,x,y;
	RegionRec updateRec;
	int isOff;
	int res;
	
	thincGC_pre(pGC);
	instr_add("POLYPOINT", TI_START);

	thincDrawable_print(dst, "");
	thincDbg(T_DBG_DRAW, ("%s: have %d points\n", func, n));
	
	stats_inc(pThinc, countPolyPoint);

	/* find enclosing box. do before in case driver func changes the
	 * region */
	x = xmin = xmax = pPoint[0].x;
	y = ymin = ymax = pPoint[0].y;

	for(i=1; i<n; i++) {
		if(mode == CoordModeOrigin) {
			x = pPoint[i].x;
			y = pPoint[i].y;
		} else { /* CoordModePrevious */
			x += pPoint[i].x;
			y += pPoint[i].y;
		}

		if(x < xmin) xmin = x;
		else if(x > xmax) xmax = x;
		if(y < ymin) ymin = y;
		else if(y > ymax) ymax = y;
	}

	xmin += dst->x;
	xmax += dst->x;
	ymin += dst->y;
	ymax += dst->y;

	/* call driver's func */
	instr_add("POLYPOINT", TI_DRAW_START);
	(*pGC->ops->PolyPoint)(dst, pGC, mode, n, pPoint);
	instr_add("POLYPOINT", TI_DRAW_END);

	thincDbg(T_DBG_DRAW, ("%s: sending update (%d,%d) -> (%d,%d)\n", func,
		xmin, ymin, xmax-xmin, ymax-ymin));

	call_clipUpdate(pGC, &updateRec, xmin, ymin, xmax-xmin, ymax-ymin,
		noret);

	/* see what kind of drawable this is */
	isOff = is_offscreen(dst);
	punt_to_raw(pThinc, isOff, res, dst, &updateRec);
	if(res != T_OK) {
		thincWarn("THINC: Couldn't process PolyPoint command\n");
	}

	REGION_UNINIT(pScreen, &updateRec);

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("POLYPOINT", TI_END);
}

/*
 * Update is collection of rectangles which enclose all the drawn lines 
 */
void thincPolylines(DrawablePtr dst, GCPtr pGC, int mode, int n, 
	DDXPointPtr pPoint)
{
	t_dbg_name(func)
	/* can't do thincGC_pre() because we may not need to unwrap the
	 * ops vector */

	instr_add("POLYLINES", TI_START);
	thincDrawable_print(dst, "");
	thincDbg(T_DBG_DRAW, ("%s: have %d lines\n", func, (n-1) ? n-1 : 1));

	stats_inc(THINCPtr(pGC->pScreen), countPolylines);

	/* we just call the mi functions, which in turn will call our low 
	 * level span functions to handle everything the line drawing. 
	 * Life's easier down there
	 */
	instr_add("POLYLINES", TI_DRAW_START);
	if(pGC->lineWidth == 0)
		miZeroLine(dst, pGC, mode, n, pPoint);
	else
		miWideLine(dst, pGC, mode, n, pPoint);
	instr_add("POLYLINES", TI_DRAW_END);
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("POLYLINES", TI_END);
}

/*
 * Update: a region enclosing the line segments
 */
void thincPolySegment( DrawablePtr dst, GCPtr pGC, int n, xSegment *pSegments)
{
	t_dbg_name(func)
	/* don't want to wrap the ops vector since we rely on mi for drawing 
	 * the lines */

	instr_add("POLYSEGMENT", TI_START);
	thincDbg(T_DBG_DRAW, ("%s\n", func));
	thincDrawable_print(dst, "");

	stats_inc(THINCPtr(pGC->pScreen), countPolySegment);

	instr_add("POLYSEGMENT", TI_DRAW_START);
	miPolySegment(dst, pGC, n, pSegments);
	instr_add("POLYSEGMENT", TI_DRAW_END);
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("POLYSEGMENT", TI_END);
}

/*
 * Update is region enclosing all the arcs
 */
void thincPolyArc( DrawablePtr dst, GCPtr pGC, int n, xArc* pArc)
{
	t_dbg_name(func)
	int i, lw, add;
	xRectangle *rects;
	RegionPtr reg;
	int isOff;
	int res;
	thincGC_pre(pGC);

	instr_add("POLYARC", TI_START);
	thincDbg(T_DBG_DRAW, ("%s. Have %d arcs\n", func, n));
	thincDrawable_print(dst, "");

	stats_inc(THINCPtr(pScreen), countPolyArc);

	instr_add("POLYARC", TI_ALLOC_START);
	rects = calloc(n, sizeof(xRectangle));
	if(!rects) {
		goto error;
	}
	instr_add("POLYARC", TI_ALLOC_END);

	lw = pGC->lineWidth;
	if(lw == 0) 
		lw = 1;

	add = lw/2;

	for(i=0; i<n; i++) {
		rects[i].x = pArc[i].x + dst->x - add;
		rects[i].y = pArc[i].y + dst->y - add;
		rects[i].width = pArc[i].width + lw;
		rects[i].height = pArc[i].height + lw;
	}

	reg = RECTS_TO_REGION(pScreen, n, rects, CT_NONE);
	free(rects);

	/* call driver's func */
	instr_add("POLYARC", TI_DRAW_START);
	(*pGC->ops->PolyArc)(dst, pGC, n, pArc);
	instr_add("POLYARC", TI_DRAW_END);

	/* clip the update */
	call_clipUpdateRegion(pGC, reg, noret);

	/* send update */
	isOff = is_offscreen(dst);
	punt_to_raw(pThinc, isOff, res, dst, reg);
	if(res != T_OK) {
		thincWarn("THINC: Couldn't process PolyArc command\n");
	}

	REGION_DESTROY(pScreen, reg);

	thincGC_post(pGC);
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("POLYARC", TI_END);
	return;

error:	
	/* call driver's function even on error */
	instr_add("POLYARC", TI_DRAW_START);
	(*pGC->ops->PolyArc)(dst, pGC, n, pArc);
	instr_add("POLYARC", TI_DRAW_END);
	thincGC_post(pGC);
	thincDbg(T_DBG_ERROR, ("%s: Error occured, called driver's function "
		"anyway\n", func));
	instr_add("POLYARC", TI_END);
}

/*
 * Update: Region enclosing all the rectangles. Same code as PolyArc.
 */
void thincPolyFillArc( DrawablePtr dst, GCPtr pGC, int n, xArc *pArc)
{
	t_dbg_name(func)

	instr_add("POLYFILLARC", TI_START);
	thincDbg(T_DBG_DRAW, ("%s. %d arcs\n", func, n));
	thincDrawable_print(dst, "");

	stats_inc(THINCPtr(pGC->pScreen), countPolyFillArc);

	instr_add("POLYFILLARC", TI_DRAW_START);
	miPolyFillArc(dst, pGC, n, pArc);
	instr_add("POLYFILLARC", TI_DRAW_END);
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("POLYFILLARC", TI_END);
}

void thincFillPolygon( DrawablePtr dst, GCPtr pGC, int shape, int mode, 
	int count, DDXPointPtr pPoint)
{
	t_dbg_name(func)
	instr_add("FILLPOLYGON", TI_START);

	thincDbg(T_DBG_DRAW, ("%s. Polygon has %d points. FillStyle %d\n", 
		func, count, pGC->fillStyle));
	thincDrawable_print(dst, "");

	stats_inc(THINCPtr(pGC->pScreen), countFillPolygon);

	instr_add("FILLPOLYGON", TI_DRAW_START);
	miFillPolygon(dst, pGC, shape, mode, count, pPoint);
	instr_add("FILLPOLYGON", TI_DRAW_END);
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("FILLPOLYGON", TI_END);
}

/****************************************************************************
 *			Pixblit Procedures 
 ****************************************************************************/

/* Draw rows of pixels starting at pPoints of width pWidths */
/* GC components: alu, clipOrg, clientClip, and fillStyle. 
 * GC mode-dependent components: 
 * -> fgPixel (for fillStyle Solid)
 * -> tile, patOrg (for fillStyle Tile)
 * -> stipple, patOrg, fgPixel (for fillStyle Stipple)
 * -> stipple, patOrg, fgPixel and bgPixel (for fillStyle OpaqueStipple).
 */
void thincFillSpans( DrawablePtr dst, GCPtr pGC, int nSpans, 
	DDXPointPtr pPoints, int *pWidths, int sorted)
{
	t_dbg_name(func)
	int height =1, i;
	static xRectangle* spans=NULL;
	static int spans_len=0;
	xRectangle* curr;
	RegionPtr update;
	int isOff;
	int res = T_OK;

	thincGC_pre(pGC);
	instr_add("FILLSPANS", TI_START);

	thincDbg(T_DBG_DRAW, ("%s: have %d spans. FillStyle %d\n", 
		func, nSpans, pGC->fillStyle));

	thincDrawable_print(dst, "");

	stats_inc(pThinc, countFillSpans);

	/* call driver's func */
	instr_add("FILLSPANS", TI_DRAW_START);
	(*pGC->ops->FillSpans)(dst, pGC, nSpans, pPoints, pWidths, sorted);
	instr_add("FILLSPANS", TI_DRAW_END);

	thincDbg(T_DBG_DRAW, ("%s: here\n", func));

	if(!spans) {
		thincDbg(T_DBG_DRAW, ("%s: !spans\n", func));
		instr_add("FILLSPANS", TI_ALLOC_START);
		spans = malloc((nSpans*2) * sizeof(xRectangle));
		instr_add("FILLSPANS", TI_ALLOC_END);
		spans_len = nSpans*2;
	} else if(spans_len < nSpans) {
		thincDbg(T_DBG_DRAW, ("%s: spans_len (%d) < nSpans (%d)\n",
			func, spans_len, nSpans));
		instr_add("FILLSPANS", TI_ALLOC_START);
		spans = realloc(spans, (nSpans*2)*sizeof(xRectangle));
		instr_add("FILLSPANS", TI_ALLOC_END);
		spans_len = nSpans*2;
	}

	if(!spans) {
		spans_len = 0;
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for spans\n", func));
		return;
	}

	curr = spans;

	/* the points coords are global */
	for(i=0; i<nSpans; i++, curr++) {
		curr->x = pPoints[i].x;
		curr->y = pPoints[i].y;
		curr->width = pWidths[i];
		curr->height = height;
	}

	/* create the region and send the update. Note that the update will
	 * not send the region's extents but each rectangle separatedly */
	update = RECTS_TO_REGION(pScreen, nSpans, spans, CT_NONE);

	/* clip the region */
	call_clipUpdateRegion(pGC, update, noret);

	isOff = is_offscreen(dst);

	/* Check if GC has weird settings */
	if(pThincGC->flags & gc_weirdo) {
		thincDbg(T_DBG_DRAW, ("%s: Unsupported GC settings, bailing "
			"out\n", func));
		punt_to_raw(pThinc, isOff, res, dst, update);
	/* can we send a solid fill? */
	} else if(pThinc->useSFill && pGC->fillStyle == FillSolid) {
		thincDbg(T_DBG_DRAW, ("%s: Using solid fill for update\n", 
			func));
		res = thincFillRegion_Solid(pThinc, dst, update, isOff,
			pGC->fgPixel);
	/* can we send a pixmap fill? */
	} else if(pThinc->usePFill && pGC->fillStyle == FillTiled) {
		thincDbg(T_DBG_DRAW, ("%s: Using Pixmap Fill for update\n", 
			func));
		res = thincFillRegion_Pixmap(pThinc, dst, update, isOff, pGC);
	/* can we send a glyph fill? */
	} else if(pThinc->useGlyph && pGC->fillStyle == FillStippled) {
		thincDbg(T_DBG_DRAW, ("%s: Using Glyph Fill for update\n", 
			func));
		res = thincFillRegion_Glyph(pThinc, dst, update,isOff,pGC);
	/* can we send a bilevel fill? */
	} else if(pThinc->useBilevel && pGC->fillStyle==FillOpaqueStippled) {
		thincDbg(T_DBG_DRAW, ("%s: Using Bilevel Fill for update\n", 
			func));
		res = thincFillRegion_Bilevel(pThinc,dst,update,isOff,pGC);
	/* oh well.. */
	} else {
		thincDbg(T_DBG_DRAW, ("%s: Couldn't optimize update, sending "
			"it RAW\n", func));
		punt_to_raw(pThinc, isOff, res, dst, update);
	}

	if(res != T_OK) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't process FillRect "
			"command\n", func));
		REGION_DESTROY(pThinc->pScreen, update);
		thincGC_post(pGC);
		return;
	}

	REGION_DESTROY(pScreen, update);

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("FILLSPANS", TI_END);
}

/* For each span, this routine should copy pWidths bits from pSrc to 
 * pDrawable at pPoints using the raster-op from the GC.
 * GC components: alu, clipOrg, and clientClip
 * XXX: Does this function copy data from pSrc to the drawable and 
 * nothing else? */
void thincSetSpans( DrawablePtr pDrawable, GCPtr pGC, char *pSrc, 
	DDXPointPtr pPoints, int *pWidths, int nSpans, int sorted)
{
	t_dbg_name(func)
	int height =1, i;
	static xRectangle *spans=NULL;
	static int spans_len=0;
	xRectangle* curr;
	RegionPtr update;
	int isOff;
	int res = T_OK;
	thincGC_pre(pGC);
	instr_add("SETSPANS", TI_START);

	thincDbg(T_DBG_DRAW, ("%s: have %d spans\n", func, nSpans));

	thincDrawable_print(pDrawable, "");
	stats_inc(pThinc, countSetSpans);

	/* call driver's func */
	instr_add("SETSPANS", TI_DRAW_START);
	(*pGC->ops->SetSpans)(pDrawable, pGC, pSrc, pPoints, pWidths, 
		nSpans, sorted);
	instr_add("SETSPANS", TI_DRAW_END);

	instr_add("SETSPANS", TI_ALLOC_START);
	if(!spans)
		spans = calloc(nSpans+2, sizeof(xRectangle));
	else if(spans_len < nSpans)
		spans = realloc(spans, (nSpans+2)*sizeof(xRectangle));
	if(!spans) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory\n",
			func));
		return;
	}
	instr_add("SETSPANS", TI_ALLOC_END);
	spans_len = nSpans+2;

	curr = spans;

	for(i=0; i<nSpans; i++, curr++) {
		curr->x = pPoints[i].x;
		curr->y = pPoints[i].y;
		curr->width = pWidths[i];
		curr->height = height;
	}

	/* create the region and send the update. Note that the update will
	 * not send the region's extents but each rectangle separatedly */
	update = RECTS_TO_REGION(pScreen, nSpans, spans, CT_NONE);

	/* clip the update */
	call_clipUpdateRegion(pGC, update, noret);

	isOff = is_offscreen(pDrawable);
	punt_to_raw(pThinc, isOff, res, pDrawable, update);
	if(res != T_OK) {
		thincWarn("THINC: Couldn't process SetSpans command\n");
	}
	
	REGION_DESTROY(pScreen, update);
	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("SETSPANS", TI_END);
}

/* XXX: This function is only used by the mi software cursor, so no need to
 * optimize it */
void thincPushPixels( GCPtr pGC, PixmapPtr pBitMap, DrawablePtr pDrawable,
	int w, int h, int x, int y)
{
	t_dbg_name(func)
	RegionRec updateRec;
	int isOff;
	int res;
	thincGC_pre(pGC);
	instr_add("PUSHPIXELS", TI_START);

	thincDbg(T_DBG_DRAW, ("%s\n", func));

	thincDrawable_print(pDrawable, "");

	stats_inc(pThinc, countPushPixels);

	/* call driver's func */
	instr_add("PUSHPIXELS", TI_DRAW_START);
	(*pGC->ops->PushPixels)(pGC, pBitMap, pDrawable, w, h,
		x, y);
	instr_add("PUSHPIXELS", TI_DRAW_END);

	x += pDrawable->x;
	y += pDrawable->y;

	thincDbg(T_DBG_DRAW, ("%s (id: %ld) (%d,%d) -> (%d,%d)\n", func, 
		pDrawable->id, x, y, w, h));

	call_clipUpdate(pGC, &updateRec, x, y, w, h, noret);

	isOff = is_offscreen(pDrawable);
	punt_to_raw(pThinc, isOff, res, pDrawable, &updateRec);
	if(res != T_OK) {
		thincWarn("THINC: Couldn't process PushPixels command\n");
	}

	REGION_UNINIT(pScreen, &updateRec);

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	instr_add("PUSHPIXELS", TI_END);
}

/* el fin. */
