/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/*
 * Header for instrumentation code
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#ifndef _THINC_INSTRUMENT_H_
#define _THINC_INSTRUMENT_H_

/***** statistics ******/

#ifdef T_GETSTATS
/* captures stats for each thinc command */
struct thincStats {
	unsigned long count;	/* how many */
	unsigned long zcount;
	unsigned long size;	/* how much uncompressed data */
	unsigned long zsize;	/* how much compressed data (if applicable) */
	/* times spent doing particular operations */
	unsigned long create;
	unsigned long destroy;
	unsigned long copy;
	unsigned long modify;
	unsigned long info;
	unsigned long exec;
	unsigned long flush;
	unsigned long write;
	unsigned long block;
	unsigned long total;	/* sum of previous times */
};

/* define some variables only when stats are being collected */
#define stats_var(var)	unsigned long var;
#define stats_var2(v1, v2) 	unsigned long v1, v2;

/* manipulate stats */
#define stats_init_simple(stat)	(stat) = 0
#define stats_init(str, which) ((str)->which = 0)
#define stats_get(str, which) ((str)->which)
#define stats_set(str, which, v) (stats_get(str, which) = (v))
#define stats_inc(str, which) (stats_get(str, which)++)
#define stats_dec(str, which) (stats_get(str, which)--)
#define stats_add(str, which, v) (stats_get(str, which) += (v))
#define stats_add_nostr(var, v) ((var) += (v))
#define stats_print(str, which) \
	thincInfo("=> " #which ": %lu\n", stats_get(str,which))

/* don't really use str but just in case we need it in the future */
#define stats_start(str, start) \
	gettime_ms(start)

#define stats_end_simple(str, end) \
	gettime_ms(end)

#define stats_end(str, start, end, which) do {	\
	stats_end_simple(str, end);				\
	stats_add(str, which, (end) - (start));		\
} while(0)

#define stats_start_nostr(start) \
	gettime_ms(start)

#define stats_end_nostr(start, end, var) do {	\
	stats_end_simple(str, end);				\
	stats_add_nostr(var, (end) - (start));		\
} while(0)

/* write something to the log when we're collecting stats. argument
 * should be enclosed in parentheses (so we can support variable args) */
#define stats_log(s)	thincInfo s
/* same as stats_log but for debugging */
#define stats_dbg(s)	thincDbg(T_DBG_STATS, s)

void stats_str_init(struct thincStats *stats);
void stats_str_print(struct thincStats *stats);

#else					/* T_GETSTATS */
#define stats_var(v)
#define stats_var2(v1, v2)
#define stats_init_simple(stat)
#define stats_init(str, which)
#define stats_get(str, which) 
#define stats_set(str, which, v)
#define stats_inc(str, which) 
#define stats_dec(str, which) 
#define stats_add(str, which, v)
#define stats_add_nostr(var, v)
#define stats_print(str, which)
#define stats_start(str, start)
#define stats_start_nostr(start)
#define stats_end_simple(str, end)
#define stats_end(str, start, end, which)
#define stats_end_nostr(start, end, var)
#define stats_log(s)
#define stats_dbg(s)
#define stats_str_init(s)
#define stats_str_print(s)
#endif					/* T_GETSTATS */

/***** instrumentation ******/
#ifdef T_INSTRUMENT

/* define this to get a looooot more events reported */
#if 0
#define T_INSTRUMENT_EXTRA
#endif

/* from <asm/msr.h>. x86 only? */
#define rdtscl(low) \
     __asm__ __volatile__("rdtsc" : "=a" (low) : : "edx")

#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))

/* can't use long long so be careful if your measurements
 * are too far apart in time */
#define gettime_tsc(t) rdtscl(t)

#define gettime_tsc2(tlow, thigh) rdtsc(tlow, thigh)

#define instr_time(t) gettime_tsc(t)
#define instr_time2(tlow, thigh) gettime_tsc2(tlow, thigh)

#define T_INSTR_STRLEN	32

/* describes an event */
struct t_irecord {
	char 		str[T_INSTR_STRLEN];
	unsigned int	ev;
	unsigned long	tlow;
	unsigned long	thigh;
};

/* some predefined events */
enum {
	TI_START
	, TI_END
	, TI_DRAW_START
	, TI_DRAW_END
	, TI_ALLOC_START
	, TI_ALLOC_END
	, TI_BUFFER_START
	, TI_BUFFER_END
	, TI_EXEC_QUEUE_START
	, TI_EXEC_QUEUE_END
	, TI_COPY_QUEUE_START
	, TI_COPY_QUEUE_END
} t_instr_evs;

/* how many records we have */
#define T_INSTR_HOWMANY		(1UL<<16)

#define instr_print_ev(ev)	\
	instr_log(("%d: %s\n", ev, t_str(ev)))

#define instr_print_rec(i)	\
	instr_log(("%s %u %lu %lu\n", instr_rec_str(i), instr_rec_ev(i),\
		instr_rec_thigh(i), instr_rec_tlow(i)))


extern struct t_irecord* t_ievents;
extern unsigned long t_icount;
extern unsigned long t_iwrap;


/* macros */
#define instr_inc() 	\
	((++t_icount == T_INSTR_HOWMANY) && (t_icount = 0) && (++t_iwrap))

#define instr_curr()		(t_ievents[t_icount])
#define instr_currp()		(&t_ievents[t_icount])
#define instr_curr_str()	(instr_curr().str)
#define instr_curr_ev()		(instr_curr().ev)
#define instr_curr_tlow()		(instr_curr().tlow)
#define instr_curr_thigh()		(instr_curr().thigh)

#define instr_rec(n)		(t_ievents[n])
#define instr_recp(n)		(&t_ievents[n])
#define instr_rec_str(n)	(instr_rec(n).str)
#define instr_rec_ev(n)		(instr_rec(n).ev)
#define instr_rec_tlow(n)		(instr_rec(n).tlow)
#define instr_rec_thigh(n)		(instr_rec(n).thigh)

/* makes sure instrumentation was properly initialized */
#define instr_check()	\
	t_bug_on(					\
		!t_ievents, 				\
		T_BUG(	"Instrumentation not properly initialized. "	\
			"Something really bad is about to happen\n")	\
	)

#define instr_strcpy(id) do {				\
	register char *__to = instr_curr_str();		\
	register char *__from = (id);			\
	register unsigned int __i = T_INSTR_STRLEN-1;	\
	while((*__to++ = *__from++) && --__i)		\
		/* nuthin' */ ;				\
	*__to = '\0';					\
} while(0)

#define instr_add(_str, _ev) do {			\
	instr_check();					\
	instr_strcpy(_str);				\
	instr_curr_ev() = (_ev);			\
	instr_time2(instr_curr_tlow(), instr_curr_thigh());	\
	instr_inc();					\
} while(0)

/* use this one for events we normally don't want, but may need
 * in some special circumstances */
#ifdef T_INSTRUMENT_EXTRA
#define instr_add_extra(_str, _ev) instr_add(_str, _ev)
#else
#define instr_add_extra(_str, _ev)
#endif


#define instr_add_time(_str, _ev, _tlow, _thigh) do { 	\
	instr_check();					\
	instr_strcpy(_str);				\
	instr_curr_ev() = (_ev);			\
	instr_curr_tlow() = (_tlow);			\
	instr_curr_thigh() = (_thigh);			\
	instr_inc();					\
} while(0)

/* write something to the log when we're doing instrumentation. argument
 * should be enclosed in parentheses (so we can support variable args) */
#define instr_log(s)	thincInfo s
/* same as instr_log but for debugging */
#define instr_dbg(s)	thincDbg(T_DBG_STATS, s)

#else					/* T_INSTRUMENT */
#define instr_time(t)
#define instr_time2(tlow, thigh)
#define gettime_tsc(t)
#define gettime_tsc2(tlow, thigh)
#define rdtscl(low)
#define rdtsc(low,high)

#define instr_print_ev(ev)
#define instr_print_rec(i)

#define instr_inc()
#define instr_strcpy(id)
#define instr_add(id, ev)
#define instr_add_time(_id, _ev, _tlow, _thigh)

#define instr_curr()	 
#define instr_currp()	 
#define instr_curr_str() 
#define instr_curr_ev()	 
#define instr_curr_tlow()
#define instr_curr_thigh()

#define instr_rec(n)	 
#define instr_recp(n)	 
#define instr_rec_str(n) 
#define instr_rec_ev(n)	 
#define instr_rec_tlow(n)
#define instr_rec_thigh(n)

#define instr_log(s)
#define instr_dbg(s)
#define instr_add_extra(_str, _ev)

#endif					/* T_INSTRUMENT */

#endif /* _THINC_INSTRUMENT_H_ */
