/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * thinc Keyboard functions
 * 
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#include "xf86.h"
#include "xf86Xinput.h"
#include "mi.h"
#include "mipointer.h"

#if 0 /* we don't use this anymore */
#ifdef XKB
#include "X11/extensions/XKB.h"
#include "X11/extensions/XKBstr.h"
#include "X11/extensions/XKBsrv.h"

static char * xkbkeymap;
static char * xkbkeycodes;
static char * xkbtypes;
static char * xkbcompat;
static char * xkbsymbols;
static char * xkbgeometry;
static Bool xkbcomponents_specified;
static char * xkbrules;
static char * xkbmodel;
static char * xkblayout;
static char * xkbvariant;
static char * xkboptions;
#endif
#endif /* old XKB stuff */

/* event handling functions */
static void thincInsertKeybEvent(thinc_evKeyb ev, unsigned long time);

static InputInfoPtr thincKeybInit(InputDriverPtr drv, IDevPtr dev, int flags);
static void thincKeybUninit(InputDriverPtr drv, InputInfoPtr info, int flags);
static int thincKbdDeviceInit(DeviceIntPtr dev, KeySymsPtr keySyms,
	CARD8 *modMap);
static int thincKeybControl(DeviceIntPtr device, int what);

static DeviceIntPtr thincKbdDevice = NULL;

/* keyboard map info */
#include "atKeynames.h"
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 1
#include "xf86Keymap.h"
#else
#include "thincKeyb.h"
#endif

static InputDriverRec ThincKeyb = {
	THINC_VERSION_MAJOR,    /* version */
        "thinckeyb",   	/* name */
        NULL,          	/* identify */
        thincKeybInit, 	/* preinit */
        thincKeybUninit,	/* uninit */
        NULL,           /* module */
        0               /* ref count */
};

static const char *OPT_DEFAULTS[] = {
        NULL
};

#ifdef XINPUT
#include "xf86Xinput.h"
#define enqueue(e)	xf86eqEnqueue(e)
#else
#define enqueue(e)	mieqEnqueue(e)
#endif

/* very simplifed map. it's supposed to do the reverse of XKeycodeToKeysym
 * (which is what the X client uses to convert from X)
 * the one defined in xf86Keymap.h is not good for our purpose :( */
static KeySym thincMap[MAX_KEYCODE] = {
    /* 0x00 */  NoSymbol,       
    /* 0x01 */  XK_Escape,      
    /* 0x02 */  XK_1,           
    /* 0x03 */  XK_2,           
    /* 0x04 */  XK_3,           
    /* 0x05 */  XK_4,           
    /* 0x06 */  XK_5,           
    /* 0x07 */  XK_6,           
    /* 0x08 */  XK_7,           
    /* 0x09 */  XK_8,           
    /* 0x0a */  XK_9,           
    /* 0x0b */  XK_0,           
    /* 0x0c */  XK_minus,       
    /* 0x0d */  XK_equal,       
    /* 0x0e */  XK_BackSpace,   
    /* 0x0f */  XK_Tab,         
    /* 0x10 */  XK_q,           
    /* 0x11 */  XK_w,           
    /* 0x12 */  XK_e,           
    /* 0x13 */  XK_r,           
    /* 0x14 */  XK_t,           
    /* 0x15 */  XK_y,           
    /* 0x16 */  XK_u,           
    /* 0x17 */  XK_i,           
    /* 0x18 */  XK_o,           
    /* 0x19 */  XK_p,           
    /* 0x1a */  XK_bracketleft, 
    /* 0x1b */  XK_bracketright,
    /* 0x1c */  XK_Return,      
    /* 0x1d */  XK_Control_L,   
    /* 0x1e */  XK_a,           
    /* 0x1f */  XK_s,           
    /* 0x20 */  XK_d,           
    /* 0x21 */  XK_f,           
    /* 0x22 */  XK_g,           
    /* 0x23 */  XK_h,           
    /* 0x24 */  XK_j,           
    /* 0x25 */  XK_k,           
    /* 0x26 */  XK_l,           
    /* 0x27 */  XK_semicolon,   
    /* 0x28 */  XK_quoteright,  
    /* 0x29 */  XK_quoteleft,	
    /* 0x2a */  XK_Shift_L,     
    /* 0x2b */  XK_backslash,   
    /* 0x2c */  XK_z,           
    /* 0x2d */  XK_x,           
    /* 0x2e */  XK_c,           
    /* 0x2f */  XK_v,           
    /* 0x30 */  XK_b,           
    /* 0x31 */  XK_n,           
    /* 0x32 */  XK_m,           
    /* 0x33 */  XK_comma,       
    /* 0x34 */  XK_period,      
    /* 0x35 */  XK_slash,       
    /* 0x36 */  XK_Shift_R,     
    /* 0x37 */  XK_KP_Multiply, 
    /* 0x38 */  XK_Alt_L,	
    /* 0x39 */  XK_space,       
    /* 0x3a */  XK_Caps_Lock,   
    /* 0x3b */  XK_F1,          
    /* 0x3c */  XK_F2,          
    /* 0x3d */  XK_F3,          
    /* 0x3e */  XK_F4,          
    /* 0x3f */  XK_F5,          
    /* 0x40 */  XK_F6,          
    /* 0x41 */  XK_F7,          
    /* 0x42 */  XK_F8,          
    /* 0x43 */  XK_F9,          
    /* 0x44 */  XK_F10,         
    /* 0x45 */  XK_Num_Lock,    
    /* 0x46 */  XK_Scroll_Lock,	
    /* 0x47 */  XK_KP_Home,	
    /* 0x48 */  XK_KP_Up,	
    /* 0x49 */  XK_KP_Prior,	
    /* 0x4a */  XK_KP_Subtract, 
    /* 0x4b */  XK_KP_Left,	
    /* 0x4c */  XK_KP_Begin,	
    /* 0x4d */  XK_KP_Right,	
    /* 0x4e */  XK_KP_Add,      
    /* 0x4f */  XK_KP_End,	
    /* 0x50 */  XK_KP_Down,	
    /* 0x51 */  XK_KP_Next,	
    /* 0x52 */  XK_KP_Insert,	
    /* 0x53 */  XK_KP_Delete,	
    /* 0x54 */  XK_Sys_Req,	
    /* 0x55 */  NoSymbol,	
    /* 0x56 */  XK_less,	
    /* 0x57 */  XK_F11,		
    /* 0x58 */  XK_F12,		
    /* 0x59 */  XK_Home,	
    /* 0x5a */  XK_Up,		
    /* 0x5b */  XK_Prior,	
    /* 0x5c */  XK_Left,	
    /* 0x5d */  XK_Begin,	
    /* 0x5e */  XK_Right,	
    /* 0x5f */  XK_End,		
    /* 0x60 */  XK_Down,	
    /* 0x61 */  XK_Next,	
    /* 0x62 */  XK_Insert,	
    /* 0x63 */  XK_Delete,	
    /* 0x64 */  XK_KP_Enter,	
    /* 0x65 */  XK_Control_R,	
    /* 0x66 */  XK_Pause,	
    /* 0x67 */  XK_Print,	
    /* 0x68 */  XK_KP_Divide,	
    /* 0x69 */  XK_Alt_R,	
    /* 0x6a */  XK_Break,	
    /* 0x6b */  XK_Meta_L,	
    /* 0x6c */  XK_Meta_R,	
    /* 0x6d */  XK_Menu,	
    /* 0x6e */  XK_F13,		
    /* 0x6f */  XK_F14,		
    /* 0x70 */  XK_F15,		
    /* 0x71 */  XK_F16,		
    /* 0x72 */  XK_F17,		
    /* 0x73 */  XK_backslash,	
    /* 0x74 */  NoSymbol,	
    /* 0x75 */  NoSymbol,	
    /* 0x76 */  NoSymbol,	
    /* 0x77 */  NoSymbol,	
    /* 0x78 */  NoSymbol,	
    /* 0x79 */  XK_Henkan,	
    /* 0x7a */  NoSymbol,	
    /* 0x7b */  XK_Muhenkan,	
    /* 0x7c */  NoSymbol,	
    /* 0x7d */  XK_backslash,	
    /* 0x7e */  NoSymbol,	
    /* 0x7f */  NoSymbol,	
};




int thincHandleKeybEv(thincClientPtr cl)
{
	t_dbg_name(func)
	int res;
	thinc_evKeyb k;
	unsigned long time;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	res = thincReadAnyMsg(cl, &k, T_EV_KEYB_SIZE);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s Couldn't read keyb event "
			"message\n", func));
		return res;
	} else if(res == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Client closed connection "
			"while reading keyb event message\n", func));
		return res;
	}

	k.key = ntohl(k.key);

	gettime_ms(time);

	thincDbg(T_DBG_INPUT, ("%s: Got keyboard event for key %d, pressed? "
		"%d\n",	func, k.key, k.down));

	/* inject the event */
	thincInsertKeybEvent(k, time);

#if 0
	if(k.down == T_DOWN) {
		cl->parent->keyb_down++;
	} else {
		cl->parent->keyb_up++;
	}
#else
	if(k.down == T_DOWN) {
		thincIncKeybUp(cl->parent);
	} else {
		thincIncKeybDown(cl->parent);
	}
#endif

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return T_OK;
}

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 1
static void thincInsertKeybEvent(thinc_evKeyb ev, unsigned long time)
{
	t_dbg_name(func)
	KeySymsPtr keySyms;
	int i, keycode;
	KeySym key = ev.key;
	xEvent e;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	if(!thincKbdDevice) {
		thincError("ERROR: No keyboard device found, can't add "
			"keyboard event\n");
		return;
	}

	keySyms = &thincKbdDevice->key->curKeySyms;

	if(ev.down == T_DOWN)
		e.u.u.type = KeyPress;
	else
		e.u.u.type = KeyRelease;

	/* find the keycode */
	keycode = 0;

	thincDbg(T_DBG_INPUT, ("%s: looking for keysym %#lx\n", func, key));
	
	/* see if it's a key we know about. TODO: What if we don't know
	 * about it? hehe, such a hack */
	for(i=0; i< NUM_KEYCODES; i++) {
		if(key == thincMap[i]) {
			keycode = MIN_KEYCODE + i;
			break;
		}
	}

	thincDbg(T_DBG_INPUT, ("%s: found keycode %#x\n", func, keycode));

	e.u.keyButtonPointer.time = time;
	e.u.u.detail = keycode;
	enqueue(&e);

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}
#else
static void thincInsertKeybEvent(thinc_evKeyb ev, unsigned long time)
{
	t_dbg_name(func)
	KeySymsPtr keySyms;
	int i, keycode;
	KeySym key = ev.key;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	if(!thincKbdDevice) {
		thincError("ERROR: No keyboard device found, can't add "
			"keyboard event\n");
		return;
	}

	keySyms = &thincKbdDevice->key->curKeySyms;

	/* find the keycode */
	keycode = 0;

	thincDbg(T_DBG_INPUT, ("%s: looking for keysym %#lx\n", func, key));
	
	/* see if it's a key we know about. TODO: What if we don't know
	 * about it? hehe, such a hack */
	for(i=0; i< NUM_KEYCODES; i++) {
		if(key == thincMap[i]) {
			keycode = MIN_KEYCODE + i;
			break;
		}
	}

	thincDbg(T_DBG_INPUT, ("%s: found keycode %#x\n", func, keycode));

	xf86PostKeyboardEvent(thincKbdDevice, keycode, (ev.down == T_DOWN));

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}
#endif

/* TODO */
static void thincSendBell(void)
{
	/*
	thincLog("Keyboard Bell\n");
	*/
}

static InputInfoPtr thincKeybInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
	t_dbg_name(func)
        /* seems to be an structure with information about the input
         * device. Steps: Allocate, fill in, process options (none for us)
         * and return so it gets registered */
	InputInfoPtr info;
#if 0	/* used in no longer used xkb stuff below */
	char *curr;
	int from;
#endif

	thincDbg(T_DBG_INPUT, ("%s\n", func));
	
	if(!(info = xf86AllocateInput(drv, 0)))
		return NULL;

	/* fill the structure */
	info->name = dev->identifier;
	info->type_name = "thincKeyb";
	info->flags = XI86_KEYBOARD_CAPABLE;
	info->device_control = thincKeybControl;
	info->read_input = NULL;
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
	info->motion_history_proc = NULL;
	info->history_size = 0;
#endif
	info->control_proc = NULL;
	info->close_proc = NULL;
	info->switch_mode = NULL;
	info->conversion_proc = NULL;
	info->reverse_conversion_proc = NULL;
	info->fd = -1;
	info->dev = NULL;
	info->private_flags = 0;
	info->always_core_feedback = 0;
	info->conf_idev = dev;

        /* collect options and process the common ones. For now we
         * don't have any options. Leaving here just in case */
        xf86CollectInputOptions(info, OPT_DEFAULTS, NULL);
        xf86ProcessCommonOptions(info, info->options);

	/* Mark device as configured */
	info->flags |= XI86_CONFIGURED;

	/* XKB Stuff */
#if 0
#ifdef XKB
	from = X_DEFAULT;
	if(noXkbExtension)
		from = X_CMDLINE;
	else if(xf86FindOption(dev->commonOptions, "XkbDisable")) {
		noXkbExtension = xf86SetBoolOption(dev->commonOptions,
			"XkbDisable", FALSE);
		from = X_CONFIG;
	}
	if(noXkbExtension)
		xf86Msg(from, "XKB disabled\n");

#define NULL_IF_EMPTY(s) (curr[0] ? s: (free(curr), (char*)NULL))

	if(!noXkbExtension && !XkbInitialMap) {
		if((curr = xf86SetStrOption(dev->commonOptions, "XkbKeymap", 
			NULL))) {
			xkbkeymap = NULL_IF_EMPTY(curr);
			xf86Msg(X_CONFIG, "XKB: Keymap: %s\n", xkbkeymap);
		} else {
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbCompat", NULL))) {
				xkbcompat = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: compat %s\n",xkbcompat);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbTypes", NULL))) {
				xkbtypes = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: types %s\n", xkbtypes);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbKeycodes", NULL))) {
				xkbkeycodes = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: keycodes %s\n",
					xkbkeycodes);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbGeometry", NULL))) {
				xkbgeometry = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: geometry %s\n",
					xkbgeometry);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbSymbols", NULL))) {
				xkbsymbols = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: symbols %s\n",
					xkbsymbols);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbRules", NULL))) {
				xkbrules = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: rules %s\n",
					xkbrules);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbModel", NULL))) {
				xkbmodel = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: model %s\n",
					xkbmodel);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbLayout", NULL))) {
				xkblayout = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: layout %s\n",
					xkblayout);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbVariant", NULL))) {
				xkbvariant = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: variant %s\n",
					xkbvariant);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbOptions", NULL))) {
				xkboptions = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: options %s\n",
					xkboptions);
			}
		}
	}
#undef NULL_IF_EMPTY
#endif	/* XKB */
#endif

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return info;
}

static void thincKeybUninit(InputDriverPtr drv, InputInfoPtr info, int flags)
{
	thincKeybControl(info->dev, DEVICE_OFF);
}

static int thincKeybControl(DeviceIntPtr device, int what)
{
	t_dbg_name(func)
	KeySymsRec keySyms;
	CARD8 modMap[MAP_LENGTH];
	DevicePtr dev = (DevicePtr)device;
	int res;

	thincDbg(T_DBG_INPUT, ("%s: what %d\n", func, what));

	switch(what) {
		case DEVICE_INIT:
			res = thincKbdDeviceInit(device, &keySyms, modMap);
			if(res != Success) {
				thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
				return res;
			}
#if 0
#ifdef XKB
			if(noXkbExtension) {
#endif
#endif
			InitKeyboardDeviceStruct(dev, &keySyms, modMap,
				(BellProcPtr)thincSendBell,
				(KbdCtrlProcPtr)NoopDDA);
#if 0
#ifdef XKB
			} else {
				XkbComponentNamesRec names;
				if(XkbInitialMap) {
					if((xkbkeymap = strchr(XkbInitialMap,
						'/')) != NULL)
						xkbkeymap++;
					else
						xkbkeymap = XkbInitialMap;
				}
				if(xkbkeymap) {
					names.keymap = xkbkeymap;
					names.keycodes = NULL;
					names.types = NULL;
					names.compat = NULL;
					names.symbols = NULL;
					names.geometry = NULL;
				} else {
					names.keymap = NULL;
					names.keycodes = xkbkeycodes;
					names.types = xkbtypes;
					names.compat = xkbcompat;
					names.symbols = xkbsymbols;
					names.geometry = xkbgeometry;
				}
				if((xkbkeymap || xkbcomponents_specified) &&
				   (xkbmodel == NULL || xkblayout == NULL)) {
					xkbrules = NULL;
				}
				XkbInitKeyboardDeviceStruct(device,
					&names, &keySyms, modMap,
					(BellProcPtr)thincSendBell,
					(KbdCtrlProcPtr)NoopDDA);
			}
#endif
#endif
			break;
		case DEVICE_ON:
			dev->on = TRUE;
			break;
		case DEVICE_OFF:
			dev->on = FALSE;
			break;
		case DEVICE_CLOSE:
			break;
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return Success;
}

/* fills in the keysyms and keymap. from KbdGetMapping (lnx_KbdMap.c) */
static int thincKbdDeviceInit(DeviceIntPtr dev, KeySymsPtr keySyms,
	CARD8 *modMap)
{
	t_dbg_name(func)
	int i;
	KeySym *k;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	thincKbdDevice = dev;

	for(i=0; i<MAP_LENGTH; i++)
		modMap[i] = NoSymbol;

	for(k=map,i = MIN_KEYCODE; i < (NUM_KEYCODES + MIN_KEYCODE);++i, k+=4) 
	{
		switch(*k) {
			case XK_Shift_L:
			case XK_Shift_R:
				modMap[i] = ShiftMask;
				break;

			case XK_Control_L:
			case XK_Control_R:
				modMap[i] = ControlMask;
				break;

			case XK_Caps_Lock:
				modMap[i] = LockMask;
				break;

			case XK_Alt_L:
			case XK_Alt_R:
				modMap[i] = AltMask;
				break;

			case XK_Num_Lock:
				modMap[i] = NumLockMask;
				break;

			case XK_Scroll_Lock:
				modMap[i] = ScrollLockMask;
				break;

			/* kana support */
			case XK_Kana_Lock:
			case XK_Kana_Shift:
				modMap[i] = KanaMask;
				break;

			/* alternate toggle for multinational support */
			case XK_Mode_switch:
				modMap[i] = AltLangMask;
				break;

		}
	}

	keySyms->minKeyCode = MIN_KEYCODE;
	keySyms->maxKeyCode = MAX_KEYCODE;
	keySyms->mapWidth = GLYPHS_PER_KEY;
	keySyms->map = map;

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return Success;
}

/***** dynamic loader stuff ******/
void thincInitKeyboard(void)
{
	xf86AddInputDriver(&ThincKeyb, NULL, 0);
}

static void thincKeybUnplug(pointer p)
{
}

static pointer thincKeybPlug(pointer module, pointer options, int *errmaj,
	int *errmin)
{
	xf86AddInputDriver(&ThincKeyb, module, 0);
	
	return module;
}

static XF86ModuleVersionInfo thincKeybVersionRec = {
	"thinckeyb",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	THINC_VERSION_MAJOR, THINC_VERSION_MINOR, THINC_PATCHLEVEL,
	ABI_CLASS_XINPUT,
	ABI_XINPUT_VERSION,
	MOD_CLASS_XINPUT,
	{0, 0, 0, 0}
};

XF86ModuleData thincKeybModuleData = {
	&thincKeybVersionRec,
	thincKeybPlug,
	thincKeybUnplug
};
