/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * ThincMouse functions
 * 
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#include "xf86.h"
#include "xf86Xinput.h"
#include "mi.h"
#include "mipointer.h"
#include "xf86OSmouse.h"
#include <netinet/in.h>

static InputInfoPtr thincMouseInit(InputDriverPtr drv, IDevPtr dev, int flags);
static void thincMouseUninit(InputDriverPtr drv, InputInfoPtr info, int flags);
static int thincMouseControl(DeviceIntPtr dev, int what);

static void thincMouseCtrl(DeviceIntPtr device, PtrCtrl *ctrl);
static void thincMousePostEvent(InputInfoPtr pInfo, int buttons,
	int dx, int dy, int dz, int dw);
static void thincMouseCommonOptions(InputInfoPtr pInfo);
static Bool thincMouseConvert(InputInfoPtr pInfo, int first, int num,
	int v0, int v1, int v2, int v3, int v4, int v5, int *x, int *y);

/* Structure describing our mouse. Contains the driver entry points required
 * during InitInput() */
static InputDriverRec ThincMouse = {
	THINC_VERSION_MAJOR,	/* version */
	"thincmouse",	/* name */
	NULL,		/* identify */
	thincMouseInit,	/* preinit */
	thincMouseUninit,	/* uninit */
	NULL,		/* module */
	0		/* ref count */
};

static const char *OPT_DEFAULTS[] = {
	NULL
};

unsigned char thincPtrAccel;

static DeviceIntPtr thincMouseDevice = NULL;

static InputInfoPtr thincMouseInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
	/* seems to be an structure with information about the input
	 * device. Steps: Allocate, fill in, process options (none for us)
	 * and return so it gets registered */
	InputInfoPtr info;
	MouseDevPtr pMse;

	if(!(info = xf86AllocateInput(drv, 0)))
		return NULL;

	/* fill the structure */
	info->name = dev->identifier;
	info->type_name = "thincMouse";
	info->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
	info->device_control = thincMouseControl;
	info->read_input = NULL;

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
	info->motion_history_proc = xf86GetMotionEvents;
	info->history_size = 0;
#endif
	info->control_proc = NULL;
	info->close_proc = NULL;
	info->switch_mode = NULL;
	info->conversion_proc = thincMouseConvert;
	info->reverse_conversion_proc = NULL;
	info->fd = -1;
	info->dev = NULL;
	info->private_flags = 0;
	info->always_core_feedback = 0;
	info->conf_idev = dev;

	/* XXX: Taken from standard xfree86 mouse.c */
	/* Allocate the MouseDevRec and initialise it. */
	/*
	 * XXX This should be done by a function in the core server since the
	 * MouseDevRec is defined in the os-support layer.
	 */
	if (!(pMse = xcalloc(sizeof(MouseDevRec), 1)))
		return info;
	info->private = pMse;
	pMse->Ctrl = thincMouseCtrl;
	pMse->PostEvent = thincMousePostEvent;
	pMse->CommonOptions = thincMouseCommonOptions;

	/* collect options and process the common ones. For now we
	 * don't have any options. Leaving here just in case */
	xf86CollectInputOptions(info, OPT_DEFAULTS, NULL);
	xf86ProcessCommonOptions(info, info->options);

	/* mark it as configured */
	info->flags |= XI86_CONFIGURED;

	/* return the configured device */
	return info;
}

static void thincMouseUninit(InputDriverPtr drv, InputInfoPtr info, int flags)
{
	thincMouseControl(info->dev, DEVICE_OFF);
}

static void thincPtrDeviceCtrl(DevicePtr dev, PtrCtrl *ctrl)
{
	thincPtrAccel = (char)ctrl->num;
}

static int thincMouseControl(DeviceIntPtr device, int what)
{
	t_dbg_name(func)
	BYTE map[6];
	DevicePtr pDev = (DevicePtr)device;
	InputInfoPtr pInfo;
	MouseDevPtr pMse;

	thincDbg(T_DBG_INPUT, ("%s: %d\n", func, what));

	pInfo = device->public.devicePrivate;
	pMse = pInfo->private;
	pMse->device = device;

	switch(what) {
		case DEVICE_INIT:
			thincDbg(T_DBG_INPUT, ("%s: initialiazing device\n",
				func));
			map[1] = 1;
			map[2] = 2;
			map[3] = 3;
			map[4] = 4;
			map[5] = 5;
			InitPointerDeviceStruct(pDev, map, 5, 
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
				miPointerGetMotionEvents, 
#else
				GetMotionHistory,
#endif
				(PtrCtrlProcPtr) thincPtrDeviceCtrl,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
				miPointerGetMotionBufferSize()
#else
				GetMotionHistorySize(), 2
#endif
			);

			/* X valuator */
			xf86InitValuatorAxisStruct(device, 0, 0, -1, 1, 0, 1);
			xf86InitValuatorDefaults(device, 0);
			/* Y valuator */
			xf86InitValuatorAxisStruct(device, 1, 0, -1, 1, 0, 1);
			xf86InitValuatorDefaults(device, 1);
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
			xf86MotionHistoryAllocate(pInfo);
#endif
			thincMouseDevice = device;
			break;
		case DEVICE_ON:
			thincDbg(T_DBG_INPUT, ("%s: turning device on\n", func));
			pDev->on = TRUE;
			device->public.on = TRUE;
			break;
		case DEVICE_OFF:
		case DEVICE_CLOSE:
			thincDbg(T_DBG_INPUT, ("%s: turning device off\n", func));
			pDev->on = FALSE;
			device->public.on = FALSE;
			break;
	}

	return Success;
}

static void thincMouseCtrl(DeviceIntPtr device, PtrCtrl *ctrl)
{
	thincError("%s: I was called!\n", __func__);
}

static void thincMousePostEvent(InputInfoPtr pInfo, int buttons,
	int dx, int dy, int dz, int dw)
{
	thincError("%s: I was called!\n", __func__);
}

static void thincMouseCommonOptions(InputInfoPtr pInfo)
{
	thincError("%s: I was called!\n", __func__);
}

static Bool thincMouseConvert(InputInfoPtr pInfo, int first, int num,
	int v0, int v1, int v2, int v3, int v4, int v5, int *x, int *y)
{
	if(first != 0 || num != 2)
		return FALSE;

	*x = v0;
	*y = v1;

	return TRUE;
}

int thincHandleMotionEv(thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_evMotion m;
	int res;
	static unsigned long count = 0;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	t_bug_on_2(
		!thincMouseDevice,
		T_BUG("Mouse hasn't been initialized"),
		return T_ERROR
	);
	/* read motion message */
	res = thincReadAnyMsg(cl, &m, T_EV_MOTION_SIZE);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s Couldn't read motion event "
			"message\n", func));
		return res;
	} else if(res == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s Client closed connection "
			"while reading motion event message\n", func));
		return res;
	}
	
	cl->parent->last_mouse_client = cl;

	m.x = ntohs(m.x);
	m.y = ntohs(m.y);

	/* if using hw cursor, update move counter */
	if(cl->parent->useHWCursor)
		cl->parent->moveCount++;

	thincDbg(T_DBG_INPUT, ("%s: got motion event at %d,%d, "
		"count %d\n", func, m.x, m.y, cl->parent->moveCount));

	/* move the pointer */
	xf86PostMotionEvent(thincMouseDevice, 1, 0, 2, m.x, m.y);
	count++;
	thincDbg(T_DBG_INPUT, ("%s: finished. posted event number %ld\n", 
		func, count));

	return T_OK;
}

int thincHandleButtonEv(thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_evButton b;
	int res, button;
	unsigned int button_state;
	int isdown;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	t_bug_on_2(
		!thincMouseDevice,
		T_BUG("Mouse hasn't been initialized"),
		return T_ERROR
	);

	res = thincReadAnyMsg(cl, &b, T_EV_BUTTON_SIZE);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s Couldn't read button "
			"event message\n", func));
		return res;
	} else if(res == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s Client closed connection "
			"while waiting for button event message\n", func));
		return res;
	}

	/* send a ping right back to wake up tcp */
	if(cl->parent->ping_freq)
		thincSendPingAll(cl->parent);

	/* extract button info */
	button = b.buttonMask & 127;
	button_state = b.buttonMask & 128;
	button_state >>= 7;

	thincDbg(T_DBG_INPUT, ("%s: Got event for button %d, pressed? %d\n", 
		func, button, button_state));

	isdown = (button_state == T_DOWN ? 1 : 0);
#if 0
	if(isdown) {
		cl->parent->mouse_down++;
	} else {
		cl->parent->mouse_up++;
	}
#else
	if(isdown) {
		thincIncMouseUp(cl->parent);
	} else {
		thincIncMouseDown(cl->parent);
	}
#endif
	xf86PostButtonEvent(thincMouseDevice, 0, button, isdown, 0, 0, 0);
	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return T_OK;
}

/* Functions for dynamic loader */
/* mostly stubs */
static void thincMouseUnplug(pointer p)
{
}

static pointer thincMousePlug(pointer module, pointer options,
	int *errmaj, int *errmin)
{
	xf86Msg(X_ERROR, "Hello\n");
	xf86AddInputDriver(&ThincMouse, module, 0);
	xf86Msg(X_ERROR, "Bye Bye\n");

	return module;
}

void thincInitMouse()
{
	xf86AddInputDriver(&ThincMouse, NULL, 0);
}

static XF86ModuleVersionInfo thincMouseVersionRec =
{
	"thincmouse",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	THINC_VERSION_MAJOR, THINC_VERSION_MINOR, THINC_PATCHLEVEL,
	ABI_CLASS_XINPUT,
	ABI_XINPUT_VERSION,
	MOD_CLASS_XINPUT,
	{0, 0, 0, 0}
};

XF86ModuleData thincMouseModuleData = {
	&thincMouseVersionRec,
	thincMousePlug,
	thincMouseUnplug
};
