/* Copyright 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Implemenation of THINC's output message stack
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop 8 */
#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

static int thinc_msgStackAddModuleGen(thincListPtr s, thincMsgModulePtr m);

thincMsgStackPtr thinc_msgStackCreate(thincClientPtr cl)
{
	thincMsgStackPtr stack;

	stack = malloc(sizeof(thincMsgStackRec));
	if(!stack)
		return NULL;
	t_list_init(&stack->stack);

	return stack;
}

void thinc_msgStackDestroy(thincClientPtr cl, thincMsgStackPtr s)
{
#if 0
	t_list_empty_simple(&s->stack);
	free(s);
#else
	thincMsgModulePtr m,n;

	t_list_for_each_entry(m, n, thincMsgModule, &s->stack, stack) {
		thinc_msgStackDestroyModule(cl, s, m);
	}
	t_bug_on(
		!t_list_is_empty(&s->stack),
		T_BUG("Message stack not empty after trying to destroy it")
	);
	free(s);
#endif
}

thincMsgModulePtr thinc_msgNewModule(thincMsgMethod method, const char *name, 
	int name_len, unsigned char priority)
{
	thincMsgModulePtr module;

	module = malloc(sizeof(*module));
	if(!module) 
		return NULL;

	module->method = method;
	module->priority = priority;

	if(name_len > T_MSGMOD_NAME_LEN) {
		/* truncate length. -1 for \0 */
		strncpy(module->name, name, T_MSGMOD_NAME_LEN-1);
		module->name[T_MSGMOD_NAME_LEN-1] = '\0';
		module->name_len = T_MSGMOD_NAME_LEN;
	} else {
		module->name_len = name_len;
		strncpy(module->name, name, name_len);
	}

	t_list_init(&module->stack);

	return module;
}

int thinc_msgStackAdd(thincClientPtr cl, thincMsgStackPtr stack,
	thincMsgMethod method, const char *name, int name_len, 
	unsigned char priority)
{
	thincMsgModulePtr m = thinc_msgNewModule(method, name, name_len,
		priority);
	if(!m) 
		return T_ERROR;

	return thinc_msgStackAddModule(cl, stack, m);
}

int thinc_msgStackAddModule(thincClientPtr cl, thincMsgStackPtr stack, 
	thincMsgModulePtr m)
{
	return thinc_msgStackAddModuleGen(&stack->stack, m);
}

static int thinc_msgStackAddModuleGen(thincListPtr s, thincMsgModulePtr m)
{
	t_dbg_name(func)
	thincMsgModulePtr c, prev;

	thincDbg(T_DBG_CLIENT, ("%s: adding module %s to msg stack\n",
		func, m->name));
	prev = NULL;
	t_list_for_each_entry_fast(c, thincMsgModule, s, stack) {
		if(m->priority < c->priority) {
			thincDbg(T_DBG_CLIENT, ("%s: adding module before "
				"module %s\n", func, c->name));
			break;
		}
		prev = c;
	}
	if(!prev) {
		t_list_add_head(mod_list(m), s);
	} else {
		t_list_add(mod_list(m), mod_list(prev), mod_list(c));
	}
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));

	return T_OK;
}

/* find a module, takes it out from the message stack and then free its
 * memory. For now we find a module by its name. Another way would be
 * using its priority. or some other id. but then the caller would have
 * to store ids. painful. */
void thinc_msgStackDelModule(thincClientPtr cl, thincMsgStackPtr stack,
	const char *name)
{
	thincMsgModulePtr m;
	thincListPtr s = &stack->stack;
	int found_it = T_FALSE;

	t_list_for_each_entry_fast(m, thincMsgModule, s, stack) {
		if(!strncmp(name, m->name, m->name_len)) {
			found_it = T_TRUE;
			break;
		}
	}

	/* shouldn't be an error */
	if(!found_it)
		return;

	thinc_msgStackDestroyModule(cl, stack, m);
}

void thinc_msgStackDestroyModule(thincClientPtr cl, thincMsgStackPtr stack,
	thincMsgModulePtr m)
{
	thinc_msgStackRemoveModule(cl, stack, m);
	free(m);
}

/* this only takes it out of the linked list */
void thinc_msgStackRemoveModule(thincClientPtr cl, thincMsgStackPtr stack, 
	thincMsgModulePtr m)
{
	t_list_del_el(mod_list(m));
}

/* stack sees messages as opaque objects, it just passes them to the
 * modules without asking */
int thinc_msgStackExec(thincClientPtr cl, thincMsgStackPtr stack, void *obj)
{
	t_dbg_name(func)
	thincListPtr s = &stack->stack;
	thincMsgModulePtr module;
	int err;
	instr_add("STACK_EXEC", TI_START);

	thincDbg(T_DBG_CLIENT, ("%s: executing message stack\n", func));

	t_list_for_each_entry_fast(module, thincMsgModule, s, stack) {
		thincDbg(T_DBG_CLIENT, ("%s: Executing module %s\n", func,
			module->name));
		err = (*module->method)(cl, obj);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Error executing "
				"module %s\n", func, module->name));
			return T_ERROR;
		} else if(err == T_STOP) {
			thincDbg(T_DBG_CLIENT, ("%s: module %s has stopped "
				"message stack execution\n", func,
				module->name));
			break;
		}
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	instr_add("STACK_EXEC", TI_END);
	return T_OK;
}

/*************** msg stack step *************/
thincMsgStackStepPtr thinc_msgStackStepCreate()
{
	thincMsgStackStepPtr stack;

	stack = malloc(sizeof(*stack));
	if(!stack)
		return NULL;
	t_list_init(&stack->stack);
	stack->curr = NULL;
	stack->handle = NULL;

	return stack;
}

void thinc_msgStackStepDestroy(thincMsgStackStepPtr s)
{
	thincMsgModulePtr m,n;

	t_list_for_each_entry(m, n, thincMsgModule, &s->stack, stack) {
		thinc_msgStackStepDestroyModule(s, m);
	}
	t_bug_on(
		!t_list_is_empty(&s->stack),
		T_BUG("Message stack not empty after trying to destroy it")
	);
	s->curr = NULL;
	s->handle = NULL;
	free(s);
}

int thinc_msgStackStepAdd(thincMsgStackStepPtr stack,
	thincMsgMethod method, const char *name, int name_len, 
	unsigned char priority)
{
	thincMsgModulePtr m = thinc_msgNewModule(method, name, name_len,
		priority);
	if(!m) 
		return T_ERROR;

	return thinc_msgStackStepAddModule(stack, m);
}

int thinc_msgStackStepAddModule(thincMsgStackStepPtr stack, 
	thincMsgModulePtr m)
{
	if(thinc_msgStackAddModuleGen(&stack->stack, m) == T_ERROR)
		return T_ERROR;
	else {
		stack->curr = t_list_first_entry(&stack->stack, 
			thincMsgModule, stack);
		return T_OK;
	}
}

void thinc_msgStackStepDestroyModule(thincMsgStackStepPtr stack,
	thincMsgModulePtr m)
{
	thinc_msgStackStepRemoveModule(stack, m);
	free(m);
}

void thinc_msgStackStepRemoveModule(thincMsgStackStepPtr stack, 
	thincMsgModulePtr m)
{
	t_list_del_el(mod_list(m));
}

/* will return T_MSG_DONE when all modules haven executed. T_MSG_NOTDONE
 * otherwise (or T_ERROR if something bad happens) */
int thinc_msgStackStepExec(thincClientPtr cl, thincMsgStackStepPtr stack, 
	void *handle)
{
	t_dbg_name(func)
	thincListPtr s = &stack->stack;
	thincMsgModulePtr module;
	int err = T_OK;
	instr_add("STACK_EXEC", TI_START);

	thincDbg(T_DBG_CLIENT, ("%s: executing message stack\n", func));

	stack->handle = handle;
	module = stack->curr;

	t_list_for_each_entry_cont_fast(module, thincMsgModule, s, stack) {
		thincDbg(T_DBG_CLIENT, ("%s: Executing module %s\n", func,
			module->name));
		err = (*module->method)(cl, handle);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Error executing "
				"module %s\n", func, module->name));
			return T_ERROR;
		} else if(err == T_MSG_NOTDONE) {
			thincDbg(T_DBG_CLIENT, ("%s: module %s not done. "
				"Stopping for now\n", func, module->name));
			stack->curr = module;
			break;
		}
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished, \n", func));
	instr_add("STACK_EXEC", TI_END);
	return err;
}
