/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/* 
 * Header for THINC's output message stack
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop 8 */

/* 
 * This is a stack of modules that process thinc messages when they 
 * need to be sent to the client. The idea is to make it dynamic,
 * so modules can be added and removed at run time as needed. The stack
 * is implemented as a linked list of modules, and it's processed by
 * a top level function that calls each module in turn. Current
 * implementation represents modules as MsgModule objects:
 * + method [fn pointer]
 * + priority
 * The function pointer is the module's execution entry point. It is passed
 * a pointer to an object representing the message that's being processed.
 * It is expected that each message has a different stack and thus
 * different modules. The priority determines the order of the stack.
 * The idea is similar to that of sequence numbers in standard rc Unix
 * scripts, where modules are called in ascending order every time
 * a message is sent. Priorities range from 0 to 255. Since the actual sending 
 * of the message is done in a module, we define a standard priority for
 * this module (T_SEND_PRIO) and define modules with lower priority to be
 * in the PRE-send stage, and modules with larger priority to be in the
 * POST-send stage.
 * The message stack is represented as a MsgStack object which contains
 * the module list and is in charge of adding, removing, and executing the
 * stack.
 * It is expected that each client has its own MsgStack for each 
 * THINC message, and is expected that the stacks are different. For example,
 * a client that is currently resized will have a RESIZE module in the
 * appropiate stacks, which won't be present in normal-sized clients.
 * The message stack is executed in priority order. The STACK object
 * calls each module's method. If the method returns T_OK, the next
 * module is called. If T_ERROR is returned, execution stops and an error is 
 * returned. If T_STOP is returned, execution stops too, but success is
 * returned.
 */

#ifndef _THINC_MSG_STACK_H_
#define _THINC_MSG_STACK_H_

/* some useful priorities */
#define T_PRIO_MIN	0
#define T_PRIO_MAX	255
#define T_PRIO_SEND	180

typedef struct thincMsgStack_S {
	thincListRec stack;
} thincMsgStackRec, *thincMsgStackPtr;

typedef int (*thincMsgMethod) (thincClientPtr, void*);

/* 31 to align with the priority :) */
#define T_MSGMOD_NAME_LEN	31

typedef struct thincMsgModule_S {
	thincMsgMethod method;
	unsigned int name_len;
	char name[T_MSGMOD_NAME_LEN];
	unsigned char priority;
	thincListRec stack;
} thincMsgModule, *thincMsgModulePtr;

#define mod_list(ptr) (&((thincMsgModulePtr)ptr)->stack)

#define t_initModList(name) t_list_head_init((name).stack)

#define t_initModule(name, string, string_len, prio) { 	\
	thinc_##name,					\
	string_len,					\
	string,						\
	prio,						\
	t_initModList(name)				\
}

/* You shouldn't add a static module to a stack unless you know it will
 * only belong to a single stack. otherwise, expect trouble */
#define thincDefModule(name, string, string_len, prio) \
	static thincMsgModule name = t_initModule(name, string, string_len, prio)

/* use this to add a module to stack that will be a copy of module module.
 * see above for warning on using static modules */
#define thinc_msgAddUsingModule(cl, stack, module) \
	thinc_msgStackAdd(cl, stack, (module)->method, (module)->name, \
		(module)->name_len, (module)->priority)

#define thinc_msgDelUsingModule(cl, stack, module) \
	thinc_msgStackDelModule(cl, stack, (module)->name)

/* msg stack api */
thincMsgStackPtr thinc_msgStackCreate(thincClientPtr cl);
void thinc_msgStackDestroy(thincClientPtr, thincMsgStackPtr);
thincMsgModulePtr thinc_msgNewModule(thincMsgMethod method, const char *name, 
	int name_len, unsigned char priority);
int thinc_msgStackAdd(thincClientPtr cl, thincMsgStackPtr stack,
	thincMsgMethod method, const char *name, int name_len, 
	unsigned char priority);
int thinc_msgStackAddModule(thincClientPtr, thincMsgStackPtr, 
	thincMsgModulePtr);
void thinc_msgStackDelModule(thincClientPtr, thincMsgStackPtr, 
	const char *name);
void thinc_msgStackDestroyModule(thincClientPtr, thincMsgStackPtr, 
	thincMsgModulePtr);
void thinc_msgStackRemoveModule(thincClientPtr, thincMsgStackPtr, 
	thincMsgModulePtr);
int thinc_msgStackExec(thincClientPtr, thincMsgStackPtr, void *);

/* another kind of message stack: while the normal one executes all modules
 * at once, this one allows you to execute them in steps. It will execute
 * one module at a time, and will only execute the next one when the current
 * one returns T_MSG_DONE. Very useful for handshake */
typedef struct thincMsgStackStep_S {
	thincListRec stack;
	thincMsgModulePtr curr;
	void *handle;
} thincMsgStackStepRec, *thincMsgStackStepPtr;

#define T_MSG_DONE		T_OK
#define T_MSG_NOTDONE		T_OK+1

thincMsgStackStepPtr thinc_msgStackStepCreate(void);
void thinc_msgStackStepDestroy(thincMsgStackStepPtr);
void thinc_msgStackStepDestroyModule(thincMsgStackStepPtr stack,
	thincMsgModulePtr m);
void thinc_msgStackStepRemoveModule(thincMsgStackStepPtr stack, 
	thincMsgModulePtr m);
int thinc_msgStackStepAddModule(thincMsgStackStepPtr stack, 
	thincMsgModulePtr m);
int thinc_msgStackStepAdd(thincMsgStackStepPtr stack,
	thincMsgMethod method, const char *name, int name_len, 
	unsigned char priority);
int thinc_msgStackStepExec(thincClientPtr, thincMsgStackStepPtr, void *);

/* use this to add a module to stack that will be a copy of module module.
 * see above for warning on using static modules */
#define thinc_msgStepAddUsingModule(stack, module) \
	thinc_msgStackStepAdd(stack, (module)->method, (module)->name, \
		(module)->name_len, (module)->priority)

#define thinc_msgStepDelUsingModule(stack, module) \
	thinc_msgStackStepDelModule(stack, (module)->name)


#endif	/* _THINC_MSG_STACK_H_ */
