/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Network utility functions for thinc
 * 
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/poll.h>

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thincCommon.h"
#include "thincUtil.h"

#ifndef _THINC_BACKLOG
#define _THINC_BACKLOG 2 /* we don't expect to have too many connections */
#endif

static int tcpThincListenFindPort(const char *addr, int port, int *find_port);

/* sets up a listening socket. It's bound to addr:port. If addr is NULL,
 * INADDR_ANY is used.
 * returns socket or -1 on error. Sets REUSEADDR socket option on new
 * socket. backlog is set to _THINC_BACKLOG_ */
#if 0
int tcpThincListen(const char *addr, int port)
#else
int tcpThincListen(const char *addr, int port, int *find_port)
#endif
{
	t_dbg_name(func)
	int sock;
	struct sockaddr_in addr_s;
	socklen_t addrlen;
	int reuse=1;

	thincDbg(T_DBG_IO, ("%s: addr %s, port %d\n", func, addr, port));

	if(*find_port)
		return tcpThincListenFindPort(addr, port, find_port);

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if(sock == -1) {
		thincError("Couldn't open listening socket: %s\n", 
			strerror(errno));
		return T_ERROR;
	}

	reuse=1;
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse))
		== -1) {
		thincError("Error setting up listening socket: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}

	addr_s.sin_family = AF_INET;
	addr_s.sin_port = htons(port);
	if(!addr)
		addr_s.sin_addr.s_addr = htonl(INADDR_ANY);
	else {
		struct hostent *h = gethostbyname(addr);
		if(h == NULL) {
			thincError("Error: Invalid listening address: %s\n",
				strerror(errno));
			close(sock);
			return T_ERROR;
		}
		memmove(&addr_s.sin_addr.s_addr, h->h_addr, 
			sizeof(struct in_addr));
	}
	addrlen = sizeof(addr_s);

	if(bind(sock, (struct sockaddr*) &addr_s, addrlen) == -1) {
		thincError("Error: Couldn't bind listening socket: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}

	if(listen(sock, _THINC_BACKLOG) == -1) {
		thincError("Error: Couldn't set up socket for listening: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}

	thincDbg(T_DBG_IO, ("%s: finished\n", func));

	return sock;
}

/* i feel shame. who would've thought code this ugly could be written? 
 * btw, the port used is returned in find_port */
static int tcpThincListenFindPort(const char *addr, int port, int *find_port)
{
	t_dbg_name(func)
	int sock, curr, done = T_NO;
	struct sockaddr_in addr_s;
	socklen_t addrlen;
	int reuse=1;

	thincDbg(T_DBG_IO, ("%s: looking for free port starting at %d\n", func, 
		port));

	/* set up things that will not change from call to call */
	addr_s.sin_family = AF_INET;
	if(!addr)
		addr_s.sin_addr.s_addr = htonl(INADDR_ANY);
	else {
		struct hostent *h = gethostbyname(addr);
		if(h == NULL) {
			thincError("Error: Invalid listening address: %s\n",
				strerror(errno));
			return T_ERROR;
		}
		memmove(&addr_s.sin_addr.s_addr, h->h_addr, 
			sizeof(struct in_addr));
	}
	addrlen = sizeof(addr_s);
	reuse=1;

	/* the fun starts */
	curr = port;
	do {
		if( (sock = socket(PF_INET, SOCK_STREAM, 0)) == -1)
			return T_ERROR;

		if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse, 
			sizeof(reuse)) == -1) {
			close(sock);
			return T_ERROR;
		}

		addr_s.sin_port = htons(curr);

		if(bind(sock, (struct sockaddr*) &addr_s, addrlen) == -1) {
			close(sock);
			if(errno == EADDRINUSE) {
				++curr;
				continue;
			}
			return T_ERROR;
		}

		if(listen(sock, _THINC_BACKLOG) == -1) {
			close(sock);
			if(errno == EADDRINUSE) {
				++curr;
				continue;
			}
			return T_ERROR;
		}
	
		/* schueeeet, we got it */
		done = T_YES;
	} while(!done);

	*find_port = curr;

	thincDbg(T_DBG_IO, ("%s: finished. found free port %d\n", func,
		*find_port));
	return sock;
}

#if 0
/* sets up a receiving socket for udp, bound to addr:port. If addr is NULL,
 * INADDR_ANY is used. 
 * It doesn't actually do listen, but it's called like that to have a common 
 * namespace across tcp/udp.
 * Returns socket or -1 on error. Sets REUSEDADDR socket option */
int udpThincListen(const char *addr, int port)
{

}
#endif

/* simple wrapper, but the idea is to eventually virtualize all protocol
 * specific stuff (maybe through function pointers?), so it may be 
 * possible that we may need to have to do special things when closing a 
 * socket in the future */
int tcpThincClose(int sock)
{
	return close(sock);
}

int thincDoSelect(const fd_set *fd_in, fd_set *fd_out, int maxfd)
{
	int res;
	struct timeval tv;

	tv.tv_sec = tv.tv_usec = 0;
again:
	memcpy(fd_out, fd_in, sizeof(fd_set));
	res = select(maxfd, fd_out, NULL, NULL, &tv);
	if(res == -1) {
		if(errno == EINTR)
			goto again;
		else {
			thincError("Error: select() failed: %s\n", 
				strerror(errno));
			return T_ERROR;
		}
	} else
		return res;
}

/* poll() the socket waiting for "events" to happen or timeout to
 * expire (< 0 -> infinite timeout). Returns -1 on error. Returns 0 on 
 * success and if revents is not NULL, it's filled with the revents 
 * returned from poll. */
int thincDoPollRevs(int sock, short events, short *revents, int timeout)
{
	struct pollfd fds[1];
	int res;
	t_dbg_name(func)

	fds[0].fd = sock;
	fds[0].events = events;
	fds[0].revents = 0;

again:
	res = poll(fds, 1, timeout);
	if(res == -1) {
		if(errno == EINTR)	/* try again */
			goto again;
		thincDbg(T_DBG_ERROR, ("%s: Error polling socket %d: %s\n", 
			func, sock, strerror(errno)));
		return T_ERROR;
	} else if(fds[0].revents & (POLLERR | POLLHUP | POLLNVAL)) {
		thincDbg(T_DBG_ERROR, ("%s: Error condition %d on socket\n", 
			func, fds[0].revents));
		if(revents)
			*revents = fds[0].revents;
		return T_ERROR;
	} else if(res == 0) { /* timeout */
		thincDbg(T_DBG_IO, ("%s: timeout waiting on socket %d\n", 
			func, sock));
		return T_TIMEOUT;
	}

	if(revents)
		*revents = fds[0].revents;

	return T_OK;
}
