/* Copyright 2006, 2007 Ricardo Baratto */
/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
 * Author:
 * Ricardo Baratto <ricardo@cs.columbia.edu>
 * Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#ifdef T_RECORD

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/param.h>
#include <arpa/inet.h>

/* for filename generation */
#include <openssl/rand.h>
#include <openssl/sha.h>

#define T_REC_MAX_FN_LEN	MAXPATHLEN
#define T_REC_NEWDIR_MODE	00770
/* functions that use the space provided by record->dir should call this
 * macro before they return */
#define t_cleanup_dir(record) ((record)->dir[(record)->dir_len] = '\0')

#define T_REC_NEWFILE_MAXTRIES	50
#if 1
#define T_REC_OPEN_FLAGS		\
	O_RDWR|O_LARGEFILE|O_NOFOLLOW
#else
#define T_REC_OPEN_FLAGS		\
	O_RDWR|O_NOFOLLOW
#endif
#define T_REC_NEWFILE_FLAGS		\
	O_CREAT|O_EXCL|T_REC_OPEN_FLAGS
#define T_REC_NEWFILE_MODE		00660

/* local stuff */
static int thincRecordInitDir(thincScreenPtr pThinc, const char *base_dir);
static int thincRecordInitFiles(thincScreenPtr pThinc, const char *base_dir);
static int thincRecordInitBuffer(thincScreenPtr pThinc);
static int thincRecordInitClient(thincScreenPtr pThinc);
static int thincRecordInitShot(thincScreenPtr pThinc);

static void thincRecordClientClose(thincClientPtr cl);
static int thincRecordClBuffer(thincClientPtr cl, thincCmdPtr cmd);
static int thincRecordClFlushCmds(thincScreenPtr pThinc, thincClientPtr cl);
static int thincRecordClFlushSched(thincScreenPtr pThinc, thincClientPtr cl);
static int thincRecordClInitMsgStacks(thincRecordPtr, thincClientPtr cl);
static void thincRecordClDestroyMsgStacks(thincClientPtr cl);

static char* thincRecordNewFileID(thincRecordPtr record);
static int thincRecordNewFile(thincRecordPtr record, const char *ext, char *id);
static int thincRecordNewCmdFile(thincRecordPtr record, char *id);
static int thincRecordNewShotFile(thincRecordPtr record, char *id);
static int thincRecordInitIndex(thincScreenPtr pThinc);
static int thincRecordInitExtraFiles(thincScreenPtr pThinc);
static int thincRecordWriteMagic(int fd);
static int thincRecordGetNewCmdFile(thincRecordPtr record);
#if 0
static int thincRecordChkMagic(int fd);
#endif

static void thincRecordDestroyShot(thincScreenPtr pThinc);
static void thincRecordClientRemove(thincScreenPtr pThinc);

static int thincSendMsgRecord(thincClientPtr cl, int type, int msg_len, 
	int flags);
static int thincSendMsgRecordCb(thincClientPtr cl, char *buf, 
	unsigned int *len, void *ptr);

/* on_error are statements you want to be executed on error. if you have
 * nothing to do, pass t_empty_on_error */
#define t_rec_chmod(fn, mode, on_error) do { 			\
	if(chmod(fn, mode) == -1) {					\
		thincDbg(T_DBG_RECORD, ("%s: couldn't change mode "	\
			"of file %s: %s\n", func, fn, strerror(errno)));\
		on_error;						\
		return T_ERROR;						\
	}								\
} while(0)

#define t_rec_chown(fn, on_error) do {				\
	if(chown(fn, getuid(), -1) == -1) {				\
		thincDbg(T_DBG_RECORD, ("%s: couldn't change ownership "\
			"of file %s: %s\n", func, fn, strerror(errno)));\
		on_error;						\
		return T_ERROR;						\
	}								\
} while(0)

#define t_rec_write(fd, buf, size, on_error) do {			\
	if(write(fd, buf, size) == -1) {				\
		thincDbg(T_DBG_RECORD, ("%s: couldn't write %u bytes "	\
			"to file: %s\n", func, size, strerror(errno)));\
		on_error;						\
		return T_ERROR;						\
	}								\
} while(0)

#define t_empty_on_error

/* chks if we can buffer sz bytes into the buffer */
#define t_rec_canBuf(record, sz) 		\
	( ((record)->buf_size + (sz) ) <= T_REC_BUFSIZE )

#define t_rec_maxBuf(record)			\
	( T_REC_BUFSIZE - (record)->buf_size )

#define t_rec_bufAdd(record, sz) do {			\
	record->c_buf += sz;				\
	record->buf_size += sz;				\
} while(0)

#define t_rec_bufReset(record) do {			\
	record->c_buf = record->buf;			\
	record->buf_size = 0;				\
} while(0)

#define T_REC_BUF_MINSZ		256

/***************************** init/shutdown **********************************/
/* commands will be recorded in base_dir/username */
int thincRecordInit(thincScreenPtr pThinc, const char *base_dir, int flags)
{
	t_dbg_name(func)
	thincRecordPtr record;

	thincDbg(T_DBG_RECORD, ("%s: owner is %s (%d)\n", func, pThinc->owner, pThinc->owner_len));
	thincDbg(T_DBG_RECORD, ("%s: initializing command recorder\n", func));

	record = calloc(1, sizeof(*record));
	if(!record) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't allocate memory "
			"for command recorder\n", func));
		return T_ERROR;
	}

	record->pThinc = pThinc;
	pThinc->record = record;


	if(flags & T_REC_ALWAYS)
		record->always = 1;
	else
		record->always = 0;

	gettime_ms_llu(record->time);

	/* initialize some things for thincRecordDestroy */
	record->cmd = -1;
	record->shot = -1;

	thincDbg(T_DBG_RECORD, ("%s: owner is %s (%d)\n", func, 
		pThinc->owner, pThinc->owner_len));
	if(thincRecordInitFiles(pThinc, base_dir) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"recorder files\n", func));
		thincRecordDestroy(pThinc);
		return T_ERROR;
	}

	if(thincRecordInitBuffer(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"recorder buffer\n", func));
		thincRecordDestroy(pThinc);
		return T_ERROR;
	}

	/* init our virtual client */
	if(thincRecordInitClient(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"recorder client\n", func));
		thincRecordDestroy(pThinc);
		return T_ERROR;
	}

	/* set up first command info before taking 1st screenshot */
	t_rec_fill_cmd(record->c_cmd, record->cmd_fn, record->cmd_fp);

	/* init screenshot stuff, and take first one in history of
	 * humanity */
	if(thincRecordInitShot(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"screen shot thang\n", func));
		thincRecordDestroy(pThinc);
		return T_ERROR;
	}

#if 0
	thincDbg(T_DBG_RECORD, ("%s: initializing state recorded\n", func));
	if(thincRecordInitState(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"state record thingie\n", func));
		thincRecordDestroy(pThinc);
		return T_ERROR;
	}
#endif

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

/* since we may get called on an aborted init gotta be extra careful */
void thincRecordDestroy(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);

	if (!record)
		return;
	
	thincDbg(T_DBG_RECORD, ("%s: destroying command recorder\n", func));

	if (record->buf && record->cmd != -1 && record->shot != -1) {
		thincRecordFlush(pThinc);
	}

	/* take one final screenshot just for fun */
	if(record->shot_stack && record->cl) {
		thincRecordTakeShot(pThinc, T_TRUE);
#if 0
		REGION_UNINIT(pThinc->pScreen, &record->display_change);
#else
		thincSStatsFree(pThinc, record->stats_id);
#endif
		/* this one "needs" the client */
		thincRecordDestroyShot(pThinc);
	}

	free(record->buf);
	record->buf = NULL;

	if(record->cl)
		thincRecordClientClose(record->cl);

#if 0
	thincRecordDestroyState(pThinc);
#endif

	/* TODO: Should save state of affairs before closing */
	close(record->index);
	close(record->cmd);
	close(record->shot);

	free(record->dir);
	free(record);
	pThinc->record = NULL;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
}

/*
 * directory and file structure used by the recorder:
 * dir:
 *   use base directory, append user name to it, create, make 00770 and
 *   chown to user. we need 00770 in case the dir is in NFS and we get
 *   root-squashed. it is recommended that base_dir be owned by root but
 *   accessible by the owner (so he/she can replay the commands :)
 * files:
 *   index file: record.idx
 *   screenshot files: ID.shot
 *   command files: ID.cmd
 *
 * ID may be anything.
 *
 * all files have magic number that identifies them as being a THINC record
 * files (T_REC_MAGIC).
 *
 * files may exist when we start, but if record.idx doesn't exist then
 * any other files will be ignored (and destroyed if needed). also
 * record.idx should have current state of affairs. if we are unable to
 * recover enough information from it about the other files, they will
 * be ignored (and destroyed if needed).
 * TODO: define what "state" means
 */
static int thincRecordInitFiles(thincScreenPtr pThinc, const char *base_dir)
{
	t_dbg_name(func)

	thincDbg(T_DBG_RECORD, ("%s: initializing recorder files\n", func));

	if(thincRecordInitDir(pThinc, base_dir) == T_ERROR)  {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"recorder directory\n", func));
		return T_ERROR;
	}

	if(thincRecordInitIndex(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"index file\n", func));
		return T_ERROR;
	}

	if(thincRecordInitExtraFiles(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"command and screenshot files\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

static int thincRecordInitDir(thincScreenPtr pThinc, const char *base_dir)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);
	char *c;
	short need_slash = T_FALSE, exists = T_FALSE;
	unsigned int dir_len, base_len;

	thincDbg(T_DBG_RECORD, ("%s: initializing recorder directory\n",
		func));

	/* FIXME: Should we do some sanity checks on the directory? */
	if(base_dir) {
		base_len = strlen(base_dir);
		if(base_dir[base_len] != '/') {
			thincDbg(T_DBG_RECORD, ("%s: user dir %s doesn't "
				"terminate in /\n", func, base_dir));
			need_slash = T_TRUE;
		}
	} else {
		base_dir = T_REC_DEFAULT_DIR;
		base_len = T_REC_DEFAULT_DIR_LEN;
	}

	/* +1 -> final /
	 */
	dir_len = base_len + pThinc->owner_len + 1;
	dir_len += (need_slash ? 1 : 0);

	thincDbg(T_DBG_RECORD, ("%s: allocating %u bytes for directory\n",
		func, dir_len+1));
	
	/* allocate space after the dir name that can be used for other
	 * functions to build filenames */
	record->dir = calloc(T_REC_MAX_FN_LEN, sizeof(char));
	if(!record->dir) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't allocate memory "
			"for command recorder\n", func));
		free(record);
		return T_ERROR;
	}
	record->dir_len = dir_len;

	c = record->dir;
	memcpy(c, base_dir, base_len);
	c += base_len;
	if(need_slash)
		*c++ = '/';

	memcpy(c, pThinc->owner, pThinc->owner_len);
	record->dir[dir_len-1] = '/';

	thincDbg(T_DBG_RECORD, ("%s: creating directory '%s' (%d) to "
		"record commands\n", func, record->dir, record->dir_len));

	if(mkdir(record->dir, T_REC_NEWDIR_MODE) == -1) {
		/* this should be ok */
		if(errno != EEXIST) {
			thincError("ERROR: Couldn't create record directory "
				"%s: %s\n", record->dir, strerror(errno));
			thincError("ERROR: Maybe you forgot to "
				"create the base record directory %s?\n",
				base_dir);
			free(record->dir);
			record->dir = NULL;
			return T_ERROR;
		} else
			exists = T_TRUE;
	}
	/* only do this if creating from scratch */
	if(exists)
		goto out;
	
	t_rec_chown(record->dir, (free(record->dir), record->dir = NULL));
	/* need to make sure this is set (in case umask gets in the way) */
	t_rec_chmod(record->dir, T_REC_NEWDIR_MODE, 
		(free(record->dir), record->dir = NULL));
out:

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

/* initialize new index file.
 * TODO: Support reusing existing index file */
static int thincRecordInitIndex(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	int idx;
	thincRecordPtr record = t_recordPtr(pThinc);
	char *fn;
	struct thincRecIndexHeader h;

	thincDbg(T_DBG_RECORD, ("%s: initializing index file\n", func));

	fn = record->dir + record->dir_len;
	memcpy(fn, T_REC_IDX_NAME, strlen(T_REC_IDX_NAME));
	fn = record->dir;

	thincDbg(T_DBG_RECORD, ("%s: opening %s\n", func, fn));

	/* FIXME: blindly overwriting existing index file */
	idx = open(fn, T_REC_OPEN_FLAGS|O_CREAT, T_REC_NEWFILE_MODE);
	if(idx == -1) {
		thincDbg(T_DBG_ERROR, ("%s: couldn't create index file: %s\n",
			func, strerror(errno)));
		t_cleanup_dir(record);
		return T_ERROR;
	}

	/* XXX: should we do this even when file exists? */
	t_rec_chown(fn, (t_cleanup_dir(record), close(idx)));
	t_rec_chmod(fn, T_REC_NEWFILE_MODE, 
		(t_cleanup_dir(record), close(idx)) );

	/* done with this thang */
	t_cleanup_dir(record);

	thincDbg(T_DBG_RECORD, ("%s: writing magic and header to index\n",
		func));
	if(thincRecordWriteMagic(idx) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't write magic "
			"number to index file\n", func));
		close(idx);
		return T_ERROR;
	}

	if(pThinc->useRawCache)
		h.raw_cache_bits = pThinc->raw_cache_bits;
	else
		h.raw_cache_bits = 0;
	h.dummy = 0;
	t_rec_write(idx, &h, T_REC_INDEX_HEADER_SZ, close(idx));

	/* no mol */
	record->index = idx;
	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

/* open a new command and screenshot file, ready to be filled with tasty
 * records */
static int thincRecordInitExtraFiles(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);
	int cmd, shot;

	thincDbg(T_DBG_RECORD, ("%s: initializing new record and screenshot "
		"files\n", func));

	shot = thincRecordNewShotFile(record, record->shot_fn);
	if(shot == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize new "
			"screenshot file\n", func));
		return T_ERROR;
	}
	record->shot = shot;
	record->shot_fp = 0;

	cmd = thincRecordNewCmdFile(record, record->cmd_fn);
	if(cmd == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize new "
			"command file\n", func));
		return T_ERROR;
	}
	record->cmd = cmd;
	record->cmd_fp = 0;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

static int thincRecordInitBuffer(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);

	thincDbg(T_DBG_RECORD, ("%s: initializing buffer with size %u\n",
		func, T_REC_BUFSIZE));

	record->buf = malloc(T_REC_BUFSIZE);
	if(!record->buf) {
		thincDbg(T_DBG_ERROR, ("%s: no memory for buffer\n", func));
		return T_ERROR;
	}

	record->buf_size = 0;
	record->c_buf = record->buf;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

/*************************** client functions *******************************/
/* this client is only active when there are no other clients, and 
 * record->always is set */
static int thincRecordInitClient(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincClientPtr cl;
	thincRecordPtr record = t_recordPtr(pThinc);
	BoxRec clip;

	thincDbg(T_DBG_RECORD, ("%s: initializing recorder client\n", func));

	cl = calloc(1, sizeof(thincClientRec));
	if(!cl) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"recorder client\n", func));
		return T_ERROR;
	}

	cl->sock = -1;
	strncpy(cl->id, "RECORDER", 8);
	cl->id[T_CLID_LEN-1] = '\0';

	cl->buffer = thincRecordClBuffer;
	cl->close_cl = thincRecordClientClose;

	/* can't use std ones cuz they'd poll our invalid socket and probably
	 * die */
	if(pThinc->doSched) {
		if(thincSchedInit(cl) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"initialize client's command scheduler. "
				"New client rejected\n", func));
			cl->close_cl(cl);
			return T_ERROR;
		}
		cl->last = 0;
		cl->flush = thincRecordClFlushSched;
	} else
		cl->flush = thincRecordClFlushCmds;

	cl->send_msg = thincSendMsgRecord;
	cl->read = thincBadRead;
	cl->write = thincBadWrite;
	t_list_init(&cl->cmdBuffer);
	cl->buffer_batch = 0;

	clip.x1 = 0;
	clip.y1 = 0;
	clip.x2 = clip.x1 + pThinc->width;
	clip.y2 = clip.y1 + pThinc->height;
	REGION_INIT(pThinc->pScreen, &cl->clClip, &clip, 1);
	cl->clipIsSub = T_FALSE;
	cl->resized = T_CL_RESZ_NONE;
	cl->parent = pThinc;

	/* init cache */
	if(pThinc->useRawCache) {
		cl->raw_cache = thinc_cache_init(pThinc->raw_cache_bits);
		if(!cl->raw_cache) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"initialize client's image cache\n", func));
			cl->close_cl(cl);
			return T_ERROR;
		}
	}

	if(thincRecordClInitMsgStacks(record, cl) == T_ERROR) {
		cl->close_cl(cl);
		return T_ERROR;
	}
	if(thincVInitRecordClient(cl) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize video "
			"for recording client\n", func));
		cl->close_cl(cl);
		return T_ERROR;
	}

	gettime_ms(cl->last_flush);

	cl->connected = T_TRUE;
	record->cl = cl;

	if(record->always)
		thincRecordClientEnable(pThinc);
	else
		record->cl_enabled = T_FALSE;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

/* NOTE: shouldn't be called by CloseClient(), but by RecordDestroy */
/* FIXME: If this gets called as a result of an error, recording will be
 * all screwed up. Should try to create a new record client or just kill
 * the server... */
static void thincRecordClientClose(thincClientPtr cl)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = cl->parent;
	thincRecordPtr record = t_recordPtr(pThinc);

	thincDbg(T_DBG_RECORD, ("%s: closing recording client\n", func));

	thincRecordClientRemove(pThinc);

	REGION_UNINIT(pThinc->pScreen, &cl->clClip);

	thincRecordClDestroyMsgStacks(cl);

	cl->parent = NULL;
	record->cl = NULL;
	free(cl);

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
}

#if 0
/* call when got a new client, _after_ it has become connected and clientCount
 * has been increased */
void thincRecordClientDisable(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);
	thincClientPtr cl = record->cl;

	thincDbg(T_DBG_RECORD, ("%s: disabling recording client\n", func));

	if(!record->always)
		/* client shouldn't be active, so nothing to do */
		return;

	/* assumes clientCount has been increased beforehand */
	if(pThinc->clientCount > 2)
		/* it should be disabled by now */
		return;

	/* remove from client list */
	t_list_del_el(&cl->link);
	pThinc->clientCount--;
	record->cl_enabled = T_FALSE;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
}

/* call when killing a real client, _after_ you decrease clientCount */
void thincRecordClientEnable(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record;
	thincClientPtr cl;

	thincDbg(T_DBG_RECORD, ("%s: enabling recording client\n", func));

	if(pThinc->clientCount)
		/* some clients still alive */
		return;

	record = t_recordPtr(pThinc); 
	/* for when we get called from CloseScreen() */
	if(!record)
		return;
	if(!record->always)
		return;

	cl = record->cl;
	if(!cl)
		return;

	t_list_add_head(&cl->link, &pThinc->clients);
	pThinc->clientCount = 1;
	record->cl_enabled = T_TRUE;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
}
#else
/* call when got a new client, _after_ it has become connected and clientCount
 * has been increased */
int thincRecordClientEnable(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);
	thincClientPtr cl;

	thincDbg(T_DBG_RECORD, ("%s: enabling recording client\n", func));

	/* just in case */
	t_bug_on_2(
		(!record || !record->cl),
		T_BUG("cannot enable recording without recording support"),
		return T_ERROR;
	);

	if(record->cl_enabled) {
		thincDbg(T_DBG_RECORD, ("%s: client already enabled\n", func));
		return T_OK;
	}

	cl = record->cl;

	t_list_add_tail(&cl->link, &pThinc->clients);
	pThinc->clientCount++;
	record->cl_enabled = T_TRUE;

	thincDbg(T_DBG_RECORD, ("%s: finished, have %d clients\n", func,
		pThinc->clientCount));
	return T_OK;
}

/* call when killing a real client, _after_ you decrease clientCount. It will
 * only disable if there are no real clients _and_ always recording has
 * been disabled (record->always is false) */
void thincRecordClientDisable(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);
	thincClientPtr cl;

	thincDbg(T_DBG_RECORD, ("%s: disabling recording client\n", func));

	/* need this as CloseClient() calls this function and in the
	 * shutdown sequence recording is destroyed before the clients
	 * are closed */
	if(!record || !record->cl) {
		thincDbg(T_DBG_RECORD, ("%s: but recording is no good!\n",
			func));
		return;
	}

	if(record->always) {
		/* client should stay active, so nothing to do */
		thincDbg(T_DBG_RECORD, ("%s: always recording, can't disable\n",
			func));
		return;
	}

	if(!record->cl_enabled) {
		thincDbg(T_DBG_RECORD, ("%s: client already disabled\n", func));
		return;
	}

	/* assumes clientCount has been decreased beforehand */
	if(pThinc->clientCount > 1) {
		/* still some people connected */
		thincDbg(T_DBG_RECORD, ("%s: but some people still "
			"connected!\n", func));
		return;
	}

	cl = record->cl;

	/* remove from client list */
	t_list_del_el(&cl->link);
	pThinc->clientCount--;
	record->cl_enabled = T_FALSE;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
}

/* if recording client is in clients list it will take it out of there.
 * disable() is more picky about when the client can be taken out */
static void thincRecordClientRemove(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);
	thincClientPtr cl;

	thincDbg(T_DBG_RECORD, ("%s: removing recording client\n", func));

	if(!record || !record->cl) {
		return;
	}

	if(!record->cl_enabled) {
		thincDbg(T_DBG_RECORD, ("%s: client not enabled, finished\n", 
			func));
		return;
	}

	cl = record->cl;

	/* remove from client list */
	t_list_del_el(&cl->link);
	pThinc->clientCount--;
	record->cl_enabled = T_FALSE;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
}
#endif

/* we use this to discard commands we don't want to record */
static int thincRecordClBuffer(thincClientPtr cl, thincCmdPtr cmd)
{
	/* dont care about cursors */
	if(cmd->type == T_FB_CURSOR)
		return T_OK;

	return thincBufferCmdClientReal(cl, cmd);
}

static int thincRecordClFlushCmds(thincScreenPtr pThinc, thincClientPtr cl)
{
	if(t_list_is_empty(&cl->cmdBuffer)) 
		return T_OK;

	return thincFlushListBuffer(pThinc, cl, &cl->cmdBuffer);
}

static int thincRecordClFlushSched(thincScreenPtr pThinc, thincClientPtr cl)
{
	return thincFlushSchedulerNoPoll(pThinc, cl);
}

static int thincRecordClInitMsgStacks(thincRecordPtr record, thincClientPtr cl)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_RECORD, ("%s: initializing message stacks\n", func));

	cl->rawStack = NULL;
	cl->bitmapStack = NULL;
	cl->pixmapStack = NULL;
	cl->fillStack = NULL;
	cl->copyStack = NULL;

	err = thinc_rawInitRecordStack(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize RAW "
			"message stack\n", func));
		return T_ERROR;
	}

	err = thinc_pixmapInitRecordStack(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize PIXMAP "
			"message stack\n", func));
		return T_ERROR;
	}

	err = thinc_bitmapInitRecordStack(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize PIXMAP "
			"message stack\n", func));
		return T_ERROR;
	}

	err = thinc_fillInitRecordStack(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize SFILL "
			"message stack\n", func));
		return T_ERROR;
	}

	err = thinc_copyInitRecordStack(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize COPY "
			"message stack\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

static void thincRecordClDestroyMsgStacks(thincClientPtr cl)
{
	if(cl->rawStack)
		thinc_rawDestroyRecordStack(cl);
	if(cl->bitmapStack)
		thinc_bitmapDestroyRecordStack(cl);
	if(cl->pixmapStack)
		thinc_pixmapDestroyRecordStack(cl);
	if(cl->fillStack)
		thinc_fillDestroyRecordStack(cl);
	if(cl->copyStack)
		thinc_copyDestroyRecordStack(cl);
}

static int thincSendMsgRecord(thincClientPtr cl, int type, int msg_len, 
	int flags)
{
	return thincRecord(cl, msg_len, type, flags, thincSendMsgRecordCb, 
		(void*)msg_len);
}

static int thincSendMsgRecordCb(thincClientPtr cl, char *buf, 
	unsigned int *len, void *ptr)
{
	int msg_len = (int)ptr;

	if(msg_len > *len) {
		T_BUG_ARG_2("Can't record message of length %d, have only %d "
			"bytes available", msg_len, *len);
		return T_ERROR;
	}

	memcpy(buf, t_msgBuffer, msg_len);
	*len = msg_len;

	return T_REC_DONE;
}

/**************************** screenshot functions ***************************/
static int thincRecordInitShot(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	int err;
	thincRecordPtr record = t_recordPtr(pThinc);

	thincDbg(T_DBG_RECORD, ("%s: initializing screenshot thang\n", func));
	
	err = thinc_shotInitStack(t_recordPtr(pThinc));
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"screenshot message stack\n", func));
		return T_ERROR;
	}

	/* init timer so we can keep taking shots */
	record->shot_timer = thincInitATimer(pThinc);
	if(!record->shot_timer) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"screenshot timer\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_RECORD, ("%s: taking first screenshot\n", func));

	err = thincRecordTakeShot(pThinc, T_TRUE);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't take first "
			"screenshot\n", func));
		return T_ERROR;
	}

	/* initialize display change tracking */
#if 0
	record->display_count = 0;
	REGION_INIT(pThinc->pScreen, &record->display_change, NullBox, 1);
#else
	record->stats_id = thincSStatsNew(pThinc);
	if(record->stats_id == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't initialize "
			"statistics\n", func));
		return T_ERROR;
	}
#endif

	/* and set the timer */
	record->shot_timer = thincSetATimer(pThinc, record->shot_timer, 
		TimerForceOld, T_REC_SHOT_TIMER, thincRecordTimerShot, 
		(pointer)record);
	if(!record->shot_timer) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't set screenshot "
			"timer\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

static void thincRecordDestroyShot(thincScreenPtr pThinc)
{
	thinc_shotDestroyStack(t_recordPtr(pThinc));
}

/***************************** record functions *******************************/

/* record a command. the way this works is as follows:
 * - call with how much data you need to record.
 * - record() will call your callback function with a memory region to where
 *   you can put your data. it will also pass the maximum amount of data that
 *   you can write there along with an opaque handle.
 * - if there's not enough space for you to write on, return T_REC_MORE
 *   and the function will be called again once more space is made available.
 * - when all data has been recorded, the callback should return T_REC_DONE.
 * - the callback should also fill in how much space it used.
 * - one command record will be generated for each call to record(), _not_
 *   to the callback.
 * - other arguments are like thincBufferMsg()
 */
int thincRecord(thincClientPtr cl, unsigned int len, int type, int flags,
	int (*rec_cb)(thincClientPtr, char *, unsigned int*, void *),
	void *handle)
{
	t_dbg_name(func)
	int err;
	thincScreenPtr pThinc = cl->parent;
	thincRecordPtr record = t_recordPtr(pThinc);
	unsigned int total, space, curr_fp;
	struct thincRecCmdHeader *ch;

	thincDbg(T_DBG_RECORD, ("%s: recording %u bytes, fd is %d\n", func, 
		len, record->cmd));

	gettime_ms_llu(record->time);

	/* chk the space situation */
	total = T_REC_CMD_HEADER_SZ + len;
	space = t_rec_maxBuf(record);
	thincDbg(T_DBG_RECORD, ("%s: time %llu, need %u bytes, have %u\n", 
		func, record->time, total, space));
	/* 
	 * > if have enough space, just write header, then go in a loop 
	 * calling rec_cb until it returns T_REC_DONE
	 * > if not enough space, flush buffer and go into the loop.
	 */
	if( (total > space) ) {
		/* so much fun with the boss... */
		thincDbg(T_DBG_RECORD, ("%s: don't have enough space, only "
			"%u bytes. flushing the bluffer\n", func, space));
		err = thincRecordFlush(pThinc);
		if(err == T_ERROR) {
			/* sangre de cristo! */
			thincDbg(T_DBG_ERROR, ("ERROR: %s: oh boy, couldn't "
				"flush the buffer. yer dead\n", func));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_RECORD, ("%s: adding command header type %#x, flags "
		"%#x\n", func, type, flags));

	ch = (struct thincRecCmdHeader*)record->c_buf;
	ch->time = htonll(record->time);
	ch->size = htonl(len);
	ch->type = type;
	ch->flags = flags;

	t_rec_bufAdd(record, T_REC_CMD_HEADER_SZ);
	space = t_rec_maxBuf(record);

	while(1) {
		err = (*rec_cb)(cl, record->c_buf, &space, handle);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: record callback "
				"err-ed\n", func));
			return T_ERROR;
		} 
		/* ok, dude wrote some data. first, update our pointers
		 * and other things. then if it still needs to write more,
		 * chk if there's enough space left in the buffer, flush
		 * it if needed, and round we go again.
		 * if it's done, then we're done, ja?
		 */
		t_rec_bufAdd(record, space);

		if(err == T_REC_DONE) {
			thincDbg(T_DBG_RECORD, ("%s: done writing, out "
				"we go. have %u bytes left in the "
				"buffer\n", func, t_rec_maxBuf(record)));
			break;
		}
		/* T_REC_MORE */
		thincDbg(T_DBG_RECORD, ("%s: callback wrote %u "
			"bytes... and it wants more!\n", func, space));
		total -= space;
		space = t_rec_maxBuf(record);

		if(total > space) {
			err = thincRecordFlush(pThinc);
			if(err == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"flush record buffer\n", func));
				return T_ERROR;
			}
			/* to simplify things we don't police the caller's
			 * length claims. hopefully it won't depend on us
			 * to keep track of how much data it has yet to 
			 * write */
			space = t_rec_maxBuf(record);
		}
	}

	/* chk if we need a new file. we need to add buf_size to account
	 * for what hasn't been flushed to disk. also we use this to update
	 * the current command information (for when we take the next
	 * screenshot) */
	curr_fp = record->cmd_fp + record->buf_size;
	if(curr_fp > T_REC_CMDFILE_MAXSZ) {
		thincDbg(T_DBG_RECORD, ("%s: need a new command file. this "
			"one's too big\n", func));
		/* this flushes the bluffer */
		err = thincRecordGetNewCmdFile(record);
		if(err == T_ERROR) {
			/* poor dude, thought everything was cool and sheit
			 * and look where it got screwed */
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get "
				"a new command file\n", func));
			return T_ERROR;
		}
		/* since got new file, have to copy everything */
		t_rec_fill_cmd(record->c_cmd, record->cmd_fn, record->cmd_fp);
	} else {
		/* just update the fp */
		thincDbg(T_DBG_RECORD, ("%s: current cmd fp %d\n", func,
			curr_fp));
		record->c_cmd.fp = curr_fp;
	}

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

int thincRecordFlush(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	thincRecordPtr record = t_recordPtr(pThinc);
	
	if(!record)
		return T_OK;

	thincDbg(T_DBG_RECORD, ("%s: flushing buffer with %u bytes to fd %d\n",
		func, record->buf_size, record->cmd));
	
	t_rec_write(record->cmd, record->buf, record->buf_size, 
		t_empty_on_error);

	record->cmd_fp += record->buf_size;

	t_rec_bufReset(record);

	thincDbg(T_DBG_RECORD, ("%s: finished, fp is now %u\n", func,
		record->cmd_fp));
	return T_OK;
}

static int thincRecordGetNewCmdFile(thincRecordPtr record)
{
	t_dbg_name(func)
	struct thincRecCmdHeader *ch;
	int cmd, err;
	char data[T_REC_CMD_NEXTFILE_SZ];

	thincDbg(T_DBG_RECORD, ("%s: current command file is full, getting "
		"new one\n", func));

	/* flush the bluffer first */
	err = thincRecordFlush(record->pThinc);
	if(err == T_ERROR)
		return T_ERROR;

	/* get new file */
	cmd = thincRecordNewCmdFile(record, record->cmd_fn);
	if(cmd == T_ERROR)
		return T_ERROR;

	thincDbg(T_DBG_RECORD, ("%s: got new fd %d\n", func, cmd));

	/* write the final header */
	ch = (struct thincRecCmdHeader*) data;
	ch->time = htonll(record->time);
	ch->size = htonl(T_REC_FILEID_LEN);
	ch->type = T_REC_CMD_TYPE_NEXTFILE;
	ch->flags = 0;

	memcpy(data + T_REC_CMD_HEADER_SZ, record->cmd_fn, 
		T_REC_FILEID_LEN);

	t_rec_write(record->cmd, data, T_REC_CMD_NEXTFILE_SZ, t_empty_on_error);

	close(record->cmd);
	record->cmd = cmd;
	record->cmd_fp = 0;

	thincDbg(T_DBG_RECORD, ("%s: finished, fd is now %d. it should be %d\n",
		func, record->cmd, cmd));
	return T_OK;
}

int thincRecordGetNewShotFile(thincRecordPtr record)
{
	t_dbg_name(func)
	struct thincRecShotHeader *sh;
	int shot;
	char data[T_REC_SHOT_NEXTFILE_SZ];

	thincDbg(T_DBG_RECORD, ("%s: current screenshot file is full, getting "
		"new one\n", func));

	/* get new file */
	shot = thincRecordNewShotFile(record, record->shot_fn);
	if(shot == T_ERROR)
		return T_ERROR;

	/* write the final header */
	sh = (struct thincRecShotHeader*)data;
	sh->time = htonll(record->time);
	sh->size = htonl(T_REC_FILEID_LEN);
	sh->width = 0;
	sh->height = 0;
	sh->info = htons(T_REC_SHOT_TYPE_NEXTFILE);

	memcpy(data+T_REC_SHOT_HEADER_SZ, record->shot_fn, T_REC_FILEID_LEN);

	t_rec_write(record->shot, data, T_REC_SHOT_NEXTFILE_SZ, 
		t_empty_on_error);

	close(record->shot);
	record->shot = shot;
	record->shot_fp = 0;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

/***************************** util functions *********************************/

#define T_REC_RANDOM_LEN	128
#define T_REC_HASH_LEN	20	/* *2 should always be >= FILEID_LEN */

/* returns the id (which you can destroy with free()). to find the length
 * of the id just strlen() it */
static char* thincRecordNewFileID(thincRecordPtr record)
{
	unsigned char random[T_REC_RANDOM_LEN];
	unsigned char hash[T_REC_HASH_LEN];
	char *file_id;
	int i;
	const int id_len = T_REC_FILEID_LEN + 1; /* for \0 */
	
	file_id = calloc(id_len, sizeof(char));
	if(!file_id)
		return NULL;

	/* get some random data */
	if(!RAND_pseudo_bytes(random, T_REC_RANDOM_LEN)) {
		free(file_id);
		return NULL;
	}
	/* hash the random data */
	SHA1(random, T_REC_RANDOM_LEN, hash);

	/* we only use half of the SHA1 for the ID. 40 chars is kinda too
	 * much */
	for(i=0; i<T_REC_FILEID_LEN>>1; ++i) {
		sprintf(&file_id[i*2], "%02x", hash[i]);
	}

	return file_id;
}

/* create and open new file (screenshot or command). use O_CREAT|O_EXCL to make
 * sure no existing files get overwritten. file will be called ID.shot where ID
 * is generated by thincRecordNewFileID(). returns file descriptor of new file
 * or T_ERROR. New file should be mode 0660. The base name of the file
 * is also copied into id (no extension, no dir tho). It should have space
 * for T_REC_FILEID_LEN bytes */
static int thincRecordNewFile(thincRecordPtr record, const char *ext, 
	char *out_id)
{
	t_dbg_name(func)
	char *id, *c, *fn;
	unsigned int len, count=0;
	int fd;

	thincDbg(T_DBG_RECORD, ("%s: creating new file with extension %s\n", 
		func, ext));

	/* need to try this until we find a filename we haven't used */
again:
	id = thincRecordNewFileID(record);
	if(!id) {
		thincDbg(T_DBG_RECORD, ("ERROR: %s: couldn't get an id for "
			"new file\n", func));
		return T_ERROR;
	}
	len = T_REC_FILEID_LEN;
	len += record->dir_len;
	len += strlen(ext);

	thincDbg(T_DBG_RECORD, ("%s: using %d bytes for new file name\n", func,
		len));

	if(len > T_REC_MAX_FN_LEN) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: resulting filename for "
			"id %s is too long (%d chars)\n", func, id, len));
		free(id);
		return T_ERROR;
	}

	fn = record->dir;
	c = fn + record->dir_len;
	memcpy(c, id, T_REC_FILEID_LEN);
	c += T_REC_FILEID_LEN;
	memcpy(c, ext, strlen(ext));
	fn[len] = '\0';

	/* get this out of the way */
	memcpy(out_id, id, T_REC_FILEID_LEN);
	free(id);

	thincDbg(T_DBG_RECORD, ("%s: trying with filename %s\n", func, fn));

	/* game time */
	fd = open(fn, T_REC_NEWFILE_FLAGS, T_REC_NEWFILE_MODE);
	if(fd == -1) {
		/* only on error. we need fn for just a tiny bit longer */
		t_cleanup_dir(record);
		/* merde */
		if(errno == EEXIST) {
			/* XXX: since id is currently random, the only way we
			 * have of stopping this loop is this horrible hack */
			if(count++ > T_REC_NEWFILE_MAXTRIES) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: tried %d "
					"times to find a unique filename and "
					"failed misserably. ja, i suck\n", 
					func, T_REC_NEWFILE_MAXTRIES));
				return T_ERROR;
			}
			thincDbg(T_DBG_RECORD, ("%s: file %s already exists, "
				"trying again\n", func, fn));
			goto again;
		}
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create new "
			"file %s: %s\n", func, fn, strerror(errno)));
		return T_ERROR;
	}

	t_rec_chown(fn, t_empty_on_error);
	t_rec_chmod(fn, T_REC_NEWFILE_MODE, t_empty_on_error);

	t_cleanup_dir(record);
	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return fd;
}

static int thincRecordNewShotFile(thincRecordPtr record, char *id)
{
	return thincRecordNewFile(record, T_REC_SHOT_EXT, id);
}

static int thincRecordNewCmdFile(thincRecordPtr record, char *id)
{
	return thincRecordNewFile(record, T_REC_CMD_EXT, id);
}

static int thincRecordWriteMagic(int fd)
{
	t_dbg_name(func)
	const char magic[] = T_REC_MAGIC;
	unsigned char i;

	for(i=0; i<4; ++i) {
		t_rec_write(fd, &magic[i], 1, t_empty_on_error);
	}

	return T_OK;
}

#if 0
/* returns T_OK if fd has thinc's magic number */
static int thincRecordChkMagic(int fd)
{
	char m;
	const char magic[] = T_REC_MAGIC;

	if(read(fd, &m, 1) == -1)
		return T_ERROR;
	if(m != magic[0])
		return T_ERROR;

	if(read(fd, &m, 1) == -1)
		return T_ERROR;
	if(m != magic[1])
		return T_ERROR;

	if(read(fd, &m, 1) == -1)
		return T_ERROR;
	if(m != magic[2])
		return T_ERROR;

	if(read(fd, &m, 1) == -1)
		return T_ERROR;
	if(m != magic[3])
		return T_ERROR;

	return T_OK;
}
#endif

/* copies current command info into passed struct */
void thinc_recordCurrCmd(thincScreenPtr pThinc, struct thincRecCmd *c_cmd)
{
	thincRecordPtr record = t_recordPtr(pThinc);

	t_rec_copy_cmd(*c_cmd, record->c_cmd);
}

#endif /* T_RECORD */
