/* Copyright 2006, 2007 Ricardo Baratto */
/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
 * Author:
 * Ricardo Baratto <ricardo@cs.columbia.edu>
 * Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

/* just a bunch of msg stacks */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#ifdef T_RECORD
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <arpa/inet.h>

/* generic things */
#define T_PRIO_RECORD		T_PRIO_SEND /* better name in this context */

/* local vars */
#ifdef T_GETSTATS
static unsigned long t_start;	/* needed for statistics */
#endif

/**************************** screenshot recording **************************/
static int thinc_shotRecord(thincClientPtr cl, void *msg);
static int thinc_shotUpdate(thincClientPtr cl, void *msg);
static int thinc_shotCompress(thincClientPtr cl, void *msg);
#ifdef T_GETSTATS
static int thinc_shotPreStats(thincClientPtr cl, void *msg);
static int thinc_shotPostStats(thincClientPtr cl, void *msg);
#endif

#define T_SHOTPRIO_RECORD	T_PRIO_SEND
#define T_SHOTPRIO_UPDATE	T_SHOTPRIO_RECORD+1
#define T_SHOTPRIO_COMPRESS	T_SHOTPRIO_RECORD-10
#ifdef T_GETSTATS
#define T_SHOTPRIO_PRESTATS	T_PRIO_MIN
#define T_SHOTPRIO_POSTSTATS	T_PRIO_MAX
#endif

thincDefModule(shotRecord, "SHOT_RECORD", 11, T_SHOTPRIO_RECORD);
thincDefModule(shotUpdate, "SHOT_UPDATE", 11, T_SHOTPRIO_UPDATE);
thincDefModule(shotCompress, "SHOT_COMPRESS", 13, T_SHOTPRIO_COMPRESS);
#ifdef T_GETSTATS
thincDefModule(shotPreStats, "SHOT_PRESTATS", 13, T_SHOTPRIO_PRESTATS);
thincDefModule(shotPostStats, "SHOT_POSTSTATS", 14, T_SHOTPRIO_POSTSTATS);
#endif

#define shot_addmodule(record, mod) { \
	err = thinc_msgAddUsingModule((record)->cl,(record)->shot_stack,(mod));\
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add module "\
			"%s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}

typedef struct thincShotObj_S {
	thincScreenPtr pThinc;	/* we get fb info from here. am lazy today */
	thincRecordPtr record;

	unsigned int 		len:32;
	unsigned long long	time;
	unsigned int		pitch:32;
	char *			data;

	/* flags */
	unsigned int		compressed:1;
	unsigned int		pad:7;
} thincShotObj;

/* this one is part of the record struct */
int thinc_shotInitStack(thincRecordPtr record)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_RECORD, ("%s: initializing screenshot message stack\n",
		func));

	record->shot_stack = thinc_msgStackCreate(record->cl);
	if(!record->shot_stack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create "
			"screenshot message stack\n", func));
		return T_ERROR;
	}

#ifdef T_GETSTATS
	shot_addmodule(record, &shotPreStats);
	shot_addmodule(record, &shotPostStats);
#endif
	shot_addmodule(record, &shotCompress);
	shot_addmodule(record, &shotRecord);
	shot_addmodule(record, &shotUpdate);

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

void thinc_shotDestroyStack(thincRecordPtr record)
{
	thinc_msgStackDestroy(record->cl, record->shot_stack);
	record->shot_stack = NULL;
}

/********************** the actual code ************************************/
/* takes a screenshot and dumps it to disk if it's the right thing to do so 
 * (or force is T_TRUE) */
int thincRecordTakeShot(thincScreenPtr pThinc, int force)
{
	t_dbg_name(func)
	thincShotObj obj;
	thincRecordPtr record = t_recordPtr(pThinc);
	int err;

	thincDbg(T_DBG_RECORD, ("%s: taking a screenshot baby\n", func));

	/* don't take a screenshot if not much has changed */
	if(!force) {
		int c_count, c_perc;
		c_count = thincGetDisplayCount(pThinc, record->stats_id);
		c_perc = thincGetDisplayPercent(pThinc, record->stats_id);
		/* different levels of tolerance for changes on the
		 * screen. somewhat arbitrary */
		if( (c_perc < 10) && (c_count < 10000) ) {
			thincInfo("Not enough has changed (%d%%, %d changes), "
				"skipping this screenshot\n", c_perc, c_count);
			return T_OK;
		} else if( (c_perc < 30) && (c_count < 1000) )  {
			thincInfo("Not enough has changed (%d%%, %d changes), "
				"skipping this screenshot\n", c_perc, c_count);
			return T_OK;
		} else {
			thincInfo("%d%% (%d) has changed, taking shot\n",
				c_perc, c_count);
			/* reset counters */
			thincResetDisplayCount(pThinc, record->stats_id);
			thincResetDisplayPercent(pThinc, record->stats_id);
		}
 		thincDbg(T_DBG_RECORD, ("%s: display has changed %d times. "
 			"%d%% of the screen has changed\n", func, c_count, 
 			c_perc));
	}

	obj.pThinc = pThinc;
	obj.record = record;
	obj.len = pThinc->widthBytes*pThinc->height;
	obj.pitch = pThinc->widthBytes;

	obj.data = pThinc->theFB;
	
	gettime_ms_llu(record->time);
	obj.time = record->time;
	obj.compressed = 0;

	err = thinc_msgStackExec(record->cl, record->shot_stack, &obj);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't save current "
			"screenshot\n", func));
		return T_ERROR;
	} 

	if(record->shot_fp > T_REC_SHOTFILE_MAXSZ) {
		thincDbg(T_DBG_RECORD, ("%s: need a new screenshot file. this "
			"one's too big\n", func));
		err = thincRecordGetNewShotFile(record);
		if(err == T_ERROR) {
			/* poor dude, thought everything was cool and sheit
			 * and look where it got screwed */
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get "
				"a new command file\n", func));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_RECORD, ("%s: finished\n", func));
	return T_OK;
}

CARD32 thincRecordTimerShot(OsTimerPtr timer, CARD32 time, pointer arg)
{
	t_dbg_name(func)
	int err;
	thincRecordPtr record = (thincRecordPtr) arg;

	thincDbg(T_DBG_RECORD, ("%s: timer expired, taking a new screenshot\n",
		func));

	err = thincRecordTakeShot(record->pThinc, T_FALSE);
	if(err == T_ERROR) {
		/* XXX: this is bad... maybe we should have a function like
		 * thincAbort() that kills the server? */
		thincDbg(T_DBG_ERROR, ("ERROR: %s: couldn't take screenshot "
			"we're totally faaaaaaaaaaaaaaaaaacked\n", func));
		/* this means we don't take anymore shots :/ */
		return 0;
	} else {
		thincDbg(T_DBG_RECORD, ("%s: all gut, finished\n", func));
		return T_REC_SHOT_TIMER;
	}
}

static int thinc_shotCompress(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, shotCompress.name)
	thincShotObj *obj = msg;
	thincScreenPtr pThinc = cl->parent;
	unsigned int i;
	int err;
	char *zbuff, *from;

	thincDbg(T_DBG_RECORD, ("%s: compressing screenshot\n", modname));

	if(thinc_compressPrepare(pThinc) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't prepare "
			"compression engine for screenshot, we go "
			"uncompressed\n",
			modname));
		/* but we cool and just don't compress this one 8) */
		return T_OK;
	}

	zbuff = thinc_compressGetBuffer(pThinc);
	thinc_compressSetParams(pThinc, T_ZPARAM_W|T_ZPARAM_H,
		pThinc->width, pThinc->height);
	from = obj->data;
	for(i=0; i<pThinc->height; ++i) {
		thincDbg(T_DBG_RECORD, ("%s: compressing line %u, data %p, pitch %d\n", modname, i, from, obj->pitch));
		thinc_compressSetParams(pThinc, T_ZPARAM_IN, from, obj->pitch);
		if(thinc_compress(pThinc) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't compress "
				"line %u\n", modname, i));
			/* we go uncompressed. no need to kill the compression 
			 * engine (it's probably fucked anyway, so why bother?) 
			 */ 
			return T_OK;
		}
		from += obj->pitch;
	}
	err = thinc_compressFlush(pThinc);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush compressed "
			"data\n", modname));
		return T_OK;
	}
	thincDbg(T_DBG_RECORD, ("%s: compressed screenshot from %lu bytes "
		"to a whooping %lu bytes\n", modname, (unsigned long)obj->len, 
		(unsigned long)err));
	obj->len = err;
	obj->data = thinc_compressGetBuffer(pThinc);
	obj->compressed = 1;
	obj->pitch = 0;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

#define t_rec_write(fd, buf, size, on_error) do {			\
	if(write(fd, buf, size) == -1) {				\
		thincDbg(T_DBG_RECORD, ("%s: couldn't write %u bytes "	\
			"to file: %s\n", modname, size, strerror(errno)));\
		on_error;						\
		return T_ERROR;						\
	}								\
} while(0)

#define t_empty_on_error

static int thinc_shotRecord(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, shotRecord.name)
	thincShotObj *obj = msg;
	thincRecordPtr record = obj->record;
	thincScreenPtr pThinc = obj->pThinc;

	struct thincRecShotHeader shot;

	thincDbg(T_DBG_RECORD, ("%s: recording screenshot\n", modname));

	shot.time = htonll(obj->time);
	shot.size = htonl(obj->len);
	shot.width = htons(pThinc->width);
	shot.height = htons(pThinc->height);
	shot.bpp = htons(pThinc->bpp);

	shot.info = 0;
	if(obj->compressed)
		shot.info |= T_REC_SHOT_COMPRESSED;

	shot.info = htons(shot.info);

	/* XXX shaya says this is bad but i'm too lazy today */
	t_rec_write(record->shot, &shot, T_REC_SHOT_HEADER_SZ,
		t_empty_on_error);

	thincDbg(T_DBG_RECORD, ("%s: writing %d bytes of data (%p))\n", 
		modname, obj->len, obj->data));
	t_rec_write(record->shot, obj->data, obj->len, t_empty_on_error);

	/* now update len so gods know how much data we actually wrote and
	 * can udpate their scheisse */
	obj->len += T_REC_SHOT_HEADER_SZ;
	
	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_shotUpdate(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, shotUpdate.name)
	struct thincRecRecord rec;
	thincShotObj *obj = msg;
	thincRecordPtr record = obj->record;

	thincDbg(T_DBG_RECORD, ("%s: writing record of this screenshot "
		"to index file\n", modname));

	/* ok, took screenshot, now generate a record in the index file */
	rec.time = htonll(obj->time);
	t_rec_fill_shot_out(rec.shot, record->shot_fn, record->shot_fp);
	t_rec_copy_cmd_out(rec.cmd, record->c_cmd);

	/* XXX shaya says this is bad but i'm too lazy today */
	t_rec_write(record->index, &rec, T_REC_RECORD_SZ, t_empty_on_error);

	record->shot_fp += obj->len;

	thincDbg(T_DBG_RECORD, ("%s: finished\n", modname));
	return T_OK;
}

#ifdef T_GETSTATS
static int thinc_shotPreStats(thincClientPtr cl, void *msg)
{
	stats_start_nostr(t_start);

	return T_OK;
}

static int thinc_shotPostStats(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, shotPostStats.name)
	unsigned long t_end;

	stats_end_simple(cl->parent, t_end);
	thincDbg(T_DBG_RECORD, ("%s: took us %lums to take the shot\n",
		modname, t_end-t_start));

	return T_OK;
}
#endif /* T_GETSTATS */

#endif /* T_RECORD */
