/* Copyright 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Resize handling functions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include <netinet/in.h>

int thincHandleReszEvent(thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_reszEvent e;
	int err, i,j, prev_state;

	thincDbg(T_DBG_CLIENT, ("%s: got resize event message from client\n",
		func));

	err = thincReadAnyMsg(cl, &e, T_RESZ_EVENT_SIZE);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read resize "
			"event message\n", func));
		return T_ERROR;
	} else if(err == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Connection closed while "
			"trying to read resize event message\n", func));
		return T_CLOSED;
	}

	cl->curr_width = ntohs(e.width);
	cl->curr_height = ntohs(e.height);

	i = cl->curr_width * cl->curr_height;
	j = cl->parent->width * cl->parent->height;

	prev_state = cl->resized;
	if(i == j) {
		thincDbg(T_DBG_CLIENT, ("%s: client's viewport back to "
			"original size\n", func));
		cl->resized = T_CL_RESZ_NONE;
	} else if(i > j) {
		thincDbg(T_DBG_CLIENT, ("%s: Client's viewport larger than "
			"original. Can't do anything\n", func));
		cl->resized = T_CL_RESZ_MORE;
	} else {
		thincDbg(T_DBG_CLIENT, ("%s: Client's viewport smaller than "
			"original. Resizing RAW updates for it\n", func));
		cl->resized = T_CL_RESZ_LESS;
	}
	if(prev_state != cl->resized) {
		if(prev_state == T_CL_RESZ_LESS)
			cl->parent->resizeCount--;
		else if(cl->resized == T_CL_RESZ_LESS)
			cl->parent->resizeCount++;
	}

	cl->factor_x = (double)cl->curr_width / cl->parent->width;
	cl->factor_y = (double)cl->curr_height / cl->parent->height;

	/* inform other parts of the server about the resize event */
	err = thincVidClientResized(cl, prev_state);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not update video "
			"information for resize event\n", func));
		return T_ERROR;
	}
	err = thinc_rawStackClResized(cl, prev_state);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not add resize "
			"module to client's RAW stack\n", func));
		return T_ERROR;
	}
	err = thinc_pixStackClResized(cl, prev_state);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not add resize "
			"module to client's PIXMAP stack\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished. Client's new size: %dx%d\n", 
		func, cl->curr_width, cl->curr_height));
	return T_OK;
}

int thincHandleReszUpdate(thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_reszUpdate u;
	thinc_rectRegion r;
	int err, i;
	WindowPtr root;
	thincScreenPtr pThinc = cl->parent;

	thincDbg(T_DBG_CLIENT, ("%s: reading resize update request from "
		"client\n", func));

	root = thincGetRootWindow(pThinc);
	if(!root) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get root "
			"window!!\n", func));
		return T_ERROR;
	}

	err = thincReadAnyMsg(cl, &u, T_RESZ_UPDATE_SIZE);
	if(err == T_ERROR || err == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read Resize "
			"Update message\n", func));
		return err;
	}

	u.numrects = ntohl(u.numrects);
	thincDbg(T_DBG_CLIENT, ("%s: have %d rectangles in request\n", func,
		u.numrects));

	for(i=0; i<u.numrects; ++i) {
		err = thincReadData(cl, &r,  T_RECTREGION_SIZE);
		if(err == T_ERROR || err == T_CLOSED) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read "
				"Resize Update message\n", func));
			return err;
		}
		r.x = ntohs(r.x);
		r.y = ntohs(r.y);
		r.width = ntohs(r.width);
		r.height = ntohs(r.height);

		thincDbg(T_DBG_CLIENT, ("%s: queueing update for rectangle "
			"(%d,%d):(%dx%d)\n", func, r.x, r.y, r.width, 
			r.height));

		err = thincBufferNewCmd(pThinc, root, T_FB_RUPDATE, r.x,
			r.y, r.width, r.height, T_RAW_VA_NONE);
		if(err != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer "
				"resize update for client\n", func));
			return T_ERROR;
		}
	}
	
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

void thincReszBox(thincClientPtr cl, BoxPtr box)
{
	box->x1 *= cl->factor_x;
	box->y1 *= cl->factor_y;
	box->x2 *= cl->factor_x;
	box->y2 *= cl->factor_y;
}
