/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * THINC util functions:
 * -> read/write from file descriptor
 * -> print to log
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#include <errno.h>
#include <sys/select.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/poll.h>
#include <netinet/in.h>

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

/* it'd be problematic to include X's headers files in here */
extern int xf86BlockSIGIO(void);
extern void xf86UnblockSIGIO(int sigstate);

/* This is a buffer used to send messages specific to a client
 * (for example, for connection-start messages, we can't use the
 * global buffer because this messages would confuse already 
 * connected clients)
 */
static char msgBuffer[T_STATIC_BUFLEN];
char *t_msgBuffer = msgBuffer + THINC_HEADERSIZE;

stats_var(writeTime)
stats_var(blockTime)
stats_var(tmpSize)

/* functions for clients which have no notion of "sending" messages */
int thincSendMsgNull(thincClientPtr cl, int type, unsigned int msg_len, 
		     int flags)
{
	return T_OK;
}

int thincBadRead(thincClientPtr cl, void *buf, unsigned int count)
{
	T_BUG_ARG("attempt to read from an unreadable client %s", cl->id);
	return T_ERROR;
}

int thincBadWrite(thincClientPtr cl, const void *buf, unsigned int count)
{
	T_BUG_ARG("attempt to write to an unwritable client %s", cl->id);
	return T_ERROR;
}

/* Send the message stored in t_msgBuffer */
int thincSendMsg(thincClientPtr cl, int type, unsigned int msg_len, int flags)
{
	thinc_header *h;
	int res;
	t_dbg_name(func)
	int len = msg_len + THINC_HEADERSIZE;

	thincDbg(T_DBG_IO, ("%s: sending message of length %d\n", func, 
		msg_len));

	if(msg_len > T_STATIC_BUFLEN) {
		T_BUG_ARG("Message is too long: %d", msg_len);
		return T_ERROR;
	}

	/* make the header for the msg */
	h = (thinc_header*)msgBuffer;

	h->type = type;
	h->flags = flags;
	/* FIXME: Still undecided on how to handle the length field */
	h->length = htons(msg_len);

	res = thincWriteData(cl, msgBuffer, len);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't write message\n", 
			func));
		return T_ERROR;
	}

	thincDbg(T_DBG_IO, ("%s: finished\n", func));

	return T_OK;
}

/* why?
 * We want to buffer write calls to minimize syscall overhead and hopefully
 * improve responsiveness of the server.
 * how?
 * A message is sent to a client by first putting it in the message buffer
 * and later on flushing the buffer.
 * The first step is accomplished by thincBufferMsg().
 * The second step by thincFlushBuffer().
 * thincBufferMsg() is used both as utility and buffer manager:
 * It first checks to see if there's enough space in the buffer for the
 * new message. If there isn't the buffer is flushed first.
 * After enough space has been found, a msg header is written to it. 
 * Finally, a pointer to where the caller should put its message is
 * returned. Yes, this is unsafe because we're trusting the caller
 * too much. However, as long as the function behaves, and only does 
 * what it said it'd do (ie. put msg_len bytes in the buffer, nothing more,
 * nothing less) things will be ok.
 * If the type is 0, thincBufferMsg() will assume that the caller wants to
 * buffer data and no message header is added.
 * Since buffering can be turned off in the configuration file, 
 * thincBufferMsg() checks first to see if buffering should be done. If we
 * shouldn't, it'll just silently fail. Therefore, the caller is responsible
 * to always revert to sending messages the normal way in case of a buffering
 * failure.
 * Note that the caller doesn't need to worry about updating any
 * pointers of the buffer. That's all taken care of.
 * NULL is returned on error (eg. buffer not initialized, length is
 * bigger than the maximum allowed, ...)
 * TODO:
 * We could put a marker at the end of every piece that's given to
 * a caller and next time we're called chk that the marker hasn't been
 * overwritten.
 * TODO: If write buffering is not being done, return a pointer to
 * the static buffer that's used as fallback. Then, when flushbuffer()
 * is called, just flush that buffer. Need to keep some state tho,
 * but it makes sending messages so much cleaner and transparent to 
 * everybody...
 */
/* FIXME: use unsigned int for lengths */
void* thincBufferMsg(thincClientPtr cl, int type, int flags, 
		     unsigned int msg_len, int data_len)
{
	t_dbg_name(func)
	int res;
	thinc_header *h;
	int space;
	void *p;
	/* how much data is being written */
	int len;
	
	thincDbg(T_DBG_IO, ("%s: sending message of length %d and data %d\n", 
		func, msg_len, data_len));

	/* chk if we're supposed to be buffering */
	if(!cl->parent->doWriteBuffer) {
		/* nope. just fail. caller is responsible to revert
		 * back to normal writes */
		thincDbg(T_DBG_IO, ("%s: finished. no buffering being "
			"done\n", func));
		return NULL;
	}

#ifdef T_RECORD
	/* this will be true for the recording client... */
	if(!cl->msgBufP)
		return NULL;
#else
	/* sanity checks */
	t_bug_on_2(
		(!cl->msgBufP),
		T_BUG("ERROR: Message buffer hasn't been initalized\n"),
		return NULL
	);
#endif

	if(type != 0) 
		len = THINC_HEADERSIZE + msg_len + data_len;
	else
		len = data_len;
	
	/* chk there's enough space in the buffer */
	space = cl->msgBufMaxSize - cl->msgBufSize;
	
	thincDbg(T_DBG_IO, ("%s: there's space for %d bytes in the buffer. "
		"can we add the msg?\n", func, space));

	if(len > cl->msgBufMaxSize) {
		thincDbg(T_DBG_IO, ("%s: Message too big for buffer\n", func));
		return NULL;
	} else if(len > space) {
		thincDbg(T_DBG_IO, ("%s: Not enough space in buffer. "
			"Flushing first\n", func));
		/* not enough space for the message. Flush the buffer */
		res = thincFlushBuffer(cl);
		if(res == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush "
				"message buffer\n", func));
			return NULL;
		}
	}

	/* ok, there's enough space. write the header, update the
	 * buffer and return the pointer */
	if(type != 0) {
		h = (thinc_header*)cl->msgBufP;
		h->type = type;
		h->flags = flags;
		p = cl->msgBufP + THINC_HEADERSIZE;
	} else
		p = cl->msgBufP;

	cl->msgBufP += len;
	cl->msgBufSize += len;

	/* if buffer gets filled by this message it'll be flushed
	 * next time around */
	thincDbg(T_DBG_IO, ("%s: finished. buffer size %d bytes. %d bytes "
		"free\n", func, cl->msgBufSize, 
		cl->msgBufMaxSize - cl->msgBufSize));

#if 0
	thinc_assert(cl->msgBufMaxSize == T_BUFSIZE);
	thinc_assert(cl->msgBufSize <= cl->msgBufMaxSize  && cl->msgBufSize >= 0);

#endif
	return p;
}

/* flushes the write buffer of client */
int thincFlushBuffer(thincClientPtr cl)
{
	t_dbg_name(func)
	/* bufsize can be at most maxsize */
	int res, len = cl->msgBufSize;

#if 0
	thinc_assert(len <= T_BUFSIZE && len >= 0);
	thinc_assert(cl->msgBufMaxSize == T_BUFSIZE);
	thinc_assert(cl->msgBufSize <= cl->msgBufMaxSize  && cl->msgBufSize >= 0);
	thinc_assert(cl->msgBuffer[T_BUFSIZE] == -1);
#endif

	if(!cl->parent->doWriteBuffer)
		return T_OK;

	if(len == 0) {
		thincDbg(T_DBG_ANNOY, ("%s: finished. Nothing to write\n", 
			func));
		return T_OK;
	}
	
	thincDbg(T_DBG_IO, ("%s: flushing %d bytes. buffmemory %p\n", 
		func, len, cl->msgBuffer));

	t_bug_on_2(
		(len > cl->msgBufMaxSize),
		T_BUG_ARG("Invalid buffer length %u to flush\n", len),
		return T_ERROR
	);

	res = thincWriteData(cl, cl->msgBuffer, len);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't write data to "
			"client\n", func));
		return T_ERROR;
	}
	
	/* update buffer info */
	cl->msgBufP = cl->msgBuffer;
	cl->msgBufSize = 0;

	thincDbg(T_DBG_IO, ("%s: finished\n", func));

	return T_OK;
}

/* Flushes the write buffer of all the clients connected. Debugging messages 
 * here are marked as annonying because we're called from an annoying 
 * function */
int thincFlushBufferAll(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	/* bufsize can be at most maxsize */
	int res;
	thincClientPtr cl;

	thincDbg(T_DBG_ANNOY, ("%s\n", func));

	if(!pThinc->doWriteBuffer)
		return T_OK;

	call_for_all(thincFlushBuffer, (cl));

	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));

	return T_OK;
}

/* read a header. returns -1 on error, 0 if connection closed, 1 on succes */
int thincReadHeader(thincClientPtr cl, thinc_header *h)
{
	int err;
	if((err = thincReadData(cl, h, THINC_HEADERSIZE)) == T_ERROR)
		return T_ERROR;
	else {
		h->length = ntohs(h->length);
		return err;
	}
}

/* reads a message from socket sock and puts it in msg. If the header is
 * in the read buffer it's also put in msg (call thincReadHeader() first if
 * you only want the msg's body). Returns 1 on success, 0 if connection
 * is closed, -1 on error.
 */
int thincReadAnyMsg(thincClientPtr cl, void *msg, unsigned int msg_size)
{
	return thincReadData(cl, msg, msg_size);
}

int thincClientRBufferEmpty(thincClientPtr cl)
{
	return (cl->rBufSize == 0);
}

/* read data from socket. If the socket is non-blocking, -1 is returned
 * if waiting for data will block the server for more than T_TIMEOUT ms
 * (Note that some data may have been read before this happens. This data
 * is lost) On success, count will be returned. If the connection is closed 
 * before all is read, T_CLOSED is returned. On error, T_ERROR is returned */
int thincReadData(thincClientPtr cl, void *data, unsigned int count)
{
	t_dbg_name(func)
	int total, res, need;
	char *curr = (char *)data;

	thincDbg(T_DBG_IO, ("%s\n", func));

	if(count <= cl->rBufSize) {
		memcpy(data, cl->rBufStart, count);
		cl->rBufSize -= count;
		cl->rBufStart += count;
		return count;
	}

	/* not enough, copy what we have and get the rest */
	memcpy(data, cl->rBufStart, cl->rBufSize);
	curr = ((char*)data) + cl->rBufSize;
	need = count - cl->rBufSize;
	cl->rBufSize = 0;
	cl->rBufStart = cl->rBuffer;

	/* if we need to read too much data just call the normal read 
	 * function */
	if(need > T_READ_BUFLEN) {
		res = thincReadDataNoBuffer(cl, curr, need);
		if(res != T_OK) 
			return res;
		return count;
	}

	/* read to the buffer and then copy out whatever's needed */
	total = 0;
	while(total < need) {
		res = (*cl->read)(cl, cl->rBufStart, T_READ_BUFLEN-total); 
		if(res == T_WOULDBLOCK) {
			res = thincDoPoll(cl->sock, POLLIN, T_CL_TIMEOUT);
			if(res == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: An error "
					"occurred while waiting for data\n"));
				return T_ERROR;
			} else if(res == T_TIMEOUT) {
				thincDbg(T_DBG_IO, ("%s: timeout on read\n",
					 func));
				return T_ERROR;
			}
			/* can read again */
			continue;
		} else if(res == T_CLOSED) {
			return T_CLOSED;
		} else if(res == T_ERROR) {
			return T_ERROR;
		}
		total += res;
		cl->rBufStart += res;
	}

	memcpy(curr, cl->rBuffer, need);
	cl->rBufSize = total - need;
	cl->rBufStart = cl->rBuffer + need;

	thincDbg(T_DBG_IO, ("%s: finished\n", func));

	return count;
}

/* returns T_WOULDBLOCK instead of blocking */
int thincReadDataNoblock(thincClientPtr cl, void *data, unsigned int count)
{
	t_dbg_name(func)
	int total, res, need;
	char *curr = (char *)data;

	thincDbg(T_DBG_IO, ("%s\n", func));

	if(count <= cl->rBufSize) {
		memcpy(data, cl->rBufStart, count);
		cl->rBufSize -= count;
		cl->rBufStart += count;
		return count;
	}

	/* not enough, copy what we have and get the rest */
	memcpy(data, cl->rBufStart, cl->rBufSize);
	curr = ((char*)data) + cl->rBufSize;
	need = count - cl->rBufSize;
	cl->rBufSize = 0;
	cl->rBufStart = cl->rBuffer;

	/* if we need to read too much data just call the normal read 
	 * function */
	if(need > T_READ_BUFLEN) {
		res = thincReadDataNoBuffer(cl, curr, need);
		if(res != T_OK) 
			return res;
		return count;
	}

	/* read to the buffer and then copy out whatever's needed */
	total = 0;
	while(total < need) {
		res = (*cl->read)(cl, cl->rBufStart, T_READ_BUFLEN-total); 
		if(res == T_WOULDBLOCK || res == T_CLOSED || res == T_ERROR) {
			thincDbg(T_DBG_IO, ("%s: finished (%d)\n", func, res));
			return res;
		}
		total += res;
		cl->rBufStart += res;
	}

	memcpy(curr, cl->rBuffer, need);
	cl->rBufSize = total - need;
	cl->rBufStart = cl->rBuffer + need;

	thincDbg(T_DBG_IO, ("%s: finished\n", func));

	return count;
}

int thincReadDataNoBuffer(thincClientPtr cl, void *buf, unsigned int count)
{
	t_dbg_name(func)
	int total, res;
	char *curr = (char *)buf;

	thincDbg(T_DBG_IO, ("%s\n", func));

	total = 0;
	while(total != count) {
		res = (*cl->read)(cl, curr, count-total);
		if(res == T_WOULDBLOCK) {
			res = thincDoPoll(cl->sock, POLLIN, 
				T_CL_TIMEOUT);
			if(res == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: An error "
					"occurred while waiting for data\n"));
				return T_ERROR;
			} else if(res == T_TIMEOUT) { 
				thincDbg(T_DBG_IO, ("%s: timeout on read\n",
					 func));
				return T_ERROR;
			}
			/* try again */
			continue;
		} else if(res == T_CLOSED) {
			return T_CLOSED;
		} else if(res == T_ERROR) {
			return T_ERROR;
		}
		total += res;
		curr += res;
	} 

	thincDbg(T_DBG_IO, ("%s: finished\n", func));
	return count;
}

int thincReadDataSecure(thincClientPtr cl, void *buf, unsigned int count)
{
	int err  = thinc_ssl_read(cl->ssl, buf, count);
	if(err == T_SSL_CLOSED) 
		return T_CLOSED;
	else if(err == T_SSL_WOULDBLOCK)
		return T_WOULDBLOCK;
	else if(err == T_SSL_ERROR)
		return T_ERROR;
	else
		return err;
}

int thincReadDataSocket(thincClientPtr cl, void *buf, unsigned int count)
{
	int err;
try:
	err = read(cl->sock, buf, count);
	if(err == 0)
		return T_CLOSED;
	else if(err == -1) {
		if( (errno == EWOULDBLOCK) || (errno == EAGAIN) )
			return T_WOULDBLOCK;
		else if(errno == EINTR)
			goto try;
		else
			return T_ERROR;
	} else
		return err;
}

/* writes count bytes in buffer to socket. Returns count on success, -1 on 
 * error. If writing to the socket will block for more than T_TIMEOUT, -1 is
 * returned too.
 */
int thincWriteData(thincClientPtr cl, const void *buf, unsigned int count)
{
	int left = count, wrote = 0;
	int sigstate;
	const char *curr = (char *) buf;
	t_dbg_name(func)
	int sock = cl->sock;
	stats_var2(t1, t2);

	thincDbg(T_DBG_IO, ("%s: writing %d bytes\n", func, count));

	sigstate = xf86BlockSIGIO();
	left = count;
	while (left > 0) {
		stats_start_nostr(t1);
		instr_add("WRITE", TI_START);
		wrote = (*cl->write)(cl, curr, left);
		instr_add("WRITE", TI_END);
		stats_end_nostr(t1, t2, writeTime);
		if(wrote == T_CLOSED) {
			thincDbg(T_DBG_ERROR, ("ERROR: Connection closed "
				"while trying to write data\n"));
			return T_ERROR;
		} else if(wrote == T_WOULDBLOCK) {
			instr_add("WRITE_BLOCK", TI_START);
			stats_start_nostr(t1);
			wrote = thincDoPoll(sock, POLLOUT, T_CL_TIMEOUT);
			stats_end_nostr(t1, t2, blockTime);
			instr_add("WRITE_BLOCK", TI_END);
			if(wrote == T_ERROR) {
				xf86UnblockSIGIO(sigstate);
				return T_ERROR;
			} else if (wrote == T_TIMEOUT) {
				xf86UnblockSIGIO(sigstate);
				thincDbg(T_DBG_IO, ("%s: Timeout on "
					"client\n", func));
				return T_ERROR;
			}
			/* try again */
			continue;
		} else if(wrote == T_ERROR) {
			xf86UnblockSIGIO(sigstate);
			thincDbg(T_DBG_ERROR, 
				("ERROR: %s: Couldn't write data: %s\n", 
				func, strerror(errno)));
			return T_ERROR;
		}
		left -= wrote;
		curr += wrote;
	}
	xf86UnblockSIGIO(sigstate);

	thincDbg(T_DBG_IO, ("%s: finished\n", func));
	return count-left;
}

int thincWriteDataSocket(thincClientPtr cl, const void *buf, 
			 unsigned int count)
{
	int wrote;
try:
	wrote = write(cl->sock, buf, count);
	if (wrote <= 0) {
		if (errno == EINTR)
			goto try;
		else if (errno == EWOULDBLOCK || errno == EAGAIN) {
			return T_WOULDBLOCK;
		} else {
			return T_ERROR;
		}
	}
	return wrote;
}

int thincWriteDataSecure(thincClientPtr cl, const void *buf, 
			 unsigned int count)
{
	int wrote;
	wrote = thinc_ssl_write(cl->ssl, buf, count);
	if(wrote == T_SSL_CLOSED) 
		return T_CLOSED;
	else if(wrote == T_SSL_WOULDBLOCK)
		return T_WOULDBLOCK;
	else if(wrote == T_SSL_ERROR)
		return T_ERROR;
	else
		return wrote;
}

/* Returns T_BIG or T_LITTLE according to the endian-ness of the host
 * machine. From ac_c_bigendian_cross (gnu.org), in turn from 
 * Harbison&Steele. */
int thincEndian() 
{
        union {
                long l;
                char c[sizeof(long)];
        } u;
        u.l = 1;
        if(u.c[sizeof(long)-1] == 1)
                return T_BIG;
        else
                return T_LITTLE;
}

/* need this include down here so that it doesn't conflict with system
 * calls */
#include "xf86.h" 

/* used to print informational messages or debugging info to X's log
 * (will show up with (II) prepended). If doing debug, we also print
 * the time */
void thincLog(char *fmt, ...)
{
	va_list args;
#ifdef T_DO_DEBUG
	time_t t;
	struct tm b;

	time(&t);
#endif
	
	va_start(args, fmt);
#ifdef T_DO_DEBUG
	localtime_r(&t, &b);
	xf86DrvMsgVerb(-1, X_INFO, 1, "(%02d:%02d:%02d) ", b.tm_hour, b.tm_min,
		b.tm_sec);
	xf86VDrvMsgVerb(-1, X_NONE, 1, fmt, args);
#else
	xf86VDrvMsgVerb(-1, X_INFO, 1, fmt, args);
#endif

	va_end(args);
}

void thincInfo(char *fmt, ...)
{
	va_list args;
	
	va_start(args, fmt);
	xf86VDrvMsgVerb(-1, X_INFO, 1, fmt, args);
	va_end(args);
}

/* this one prints also a message to stdout */
void thincMsg(char *fmt, ...)
{
	va_list args;
	
	va_start(args, fmt);
	xf86VDrvMsgVerb(-1, X_INFO, 0, fmt, args);
	va_end(args);
}

void thincWarn(char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	xf86VDrvMsgVerb(-1, X_WARNING, 1, fmt, args);
	va_end(args);
}

/* use to print an error to X's log */
void thincError(char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	xf86VDrvMsgVerb(-1, X_ERROR, 1, fmt, args);
	va_end(args);
}
