/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Window functions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#ifdef T_APPSHARE

#define T_WIN_FIRST_ID	(1<<31)

static void thincUninitWindowPriv(WindowPtr pWin);
static int thincInitWindowPriv(WindowPtr pWin);

#endif	/* T_APPSHARE */

/************ WINDOW FUNCTIONS ********************/
WindowPtr thincGetRootWindow(thincScreenPtr pThinc)
{
	return WindowTable[pThinc->pScreen->myNum];

#if 0
	WindowPtr root;

	root = WindowTable[pScreen->myNum];
	if(!root) {
		thincDbg(9, ("%s: Couldn't get root window\n", func));
	} else if(root == pWindow) {
		thincDbg(9, ("%s: Creating root window\n", func));
	} else if(root == pWindow->parent) {
		thincDbg(9, ("%s: Creating top-level window\n", func));
	} else {
		thincDbg(9, ("%s: Creating subwindow\n", func));
	}
#endif
}

/* PaintWindowBackground/Border
 * Use 'what' to distinguish -> PW_BACKGROUND or PW_BORDER
 * and use the window's border fill style or its background fill style to
 * paint the given region. 
 * Both fill styles consist of a union, PixUnion:
   typedef union _PixUnion {
    PixmapPtr		pixmap;
    unsigned long	pixel;
   } PixUnion;
 * and a separate variable which indicates which entry is valid.
 * For border, the fill style is defined by pWindow->border
 * if pWindow->borderIsPixel != 0
 * then the union contains a pixel value, else a tile (pixmap)
 * For background, pWindow->background has the fill style and 4 possible
 * values for the union are contained in pWindow->backgroundState:
 * None => Region should be left unfilled
 * ParentRelative => background of the parent is inherited
 * BackgroundPixel => the Pixel in PixUnion should be used
 * BackgroundPixmap => the Pixmap in PixUnion should be used
 */
void thincPaintWindowBorder(WindowPtr pWindow, RegionPtr pRegion, int what)
{
	t_dbg_name(func)
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	thincDbg(T_DBG_WIN, ("%s: Fill it using %d\n", func, 
		pWindow->borderIsPixel));

	stats_inc(pThinc, countPaintWindowBorder);

	thinc_unwrap(pThinc, pScreen, PaintWindowBorder);
	(*pScreen->PaintWindowBorder)(pWindow, pRegion, what);
	thinc_wrap(pThinc, pScreen, PaintWindowBorder, thincPaintWindowBorder);

	/* we chk for this after the driver's func has been called, 
	 * 'cause we want it to execute anyway */
	if(what != PW_BORDER) {
		T_BUG_ARG("Unknown 'what' param %d", what);
		return;
	}

	/* TODO: Should we send a solid/pixmap fill udpate here?? */
	if(pWindow->borderIsPixel) {
		thincDbg(T_DBG_WIN, ("%s: border filled with value %ld\n",
			func, pWindow->border.pixel));
	} else { /* border is filled with pixmap */
		thincDbg(T_DBG_WIN, ("%s: border filled with pixmap of "
			"dimensions %dx%d\n", func,
			pWindow->border.pixmap->drawable.width,
			pWindow->border.pixmap->drawable.height));
	}
	thincBufferNewCmdReg(pThinc, pWindow, T_FB_RUPDATE, pRegion, 
		T_RAW_VA_NONE);

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));
}

/* see comments above PaintWindowBorder */
void thincPaintWindowBackground(WindowPtr pWindow, RegionPtr pRegion, int what)
{
	const char *func = __func__;
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);
	/* pixmap's data and dimensions (for BackgroundPixmap) */
	char *pixmap;
	int w,h, x, y;
	BoxPtr ext;

	thincDbg(T_DBG_WIN, ("%s: Paint %#lx using %d\n", func, 
		t_winXID(pWindow), pWindow->backgroundState));

	stats_inc(pThinc, countPaintWindowBackground);

	/* call driver's function */
	thinc_unwrap(pThinc, pScreen, PaintWindowBackground);
	(*pScreen->PaintWindowBackground)(pWindow, pRegion, what);
	thinc_wrap(pThinc, pScreen, PaintWindowBackground, 
		thincPaintWindowBackground);

	/* this one needs special care to avoid repeated calls */
	if(pWindow->backgroundState == ParentRelative) {
		/* find the window with the background state for this
		 * one and execute the rest using that window */
		do {
			pWindow = pWindow->parent;
		} while(pWindow->backgroundState == ParentRelative);
	}

	/* we chk for this after the driver's func has been called, 
	 * 'cause we want the func to execute anyway */
	if(what != PW_BACKGROUND) {
		T_BUG_ARG("Unknown 'what' param %d", what);
		return;
	}

	ext = REGION_EXTENTS(pScreen, pRegion);
	if( (box_width_zerook(ext) == 0) || (box_height_zerook(ext) == 0) ) {
		thincDbg(T_DBG_WIN, ("%s: background region empty, "
			"finished\n", func));
		return;
	}

	print_box(T_DBG_WIN, ext, "painting: ");

	/* Chk the window's background fill style and send appropiate
	 * message */
	switch (pWindow->backgroundState) {
		case None: 
			/* None -> Don't do anything */
			thincDbg(T_DBG_WIN, ("%s: Region shouldn't be filled, "
				"no need for update\n", func));
			break;
		case BackgroundPixel:
			/* Send a T_FB_SFILL (solid fill) command */
			if(pThinc->useSFill) {
				thincDbg(T_DBG_WIN, 
					("%s: solid fill, pixel value "
					"is %ld\n", func,
					pWindow->background.pixel));
				thincBufferNewCmdReg(pThinc, pWindow,
					T_FB_SFILL, pRegion, 
					pWindow->background.pixel);
			} else {
				thincBufferNewCmdReg(pThinc, pWindow,
					T_FB_RUPDATE, pRegion, T_RAW_VA_NONE);
			}
			break;
		case BackgroundPixmap:
			/* chk first if we're allowed to send pixmap fills */
			if(!pThinc->usePFill) {
				thincBufferNewCmdReg(pThinc, pWindow,
					T_FB_RUPDATE, pRegion, T_RAW_VA_NONE);
				break;
			}

			/* Send a T_FB_PFILL (pixmap fill) command */
			x = pWindow->drawable.x;
			y = pWindow->drawable.y;
			w = pWindow->background.pixmap->drawable.width;
			h = pWindow->background.pixmap->drawable.height;
			thincDbg(T_DBG_WIN, 
				("%s: pixmap dimensions: w:%d, h:%d\n", 
				func, w, h));
			/* if pixmap is too big, it's not worth sending it as a 
			 * tile */
			if(w*h > T_PIXMAP_MAXS) {
				thincDbg(T_DBG_WIN, 
					("%s: pixmap too big, sending "
					"normal update\n", func));
				thincBufferNewCmdReg(pThinc, pWindow,
					T_FB_RUPDATE, pRegion, T_RAW_VA_NONE);
				/* Finished */
			/* if region is too small don't sent the pixmap */
			/* FIXME: Use box_width/box_height() */
			} else if( (ext->x2-ext->x1)*(ext->y2-ext->y1) < w*h) {
				thincDbg(T_DBG_WIN, 
					("%s: region too small %d,%d "
					"sending normal update\n", func, 
					ext->x2-ext->x1,
					ext->y2-ext->y1));
				thincBufferNewCmdReg(pThinc, pWindow,
					T_FB_RUPDATE, pRegion, T_RAW_VA_NONE);
			} else {
				pixmap = thinc_getPixmap(
					pWindow->background.pixmap);
				thincBufferNewCmdReg(pThinc, pWindow,
					T_FB_PFILL, pRegion, 
					pixmap, x, y, w, h);
			}
			break;
		default:
			thincError("%s: Found unknown backgroundState in "
				"window\n", func);
			break;
	}

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));
}

/* Called when a window is moved to actually move the pixels of
 * the window 
 * -> ptOldOrg -> old upper-left corner location.
 * -> pRegionSrc -> region it's coming from. It's what we already have.
 */
void thincCopyWindow(WindowPtr pWindow, DDXPointRec ptOrigSrc, 
	RegionPtr pRegionSrc)
{
	t_dbg_name(func)
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);
	RegionRec copyReg;
	int dx, dy;

	thincDbg(T_DBG_WIN, ("%s\n", func));

	stats_inc(pThinc, countCopyWindow);

	dx = pWindow->drawable.x - ptOrigSrc.x;
	dy = pWindow->drawable.y - ptOrigSrc.y;

	if(dx == 0 && dy == 0) {
		thincDbg(T_DBG_WIN, ("%s: not really moving (dx & dy == 0)\n", 
			func));
		return;
	}

	/* clip the region with the window's clip region. Need to do this
	 * before, in case the driver's function changes the source region.
	 * However, we send the copy command _after_ the copy has
	 * occurred locally, in case we need to send additional raw
	 * updates (such as in the case of the client's viewport being
	 * smaller than X's viewport. See thincSendFBCopy for details) */
	REGION_INIT(pScreen, &copyReg, &miEmptyBox, 0);
	REGION_TRANSLATE(pScreen, pRegionSrc, dx, dy);
	REGION_INTERSECT(pScreen, &copyReg, &pWindow->borderClip, pRegionSrc);

	if(!REGION_NOTEMPTY(pScreen, &copyReg)) {
		thincDbg(T_DBG_WIN, ("%s: finished. Region to copy is "
			"completely clipped\n", func));
		return;
	}

	/* Translate back the source region so driver doesn't have problems */
	REGION_TRANSLATE(pScreen, pRegionSrc, -dx, -dy);

	if(pThinc->useCopy) {
		/* Send the copy */
		REGION_TRANSLATE(pScreen, &copyReg, -dx, -dy);

		thincDbg(T_DBG_WIN, ("%s: Sending copy for region (%d, %d) "
			"(%d, %d) dx %d, dy %d\n", func, copyReg.extents.x1, 
			copyReg.extents.y1, copyReg.extents.x2, 
			copyReg.extents.y2, dx, dy));
		
		/* Send the copy */
		thincBufferNewCmdReg(pThinc, pWindow, T_FB_COPY, &copyReg, 
			dx, dy);

		/* call driver's function */
		thinc_unwrap(pThinc, pScreen, CopyWindow);
		(*pScreen->CopyWindow) (pWindow, ptOrigSrc, pRegionSrc);
		thinc_wrap(pThinc, pScreen, CopyWindow, thincCopyWindow);
	} else {
		/* call driver's function */
		thinc_unwrap(pThinc, pScreen, CopyWindow);
		(*pScreen->CopyWindow) (pWindow, ptOrigSrc, pRegionSrc);
		thinc_wrap(pThinc, pScreen, CopyWindow, thincCopyWindow);

		thincDbg(T_DBG_WIN, ("%s: Sending raw update for region "
			"(%d, %d) %d, %d) dx %d, dy %d\n", func, 
			copyReg.extents.x1, 
			copyReg.extents.y1, copyReg.extents.x2, 
			copyReg.extents.y2, dx, dy));

		thincBufferNewCmdReg(pThinc, pWindow, T_FB_RUPDATE, &copyReg,
			T_RAW_VA_NONE);
	}

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));
}

#ifdef T_APPSHARE
Bool thincCreateWindow(WindowPtr pWindow)
{
	t_dbg_name(func)
	Bool res;
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);
	DrawablePtr d;

	d = &pWindow->drawable;
	thincDbg(T_DBG_WIN, ("%s: new window's size: %d,%d->%d,%d\n", 
		func, d->x, d->y, d->width, d->height));

	/* call driver's function */
	thinc_unwrap(pThinc, pScreen, CreateWindow);
	res = (*pScreen->CreateWindow) (pWindow);
	thinc_wrap(pThinc, pScreen, CreateWindow, thincCreateWindow);

	if(!res) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Driver's function failed\n",
			func));
		return res;
	}

	if(thincInitWindowPriv(pWindow) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"Window private struct\n", func));
		/* fail even if the wrapped create window worked */
		return FALSE;
	}

	thincDbg(T_DBG_WIN, ("%s: finished. window's id: %#lx\n", func,
		t_winID(pWindow)));

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));

	return res;
}

Bool thincDestroyWindow(WindowPtr pWindow)
{
	t_dbg_name(func)
	Bool res;
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	thincDbg(T_DBG_WIN, ("%s: destroying window %#lx\n", func,
		t_winID(pWindow)));

#if 0
	WindowPtr root = WindowTable[pScreen->myNum];

	if(pWindow == root) {
		thincDbg(9, ("%s: WHAAAAT?? the root window is being "
			"destroyed!! aaaaggggghhhhh\n", func));
	} else if(pWindow->parent == root) {
		thincDbg(9, ("%s: oh no! A window is getting destroyed!\n",
			func));
	} else {
		thincDbg(9,("%s: a subwindow is being destroyed\n", func));
	}
#endif

	thincUninitWindowPriv(pWindow);

	thinc_unwrap(pThinc, pScreen, DestroyWindow);
	res = (*pScreen->DestroyWindow) (pWindow);
	thinc_wrap(pThinc, pScreen, DestroyWindow, thincDestroyWindow);

	thincDbg(T_DBG_WIN, ("%s: finished\n", func));

	return res;
}

/* moves or resizes a window. Note: The actual graphics involved 
 * in moving the pixels on the screen and drawing the border are 
 * handled by CopyWindow()
 * XXX: What's x and y?? */
Bool thincPositionWindow(WindowPtr pWindow, int x, int y)
{
	Bool res;
	t_dbg_name(func)
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	/* call driver's function */
	thinc_unwrap(pThinc, pScreen, PositionWindow);
	res = (*pScreen->PositionWindow) (pWindow, x, y);
	thinc_wrap(pThinc, pScreen, PositionWindow, thincPositionWindow);

	return res;
}

Bool thincMapWindow(WindowPtr pWindow)
{
	Bool res;
	t_dbg_name(func)
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	/* call driver's function */
	thinc_unwrap(pThinc, pScreen, RealizeWindow);
	res = (*pScreen->RealizeWindow) (pWindow);
	thinc_wrap(pThinc, pScreen, RealizeWindow, thincMapWindow);

	return res;
}

Bool thincUnmapWindow(WindowPtr pWindow)
{
	Bool res;
	t_dbg_name(func)
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	/* call driver's function */
	thinc_unwrap(pThinc, pScreen, UnrealizeWindow);
	res = (*pScreen->UnrealizeWindow) (pWindow);
	thinc_wrap(pThinc, pScreen, UnrealizeWindow, thincUnmapWindow);

	return res;
}

Bool thincChangeWindowAttributes(WindowPtr pWindow, unsigned long mask)
{
	Bool res;
	t_dbg_name(func)
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	/* unwrap and call driver's function */
	thinc_unwrap(pThinc, pScreen, ChangeWindowAttributes);
	res = (*pScreen->ChangeWindowAttributes)(pWindow, mask);
	thinc_wrap(pThinc, pScreen, ChangeWindowAttributes, 
		thincChangeWindowAttributes);

	return res;
}

/* Paints the border and generates exposure events for the window. 
 * pRegion is an unoccluded region of the window, and pBSRegion is 
 * an occluded region that has backing store. Since exposure events 
 * include a rectangle describing what was exposed, this routine may 
 * have to send back a series of exposure events, one for each 
 * rectangle of the region. The count field in the expose event is a 
 * hint to the client as to the number of regions that are after this
 * one.
 * Sample implementation: Xserver/mi/miexpose.c
 */
/* the local function calls PaintBg and PaintBorder so no need
 * to send an update here */
void thincWindowExposures(WindowPtr pWindow, RegionPtr pRegion,
	RegionPtr pBSRegion)
{
	t_dbg_name(func)
	ScreenPtr pScreen = pWindow->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	/* unwrap and call driver's function */
	thinc_unwrap(pThinc, pScreen, WindowExposures);
	(*pScreen->WindowExposures) (pWindow, pRegion, pBSRegion);
	thinc_wrap(pThinc, pScreen, WindowExposures, thincWindowExposures);
}

/*** deal with window private structure ***/
static int thincInitWindowPriv(WindowPtr pWindow)
{
	t_dbg_name(func)
	static unsigned long id = T_WIN_FIRST_ID;
	thincWindowPtr pPriv = THINCWINPtr(pWindow);

	thincDbg(T_DBG_WIN, ("%s: initializing window %#lx\n", func,
		t_winXID(pWindow)));

	pPriv->id = id--;
	pPriv->flags = 0;
	t_list_init(&pPriv->list);
	pPriv->app = NULL;

	if(thincAppWindowAdd(pWindow) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"window %#lx\n", func, t_winXID(pWindow)));
		return T_ERROR;
	}

	thincDbg(T_DBG_WIN, ("%s: finished. id %#lx\n", func, pPriv->id));
	return T_OK;
}

static void thincUninitWindowPriv(WindowPtr pWin)
{
	t_dbg_name(func)

	thincAppWindowRemove(pWin);
}
#endif	/* T_APPSHARE */
