/* Copyright 2007 Ricardo Baratto */
/* Copyright 2007 Columbia University */
/* This file is part of THINC.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER(S) OR AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the copyright holder(s)
 * and author(s) shall not be used in advertising or otherwise to promote
 * the sale, use or other dealings in this Software without prior written
 * authorization from the copyright holder(s) and author(s).
 */
/* 
 * "Virtual" display driver. Mostly in charge of bootstrapping the real
 * thinc code
 *
 * Probe, PreInit & ScreenInit function heavily based on xfree86's
 * fbdev driver
 * Structure, names, comments and jokes based on generic.c
 * Also inspired by examples from Design doc and vmware's driver
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

#include "thincdev.h"

/*** Version Information for XFree86 ***/
#define THINCDEV_NAME			"THINCDEV"
#define THINCDEV_DRIVER_NAME		"thincdev"
#define THINCDEV_VERSION_NAME		"0.1"
#define THINCDEV_VERSION_MAJOR		0
#define THINCDEV_VERSION_MINOR		1
#define THINCDEV_PATCHLEVEL		0
#define THINCDEV_VERSION_CURRENT	((THINCDEV_VERSION_MAJOR << 24) | \
					(THINCDEV_VERSION_MINOR << 16) | \
					THINCDEV_PATCHLEVEL)

const char* thincdevVersionName = "THINC Driver Version 0.1.0";

/*** local functions ***/

static void THINCDEVIdentify(int);
static Bool THINCDEVProbe(DriverPtr, int);
static Bool THINCDEVPreInit(ScrnInfoPtr, int);
static Bool THINCDEVScreenInit(int, ScreenPtr, int, char **);
static Bool THINCDEVEnterVT(int, int);
static void THINCDEVLeaveVT(int, int);
static int THINCDEVValidMode(int, DisplayModePtr, Bool, int);
static Bool THINCDEVCloseScreen(int scrnIndex, ScreenPtr pScreen);

/*** driverPrivate funcs ***/
static Bool THINCDEVGetRec(ScrnInfoPtr pScrn);
static void THINCDEVFreeRec(ScrnInfoPtr pScrn);
static void THINCDEVFreeScreen(int scrnIndex, int flags);

/*** functions modeled after vfbdev.c ***/
static void THINCDEVSetVideoModes(ScrnInfoPtr pScrn);
static void THINCDEVLoadPalette(ScrnInfoPtr pScrn, int numColors, int *indices,
						   LOCO * colors, VisualPtr pVisual);
static Bool THINCDEVSaveScreen(ScreenPtr pScreen, int mode);

/* for mode switching */
static Bool THINCDEVSwitchMode(int scrnIndex, DisplayModePtr mode, int flags);
/* Define the AdjustFrame() function if the driver supports this. (?) */
static void THINCDEVAdjustFrame(int scrnIndex, int x, int y, int flags);

/*** overall driver information ***/
DriverRec THINCDEV = {
	THINCDEV_VERSION_CURRENT,
	THINCDEV_DRIVER_NAME,
	THINCDEVIdentify,
	THINCDEVProbe,
	NULL,	/* No options yet */
	NULL,
	0
};

/*** dummy chipset info ***/
static SymTabRec ThincDevChipsets[] = {
	{0, "thincdev"},
	{-1, NULL}
};

/*** our private struct ***/
typedef struct {
	unsigned char *thincdevMem;		/* the framebuffer */
	unsigned long size;			/* framebuffer's size */
	EntityInfoPtr pEnt;
	CloseScreenProcPtr CloseScreen;
} THINCDEVRec, *THINCDEVPtr;

/*** Outside dependencies ***/
static const char *fbSymbols[] = {
	"fbScreenInit",
	"fbPictureInit",
	NULL
};

/*** Get pointer to driverPrivate from ScrnInfoPtr ***/
#define THINCDEVPTR(p) ((THINCDEVPtr)((p)->driverPrivate))

static MODULESETUPPROTO(ThincDevSetup);

static XF86ModuleVersionInfo ThincDevVersionRec = {
	THINCDEV_DRIVER_NAME,
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	THINCDEV_VERSION_MAJOR, THINCDEV_VERSION_MINOR, THINCDEV_PATCHLEVEL,
	ABI_CLASS_VIDEODRV,
	ABI_VIDEODRV_VERSION,
	MOD_CLASS_VIDEODRV,
	{0, 0, 0, 0}
};

XF86ModuleData thincdevModuleData = 
	{ &ThincDevVersionRec, ThincDevSetup, NULL };

static pointer ThincDevSetup(pointer module, pointer options, int *errmaj,
	int *errmin)
{
	const char *func = __func__;
	/* we don't like being loaded more than once */
	static Bool Initialised = FALSE;

	thincdevDbg("%s(): starting\n", func);

	if (!Initialised) {
		/* Have to figure out... For now, just calling
		 * xf86AddDriver() to add ourselves to list of
		 * X drivers (xf86DriverList[] - Design:7,99) */
		thincdevDbg("%s(): Adding us to list of drivers\n", func);
		Initialised = TRUE;
		xf86AddDriver(&THINCDEV, module, 0);
		LoaderRefSymLists(fbSymbols, NULL);

		thincdevDbg("%s(): Load ok, finished\n", func);
		return (pointer) TRUE;
	} else {
		/* tsk, tsk.. trying to load us more than once? 
		 * not gonna happen baby... */
		thincdevError(-1, "%s(): help, help! Somebody tried to load "
			"us more than once!!\n", func);

		if (errmaj)
			*errmaj = LDR_ONCEONLY;
		return NULL;
	}
}

/*** Print identification string ***/
static void THINCDEVIdentify(int flags)
{
	const char *func = __func__;
	
	thincdevDbg("%s(): starting\n", func);
	
	xf86PrintChipsets(THINCDEV_NAME,
		"Driver for remote display technologies THINCDEV",
		ThincDevChipsets);

	thincdevDbg("%s(): finished\n", func);
}

/*** First stage: Probe devices ***/
static Bool THINCDEVProbe(DriverPtr drv, int flags)
{
	const char *func = __func__;

	Bool foundScreen = TRUE;
	int numDevSections, i;
	GDevPtr *devSections;
	ScrnInfoPtr pScrn;

	thincdevDbg("%s(): starting\n", func);

	if (flags & PROBE_DETECT) {
		thincdevDbg("%s(): PROBE_DETECT, bailing out\n", func);
		return FALSE;
	}

	/*** find Device sections in config file that match our driver ***/
	numDevSections = xf86MatchDevice(THINCDEV_DRIVER_NAME,&devSections);

	if(numDevSections <= 0) {
		thincdevError(-1, "%s(): No matching device section found "
			"in configuration file\n", func);
		return FALSE;
	}

	thincdevDbg("%s(): found %d dev sections\n", func, numDevSections);


	/* Taken from vfbdev driver: If I understood correctly, we only have
	 * to claim to X that we know there's an ISA or FB device in the 
	 * system and it'll believe us. 
	 * ISA is actually only supposed to be used for multihead config.
	 * (sort of explained in "howto", should put in manpage)
	 */
	for (i = 0; i < numDevSections; i++) {
		int entity;
		pScrn = NULL;
		if (	(devSections[i]->busID) &&
			(xf86ParseIsaBusString(devSections[i]->busID))) {
			/* found one for us */
			thincdevDbg("%s(): %d: Found isa bus string %s. "
				"Claiming it\n", func,
				i, devSections[i]->busID);
			entity = xf86ClaimIsaSlot(drv,0,devSections[i],TRUE);
			pScrn = xf86ConfigIsaEntity(pScrn, 0, entity, 
				NULL, RES_SHARED_VGA, NULL, NULL, NULL, NULL);
		} else {
			thincdevDbg("%s(): %d: Going with fb\n", func, i);
			entity = xf86ClaimFbSlot(drv,0,devSections[i],TRUE);
			pScrn = xf86ConfigFbEntity(pScrn, 0, entity, 
				NULL, NULL, NULL, NULL);
		}

		/* chk if we found something */
		if (pScrn) {
			thincdevDbg("%s(): [%d]: Found screen, configuring "
				"it\n", func, i);
			pScrn->driverVersion = THINCDEV_VERSION_CURRENT;
			pScrn->driverName = THINCDEV_DRIVER_NAME;
			pScrn->name = THINCDEV_NAME;
			pScrn->Probe = THINCDEVProbe;
			pScrn->PreInit = THINCDEVPreInit;
			pScrn->ScreenInit = THINCDEVScreenInit;
			pScrn->EnterVT = THINCDEVEnterVT;
			pScrn->LeaveVT = THINCDEVLeaveVT;
			pScrn->AdjustFrame = THINCDEVAdjustFrame;
			pScrn->SwitchMode = THINCDEVSwitchMode;
			pScrn->ValidMode = THINCDEVValidMode;
			pScrn->FreeScreen = THINCDEVFreeScreen;
			foundScreen = TRUE;
		}
	}

	xfree(devSections);
	thincdevDbg("%s(): Finished, found a screen? %s\n", func, 
		foundScreen ? "yes" : "no");
	return foundScreen;
}

/*** Second stage ***/
static Bool THINCDEVPreInit(ScrnInfoPtr pScrn, int flags)
{
	const char *func = __func__;
	int default_depth, default_bpp;
	int scridx = pScrn->scrnIndex;
	THINCDEVPtr fPtr;

	thincdevDbg("%s(): [%d] Screen %d\n", func, scridx, scridx);

	/* check flags first */
	if (flags & PROBE_DETECT)
		/* only want us to do PROBE */
		return FALSE;

	if (pScrn->numEntities != 1)
		return FALSE;

	/*** fill in monitor field ***/
	pScrn->monitor = pScrn->confScreen->monitor;

	/* create private struct */
	THINCDEVGetRec(pScrn);
	fPtr = THINCDEVPTR(pScrn);
	fPtr->pEnt = xf86GetEntityInfo(pScrn->entityList[0]);
	if (fPtr->pEnt->resources)
		return FALSE;

	/* Set depth/bpp. Must specify preferred and which ones we also 
	 * support
	 * Default depth is 24, and bpp 32
	 */
	default_depth = 24;
	default_bpp = 32;
	if(!xf86SetDepthBpp(pScrn, default_depth, default_bpp, default_bpp, 
		Support32bppFb)) {
		thincdevError(scridx, "%s(): Error in SetDepthBpp()", func);
		return FALSE;
	}

	xf86PrintDepthBpp(pScrn);

	/* set r/g/b weights. Taken from matrox driver, all zero means
	 * use defaults.
	 * From DESIGN:13:
	 * rgb weights only necessary if depth > 8
	 * Also, if depth == 8, must set rgbBits field
	 */
	thincdevDbg("%s(): [%d] Color settings\n", func, scridx);
	if (pScrn->depth > 8) {
		rgb zeros = { 0, 0, 0 };
		thincdevDbg("%s(): [%d] Depth > 8, setting rgb weights\n", func, scridx);

		/* XXX: should we chk that we support the weights set? */
		if (!xf86SetWeight(pScrn, zeros, zeros)) {
			thincdevError(scridx, "%s(): xf86SetWeight() failed\n",
				func);
			return FALSE;
		}
	} else {	/* if(pScrn->depth <= 8) */
		thincdevDbg("%s(): [%d] Depth <= 8, setting rgbBits\n", func, scridx);
		/* XXX: generic uses 6 bits per RGB, but mga/fbdev uses 8? */
		pScrn->rgbBits = 8;
	}

	/* Set Visual. -1 is stick with the server's default 
	 * Visual is one of: StaticGray, GrayScale, StaticColor,
	 * PseudoColor, TrueColor, DirectColor
	 */
	thincdevDbg("%s(): [%d] Setting default visual\n", func, scridx);
	if (!xf86SetDefaultVisual(pScrn, -1)) {
		thincdevError(scridx, "%s(): xf86SetDefaultVisual() failed\n",
			func);
		return FALSE;
	}
	/* DirectColor at larger than 8 seems to be a bad beast, so just
	 * bail on it */
	if (pScrn->depth > 8 && pScrn->defaultVisual != TrueColor) {
		thincdevError(scridx, "%s(): defaultVisual [%d] not supported"
			"at depth %d\n", func, pScrn->defaultVisual,
			pScrn->depth);
		return FALSE;
	}

	/*  Set gamma  */
	if (pScrn->depth > 1) {
		Gamma zeros = { 0.0, 0.0, 0.0 };
		thincdevDbg("%s(): [%d] Setting gamma\n", func, scridx);
		if (!xf86SetGamma(pScrn, zeros)) {
			thincdevError(scridx, "%s(): xf86SetGamma() failed\n",
				func);
			return FALSE;
		}
	}

	/* Video mode stuff. Really obscure for me, almost copy-paste
	 * from vfbdev.c :/ */
	/* XXX: chk vncfb for an optional way to do this using ClockRanges.
	 * which one's better?? */
	/* <obscure> */
	pScrn->progClock = TRUE;	/* ??? */

	/* Set chipset name */
	if (!pScrn->chipset)
		pScrn->chipset = (char *) ThincDevChipsets[0].name;

	/* arbitrary.. 32 megs should be enough (What's it used for?) */
	pScrn->videoRam = 32 * 1024 * 1024;

	thincdevDbg("%s(): [%d] Setting video modes\n", func, scridx);
	THINCDEVSetVideoModes(pScrn);

#if 0
	thincdevMsg(pScrn->scrnIndex, "Checking modes against monitor\n");
		
	{
		DisplayModePtr mode, first = mode = pScrn->modes;
		if (mode != NULL) {
			do {
				mode->status = xf86CheckModeForMonitor(mode, 
					pScrn->monitor);
				mode = mode->next;
			} while (mode != NULL && mode != first);
		}
		xf86PruneDriverModes(pScrn);
	}
#endif

	pScrn->currentMode = pScrn->modes;
	if (!pScrn->currentMode) {
		thincdevError(scridx, "%s(): No usable video mode found. "
			"Aborting.\n", func);
		return FALSE;
	}

	pScrn->virtualX = pScrn->currentMode->HDisplay;
	pScrn->virtualY = pScrn->currentMode->VDisplay;
	pScrn->displayWidth = pScrn->virtualX;

	thincdevDbg("%s(): HDisplay %d, VDisplay %d\n", func, 
		pScrn->currentMode->HDisplay, pScrn->currentMode->VDisplay);

	xf86PrintModes(pScrn);
	/* </OBSCURE-STUFF> */

	/*** Set display resolution ***/
	thincdevDbg("%s(): [%d] Setting DPI\n", func, scridx);
	xf86SetDpi(pScrn, 0, 0);

	/*** load "fb" module ***/
	/* XXX: Should we load xf1bpp and xf4bpp for depth < 8 ??? */
	if (xf86LoadSubModule(pScrn, "fb") == NULL) {
		thincdevError(scridx, "%s(): Couldn't load fb submodule\n",
			func);
		THINCDEVFreeRec(pScrn);
		return FALSE;
	}

	/* collect options */
	xf86CollectOptions(pScrn, NULL);

	/* finished */
	thincdevDbg("%s(): [%d] Finished.\n", func, scridx);
	return TRUE;
} /* end PreInit */


/*** THINCDEVScreenInit -- Called at the start of each server generation 
 * for each Screen ***/
/* [Based heavily on "vfbdev" implementation.] */
static Bool THINCDEVScreenInit(int scrnIndex, ScreenPtr pScreen, int argc,
	char **argv)
{
	const char *func = __func__;
	ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
	THINCDEVPtr fPtr = THINCDEVPTR(pScrn);
	VisualPtr visual;
	int flags, width, height;

	width = pScrn->virtualX;
	height = pScrn->virtualY;

	thincdevDbg("%s(): [%d] starting. Screen dimensions: %dx%d\n", func,
		scrnIndex, width, height);

	/*** reset visual list, so we can set it up ***/
	miClearVisualTypes();

	/*** set up the visuals ***/
	/* [from vfbdev.c] */
	/* * XXX: Not sure why we need two separate cases... */
	/* <obscure> */
	if (pScrn->bitsPerPixel > 8) {
		if (!miSetVisualTypes(pScrn->depth, TrueColorMask, 
			pScrn->rgbBits, TrueColor)) {
			thincdevError(scrnIndex, "%s(): miSetVisualTypes "
				"(TrueColor) failed\n", func);
			return FALSE;
		}
	} else {
		if (!miSetVisualTypes(pScrn->depth, 
			miGetDefaultVisualMask(pScrn->depth),
			pScrn->rgbBits, pScrn->defaultVisual)) {
			thincdevError(scrnIndex, "%s(): miSetVisualTypes"
				"(defaultVisual) failed\n", func);
			return FALSE;
		}
	}

	/*** set pixmap depths - from mi layer ***/
	if (!miSetPixmapDepths()) {
		thincdevError(scrnIndex, "%s(): miSetPixmapDepth() failed\n",
			func);
		return FALSE;
	}
	/* </obscure> */

	/*** allocate memory for frame buffer ***/
	fPtr->size = BitmapBytePad(width * pScrn->bitsPerPixel) * height;

	thincdevDbg("%s(): [%d] Allocating %ld bytes for framebuffer\n", func,
		scrnIndex, fPtr->size);
		
	fPtr->thincdevMem = xalloc(fPtr->size);
	if (fPtr->thincdevMem == NULL) {
		thincdevError(scrnIndex, "%s(): couldn't allocate "
			"framebuffer.  Aborting.\n", func);
		return FALSE;
	} else
		thincdevDbg("%s(): [%d] fb allocation succeeded\n", func,
			scrnIndex);

#if 0	/* for oreno */
	thincdevError(scrnIndex, "framebuffer address: %p size: %lu\n",
		fPtr->thincdevMem, fPtr->size);
#endif

	/*** call fbScreenInit -- from "fb" layer ***/
	thincdevDbg("%s(): [%d] calling fbScreenInit()\n", func, 
		scrnIndex);

	if (!fbScreenInit(pScreen, fPtr->thincdevMem, width, height, 
		pScrn->xDpi, pScrn->yDpi, pScrn->displayWidth, 
		pScrn->bitsPerPixel)) {
		thincdevError(scrnIndex, "%s(): fbScreenInit() failed\n",
			func);
		return FALSE;
	}

	/* <obscure>
	 * obscure because don't understand technical implications of
	 * DirectColor */
	if (pScrn->bitsPerPixel > 8) {
		thincdevDbg("%s(): [%d] fixing RGB ordering\n", func, 
			scrnIndex);
		/* Fixup RGB ordering */
		visual = pScreen->visuals + pScreen->numVisuals;
		while (--visual >= pScreen->visuals) {
			if ((visual->class | DynamicClass) == DirectColor) {
				visual->offsetRed = pScrn->offset.red;
				visual->offsetGreen = pScrn->offset.green;
				visual->offsetBlue = pScrn->offset.blue;
				visual->redMask = pScrn->mask.red;
				visual->greenMask = pScrn->mask.green;
				visual->blueMask = pScrn->mask.blue;
			}
		}
	}
	/* </obscure> */

	/* for RENDER extension. Must happen before THINC initialization */
#if 1
	fbPictureInit(pScreen, NULL, 0);
#endif

	/* set black and white colormap indices */
	thincdevDbg("%s(): [%d] setting black/white colormap indices\n",
		func, scrnIndex);
	xf86SetBlackWhitePixels(pScreen);

	/* backing store */
	/* if we don't want it, do: 
	 * pScreen->backingStoreSupport = NotUseful
	 */
	thincdevDbg("%s(): [%d] initializing backing store\n", func,
		scrnIndex);
	miInitializeBackingStore(pScreen);
	xf86SetBackingStore(pScreen);

	/* set function pointers before initializing thinc */
	pScreen->SaveScreen = THINCDEVSaveScreen;
	fPtr->CloseScreen = pScreen->CloseScreen;
	pScreen->CloseScreen = THINCDEVCloseScreen;

	/* Load THINC module */
	if (xf86LoaderCheckSymbol("ThincInit")) {
		Bool (*thincInitFunc) (ScreenPtr pScreen, unsigned char *fb) = 
			NULL;
		Bool (*thincCursorFunc) (ScreenPtr pScreen) = NULL;

		thincdevDbg("%s: [%d] Loading THINC module\n", func,
			scrnIndex);

		thincInitFunc = (Bool (*)(ScreenPtr, unsigned char*))
			LoaderSymbol("ThincInit");

		/* init the thinc module */
		if (thincInitFunc) {
			if ((*thincInitFunc)(pScreen,fPtr->thincdevMem)
				== FALSE) {
				thincdevError(scrnIndex, "%s(): "
					"Couldn't initialize THINC Module\n",
					func);
				return FALSE;
			}
		}

		thincdevDbg("%s(): [%d] THINC module succesfully initialized, "
			"initializing the cursor\n", func, scrnIndex);
		

		/* initialize cursor. Needs to be done before THINC can
		 * initialize its hardware cursors */
		miDCInitialize(pScreen, xf86GetPointerScreenFuncs());

		/* do thinc's cursor initialization */
		thincCursorFunc = (Bool (*)(ScreenPtr)) 
			LoaderSymbol("ThincCursorInit");
		if(thincCursorFunc) {
			if((*thincCursorFunc)(pScreen) == FALSE) {
				thincdevError(scrnIndex, "%s(): "
					"Couldn't initialize THINC's cursor\n",
					func);
				return FALSE;
			}
		}

		thincdevDbg("%s(): [%d] THINC initialization finished\n",
			func, scrnIndex);
	} else {
		thincdevDbg("%s(): [%d] THINC module not found\n", func,
			scrnIndex);

		/* just initialize the software cursor */
		miDCInitialize(pScreen, xf86GetPointerScreenFuncs());
	}

	/* call miCreateDefColormap to initialize colormap */
	thincdevDbg("%s(): [%d] Creating default colormap\n", func,
		scrnIndex);
	if (!miCreateDefColormap(pScreen)) {
		thincdevError(scrnIndex, "%s(): Couldn't create default "
			"colormap\n", func);
		return FALSE;
	}

	thincdevDbg("%s(): [%d] xf86HandleColormaps()\n", func, scrnIndex);
	flags = CMAP_PALETTED_TRUECOLOR;
	if (!xf86HandleColormaps(pScreen, 256, 8, THINCDEVLoadPalette, NULL, 
		flags)) {
		thincdevError(scrnIndex, "%s():xf86HandleColormaps failed\n",
			func);
		return FALSE;
	}

	thincdevDbg("%s(): [%d] Final touches, adrenaline going up...\n", 
		func, scrnIndex);

	/* can't believe it!! The End... */
	thincdevDbg("%s(): [%d] finished, woohoo!\n", func, scrnIndex);
	return TRUE;
}

/* These two should only be dummy. We don't expect to have
 * them called (since the user won't be changing vts on the
 * server)
 */
static Bool THINCDEVEnterVT(int scrnIndex, int flags)
{
	const char *func = __func__;
	thincdevDbg("%s()\n", func);

#if 0
	if(!xf86Screens[scrnIndex]->EnterVT) {
		thincdevDbg("No EnterVT!!!!\n");
		return FALSE;
	} 
	xf86Screens[scrnIndex]->EnterVT(scrnIndex, flags);
#endif

	return TRUE;
}

static void THINCDEVLeaveVT(int scrnIndex, int flags)
{
	const char *func = __func__;
	thincdevDbg("%s()\n", func);
	return;
}

static Bool THINCDEVCloseScreen(int scrnIndex, ScreenPtr pScreen)
{
	const char *func = __func__;
	ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
	THINCDEVPtr fPtr = THINCDEVPTR(pScrn);

	thincdevDbg("%s: [%d] starting...\n", func, scrnIndex);
	xfree(fPtr->thincdevMem);
	pScrn->vtSema = FALSE;

	pScreen->CloseScreen = fPtr->CloseScreen;

	thincdevDbg("%s(): [%d] finished\n", func, scrnIndex);

	/* call fb function */
	return (*pScreen->CloseScreen) (scrnIndex, pScreen);
}

static void THINCDEVFreeScreen(int scrnIndex, int flags)
{
	const char *func = __func__;

	thincdevDbg("%s(): [%d]\n", func, scrnIndex);

	THINCDEVFreeRec(xf86Screens[scrnIndex]);
	
	thincdevDbg("%s(): [%d] Finished\n", func, scrnIndex);
}

static int THINCDEVValidMode(int index, DisplayModePtr mode, Bool verbose,
	int flags)
{
	const char *func = __func__;

	thincdevDbg("%s(): Validating mode %s with index %d\n", func,
		mode->name, index);
	return MODE_OK;
}

/* for mode switching */
static Bool THINCDEVSwitchMode(int scrnIndex, DisplayModePtr mode, int flags)
{
	const char *func = __func__;

	thincdevDbg("%s(): [%d] Switching to mode %s\n", func, scrnIndex,
		mode->name);

	return TRUE;
}

/* Define the AdjustFrame() function if the driver supports this. (?) */
static void THINCDEVAdjustFrame(int scrnIndex, int x, int y, int flags)
{
	const char *func = __func__;

	thincdevDbg("%s(): [%d] Adjusting frame to (%d,%d)\n", func,
		scrnIndex, x, y);

	return;
}

/* (de)allocate driver's screen specific data (ThincDevRec), and hook it 
 * into the ScrnInfoRec's driverPrivate field
 * (Design:99-100)
 */
static Bool THINCDEVGetRec(ScrnInfoPtr pScrn)
{
	const char *func = __func__;

	thincdevDbg("%s(): [%d]\n", func, pScrn->scrnIndex);

	if (pScrn->driverPrivate != NULL)
		return TRUE;

	/* xnfalloc(): zeroes out and makes server exit if fails */
	pScrn->driverPrivate = xnfalloc(sizeof(THINCDEVRec));

	thincdevDbg("%s(): [%d] Finished\n", func, pScrn->scrnIndex);

	return TRUE;
}

static void THINCDEVFreeRec(ScrnInfoPtr pScrn)
{
	const char *func = __func__;

	thincdevDbg("%s(): [%d]\n", func, pScrn->scrnIndex);

	if (!pScrn->driverPrivate)
		return;

	xfree(pScrn->driverPrivate);
	pScrn->driverPrivate = NULL;

	thincdevDbg("%s: [%d] Finished\n", func, pScrn->scrnIndex);

	return;
}

/* from vfbdev.c */
static void THINCDEVSetVideoModes(ScrnInfoPtr pScrn)
{
	const char *func = __func__;
	char **modename;
	DisplayModePtr mode, this, last = NULL;

	thincdevDbg("%s(): [%d] Starting...\n", func, pScrn->scrnIndex);

	if (pScrn->display->modes == NULL) {
		thincdevDbg("%s(): [%d] No modes in display, returning\n",
			func, pScrn->scrnIndex);
		return;
	}
	thincdevDbg("%s(): [%d] set pScrn->display->modes\n",
		func, pScrn->scrnIndex);

	/* loop through display modes checking if they're known to X. After
	 * we return, the modes are actually checked against the monitor 
	 * capabilities. Note however, that for us, you shouldn't specify
	 * any characteristics (eg vert/hz refresh) in the monitor config 
	 * section (that way X lets us use any video mode) -RB */
	for (modename = pScrn->display->modes; *modename != NULL; modename++) {
		for (mode = pScrn->monitor->Modes; mode != NULL; 
			mode = mode->next)
			if (strcmp(mode->name, *modename) == 0)
				break;
		if (mode == NULL) {
			thincdevMsg(pScrn->scrnIndex, 
				"\tmode \"%s\" not found\n", *modename);
			continue;
		}
		thincdevMsg(pScrn->scrnIndex, "\tmode \"%s\" ok\n", *modename);
		if (pScrn->modes == NULL) {
			pScrn->modes = xnfalloc(sizeof(DisplayModeRec));
			this = pScrn->modes;
			memcpy(this, mode, sizeof(DisplayModeRec));
			this->next = this;
			this->prev = this;
		} else {
			this = xnfalloc(sizeof(DisplayModeRec));
			memcpy(this, mode, sizeof(DisplayModeRec));
			this->next = pScrn->modes;
			this->prev = last;
			last->next = this;
			pScrn->modes->prev = this;
		}
		last = this;
	}
	thincdevDbg("%s(): [%d] Finished\n", func, pScrn->scrnIndex);
}

static void THINCDEVLoadPalette(ScrnInfoPtr pScrn, int numColors, 
	int *indices,  LOCO * colors, VisualPtr pVisual)
{
	const char *func = __func__;
	thincdevDbg("%s(): [%d]\n", func, pScrn->scrnIndex);

	return;
}

static Bool THINCDEVSaveScreen(ScreenPtr pScreen, int mode)
{
	const char *func = __func__;
	thincdevDbg("%s(): [%d]\n", func, pScreen->myNum);

	return TRUE;
}

/********************
 * helper functions *
 ********************/

/* info msgs */
void thincdevMsg(int index, const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	xf86VDrvMsgVerb(index, X_INFO, 1, fmt, args);
	va_end(args);
}

/* error messages */
void thincdevError(int index, const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	xf86VDrvMsgVerb(index, X_ERROR, 1, fmt, args);
	va_end(args);
}

/* debug msgs */
void thincdevDbg(const char *fmt, ...)
{
#ifdef THINCDEV_DEBUG
	va_list args;
	va_start(args, fmt);
	xf86VDrvMsgVerb(-1, X_INFO, 1, fmt, args);
	va_end(args);
#endif
}
